//-------------------------------------------------------------------------
// a simple test program to receive interrupts from the 
// WINNT-PCICC32 driver from ARW Elektronik, Germany
//
// this example accesses the WINNT driver via libcc32.dll
//
// (c) 2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started from CatchInterrupts_CC32_NT            AR           28.02.2001
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <stdlib.h>  // for atexit()
#include <windows.h>
#include <winioctl.h>
#include <conio.h>   // because of _getch()
#include "..\pcicc32_ni\libcc32.h"    // include the structure definitions of DLL

//-------------------------------------------------------------------------
// DEFINES
//
#define _WINNT          // for WINNT only

#ifdef _WINNT
#define DEFDEVICENAME "\\\\.\\PCICC32:\\CC32_xx"  // a device name 'template' for WINNT/2000/...
#else  // no _WINNT
#define DEFDEVICENAME "\\\\.\\C:\\windows\\system\\Vpcic32D.vxd" // the same for WIN95/98
#endif 

#define DEFAULT_MODULE_NUMBER    1    // number of connected CC32 module

#define MESSAGESTRING "A program to receive interrupts from the PCICC32 WINNT driver. (%s)\n"


//-------------------------------------------------------------------------
// what the program provides
//
void hlpMsg(void)
{
	printf(MESSAGESTRING, __DATE__);
    printf("usage: CatchInterruptsNT [-?]\n");
	printf("       ? - get help\n");
	printf("\n");
	printf("This program addresses interface #%d\n", DEFAULT_MODULE_NUMBER);
	printf("\n");
}

void stopAtEnd(void)
{
	printf("Press any key to proceed.\n");
	_getch();
}

//-------------------------------------------------------------------------
// the main program
//
int main(int argc, char *argv[])
{
	int result;
	unsigned long dwStatus;
	CC32_HANDLE  hHandle;   // the handle to my device path

	if (argc != 1)
	{
		hlpMsg();
		exit(0);
	}

	// to stop at each exit
	atexit(stopAtEnd);

	printf(MESSAGESTRING, __DATE__);
	printf("Ctrl-C to stop.\n");

	/* open a path to a device. */
    result = cc32_open(DEFDEVICENAME, DEFAULT_MODULE_NUMBER, &hHandle);
	if (result)
	{
		printf("Can't open interface \"%s\" to CC32-module \"%d\"!\n", DEFDEVICENAME, DEFAULT_MODULE_NUMBER);
		exit(result);
	}
	else
	{	
		// enable interrupts
		cc32_enable_interrupt(hHandle);

		// get status
		result = 0;
		while (!result)
		{
			result = cc32_get_interrupt(hHandle, &dwStatus);

			printf("Got status 0x%08x.\n", dwStatus);

			if (!result)
				cc32_enable_interrupt(hHandle);
		}

		// disable interrupts
		cc32_disable_interrupt(hHandle);

		// close device -------------------------------------
		if (hHandle != INVALID_HANDLE_VALUE)
		{
			result = cc32_close(hHandle); 
			if (result)
			{
				printf("Can't close interface!\n");
				exit(result);
			}
		}

		return 0;
	}
}