//-------------------------------------------------------------------------
// CatchInterruptsW95 - a simple test program to catch interrupts generated
// by the PCICC32 interface of ARW Elektronik, Germany
//
// (c) 2001 ARW Elektronik - written by Klaus Hitschler (AR)
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps                                     AR           17.03.2001
//

//--------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <stdio.h>
#include <conio.h>    // wg. kbhit()
#include "..\vpcic32d\vpcic32d.h"

//--------------------------------------------------------------------------------------
// DEFINES
//
#define DRIVER_PATH "\\\\.\\c:\\windows\\system\\vpcic32d.vxd"
#define MY_INTERFACE 1 

#define wordNAF(base, N, A, F) ((WORD *) ((DWORD)base + (N << 10) + (A << 6) + ((F & 0xF) << 2)))
#define longNAF(base, N, A, F) ((DWORD *)((DWORD)base + (N << 10) + (A << 6) + ((F & 0xF) << 2)))

//--------------------------------------------------------------------------------------
// GLOBALS
//
char *cszProgName;
int verbose = 0;
DWORD dwOldHandler = 0;  // temporary storage for pointer to old irq handler

//--------------------------------------------------------------------------------------
// FUNCTIONS
//

//--------------------------------------------------------------------------------------
// the general help - invoke with -?
//
void hlpMsg(void)
{
	printf("catchInterruptsW95 - a simple test program for CC32-modules (WIN95/98 only).\n");
	printf("(c) ARW Elektronik (AR) - %s\n", __DATE__);
	printf("\n");
	printf("usage: catchInterruptsW95 [-?] [-p=path] [-m=ModuleNumber] [-v]\n");
	printf("       -p= give the driver path (default: %s).\n", DRIVER_PATH);
	printf("       -m= give the number of the CC32 module (default: %d).\n", MY_INTERFACE);
	printf("       -v  make the program verbose.\n");
	printf("       -?  this help.\n");
}

//--------------------------------------------------------------------------------------
// a function to get the base address of the CAMAC range
//
static int attachWindow(HANDLE nHandle, DWORD dwInterface, void **base)
{
	VPCIC32D_WINDOW window;
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

    *base = NULL;
	if (nHandle == (HANDLE)-1) return -1;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_ATTACH_CC32,
						&dwInterface, sizeof(dwInterface), 
						&window,      sizeof(window),
								 &DIOC_count, NULL);

    if (!result) 
		return GetLastError();
	else
	{
		*base = window.pvWindowBase;
		return 0;
	}
}

//--------------------------------------------------------------------------------------
// a function to release a attached range
//
static int detachWindow(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

	if (nHandle == (HANDLE)-1) return -1;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_DETACH_CC32,
						&dwInterface, sizeof(dwInterface), 
						NULL,      0, &DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
		return 0;
}

//--------------------------------------------------------------------------------------
// the irq-enabler
//
static int enableIrq(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;
	VPCIC32D_IRQ_CONTROL irqControl;

	if (nHandle == (HANDLE)-1) return -1;

	irqControl.dwInterface = dwInterface;
	irqControl.wEnable     = 1;  // enable

	// enable the interrupt
	result = DeviceIoControl(nHandle, VPCIC32_CONTROL_INTERRUPTS,
						&irqControl, sizeof(irqControl), 
						&irqControl, sizeof(irqControl), 
						&DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
		return 0;
}

//--------------------------------------------------------------------------------------
// the irq-disabler
//
static int disableIrq(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;
	VPCIC32D_IRQ_CONTROL irqControl;

	if (nHandle == (HANDLE)-1) return -1;

	irqControl.dwInterface = dwInterface;
	irqControl.wEnable     = 0;  // disable

	// enable the interrupt
	result = DeviceIoControl(nHandle, VPCIC32_CONTROL_INTERRUPTS,
						&irqControl, sizeof(irqControl), 
						&irqControl, sizeof(irqControl), 
						&DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
		return 0;
}

//--------------------------------------------------------------------------------------
// attach a user level interrupt handler
//
static int attachIrqHandler(HANDLE nHandle, DWORD dwInterface, void (*handler)(DWORD))
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;
	VPCIC32D_IRQ_HANLDER irqHandler;

	if (nHandle == (HANDLE)-1) return -1;

	irqHandler.dwInterface  = dwInterface;
	irqHandler.dwIrqHandler = (DWORD)handler; // attach

	// enable the interrupt
	result = DeviceIoControl(nHandle, VPCIC32_INSTALL_IRQ_HANDLER,
						&irqHandler, sizeof(irqHandler), 
						&irqHandler, sizeof(irqHandler), 
						&DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
	{
		dwOldHandler = irqHandler.dwIrqHandler;
		return 0;
	}
}

//--------------------------------------------------------------------------------------
// attach a user level interrupt handler
//
static int detachIrqHandler(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;
	VPCIC32D_IRQ_HANLDER irqHandler;

	if (nHandle == (HANDLE)-1) return -1;

	irqHandler.dwInterface  = dwInterface;
	irqHandler.dwIrqHandler = dwOldHandler; // detach

	// enable the interrupt
	result = DeviceIoControl(nHandle, VPCIC32_INSTALL_IRQ_HANDLER,
						&irqHandler, sizeof(irqHandler), 
						&irqHandler, sizeof(irqHandler), 
						&DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
	{
		dwOldHandler = irqHandler.dwIrqHandler;
		return 0;
	}
}

//--------------------------------------------------------------------------------------
// the user level irq-handler
//
void IrqHandler(DWORD dwInterruptVector)
{
	printf("catchInterruptsW95: Interrupt 0x%08x occured!\n", dwInterruptVector);
}

//--------------------------------------------------------------------------------------
// the frame for all the rest
//
int main(int argc, char *argv[])
{
	HANDLE		nHandle;
	void        *base;
	int         error;
	char        c, *ptr;
    int         i;
	int nModuleNumber = MY_INTERFACE;
	char        *driver_path = DRIVER_PATH;

	cszProgName = argv[0];
	
	for (i = 1; i < argc; i++)
	{
		ptr = argv[i];

		if ((*ptr == '-') || (*ptr == '/')) ptr++;
		c = *ptr;
		ptr++;
		if (*ptr == '=') ptr++;

		switch(tolower(c))
		{
			case '?': hlpMsg();
					  exit(0);
			case 'm': sscanf(ptr, "%d", &nModuleNumber);
				      break;
			case 'p': driver_path = ptr;
				      break;
			case 'v': verbose = 1;
				      break;
			default:  printf("pcic32_test: Illegal argument '%c'! (Try -?)\n", c);
				      exit(-1);
		}
	}

	printf("catchInterruptsW95: Try to test CC32-module #%d on path %s.\n", nModuleNumber, driver_path);

	nHandle = CreateFile(driver_path, 
						0, 
						0, 
						NULL, 
						0, 
						FILE_FLAG_DELETE_ON_CLOSE, 
						NULL);
	
	if (nHandle != INVALID_HANDLE_VALUE) 
	{ 
        error = attachWindow(nHandle, nModuleNumber, &base);
		if (error)
			printf("catchInterruptsW95: error @ attachWindow() = %d\n", error);
        else
		{
			printf("catchInterruptsW95: window base address into CC32 = 0x%08x\n", base);

			error = attachIrqHandler(nHandle, nModuleNumber, IrqHandler);
			if (error)
				printf("catchInterruptsW95: error @ attachIrqHandler() = %d\n", error);
			else
			{
				error = enableIrq(nHandle, nModuleNumber);
				if (error)
					printf("catchInterruptsW95: error @ enableIrq() = %d\n", error);
				else
				{
					// enable LAM interrupts
					*longNAF(base, 28, 1, 16) = 0xFFFFFFFF;

					// wait for interrupts
					do
					{
						// generate a LAM interrupt from the test board
						// *longNAF(base, 1, 3, 16) = 0x55aa55aa;

						// catch the callback function
						SleepEx(100, TRUE);
					}
					while (!kbhit());

					printf("\n");
					error = disableIrq(nHandle, nModuleNumber);
					if (error)
						printf("catchInterruptsW95: error @ disableIrq() = %d\n", error);
					error = detachIrqHandler(nHandle, nModuleNumber);
					if (error)
						printf("catchInterruptsW95: error @ attachIrqHandler() = %d\n", error);				
				}
			}
		}
        error = detachWindow(nHandle, nModuleNumber);
		if (error)
			printf("error @ detachWindow() = %d\n", error);

		CloseHandle(nHandle);
	}
	else
		printf("catchInterruptsW95: Can't open \"%s\" (Error %d)\n", DRIVER_PATH, GetLastError());

	printf("catchInterruptsW95: Press return to continue ...\n");

	getchar();

	return 0;
}

//--------------------------------------------------------------------------------------
// the end
//--------------------------------------------------------------------------------------
