
//-------------------------------------------------------------------------
// a simple test program to test the main functions of the 
// PCICC32 drivers for WIN95 or WINNT from ARW Elektronik, Germany
//
// this example uses the 'pcicc32_ni.dll' (link with .\pcicc32_ni.lib)
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           16.04.2000
// added IRQ and 'UNTIL_NOT_Q' handling for WINNT  AR           24.03.2001
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <windows.h>

extern "C"				// it's no C++ DLL
{
#include "..\pcicc32_ni\LibCC32.h"    // include the prototypes of the DLL calls
}


//-------------------------------------------------------------------------
// DEFINES
//
#define _WINNT          // for WINNT only

#ifdef _WINNT
#define DEFDEVICENAME "\\\\.\\PCICC32:\\CC32_xx"  // a device name 'template' for WINNT
#else  // no _WINNT
#define DEFDEVICENAME "\\\\.\\C:\\windows\\system\\Vpcic32D.vxd" // the same for WIN95/98
#endif 

#define MODULE_NUMBER    1    // number of connected CC32 module


//-------------------------------------------------------------------------
// MAIN
//
int main(int argc, char *argv[])
{
	int result = 0;
	CC32_HANDLE  hHandle;   // the handle to my device path
	unsigned long lbuf;		// holds the long (32 bits) reads
	unsigned short wbuf;	// holds the short (16 bit) reads
	char bTimeout;			// holds a pending timeout
	char bInterrupt;		// holds a pending LAM request
	char bX;                // holds the X response
	char bQ;				// holds the Q response
	int i;
	int nError = 0;             
	unsigned long *dwBuffer;
	unsigned long bufferSize;

	printf("A small test program to access the ARW CC32 CAMAC interface using the\n");
	printf("'pcicc32_ni.dll' - %s\n", __DATE__);

	/* open a path to a device. */
    result = cc32_open(DEFDEVICENAME, MODULE_NUMBER, &hHandle);
	if (result)
	{
		printf("Can't open interface %s!\n", DEFDEVICENAME);
		exit(result);
	}

	// write a reset to CC32
	cc32_write_word(hHandle, 31, 0, 16, 0x55aa);

	// read the CC32 local status
	lbuf = cc32_read_long_all(hHandle, 0, 0, 0);
	printf("Read %d bytes (0x%08x) with X and Q.\n", 4, lbuf);

	// enable all LAM channels
	cc32_write_long(hHandle, 28, 1, 16, 0xFFFFFFFF);

	// with the test board: generate LAMs
	cc32_write_long(hHandle, 1, 3, 16, 0x55aa55aa);
	
	// reset the counter at the test board
	cc32_write_long(hHandle, 1, 5, 16, 0x55aa55aa);

	// read succesive data from the test board
	for (i = 0; i < 20; i++)
	{
		lbuf = cc32_read_long(hHandle, 1, 5, 0, &bX, &bQ);
		printf("Read %d bytes (0x%08x), X=%d, Q=%d.\n", 4, lbuf, bX, bQ);
	}

	// read succesive data from the test board
	for (i = 0; i < 20; i++)
	{
		wbuf = cc32_read_word(hHandle, 1, 5, 0);
		printf("Read %d bytes (0x%04x).\n", 2, wbuf);
	}


	// read a block of data from test board with waiting for 'NOT-Q' ------- begin -------
	// reset the counter at the test board
	cc32_write_long(hHandle, 1, 5, 16, 0x55aa55aa);


	// set access parameter to autoread
	cc32_access_switch(hHandle, SW_UNTIL_NOT_Q | SW_AUTOREAD);

	// allocate buffer for data
	bufferSize = sizeof(unsigned long) * 0x10000 * 2;
	dwBuffer = (unsigned long *)malloc(bufferSize);
	if (dwBuffer == NULL)
	{
		nError = GetLastError();

		printf("Can't allocate buffer. (0x%08x)\n", nError);

		// close device -------------------------------------
		if (hHandle != INVALID_HANDLE_VALUE)
			CloseHandle(hHandle);

		return nError;
	}

	// read succesive data from the test board
	bufferSize >>= 2;  // make count of longs
	nError = cc32_read_long_buffer(hHandle, 1, 0, 5, dwBuffer, &bufferSize);
	printf("Read 0x%08x bytes (0x%08x / 0x%08x / 0x%08x)\n", bufferSize, *dwBuffer, *(dwBuffer + 1), *(dwBuffer + 2));

	// delete the buffer
	if (dwBuffer)
		free(dwBuffer);
	// read a block of data from test board with waiting for 'NOT-Q' ------ end ----------

	// set access parameter to standard
	cc32_access_switch(hHandle, 0);

	// get the PCIADA status and clears a pending timout
	cc32_poll_error(hHandle, &bTimeout, &bInterrupt);
	printf("bTimeout = %d, bInterrupt = %d\n", bTimeout, bInterrupt);

	// read all pending LAM channels
	lbuf = cc32_read_long_all(hHandle, 28, 0, 0);
	printf("Read LAM status %d bytes (0x%08x).\n", 4, lbuf);

	// reset all LAM channels
	cc32_write_long(hHandle, 28, 0, 16, 0);

	// get the PCIADA status again
	cc32_poll_error(hHandle, &bTimeout, &bInterrupt);
	printf("bTimeout = %d, bInterrupt = %d\n", bTimeout, bInterrupt);


	/* close the opened path */
	result = cc32_close(hHandle); 
	if (result)
	{
		printf("Can't close interface!\n");
		exit(result);
	}

	return 0;
}