//-------------------------------------------------------------------------
// a simple test program to test the main functions of the 
// WINNT-PCICC32 driver from ARW Elektronik, Germany
//
// this example accesses dircetly the WINNT driver through 'ioctls'
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           16.04.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include <pciCC32.h>    // include the structure definitions and IOCTL codes

#define DEFDEVICENAME "\\\\.\\PCICC32:\\CC32_xx"  // a device name 'template'
#define MODULE_NUMBER 1                           // module number of CC32

HANDLE hHandle = INVALID_HANDLE_VALUE;  // I use a global handle to my device path

//-------------------------------------------------------------------------
// create a DeviceName out of DEFDEVICENAME and the device number
//
static char *DeviceName(int nIfcNum)
{
	static char buffer[255];
	char *ptr = buffer;

	strcpy(buffer, DEFDEVICENAME);

	while (*ptr) ptr++;
	
	do
	{
		ptr--;
	} while (*ptr != '_');  // step back until '-'

	ptr++;
	
	if (nIfcNum >= 10)      // add the interface number
	{
		*ptr++   = '1';
		nIfcNum -= 10;
	}

	*ptr++ = '0' + nIfcNum;
	*ptr = 0;

	// a little help for the users
    return buffer;
}

//-------------------------------------------------------------------------
// open my path to the interface
//
static HANDLE openPath(char *name)
{
    return (HANDLE)CreateFile( 
                     name, 
                     GENERIC_READ | GENERIC_WRITE, 
                     0, 
                     NULL, 
                     OPEN_EXISTING, 
                     FILE_ATTRIBUTE_NORMAL, 
                     NULL 
                     );
}

//-------------------------------------------------------------------------
// init the interface
//
int Init_InterfaceNT(int nIfcNum)
{
  int error = 0;

  // start the driver ------------------
  if (hHandle == INVALID_HANDLE_VALUE)
  {
		if ((hHandle = openPath(DeviceName(nIfcNum))) == (HANDLE)-1)
			return GetLastError();
  }

  return 0;
}


//-------------------------------------------------------------------------
// set the access parameter - (note: block transfer is prepared only)
//
void setAccessParameter(HANDLE hHandle, UCHAR bAccessType, UCHAR bBlockTransfer)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
    PCICC32_ACCESS_COMMAND access_parameter;

	access_parameter.bAccessType		= bAccessType;
	access_parameter.bBlockTransfer		= bBlockTransfer;

	result = DeviceIoControl(hHandle, PCICC32_SET_ACCESS_PARA,
					&access_parameter, (DWORD)sizeof(access_parameter), NULL,
						   0, &DIOC_count, NULL);	 

	if (!result)
	{
		printf("Can't set access parameter! (0x%08x)\n", result);
		return;
	}
}

//-------------------------------------------------------------------------
// get the polling status of PCIADA interface timeout and LAM interrupts
//
void getStatus(HANDLE hHandle, char *bTimeout, char *bInterrupt)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
    PCICC32_STATUS status;

	*bTimeout   = 0;
	*bInterrupt = 0;

	result = DeviceIoControl(hHandle, PCICC32_GET_STATUS,
					NULL, 0, &status,
						   sizeof(status), &DIOC_count, NULL);	 

	if (!result)
	{
		printf("Can't get CC32 status! (0x%08x)\n", result);
		return;
	}	

	*bTimeout   = (status.bTimeout)   ? 1 : 0;
	*bInterrupt = (status.bInterrupt) ? 1 : 0;
}

//-------------------------------------------------------------------------
// clear a pending timeout status
//
void clearStatus(HANDLE hHandle)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	result = DeviceIoControl(hHandle, PCICC32_CLEAR_STATUS,
					NULL, 0, NULL, 0, &DIOC_count, NULL);	 

	if (!result)
	{
		printf("Can't get CC32 status! (0x%08x)\n", result);
		return;
	}	
}

//-------------------------------------------------------------------------------------------
// read a word/long from CAMAC
//
DWORD ReadCAMAC(HANDLE hHandle, unsigned long adr, void *buffer, int nBytesToRead)
{
  DWORD bytesRead;

  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);

  if (!ReadFile(hHandle, buffer, nBytesToRead, &bytesRead, NULL))
	  printf("Can't read (0x%08x)\n", GetLastError());

  return bytesRead;
}

//-------------------------------------------------------------------------------------------
// write a word/long to CAMAC
//
DWORD  WriteCAMAC(HANDLE hHandle, unsigned long adr, void *buffer, int nBytesToWrite)
{
  DWORD bytesWritten;

  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, buffer, nBytesToWrite, &bytesWritten, NULL))
	  printf("Can't write (0x%08x)\n", GetLastError());

  return bytesWritten;
}


//-------------------------------------------------------------------------
// the main program
//
int main(int argc, char *argv[])
{
	int result;
	int i;

	printf("A simple test program to test the PCICC32 WINNT driver. (%s)\n", __DATE__);

	result = Init_InterfaceNT(MODULE_NUMBER);  // open the device path to the 1st interface
	if (result)
	{
		printf("Can't Init_InterfaceNT() - returns error 0x%08x\n", result);
		return(result);
	}
	else
	{	
		long  lbuf = 0;
		short wbuf = 0;
		DWORD len  = 0;
		char bTimeout   = 0;
		char bInterrupt = 0;

		// set the future access width to long
		setAccessParameter(hHandle, LONG_ACCESS, 0);

		// write a reset to CC32
		lbuf = 0x55aa55aa;
		len = WriteCAMAC(hHandle, NAF(31,0,16), &lbuf, 4);
		printf("Reset done.\n", len);

		// read the CC32 local status
		len = ReadCAMAC(hHandle, NAF(0,0,0), &lbuf, 4);
		printf("Read %d bytes (0x%08x)\n", len, lbuf);

		// enable all LAM channels
		lbuf = 0xFFFFFFFF;
		len = WriteCAMAC(hHandle, NAF(28,1,16), &lbuf, 4);
		printf("Write %d bytes. (0x%08x)\n", len, lbuf);

		// with the test board: generate LAMs
		lbuf = 0x55aa55aa;
		len = WriteCAMAC(hHandle, NAF(1,3,16), &lbuf, 4);
		printf("Write %d bytes. (0x%08x)\n", len, lbuf);

		
		// reset the counter at the test board
		lbuf = 0x55aa55aa;
		len = WriteCAMAC(hHandle, NAF(1,5,16), &lbuf, 4);
		printf("Write %d bytes. (0x%08x)\n", len, lbuf);

		// read succesive data from the test board
		for (i = 0; i < 20; i++)
		{
			len = ReadCAMAC(hHandle, NAF(1,5,0), &lbuf, 4);
			printf("Read %d bytes (0x%08x)\n", len, lbuf);
		}

		// set the future access width to word
		setAccessParameter(hHandle, WORD_ACCESS, 0);

		// read succesive data from the test board
		for (i = 0; i < 20; i++)
		{
			len = ReadCAMAC(hHandle, NAF(1,5,0), &wbuf, 2);
			printf("Read %d bytes (0x%04x)\n", len, wbuf);
		}


		// get the PCIADA status
		getStatus(hHandle, &bTimeout, &bInterrupt);
		printf("bTimeout = %d, bInterrupt = %d\n", bTimeout, bInterrupt);

		// clear a pending timout
		clearStatus(hHandle);

		// set the future access width to long
		setAccessParameter(hHandle, LONG_ACCESS, 0);

		// read all pending LAM channels
		len = ReadCAMAC(hHandle, NAF(28,0,0), &lbuf, 4);
		printf("Read LAM status %d bytes (0x%08x)\n", len, lbuf);

		// reset all LAM channels
		lbuf = 0;
		len = WriteCAMAC(hHandle, NAF(28,0,16), &lbuf, 4);
		printf("Write LAM reset %d bytes. (0x%08x)\n", len, lbuf);

		// get the PCIADA status again
		getStatus(hHandle, &bTimeout, &bInterrupt);
		printf("bTimeout = %d, bInterrupt = %d\n", bTimeout, bInterrupt);


		// close device -------------------------------------
		if (hHandle != INVALID_HANDLE_VALUE)
			CloseHandle(hHandle);

		return 0;
	}
}