#ifndef __PCICC32_H__
#define __PCICC32_H__
//-------------------------------------------------------------------------
// WINNT driver for PCICC32 interface from ARW Elektronik, Germany --------
// IO definitions and common data structures between application and driver
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           16.04.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
// #include <devioctl.h> must be declared before inclusion when used for driver
// #include <winioctl.h> must be declared before inclusion when used for applications

//-------------------------------------------------------------------------
// DEFINES
//

//----------------------------------------------------------------------------------------
// macros for simple CAMAC NAF address calculation
//
#define NAF(n, a, f)  ((ULONG)((n << 10) + (a << 6) + ((f & 0xf) << 2)))

// to get a compatible view to WIN95 driver 
#define USER_CONTROL_CODE(x) (x)  

// VPCIC32D_ATTACH_CC32 and VPCIC32D_DETACH_CC32 are incompatible to WINNT - please use read and write commands
				
// get the interrupt and timeout status from a CC32 interface (0x00220008)
#define PCICC32_GET_STATUS       CTL_CODE(\
								 FILE_DEVICE_UNKNOWN,\
								 USER_CONTROL_CODE(2),\
								 METHOD_BUFFERED,\
								 FILE_ANY_ACCESS)

// clear the timeout status of a CC32 interface (0x0022000C)
#define PCICC32_CLEAR_STATUS     CTL_CODE(\
								 FILE_DEVICE_UNKNOWN,\
								 USER_CONTROL_CODE(3),\
								 METHOD_BUFFERED,\
								 FILE_ANY_ACCESS)

// set the access parameter for this file (vmemm) (0x00220010)
#define PCICC32_SET_ACCESS_PARA	 CTL_CODE(\
								 FILE_DEVICE_UNKNOWN,\
								 USER_CONTROL_CODE(4),\
								 METHOD_BUFFERED,\
								 FILE_ANY_ACCESS)

// set to check for control-code overflow
#define PCICC32_LAST_CTL_CODE PCICC32_SET_ACCESS_PARA

// switches for access mode
#define WORD_ACCESS (UCHAR)2   //            word
#define LONG_ACCESS (UCHAR)4   //            long

// this structure is output from VPCIC32_GET_STATUS call
typedef struct
{
	ULONG  dwInterface;         // CC32 module number (for compatibility to win95/98 only - not used)
	USHORT bTimeout;            // denotes a pending PCIADA timeout
	USHORT bInterrupt;          // denotes a pending LAM interrupt
} PCICC32_STATUS;    

// this structure sets the access parameter for following reads or writes to this path
typedef struct 
{
	ULONG  dwInterface;        // CC32 module number (for compatibility to win95/98 only - not used)   
	USHORT bAccessType;		   // set the current access type (WORD_ACCESS, LONG_ACCESS)
	USHORT bBlockTransfer;	   // set != 0 when block transfer (currently not available)
} PCICC32_ACCESS_COMMAND; 

#endif // __PCICC32_H__
