(****************************************************************************

            TPXMS.PAS  v2.00   written by Bastian Maerkisch

        original code v1.01 written by Vernon E. Davis 10/06/89


 Revision History:

   Date    Rev.  Description
 --------  ----  -----------------------------------------------------------
 05.01.95  2.00  initial version by Bastian Maerkisch
                 changed inline code to inline assembler
                 standardized naming conventions
                 fixed some error return misses
 27.05.02  2.01  add A20IsEnabled function
 ****************************************************************************)

UNIT _TPXMS;
INTERFACE
Type
  DoubleWord = LongInt;
  BCD        = Word;

  XMSMemMoveStruct =
  Record
     Length       : DoubleWord;
     SourceHandle : Word;
     SourceOffset : DoubleWord;
     DestHandle   : Word;
     DestOffset   : DoubleWord
  end;

  XMSHandleInfo =
  Record
     LockCount   : Byte;
     FreeHandles : Byte;
     BlockSize   : Word
  end;

  UMBHandle =
  Record
     Segment : Word;
     Size    : Word
  end;

Var
  (* control variables *)
  isXMS       : Boolean;
  XMSResult   : Word;
  XMSError    : Byte;

(* functions and procedures *)
  (* version control *)
  Function  XMSQueryVersion  : BCD;
  Function  XMSQueryRevision : BCD;

  (* high memory area routines *)
  Procedure HMAAllocMem( malloc : Word );
  Procedure HMAFreeMem;

  (* A20 control routines *)
  Procedure A20GlobalEnable;
  Procedure A20GlobalDisable;
  Procedure A20LocalEnable;
  Procedure A20LocalDisable;
  Function  A20QueryStatus : Boolean; (* deprecated *)
  Function  A20IsEnabled : Boolean;

  (* extended memory block routines *)
  Function  XMSQueryFreeMem   : Word;
  Function  XMSQueryFreeBlock : Word;
  Procedure XMSAllocMem( Var handle : Word; malloc : Word );
  Procedure XMSFreeMem( handle : Word );
  Procedure XMSMoveMem( Var MoveStructure : XMSMemMoveStruct );
  Function  XMSLockMem( handle : Word ) : DoubleWord;
  Procedure XMSUnlockMem( handle : Word );
  Procedure XMSQueryHandleInfo( handle : Word; Var HandleInfo : XMSHandleInfo );
  Procedure XMSReAllocMem( handle, malloc : Word );

  (* upper memory block routines *)
  Procedure UMBAllocMem( Var handle : UMBHandle; malloc : Word );
  Procedure UMBFreeMem( handle : UMBHandle );

  (* misc. support routines *)
  Function  XMSErrorMsg( Error : Word ) : String;
  Function  BCD2VerStr( b : BCD ) : String;

IMPLEMENTATION
Var
  XMM_Control : Pointer;

Procedure XMM_Fail; far; assembler;
asm
  xor ax, ax
  mov bl, 80h
end;

Procedure XMSQueryXMM_Control; assembler;
(* get XMM Control function address and store it in XMM_Control *)
asm
  mov ax, 4310h
  int 2fh
  mov word ptr [XMM_Control  ], bx
  mov word ptr [XMM_Control+2], es
  mov XMSResult, 1
  mov XMSError,  0
end;

Function XMSQueryExists : Boolean; assembler;
(* check if XMM is installed *)
asm
  mov ax, 4300h
  int 2fh
  mov bl, al
  mov al, 0
  cmp bl, 80h
  jne @@1
  call XMSQueryXMM_Control
  mov al, 1
 @@1:
  mov XMSResult, 1
  mov XMSError,  0
end;

Function XMSQueryVersion : BCD; assembler;
(* returns version level in BCD *)
asm
  mov di, Offset XMM_Control
  mov ax, 0000
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  0
  (* result in ax *)
end;

Function  XMSQueryRevision : BCD; assembler;
(* returns internal revision level in BCD *)
asm
  mov di, Offset XMM_Control
  mov ax, 0000
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  0
  mov ax, bx       (* result in bx *)
end;

Procedure HMAAllocMem( malloc : Word ); assembler;
(* allocate high memory area *)
asm
  mov di, Offset XMM_Control
  mov ax, 0100h
  mov dx, [malloc]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure HMAFreeMem; assembler;
(* free high memory area *)
asm
  mov di, Offset XMM_Control
  mov ax, 0200h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure A20GlobalEnable; assembler;
asm
  mov di, Offset XMM_Control
  mov ax, 0300h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure A20GlobalDisable; assembler;
asm
  mov di, Offset XMM_Control
  mov ax, 0400h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure A20LocalEnable; assembler;
asm
  mov di, Offset XMM_Control
  mov ax, 0500h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure A20LocalDisable; assembler;
asm
  mov di, Offset XMM_Control
  mov ax, 0600h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Function A20QueryStatus : Boolean; assembler; (* deprecated *)
(* returns  true if A20 is physically enabled *)
asm
  mov di, Offset XMM_Control
  mov ax, 0700h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  bl
  (* result in ax *)
end;

Function A20IsEnabled : Boolean; assembler;
(* returns  true if A20 is physically enabled *)
asm
  mov di, Offset XMM_Control
  mov ax, 0700h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  bl
  (* result in ax *)
end;

Function  XMSQueryFreeMem : Word;  assembler;
(* returns free extended memory in kB *)
asm
  mov di, Offset XMM_Control
  mov ax, 0800h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  0
  mov ax, dx  (* result in dx *)
end;

Function  XMSQueryFreeBlock : Word; assembler;
(* returns size of the largest available extended memory block in kB *)
asm
  mov di, Offset XMM_Control
  mov ax, 0800h
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, 1
  mov XMSError,  0
  (* result in ax *)
end;

Procedure XMSAllocMem( Var handle : Word; malloc : Word ); assembler;
(* allocate extended memory *)
asm
  mov di, Offset XMM_Control
  mov ax, 0900h
  mov dx, [malloc]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
  les di, handle
  mov es:[di], dx    (* result in dx *)
end;

Procedure XMSFreeMem( handle : Word ); assembler;
(* free extended memory *)
asm
  mov di, Offset XMM_Control
  mov ax, 0A00h
  mov dx, [handle]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure XMSMoveMem( Var MoveStructure : XMSMemMoveStruct ); assembler;
(* move routine for extended memory transfer *)
asm
  push ds
  push ds
  pop es
  mov di, Offset XMM_Control
  mov ax, 0B00h
  lds si, MoveStructure
  push bp
  call dword ptr es:[di]
  pop bp
  pop ds
  mov XMSResult, ax
  mov XMSError,  bl
end;

Function  XMSLockMem( handle : Word ) : DoubleWord; assembler;
(* lock extended memory block              *)
(* function returns the real block address *)
asm
  mov di, Offset XMM_Control
  mov ax, 0C00h
  mov dx, [handle]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl   (* only valid if ax=0000h *)
  mov ax, bx          (* result in dx:bx *)
end;

Procedure XMSUnlockMem( handle : Word ); assembler;
(* unlock extended memory block *)
asm
  mov di, Offset XMM_Control
  mov ax, 0D00h
  mov dx, [handle]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure XMSQueryHandleInfo( handle : Word; Var HandleInfo : XMSHandleInfo ); assembler;
(* get information about an extended memory block *)
asm
  mov di, Offset XMM_Control
  mov ax, 0E00h
  mov dx, [handle]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl   (* only valid if ax=0000h *)
  les di, HandleInfo
  mov es:[di  ], bh
  mov es:[di+1], bl
  mov es:[di+2], dx
end;

Procedure XMSReAllocMem( handle, malloc : Word ); assembler;
(* change size of an extended memory block *)
asm
  mov di, Offset XMM_Control
  mov ax, 0F00h
  mov dx, [handle]
  mov bx, [malloc]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;

Procedure UMBAllocMem( Var handle : UMBHandle; malloc : Word ); assembler;
(* allocate upper memory block *)
asm
  mov di, Offset XMM_Control
  mov ax, 1000h
  mov dx, [malloc]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl   (* only valid if ax=0000h *)
  les di, handle
  mov cx, [malloc]
  mov es:[di  ], bx
  mov es:[di+2], cx
  cmp ax, 0001
  je @@1
  mov es:[di], dx
 @@1:
end;

Procedure UMBFreeMem( handle : UMBHandle ); assembler;
(* free upper memory block *)
asm
  mov di, Offset XMM_Control
  mov ax, 1100h
  mov dx, [handle.segment]
  push bp
  call dword ptr [di]
  pop bp
  mov XMSResult, ax
  mov XMSError,  bl
end;


Function XMSErrorMsg( Error : Word ) : String;
Var
  XMSMsg : String;
begin
   Case Error of
     $00 : XMSMsg := '';
     $80 : XMSMsg := '80 : XMS Function not implemented';
     $81 : XMSMsg := '81 : VDISK detected';
     $82 : XMSMsg := '82 : A20 Error';
     $8E : XMSMsg := '8E : General Driver Error';
     $8F : XMSMsg := '8F : Unrecoverable Driver Error';
     $90 : XMSMsg := '90 : HMA does not exist';
     $91 : XMSMsg := '91 : HMA in use by another process';
     $92 : XMSMsg := '92 : Memory requested less than /HMAMIN= parameter';
     $93 : XMSMsg := '93 : HMA not allocated';
     $94 : XMSMsg := '94 : A20 is enabled';
     $A0 : XMSMsg := 'A0 : All of Extended Memory is allocated';
     $A1 : XMSMsg := 'A1 : No Extended Memory Handles available';
     $A2 : XMSMsg := 'A2 : Extended Memory Handle is invalid';
     $A3 : XMSMsg := 'A3 : Extended Move Structure: Source Handle is invalid';
     $A4 : XMSMsg := 'A4 : Extended Move Structure: Source Offset is invalid';
     $A5 : XMSMsg := 'A5 : Extended Move Structure: Destination Handle is invalid';
     $A6 : XMSMsg := 'A6 : Extended Move Structure: Destination Offset is invalid';
     $A7 : XMSMsg := 'A7 : Extended Move Structure: Length is invalid';
     $A8 : XMSMsg := 'A8 : Extended Move Structure: Move has invalid overlap';
     $A9 : XMSMsg := 'A9 : Parity Error';
     $AA : XMSMsg := 'AA : Block is not locked';
     $AB : XMSMsg := 'AB : Block is locked';
     $AC : XMSMsg := 'AC : Block Lock Count has overflowed';
     $AD : XMSMsg := 'AD : Block Lock has failed';
     $B0 : XMSMsg := 'B0 : A smaller Upper Memory Block is available';
     $B1 : XMSMsg := 'B1 : No Upper Memory Blocks are available';
     $B2 : XMSMsg := 'B2 : Upper Memory Block Segment Number is invalid';
   else
      XMSMsg := 'Unknown Error has occured';
   end;
   If XMSMsg <> '' then
      XMSMsg := 'XMS Error $' + XMSMsg;
   XMSErrorMsg := XMSMsg;
end;

Function  BCD2VerStr( b : BCD ) : String;
Var
  temp   : String[5];
  nibble : Byte;
begin
  temp := '';
  nibble := (b and $F000) shr 12;
  If nibble > 0 then temp := Char(nibble+Ord('0'));
  nibble := (b and $0F00) shr 08;
  temp := temp + Char(nibble+Ord('0')) + '.';
  nibble := (b and $00F0) shr 04;
  temp := temp + Char(nibble+Ord('0'));
  nibble := (b and $0F00);
  temp := temp + Char(nibble+Ord('0'));
  BCD2VerStr := temp;
end;

BEGIN
  XMM_Control := Ptr( 0, 0 );
  XMSResult   := 1;
  XMSError    := 0;
  isXMS       := XMSQueryExists;
  If not isXMS then XMM_Control := @XMM_Fail;
END.
