//-------------------------------------------------------------------------
// WIN95-driver for the  PCICC32 CAMAC-Interface of ARW Elektronik --------
// the PLX9050 Bug fix code
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// derived from pcivmeb.c                          AR           03.03.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

#include <debug.h>
#include <vxdwraps.h>

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)         // No stack checking for routines in this module

#include <vpicd.h>
#include "pcic32b.h"

//-------------------------------------------------------------------------
// DEFINES
//
#define CONFIG_ADDRESS  0x00000CF8 // io address of config address 
#define CONFIG_DATA     0x00000CFC // io address of config data
#define PCR_MEMORY_BUG  0x04       // 1st PCR address of potential bug
#define PCR_IO_BUG      0x05       // 2nd PCR address of potential bug
#define PCR_MEMORY_OK   0x08       // 1st PCR address of no bug 
#define PCR_IO_OK       0x09       // 2nd PCR address of no bug
#define PCR_USR_MEMBASE 0x07       // PCR address of USR MemBase (diff to VME)

#define MAX_BUS         32         // maximum number of busses to search
#define MAX_DEVICE      32         // maximum number of devices to search

//-------------------------------------------------------------------------
// TYPEDEFS
//

//-------------------------------------------------------------------------
// EXTERNALS
//
extern void  __outpd(WORD IOadr, DWORD val);
extern DWORD __inpd(WORD IOadr);

//-------------------------------------------------------------------------
// LOKALE FUNKTIONEN
//
static DWORD PCI_PCRRead(WORD typ, WORD reg, WORD func, WORD dev, WORD bus)
{
  DWORD adr;
  DWORD sft;
  
  sft   = bus;
  sft <<= 16;
  adr   = sft;
  sft   = dev;
  sft <<= 11;
  adr  |= sft;
  sft   = func;
  sft <<= 8;
  adr  |= sft;
  sft   = reg;
  sft <<= 2;
  adr  |= sft;
  adr  |= (typ & 3);
  adr  |= 0x80000000;

  __outpd(CONFIG_ADDRESS, adr);
  return __inpd(CONFIG_DATA);
}

static void PCI_PCRWrite(WORD typ, WORD reg, WORD func, WORD dev, WORD bus, DWORD data)
{
  DWORD adr;
  DWORD sft;
  
  sft   = bus;
  sft <<= 16;
  adr   = sft;
  sft   = dev;
  sft <<= 11;
  adr  |= sft;
  sft   = func;
  sft <<= 8;
  adr  |= sft;
  sft   = reg;
  sft <<= 2;
  adr  |= sft;
  adr  |= (typ & 3);
  adr  |= 0x80000000;

  __outpd(CONFIG_ADDRESS, adr);
  __outpd(CONFIG_DATA, data);  
}

//---------------------------------------------------------------
// searches for BusNo and DevNo for registered device
static WORD SearchBusDevice(DWORD dUSR_MemBase, WORD *pwBusNo, WORD *pwDevNo)
{
  WORD  d;
  WORD  b;
  DWORD compare;
  
  for (b = 0; b < MAX_BUS; b++)
  {
    for (d = 0; d < MAX_DEVICE; d++)
    {
      compare = PCI_PCRRead(0, PCR_USR_MEMBASE, 0, d, b) & (DWORD)~3;
      
      if (compare == dUSR_MemBase)
      {
        Debug_Printf_DebMax("SearchBusDevice: %d %d 0x%08x\r\n", b, d, compare);
      
        *pwBusNo = b;
        *pwDevNo = d;
        return 0;       // found
      }
    }
  }

  return 1;
}

//---------------------------------------------------------------
// function to call for bug fix
//
// returns the fixed address for LCR
//
void PLX9050BugFix(VPCIC32_STATIC_STATUS *psVss)
{
  DWORD dwData;
  WORD  wData;
  WORD  wBusNo;
  WORD  wDevNo;

  if ((psVss->dLCR_MemBase & 0x80) || (psVss->wLCR_IOBase & 0x80))
  { 
    if (SearchBusDevice(psVss->dUSR_MemBase, &wBusNo, &wDevNo))
      return;          // return immediate if nothing found??    

    if ((dwData = PCI_PCRRead(0, PCR_MEMORY_BUG, 0, wDevNo, wBusNo)) & 0x80)
    {      
      psVss->dLCR_MemBase = PCI_PCRRead(0, PCR_MEMORY_OK, 0, wDevNo, wBusNo);

      PCI_PCRWrite(0, PCR_MEMORY_BUG, 0, wDevNo, wBusNo, psVss->dLCR_MemBase);
      // PCI_PCRWrite(0, PCR_MEMORY_OK,  0, wDevNo, wBusNo, dwData);
    }
    else
      psVss->dLCR_MemBase = dwData;
      
    /* remove PCI codings because direkt PCI read */
    psVss->dLCR_MemBase &= 0xfffffff0;  
            
    if ((wData = PCI_PCRRead(0, PCR_IO_BUG, 0, wDevNo, wBusNo)) & 0x80)
    {
      psVss->wLCR_IOBase = PCI_PCRRead(0, PCR_IO_OK, 0, wDevNo, wBusNo);

      PCI_PCRWrite(0, PCR_IO_BUG, 0, wDevNo, wBusNo, psVss->wLCR_IOBase);
      // PCI_PCRWrite(0, PCR_IO_OK,  0, wDevNo, wBusNo, wData);
    }
    else
      psVss->wLCR_IOBase = wData;

    /* remove PCI codings because direkt PCI read */
    psVss->wLCR_IOBase &= 0xfffffffc;
  }
}


//---------------------------------------------------------------
// function to read parts of the pcr memory
//
BOOL dwReadPCR(DWORD dUSR_MemBase, WORD wOffset, DWORD *pdwValue)
{
  WORD wBusNo;
  WORD wDevNo;
  *pdwValue = 0;
  
  if (SearchBusDevice(dUSR_MemBase, &wBusNo, &wDevNo)) return FALSE;

  *pdwValue = PCI_PCRRead(0, wOffset, 0, wDevNo, wBusNo);

  return TRUE; 
}
