//-------------------------------------------------------------------------
// WIN95-Treiber for the  pcicc32 CAMAC Interface of ARW Elektronik -------
//
// all about configuration and registry
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
//
// What                                            who          when
// first steps, derived from pcivmer.c             AR           03.03.2000
// PLX8052 integration fixed                       AR           13.03.2000
// read interrupt directly from PCR                AR           20.02.2000
// change in device-ID                             AR           10.04.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <vmmreg.h>
#include <debug.h>
#include <vxdwraps.h>
#include <vpicd.h>
#include <configmg.h>

#include "pcic32r.h"    
#include "pcic32b.h"
#include "pcic32.h"    // driver internal structures and constants

//------------------------------------------------------------------------
// PRAGMAS
//
#undef  CURSEG
#define CURSEG() LCODE

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

//-------------------------------------------------------------------------
// DEFINES
//
#define HW_REVISION "001"                    // muss >= hardware rev. sein

#ifdef GET_OUT_OF_REGISTRY
#define HARDWAREID  "PCI\\VEN_10B5&DEV_2258&SUBSYS_22589050"
#else
#define HARDWAREID (DWORD)0x22589050
#endif

#define REGSTR_MAX_LENGTH 255  // maximum expected length

//-------------------------------------------------------------------------
// TYPEDEFS
//
typedef DWORD HKEY;

//-------------------------------------------------------------------------
// EXTERNALS
//
extern VMMLIST list;

//-------------------------------------------------------------------------
// LOKALE FUNKTIONEN
//
static char *kstrcpy(char *dst, char *src)
{
  while (*src != 0)
    *dst++ = *src++;
  *dst = 0;

  return dst; 
}

static BOOL kstrcmp(char *st0, char *st1)
{
  while (*st0 != 0)
  {
    if (*st0++ != *st1++)
      return FALSE;
  }

  return TRUE; 
}

//-------------------------------------------------------------------------
// request the subsystem id
//
#ifdef GET_OUT_OF_REGISTRY
static DWORD eval_SubSystem(DEVNODE devnode)
{
  DWORD cret;
  DWORD len = 255;
  char buff[255];

  // HardwareID holen ---------------------------------
  cret = CM_Read_Registry_Value(devnode, NULL, REGSTR_VAL_HARDWAREID,
                    REG_SZ, (void *)buff, &len, CM_REGISTRY_HARDWARE);

  if (cret != CR_SUCCESS) 
    return cret;

  if (kstrcmp(HARDWAREID, buff) == TRUE)
    return CR_SUCCESS;
  else  
    return CR_NO_SUCH_VALUE;
}
#endif

//-------------------------------------------------------------------------
// FUNKTIONEN
//
//------------------------------------------------------------------
// gets info about the devices out of registry
//
static DWORD GetCMInfo(char *name, VPCIC32_STATIC_STATUS *psVss)
{
  DEVNODE   dn0;
  CMCONFIG  ccBuffer;
  CONFIGRET cret;
  BYTE      buff[8];
  DWORD     len = 8;
  
  if((cret = CM_Locate_DevNode(&dn0, name, 0)) == CR_SUCCESS)
  {
    Out_Debug_String("CM_Locate_DevNode.\r\n");

#ifdef GET_OUT_OF_REGISTRY
    if ((cret = eval_SubSystem(dn0)) != CR_SUCCESS)
      return cret;
#endif

    if ((cret = CM_Get_Alloc_Log_Conf(&ccBuffer, dn0,
                                         CM_GET_ALLOC_LOG_CONF_ALLOC))
                                                           == CR_SUCCESS)
    {
      Out_Debug_String("CM_Get_Alloc_Log_Conf\r\n");

      psVss->wNumMemWindows = ccBuffer.wNumMemWindows;
      psVss->wNumIOPorts    = ccBuffer.wNumIOPorts;
      psVss->wNumIRQs       = ccBuffer.wNumIRQs;
      psVss->wNumDMAs       = ccBuffer.wNumDMAs;
      
#ifdef GET_OUT_OF_REGISTRY
      // revision lesen -----------------------------------
      cret = CM_Read_Registry_Value(dn0, NULL, REGSTR_VAL_HWREV,
                                          REG_SZ, (void *)buff, &len,
                                             CM_REGISTRY_HARDWARE);
#endif

      if (cret == CR_SUCCESS)                                     
      {
        DWORD dwValue;
        
        Out_Debug_String("CM_Read_Registry_Value.\r\n");

#ifdef GET_OUT_OF_REGISTRY
        buff[7] = 0;        
        kstrcpy(psVss->cszHWRevision, buff);
#else
        kstrcpy(psVss->cszHWRevision, HW_REVISION);
#endif

        // info extrahieren -------------------------------
        psVss->dLCR_MemBase   = ccBuffer.dMemBase[0];
        psVss->dLCR_MemLength = ccBuffer.dMemLength[0];
                                        
        psVss->wLCR_IOBase    = ccBuffer.wIOPortBase[0];
        psVss->wLCR_IOLength  = ccBuffer.wIOPortLength[0];     

        // psVss->wIRQ           = ccBuffer.bIRQRegisters[0];

        psVss->dUSR_MemBase   = ccBuffer.dMemBase[1];
        psVss->dUSR_MemLength = ccBuffer.dMemLength[1];

		Debug_Printf_DebMax("LCR 0x%08x (%d), USR 0x%08x (%d)\r\n", 
			psVss->dLCR_MemBase, psVss->dLCR_MemLength, 
			   psVss->dUSR_MemBase, psVss->dUSR_MemLength);

		// interrupt ist special
		{
			DWORD dwPCRreg3c;

			dwReadPCR(psVss->dUSR_MemBase, 0x3C >> 2, &dwPCRreg3c);
			psVss->wIRQ = dwPCRreg3c & 0x000000ff;
		}

#ifndef GET_OUT_OF_REGISTRY
        dwReadPCR(psVss->dUSR_MemBase, 0x2C >> 2, &dwValue);
        if (dwValue != HARDWAREID) return CR_NO_SUCH_VALUE;
#endif

		if (psVss->dUSR_MemLength != VPCIC32_USR_WINDOW)
			return CR_NO_SUCH_VALUE;

        // potential BUG fix ------------------------------
        PLX9050BugFix(psVss);

        Out_Debug_String("GetCMInfo().OK\r\n");

        return CR_SUCCESS;
      }
      else
        return cret;
    }
    else
      return cret;
  }
  else
    return cret;
}

//------------------------------------------------------------------
// prepare a VPCIC32_STATIC_STATUS structure for later use
//
static void softInit(VPCIC32_STATIC_STATUS *psVss)
{
  psVss->dwModuleNumber = -1;        // the number of the connected CC32
  psVss->dwLinkCount    = 0;         // how often this interface is requested  

  psVss->wNumMemWindows = 0;         // aus der aktuellen konfiguration
  psVss->wNumIOPorts    = 0;
  psVss->wNumIRQs       = 0;
  psVss->wNumDMAs       = 0;
      
  psVss->dLCR_MemBase   = (DWORD)BOGUSADDRESS;         // aus der aktuellen konfiguration
  psVss->dLCR_MemLength = 0;
  
  psVss->wLCR_IOBase    = 0;
  psVss->wLCR_IOLength  = 0;

  psVss->dUSR_MemBase   = (DWORD)BOGUSADDRESS;
  psVss->dUSR_MemLength = 0;

  psVss->dwShared_base  = (DWORD)BOGUSADDRESS;         // user access address

  psVss->pvLcr          = (PVOID)BOGUSADDRESS;         // physikalische adresse des lcr
  psVss->pvIfr          = (PVOID)BOGUSADDRESS;         // physikalische adresse des CC32

  psVss->pwCNTRL        = (WORD *)BOGUSADDRESS;        // pointer to CNTRL (0x50) in LCR
  psVss->pwINTCSR       = (WORD *)BOGUSADDRESS;        // pointer to INTCSR (0x4c) in LCR

  psVss->dwIrqHandle    =
  psVss->dwIrqHandler   = -1;
  psVss->TCB_S          = (PTCB)BOGUSADDRESS;

  psVss->wIRQ           = -1;

  psVss->nFPGAVersion   = -1;        // the FPGA version of the connected CC32

  psVss->dwIfcStatus    = E_EMPTY;

  psVss->cszHWRevision[0] = 0;
}

//------------------------------------------------------------------
// looks the registry for interfaces
//
DWORD ScanRegistry(char *KeyName, VMMLIST list, int *piNumIfc)
{
  DWORD     ret;
  HKEY      hKey;
  DWORD     i;
  char      buffer[REGSTR_MAX_LENGTH];
  CONFIGRET cret;

  buffer[0] = 0;
  ret = _GetRegistryKey(REGTYPE_ENUM, KeyName, REGKEY_OPEN, &hKey);

  Debug_Printf_DebMax("_GetRegistryKey(%s)\r\n", KeyName);

  if(ret == ERROR_SUCCESS)
  {
    i = 0;  
    do
    {     
      ret = _RegEnumKey(hKey, i++,
              kstrcpy(kstrcpy(buffer, KeyName),"\\"),
                                  (DWORD)REGSTR_MAX_LENGTH);

      Debug_Printf_DebMax("_RegEnumKey(%s)\r\n",buffer);

      if (ret == ERROR_SUCCESS)
      {
		  VPCIC32_STATIC_STATUS *psVss = (VPCIC32_STATIC_STATUS *)List_Allocate(list);

		  softInit(psVss);  // prepare the structure

		  cret = GetCMInfo(buffer, psVss);  
		  if (cret == CR_SUCCESS)
		  {
				// resource usage OK? ----------------     
				if ((psVss->wNumIRQs       != 1) || 
				  	(psVss->wNumDMAs       != 0) ||
					(psVss->wNumMemWindows != 3) ||
					(psVss->wNumIOPorts    != 2))
					List_Deallocate(list, (VMMLISTNODE)psVss);
				else
				{
					List_Attach(list, (VMMLISTNODE)psVss);
					(*piNumIfc)++;
				}
		  }
		  else
			  List_Deallocate(list, (VMMLISTNODE)psVss);
      }
    } while (ret == ERROR_SUCCESS);

    _RegCloseKey(hKey);

    return ERROR_SUCCESS;
  }

  return ret;
}

//-----------------------------------------------------------------
// allocates linear addresses
//
DWORD MakeLinearAddresses(VMMLIST list)
{
  VPCIC32_STATIC_STATUS *psVss; 

  psVss = (VPCIC32_STATIC_STATUS *)List_Get_First(list);
  
  while (psVss)
  {
    psVss->dwIfcStatus = E_NO_ERROR;

    // 1. map LCR -----------------------------
    if (psVss->dwIfcStatus == E_NO_ERROR)
    {
      psVss->pvLcr =
        _MapPhysToLinear(psVss->dLCR_MemBase, psVss->dLCR_MemLength, 0);

      if (psVss->pvLcr == (PVOID)BOGUSADDRESS)
        psVss->dwIfcStatus = E_NO_ADDRESS;      
    }

    // 2. map IFR -----------------------------  
    if (psVss->dwIfcStatus == E_NO_ERROR)
    {
      psVss->pvIfr =
        _MapPhysToLinear(psVss->dUSR_MemBase, psVss->dUSR_MemLength, 0);
        
      if (psVss->pvIfr == (PVOID)BOGUSADDRESS)
        psVss->dwIfcStatus = E_NO_ADDRESS;
    }

    // 3. initialize some fast access addresses --
    if (psVss->pvLcr != (PVOID)BOGUSADDRESS)
    {
		psVss->pwCNTRL	  = (WORD *)((BYTE *)psVss->pvLcr + 0x50);  // zeiger auf csr register
		psVss->pwINTCSR   = (WORD *)((BYTE *)psVss->pvLcr + 0x4c);  // zeiger auf irq status
    }

    // 4. set status to not connected ------------
    if (psVss->dwIfcStatus == E_NO_ERROR)
      psVss->dwIfcStatus = E_NOT_CONNECTED;

    // the next one
	psVss = (VPCIC32_STATIC_STATUS *)List_Get_Next(list, (VMMLISTNODE)psVss);
  }
  
  return ERROR_SUCCESS;
}

//---------------------------------------------------------------------
// checks a connection with a small test pattern
//
static DWORD TestConnection(VPCIC32_STATIC_STATUS *psVss)
{  
  WORD wRet;
  WORD retCode;
  WORD wStore0;
  WORD wStore1;
  int i;

  WORD *wTest0 = WORD_NAF(psVss->pvIfr, 30, 0, 0);
  WORD *wTest1 = WORD_NAF(psVss->pvIfr, 30, 1, 0);

  Out_Debug_String("TestConnection().Start\r\n");

  wStore0 = *WORD_NAF(psVss->pvIfr, 30, 0, 0);
  wStore1 = *WORD_NAF(psVss->pvIfr, 30, 1, 0);

  for (i = 0; i < 10000; i++)
  {
    *wTest0 = 0x5555;
    *wTest1 = 0xAAAA;
    wRet    = *wTest0;
    if (wRet != 0x5555) return E_CON_ERROR;
    
    *wTest0 = 0xAAAA;
    *wTest1 = 0x5555;
    wRet    = *wTest0;
    if (wRet != 0xAAAA) return E_CON_ERROR;

    *wTest0 = 0x0000;
    *wTest1 = 0xFFFF;
    wRet    = *wTest0;
    if (wRet != 0x0000) return E_CON_ERROR;

    *wTest0 = 0xFFFF;
    *wTest1 = 0x0000;
    wRet    = *wTest0;
    if (wRet != 0xFFFF) return E_CON_ERROR;
  }

  *WORD_NAF(psVss->pvIfr, 30, 0, 0) = wStore0;
  *WORD_NAF(psVss->pvIfr, 30, 1, 0) = wStore1;

  Out_Debug_String("TestConnection().OK\r\n");

  return E_NO_ERROR;
}

//--------------------------------------------------------------
// prueft die Verbindung zu CC32- ------------------------------
//
static DWORD checkCC32(VPCIC32_STATIC_STATUS *psVss)
{
  WORD  wCntrl;
  DWORD error;
  WORD  wStatus;

  if (psVss->pwCNTRL == (WORD *)BOGUSADDRESS)
    return E_INCOMPATIBLE;

  wCntrl = *psVss->pwCNTRL; // store

  Debug_Printf_DebMax("CheckCC32 %04x\r\n", wCntrl);

  *psVss->pwCNTRL  = RELEASE_CC32;

  if (wCntrl & 0x0800)
  {
	  wStatus = *WORD_NAF(psVss->pvIfr, 0, 0, 0);

	  Debug_Printf_DebMax("CC32-Status %04x\r\n", wStatus);

	  if ((wStatus & 0xF000) != 0x8000)
		error = E_INCOMPATIBLE;
	  else
	  {
	    error = TestConnection(psVss);
	
	    if (error == E_NO_ERROR)
		{
			psVss->dwModuleNumber = (wStatus >> 4) & 0xF;
			psVss->nFPGAVersion   = (wStatus >> 8) & 0xF;
		}
	  }
  }

  *psVss->pwCNTRL = wCntrl;  // restore

  return error;
}  

//-----------------------------------------------------------------
// checks the CC32 interface and gets some information about
//
VPCIC32_STATIC_STATUS *SearchCC32(VMMLIST list, int nModuleNumber)
{
  VPCIC32_STATIC_STATUS *psVss;

  Out_Debug_String("SearchCC32()\r\n");

  psVss = (VPCIC32_STATIC_STATUS *)List_Get_First(list);
  
  while (psVss)
  {
    if ((E_NO_ERROR == checkCC32(psVss)) && (psVss->dwModuleNumber == nModuleNumber))
		return psVss;

	psVss = (VPCIC32_STATIC_STATUS *)List_Get_Next(list, (VMMLISTNODE)psVss);
  }

  return (VPCIC32_STATIC_STATUS *)BOGUSADDRESS;
}

//---------------------------- ENDE --------------------------------------
