#ifndef __PCIC32_H__
#define __PCIC32_H__
//-------------------------------------------------------------------------
// WIN95 driver for PCICC32 CAMAC interface from ARW Elektronik, Germany --
//
// pcic32.h - internal structures and constants of the driver
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps                                     AR           07.03.2000
// enhanced with interrupt and AUTOREAD functions  AR           17.03.2001
//

//----------------------------------------------------------------------------------------
// the maximum of potential CC32 interfaces
//
#define VPCIC32D_MAX_INTERFACES    16
#define VPCIC32_USR_WINDOW         0x8000       // size of cc32 window
#define VPCIC32_USR_PAGES          (VPCIC32_USR_WINDOW >> 12) // number of pages

//----------------------------------------------------------------------------------------
// how to get or release access via lcr 
//
#define RELEASE_CC32               0x4186       // grant access to usr window
#define INHIBIT_CC32               0x4086       // disable access

//----------------------------------------------------------------------------------------
// ERROR RETURNS in dIfcStatus
//
#define E_NO_ERROR         0   // all OK
#define E_INCOMPATIBLE     1   // incompatible hardware
#define E_NO_ADDRESS       2   // cant get lcr or ifr addresses
#define E_NOT_CONNECTED    3   // no VMEMM hardware connected
#define E_CON_ERROR        4   // data transfer failure
#define E_EMPTY            -1  // no PCI interface associated

//----------------------------------------------------------------------------------------
// macros for simple address calculation
//
#define WORD_NAF(base, n, a, f)  ((WORD *) ((DWORD)base + (n << 10) + (a << 6) + ((f & 0xf) << 2)))
#define DWORD_NAF(base, n, a, f) ((DWORD *)((DWORD)base + (n << 10) + (a << 6) + ((f & 0xf) << 2)))

//------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct                // achtung: sehr empfindlich auf alignment
{
  DWORD dwModuleNumber;       // the number of the connected CC32
  DWORD dwLinkCount;          // how often this interface is requested  

  WORD  wNumMemWindows;       // aus der aktuellen konfiguration
  WORD  wNumIOPorts;
  WORD  wNumIRQs;
  WORD  wNumDMAs;
      
  DWORD dLCR_MemBase;         // aus der aktuellen konfiguration
  DWORD dLCR_MemLength;
  
  WORD  wLCR_IOBase;
  DWORD wLCR_IOLength;

  DWORD dUSR_MemBase;
  DWORD dUSR_MemLength;

  DWORD dwShared_base;        // user access address from ring 3

  PVOID pvLcr;                // physikalische adresse des lcr
  PVOID pvIfr;                // physikalische adresse des CC32

  WORD  *pwCNTRL;             // pointer to CNTRL (0x50) in LCR
  WORD  *pwINTCSR;            // pointer to INTCSR (0x4c) in LCR

  WORD  wIRQ;                 // the number of assigned irq
  WORD  nFPGAVersion;         // the FPGA version of the connected CC32

  DWORD dwIrqHandle;          // handle returned by VPICD_Virtualize_IRQ()
  VID   vid;                  // virtual interrupt descriptor
  PTCB  TCB_S;                // ptr to thread control block
  DWORD dwIrqHandler;         // user irq handler routine

  DWORD dwIfcStatus;          // what's the status of connection

  char  cszHWRevision[10];
} VPCIC32_STATIC_STATUS;


#ifndef BOGUSADDRESS
#define BOGUSADDRESS 0xFFFFFFFF
#endif

#define AR "To my daughters: Anna and Ruth, from Klaus Hitschler"

#endif // __PCIC32_H__