//-------------------------------------------------------------------------
// WIN95-driver for the  PCICC32 CAMAC-Interface of ARW Elektronik --------
// interrupt related stuff
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           17.03.2001
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>
#include <vwin32.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

#include <debug.h>
#include <vxdwraps.h>

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)         // No stack checking for routines in this module

#include <vpicd.h>
#include <winerror.h>

#include "vpcic32d.h" // shared info between driver and application
#include "pcic32i.h"  // interrupt related stuff
#include "pcic32.h"   // driver internal structures and constants

//-------------------------------------------------------------------------
// DEFINES
//
#define WITH_IRQ      // during development sometimes switched off

//-------------------------------------------------------------------------
// TYPEDEFS
//

//-------------------------------------------------------------------------
// EXTERNALS
//
extern VPCIC32_STATIC_STATUS *cc32[VPCIC32D_MAX_INTERFACES];

//-------------------------------------------------------------------------
// LOKALE FUNKTIONEN
//

//-------------------------------------------------------------------------
// Set initial values for interrupt register
//
void initInterrupts(VPCIC32_STATIC_STATUS *psVss)
{
	if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
		*psVss->pwINTCSR = 0x0009;
}

//-------------------------------------------
// disable all irqs on PCIADA
//
void DisableIRQ(WORD *pwINTCSR)
{
  volatile WORD *wAdr = pwINTCSR;
  volatile dummy;

  Out_Debug_String("DisableIRQ()\r\n");

  *wAdr       &= ~0x40;  
  dummy        = *wAdr;
}

//-------------------------------------------
// enable all irqs on PCIADA
//
void EnableIRQ(WORD *pwINTCSR)
{
  volatile WORD *wAdr = pwINTCSR;
  volatile dummy;

  Out_Debug_String("EnableIRQ()\r\n");

#ifdef WITH_IRQ
  *wAdr       |= 0x40;
  dummy        = *wAdr;
#else
  DisableIRQ(pvLCR);
#endif
}

//-------------------------------------------
// read the vector and level of IRQ1
//
DWORD dwReadIrq1Src(VPCIC32_STATIC_STATUS *psVss)
{
  register DWORD dwResult;

  Out_Debug_String("dwReadIrq1Src()\r\n");

  // read and clear pending interrupt
  dwResult = *DWORD_NAF(psVss->pvIfr, 28, 2, 0);
  *WORD_NAF(psVss->pvIfr, 28, 0, 16) = 0; // clear LAM-FF
  
  return dwResult;
}  

//------------------------------------------------------------------------
// set PCIADA specials like AUTOREAD (caution: No sharing allowed)
//
DWORD vpcic32d_control_interrupts(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  VPCIC32D_IRQ_CONTROL  *psVicIn  = (VPCIC32D_IRQ_CONTROL *)ptr->lpvInBuffer;
  DWORD dwModuleNumber = psVicIn->dwInterface;
  VPCIC32D_IRQ_CONTROL  *psVicOut = (VPCIC32D_IRQ_CONTROL *)ptr->lpvOutBuffer;

  Out_Debug_String("vpcic32d_control_interrupts()\r\n");

  *(DWORD *)ptr->lpcbBytesReturned = 0;

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES) 
  {
	  psVss = cc32[dwModuleNumber];

	  if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
	  {
		  if (psVicIn->wEnable)
			  EnableIRQ(psVss->pwINTCSR);
		  else
			  DisableIRQ(psVss->pwINTCSR);

		  // return info - if a output buffer exists
		  if (ptr->cbOutBuffer >= sizeof(VPCIC32D_IRQ_CONTROL))
		  {
			  psVicOut->dwInterface = dwModuleNumber;
			  psVicOut->wEnable    = (*psVss->pwINTCSR & 0x0040) ? 1 : 0;

			  *(DWORD *)ptr->lpcbBytesReturned = sizeof(VPCIC32D_IRQ_CONTROL);
		  }

		  return ERROR_SUCCESS;
	  }
  }

  return ERROR_NO_SYSTEM_RESOURCES;
}

//------------------------------------------------------------------------
// installs or deinstalls a user irq handler routine
//
DWORD vpcic32d_install_irq_handler(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  VPCIC32D_IRQ_HANLDER  *psVihIn  = (VPCIC32D_IRQ_HANLDER *)ptr->lpvInBuffer;
  DWORD dwModuleNumber = psVihIn->dwInterface;
  VPCIC32D_IRQ_HANLDER  *psVihOut = (VPCIC32D_IRQ_HANLDER *)ptr->lpvOutBuffer;

  Out_Debug_String("vpcic32d_install_irq_handler()\r\n");

  *(DWORD *)ptr->lpcbBytesReturned    = 0;

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES) 
  {
	  DWORD dwOldHandler;

	  psVss = cc32[dwModuleNumber];

	  if (!(psVss->wNumIRQs))  // no interrupt line available
		  return ERROR_IRQ_BUSY;
    
	  psVss->TCB_S = Get_Cur_Thread_Handle();
	  dwOldHandler = psVss->dwIrqHandler;

	  psVss->dwIrqHandler = psVihIn->dwIrqHandler; 

	  // return info - if a output buffer exists
	  if (ptr->cbOutBuffer >= sizeof(VPCIC32D_IRQ_HANLDER))
	  {
		  psVihOut->dwInterface  = dwModuleNumber;
		  psVihOut->dwIrqHandler = dwOldHandler;

		  *(DWORD *)ptr->lpcbBytesReturned = sizeof(VPCIC32D_IRQ_HANLDER);
	  }

	  return ERROR_SUCCESS;
  }

  return ERROR_NO_SYSTEM_RESOURCES;
}

//-------------------------------------------
// clear the 2nd irq on PCIADA
//
void ClearIRQ2(WORD *pwCNTRL)
{
  volatile WORD *wAdr = pwCNTRL;
  volatile dummy;

  Out_Debug_String("ClearIRQ2()\r\n");

  *wAdr      &= ~0x100;
  *wAdr      |=  0x100;
  dummy       = *wAdr;  // read back to force processor to write immediately
}

//-------------------------------------------
// interface to APC function ...
//
BOOL VXDINLINE __declspec(naked)
  _VWIN32_QueueUserApc(DWORD userproc, DWORD refdata, PTCB thread)
{
  VxDJmp(_VWIN32_QueueUserApc)
}

//-------------------------------------------
// the asynchrones procedure call
//
static void vAPC(DWORD dwModuleNumber, DWORD dwMsg)
{
#ifdef WITH_IRQ
	VPCIC32_STATIC_STATUS *psVss = cc32[dwModuleNumber];

	Debug_Printf_DebMax("vAPC(0x%08x, 0x%08x)\r\n", psVss->dwIrqHandler, dwMsg); 

	if (psVss->dwIrqHandler != BOGUSADDRESS)
		_VWIN32_QueueUserApc(psVss->dwIrqHandler, dwMsg, psVss->TCB_S);
#endif
}

//-------------------------------------------
// process the isr
//
static BOOL Process_ISR(DWORD hirq, DWORD dwRefData)
{
	register VPCIC32_STATIC_STATUS *psVss = (VPCIC32_STATIC_STATUS *)dwRefData;
	register WORD test = *psVss->pwINTCSR;

	Debug_Printf_DebMax("Process_ISR(0x%04x)\r\n", test); 

	if (*psVss->pwCNTRL & 0x0100)      // PCIADA switched on?
	{
		if ((test & 0x0068) == 0x0068) // PCI Irq is higher priority
		{
			ClearIRQ2(psVss->pwCNTRL);
#ifdef WITH_IRQ
			vAPC(psVss->dwModuleNumber, CONNECTION_TIMEOUT);  // bTimeOut
			VPICD_Phys_EOI(hirq);
#endif
			return TRUE;
		}

		if ((test & 0x0045) == 0x0045)      // CC32 Irq is lower priority
		{
#ifdef WITH_IRQ
			vAPC(psVss->dwModuleNumber, dwReadIrq1Src(psVss));
			VPICD_Phys_EOI(hirq);
#else
			DisableIRQ(psVss->pwINTCSR);   // normally managed via LAM-Flip-Flop		
#endif
			return TRUE;
		}
	}

	// VPICD_Phys_EOI(hirq);         // (don't) know what to do when no source ?? 
	return FALSE;
}


//-------------------------------------------
// interface to Process_isr
//
static void _declspec(naked)ISR_PCICC32(void)
{
  _asm
  {
    push  edx  ; referenzdaten
    push  eax  ; irq number
    call  Process_ISR
    add   esp,8
    cmp   eax,1
    ret
  }
}

//-------------------------------------------
// disable the interrupt in WIN95
//
BOOL Disable_Phys_IRQ(VPCIC32_STATIC_STATUS *psVss)
{
	if (psVss->pvLcr != (void *)BOGUSADDRESS)
		DisableIRQ(psVss->pwINTCSR);

	#ifdef WITH_IRQ
	if (psVss->dwIrqHandle != BOGUSADDRESS)
	{
		Out_Debug_String("Physically_Mask()\r\n");
		VPICD_Physically_Mask(psVss->dwIrqHandle);      
		VPICD_Force_Default_Behavior(psVss->dwIrqHandle);
		psVss->dwIrqHandle = BOGUSADDRESS;
	}
	#endif

	psVss->dwIrqHandler = BOGUSADDRESS;

	return TRUE;
}

//-------------------------------------------
// enable the interrupts in WIN95
//
BOOL Enable_Phys_IRQ(VPCIC32_STATIC_STATUS *psVss)
{
  if (psVss->pvLcr != (void *)BOGUSADDRESS)
  {
#ifdef WITH_IRQ
    psVss->vid.VID_IRQ_Number       = psVss->wIRQ;  
    psVss->vid.VID_Options          = VPICD_OPT_REF_DATA | VPICD_OPT_CAN_SHARE | VPICD_OPT_SHARE_PMODE_ONLY;	
    psVss->vid.VID_Hw_Int_Proc      = (DWORD)ISR_PCICC32;
    psVss->vid.VID_Virt_Int_Proc    = 0;	
    psVss->vid.VID_EOI_Proc         = 0;	
    psVss->vid.VID_Mask_Change_Proc = 0;	
    psVss->vid.VID_IRET_Proc        = 0;
    psVss->vid.VID_IRET_Time_Out    = 500;
    psVss->vid.VID_Hw_Int_Ref       = (DWORD)psVss;	

    psVss->dwIrqHandle  = VPICD_Virtualize_IRQ(&psVss->vid);
    psVss->dwIrqHandler = BOGUSADDRESS;

    if (psVss->dwIrqHandle != BOGUSADDRESS)
    {
       Out_Debug_String("Physically_Unmask()\r\n");
       VPICD_Physically_Unmask(psVss->dwIrqHandle);
       
       return TRUE;
    }
#else
	return TRUE;
#endif
    
    return FALSE;
  }
  else
    return FALSE;
}


//----------------------------------------------------------------
