//-------------------------------------------------------------------------
// pcic32_test - a simple test program to test the driver for the 
//               PCI to CAMAC interface PCICC32 from ARW Elektronik, Germany
//
// (c) 2000 ARW Elektronik - written by Klaus Hitschler (AR)
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps                                     AR           10.03.2000
//

//--------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <stdio.h>
#include "..\vpcic32d\vpcic32d.h"

//--------------------------------------------------------------------------------------
// DEFINES
//
#define DRIVER_PATH "\\\\.\\c:\\windows\\system\\vpcic32d.vxd"
#define MY_INTERFACE 1 

#define wordNAF(base, N, A, F) ((WORD *) ((DWORD)base + (N << 10) + (A << 6) + ((F & 0xF) << 2)))
#define longNAF(base, N, A, F) ((DWORD *)((DWORD)base + (N << 10) + (A << 6) + ((F & 0xF) << 2)))

//--------------------------------------------------------------------------------------
// GLOBALS
//
char *cszProgName;
int verbose = 0;

//--------------------------------------------------------------------------------------
// FUNCTIONS
//

//--------------------------------------------------------------------------------------
// the general help - invoke with -?
//
void hlpMsg(void)
{
	printf("pcic32_test - a simple test program for CC32-modules (WIN95/98 only).\n");
	printf("(c) ARW Elektronik (AR) - %s\n", __DATE__);
	printf("\n");
	printf("usage: pcic32_test [-?] [-p=path] [-m=ModuleNumber] [-v]\n");
	printf("       -p= give the driver path (default: %s).\n", DRIVER_PATH);
	printf("       -m= give the number of the CC32 module (default: %d).\n", MY_INTERFACE);
	printf("       -v  make the program verbose.\n");
	printf("       -?  this help.\n");
}

//--------------------------------------------------------------------------------------
// a function to get the base address of the CAMAC range
//
static int attachWindow(HANDLE nHandle, DWORD dwInterface, void **base)
{
	VPCIC32D_WINDOW window;
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

    *base = NULL;
	if (nHandle == (HANDLE)-1) return -1;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_ATTACH_CC32,
						&dwInterface, sizeof(dwInterface), 
						&window,      sizeof(window),
								 &DIOC_count, NULL);

    if (!result) 
		return GetLastError();
	else
	{
		*base = window.pvWindowBase;
		return 0;
	}
}

//--------------------------------------------------------------------------------------
// a function to release a attached range
//
static int detachWindow(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

	if (nHandle == (HANDLE)-1) return -1;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_DETACH_CC32,
						&dwInterface, sizeof(dwInterface), 
						NULL,      0, &DIOC_count, NULL);
    if (!result) 
		return GetLastError();
	else
		return 0;
}

//--------------------------------------------------------------------------------------
// get the PCIADA status from the PCI side of the interface
//
static int getStatus(HANDLE nHandle, DWORD dwInterface, int *nTimeout, int *nLAM)
{
	VPCIC32D_STATUS status;
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

	if (nHandle == (HANDLE)-1) return -1;

	*nTimeout = *nLAM = 0;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_GET_STATUS,
						&dwInterface, sizeof(dwInterface), 
						&status,      sizeof(status), &DIOC_count, NULL);

    if (!result) 
		return GetLastError();
	else
	{
		*nTimeout = status.bTimeout;
		*nLAM     = status.bInterrupt;

		return 0;
	}
}

//--------------------------------------------------------------------------------------
// clear the PCIADA status
//
static int clearStatus(HANDLE nHandle, DWORD dwInterface)
{
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

	if (nHandle == (HANDLE)-1) return -1;

	// attach a window into the CC32 space ----------------
	result = DeviceIoControl(nHandle, VPCIC32_CLEAR_STATUS,
						&dwInterface, sizeof(dwInterface), 
						NULL,      0, &DIOC_count, NULL);

    if (!result) 
		return GetLastError();
	else
		return 0;
}


//--------------------------------------------------------------------------------------
// the frame for all the rest
//
int main(int argc, char *argv[])
{
	HANDLE		nHandle;
	void        *base;
	int         error;
	char        c, *ptr;
    int         i;
	int nModuleNumber = MY_INTERFACE;
	char        *driver_path = DRIVER_PATH;
	DWORD *adresse;
	DWORD n, a, f;
	int nTimeout, nLAM;

	cszProgName = argv[0];
	
	for (i = 1; i < argc; i++)
	{
		ptr = argv[i];

		if ((*ptr == '-') || (*ptr == '/')) ptr++;
		c = *ptr;
		ptr++;
		if (*ptr == '=') ptr++;

		switch(tolower(c))
		{
			case '?': hlpMsg();
					  exit(0);
			case 'm': sscanf(ptr, "%d", &nModuleNumber);
				      break;
			case 'p': driver_path = ptr;
				      break;
			case 'v': verbose = 1;
				      break;
			default:  printf("pcic32_test: Illegal argument '%c'! (Try -?)\n", c);
				      exit(-1);
		}
	}

	printf("pcic32_test: Try to test CC32-module #%d on path %s.\n", nModuleNumber, driver_path);

	nHandle = CreateFile(driver_path, 
						0, 
						0, 
						NULL, 
						0, 
						FILE_FLAG_DELETE_ON_CLOSE, 
						NULL);
	
	if (nHandle != INVALID_HANDLE_VALUE) 
	{ 
        error = attachWindow(nHandle, nModuleNumber, &base);
		if (error)
			printf("pcic32_test: error @ attachWindow() = %d\n", error);
        else
		{
			printf("pcic32_test: window base address into CC32 = 0x%08x\n", base);

			for (i = 0; i < 0xFFFFFF; i++)
			{
				n = (i % 24) + 1;
				a = (i % 16);
				f = (i % 16);

				adresse = longNAF(base, n, a, f);
				if (verbose) printf("n: %2d, a: %2d, f: %2d, adresse: 0x%08x\r", n, a, f, adresse);
				*adresse = i;
			}

			error = getStatus(nHandle, nModuleNumber, &nTimeout, &nLAM);
			if (error)
				printf("pcic32_test: error @ getStatus() = %d\n", error);
			else
			{
				printf("pcic32_test: nTimeout: %d, nLAM: %d\n", nTimeout, nLAM);

				if (nTimeout)
				{
					error = clearStatus(nHandle, nModuleNumber);
					if (error)
						printf("pcic32_test: error @ clearStatus() = %d\n", error);
					else
					{
						printf("pcic32_test: status cleared, get status again.\n");

						error = getStatus(nHandle, nModuleNumber, &nTimeout, &nLAM);
						if (error)
							printf("pcic32_test: error @ getStatus() = %d\n", error);
						else
							printf("pcic32_test: nTimeout: %d, nLAM: %d\n", nTimeout, nLAM);
					}
				}
			}
		}
        error = detachWindow(nHandle, nModuleNumber);
		if (error)
			printf("error @ detachWindow() = %d\n", error);

		CloseHandle(nHandle);
	}
	else
		printf("pcic32_test: Can't open \"%s\" (Error %d)\n", DRIVER_PATH, GetLastError());

	printf("pcic32_test: Press return to continue ...\n");

	getchar();

	return 0;
}

//--------------------------------------------------------------------------------------
// the end
//--------------------------------------------------------------------------------------
