
//-------------------------------------------------------------------------
// WIN95 driver for PCICC32 CAMAC interface from ARW Elektronik, Germany --
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps derived from pcic32c.c              AR           03.03.2000
// PLX8052 integration error fixed                 AR           13.03.2000
// version 2.1 released                            AR           20.03.2000
// new SubSystem-ID, version 3.0 released          AR           23.03.2000
// major change in device-ID                       AR           10.04.2000
// release of version 4.0                          AR           10.04.2000
// if last found PCIADA was not REV_02 -> failed   AR           07.05.2000
// release of version 4.1                          AR           07.05.2000
//

#define VERSION 0x0401      // version of this driver

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#include <vwin32.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <vmmreg.h>
#include <vxdwraps.h>
#include <debug.h>
#include <configmg.h>
#include <winerror.h>

#include "vpcic32d.h"  // shared info between driver and application
#include "pcic32r.h"
#include "pcic32.h"    // driver internal structures and constants

//------------------------------------------------------------------------
// PRAGMAS
//
#undef  CURSEG
#define CURSEG() LCODE

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

//-------------------------------------------------------------------------
// DEFINES
//
#define PCINAME_95 "PCI\\VEN_10B5&DEV_2258"                        // start in devnode WIN95
#define PCINAME_98 "PCI\\VEN_10B5&DEV_2258&SUBSYS_22589050&REV_01" // start in devnode WIN98
#define PCINAME_99 "PCI\\VEN_10B5&DEV_2258&SUBSYS_22589050&REV_02" // start in devnode WIN98

#define MAKE_PAGE(x) (((DWORD)x) >> 12)         // berechnet die page zu einer lin adresse

//-------------------------------------------------------------------------
// EXTERNALS
//

//------------------------------------------------------------------------
// GLOBALS
//
int       iNumIfc           = 0;        // number of found PCIADAs
VMMLIST   list              = 0;        // linked list of interfaces
VPCIC32_STATIC_STATUS *cc32[VPCIC32D_MAX_INTERFACES] = {0}; // fast index to status

//------------------------------------------------------------------------
// FUNCTIONS
//

//------------------------------------------------------------------------
// some helper functions
//
//-------------------------------------------------------------------------
// functions which are not acailable in ddk
//
ULONG VXDINLINE PageReserve(ULONG page, ULONG npages, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push npages
  _asm push page
  VMMCall(_PageReserve);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

ULONG VXDINLINE PageCommitPhys(ULONG page, ULONG npages, ULONG physpg, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push physpg
  _asm push npages
  _asm push page
  VMMCall(_PageCommitPhys);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

ULONG VXDINLINE PageDecommit(ULONG page, ULONG npages, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push npages
  _asm push page
  VMMCall(_PageDecommit);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

static void enableInterface(VPCIC32_STATIC_STATUS *psVss)
{
	*psVss->pwCNTRL = RELEASE_CC32;
}

static void disableInterface(VPCIC32_STATIC_STATUS *psVss)
{
	*psVss->pwCNTRL = INHIBIT_CC32;
}

static void clearPCIADAInterrupt(VPCIC32_STATIC_STATUS *psVss)
{
	WORD wStore = *psVss->pwCNTRL;

	disableInterface(psVss);

	*psVss->pwCNTRL = wStore;
}

static void getIrqStatus(VPCIC32_STATIC_STATUS *psVss, WORD *bTimeout, WORD *bIrq)
{
	WORD wStore = *psVss->pwINTCSR;

	if (wStore & 4) 
		*bIrq = 1;
	else
		*bIrq = 0;

	if (wStore & 0x20)
		*bTimeout = 1;
	else
		*bTimeout = 0;
}

//------------------------------------------------------------------------
// the ioctl related functions
//

//------------------------------------------------------------------------
// poll the status from pciada
//
static DWORD vpcic32d_get_status(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  DWORD dwModuleNumber  = ((VPCIC32D_DEVICE *)ptr->lpvInBuffer)->dwInterface;
  VPCIC32D_STATUS *psVs = (VPCIC32D_STATUS *)ptr->lpvOutBuffer;

  Out_Debug_String("vpcic32d_get_status()\r\n");

  if (ptr->cbOutBuffer < sizeof(VPCIC32D_STATUS))
  {
	  *(DWORD *)ptr->lpcbBytesReturned = 0;
	  return ERROR_INVALID_PARAMETER;
  }

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES) 
  {
	  psVss = cc32[dwModuleNumber];

	  if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
	  {
		  psVs->dwInterface = dwModuleNumber;
		  getIrqStatus(psVss, &psVs->bTimeout, &psVs->bInterrupt);

		  *(DWORD *)ptr->lpcbBytesReturned = sizeof(VPCIC32D_STATUS);
		  return ERROR_SUCCESS;
	  }
  }

  return ERROR_NO_SYSTEM_RESOURCES;
}

//------------------------------------------------------------------------
// clear a pending irq from pciada
//
static DWORD vpcic32d_clear_status(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  DWORD dwModuleNumber  = ((VPCIC32D_DEVICE *)ptr->lpvInBuffer)->dwInterface;

  Out_Debug_String("vpcic32d_clear_status()\r\n");

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES) 
  {
	  psVss = cc32[dwModuleNumber];

	  if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
	  {
		  clearPCIADAInterrupt(psVss);
		  *(DWORD *)ptr->lpcbBytesReturned = 0;

		  return ERROR_SUCCESS;
	  }
  }

  return ERROR_NO_SYSTEM_RESOURCES;
}

//------------------------------------------------------------------------
// get a user access address from ring 3 to camac range
//
static DWORD vpcic32d_attach_cc32(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  VPCIC32_STATIC_STATUS *psVssTemp;
  DWORD dwModuleNumber = ((VPCIC32D_DEVICE *)ptr->lpvInBuffer)->dwInterface;
  VPCIC32D_WINDOW *psVw = (VPCIC32D_WINDOW *)ptr->lpvOutBuffer;

  Out_Debug_String("vpcic32d_attach_cc32()\r\n");

  if (ptr->cbOutBuffer < sizeof(VPCIC32D_WINDOW))
  {
	  *(DWORD *)ptr->lpcbBytesReturned = 0;
	  return ERROR_INVALID_PARAMETER;
  }

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES)
  {
	  psVssTemp = cc32[dwModuleNumber];

	  if (psVssTemp == (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
	  {
		  psVss = SearchCC32(list, dwModuleNumber);
		  
		  if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
		  {
			  cc32[dwModuleNumber] = psVssTemp = psVss;

			  // map device memory range
              psVss->dwShared_base = PageReserve(PR_SHARED, MAKE_PAGE(VPCIC32_USR_WINDOW), PR_FIXED);

              if (PageCommitPhys(MAKE_PAGE(psVss->dwShared_base), VPCIC32_USR_PAGES, 
				                            MAKE_PAGE(psVss->dUSR_MemBase), (PC_INCR | PC_USER | PC_WRITEABLE)) == 0)
			  {
				  _PageFree((PVOID)psVss->dwShared_base, 0);        
				  psVss->dwShared_base = (DWORD)BOGUSADDRESS;
				  return ERROR_NO_SYSTEM_RESOURCES;
			  }

			  enableInterface(psVssTemp);
		  }
		  else
			  return ERROR_CANTOPEN;
	  }

	  psVssTemp->dwLinkCount++;

	  // give back device memory range
	  psVw->dwInterface  = dwModuleNumber;
	  psVw->pvWindowBase = (void *)psVss->dwShared_base;
  	  *(DWORD *)ptr->lpcbBytesReturned = sizeof(VPCIC32D_WINDOW);
  }
  else 
	  return ERROR_NO_SYSTEM_RESOURCES;

  return ERROR_SUCCESS;
}

//------------------------------------------------------------------------
// release a user access to ...
//
static DWORD vpcic32d_detach_cc32(PDIOCPARAMETERS ptr)
{
  VPCIC32_STATIC_STATUS *psVss;
  DWORD dwModuleNumber = ((VPCIC32D_DEVICE *)ptr->lpvInBuffer)->dwInterface;

  Out_Debug_String("vpcic32d_detach_cc32()\r\n");

  if (dwModuleNumber < VPCIC32D_MAX_INTERFACES) 
  {
	  psVss = cc32[dwModuleNumber];

	  if (psVss != (VPCIC32_STATIC_STATUS *)BOGUSADDRESS)
	  {
		  psVss->dwLinkCount--;

		  if (psVss->dwLinkCount <= 0)
		  {
		      // deinit the interface
			  disableInterface(psVss);

			  // free access
			  if (PageDecommit(MAKE_PAGE(psVss->dwShared_base), VPCIC32_USR_PAGES, 0) == 0)
				  return ERROR_NO_SYSTEM_RESOURCES;
			  _PageFree((PVOID)psVss->dwShared_base, 0);   
			  
			  psVss->dwShared_base = (DWORD)BOGUSADDRESS;
			  cc32[dwModuleNumber] = (VPCIC32_STATIC_STATUS *)BOGUSADDRESS;
			  *(DWORD *)ptr->lpcbBytesReturned = 0;
		  }
	  }
  }
  else
	  return ERROR_NO_SYSTEM_RESOURCES;

  return ERROR_SUCCESS;
}

//------------------------------------------------------------------------
// called every time a path is opened
//
static DWORD vpcic32d_dioc_open(PDIOCPARAMETERS ptr)
{
  Out_Debug_String("vpcic32d_dioc_open()\r\n");

  if (ptr->cbOutBuffer < sizeof(WORD))
	  *(DWORD *)ptr->lpcbBytesReturned = 0;
  else
  {
	  *(DWORD *)ptr->lpcbBytesReturned = sizeof(WORD);
	  *(WORD  *)ptr->lpvOutBuffer      = VERSION;
  }

  return ERROR_SUCCESS;
}

//------------------------------------------------------------------------
// called every time a path is closed
//
static DWORD vpcic32d_dioc_closehandle(PDIOCPARAMETERS ptr)
{
  Out_Debug_String("vpcic32d_dioc_closehandle()\r\n");
   
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------
// called when driver is launched
//
BOOL _stdcall CVPCIC32D_Device_Init(void)
{
	DWORD retcode;
	DWORD success = ERROR_FILE_NOT_FOUND;
	int i;

	Out_Debug_String("CVPCIC32D_Device_Init()\r\n");

	for (i = 0; i < VPCIC32D_MAX_INTERFACES; i++)
		cc32[i] = (VPCIC32_STATIC_STATUS *)BOGUSADDRESS;

	list = List_Create(LF_ASYNC | LF_ALLOC_ERROR, sizeof(VPCIC32_STATIC_STATUS));

	iNumIfc = 0;
	retcode = ScanRegistry(PCINAME_95, list, &iNumIfc);
	if (retcode == ERROR_SUCCESS) 
		success = retcode;
	retcode = ScanRegistry(PCINAME_98, list, &iNumIfc);
	if (retcode == ERROR_SUCCESS) 
		success = retcode;
	retcode = ScanRegistry(PCINAME_99, list, &iNumIfc);
	if (retcode == ERROR_SUCCESS) 
		success = retcode;
	
	if (success != ERROR_SUCCESS)
	{
		success = ERROR_CANTOPEN;    
		goto getOut;
	}

	// if no PCIADA which fits is available
	if (!iNumIfc)
	{
		retcode = ERROR_NO_SYSTEM_RESOURCES;
		goto getOut;
	}

	MakeLinearAddresses(list);

	Debug_Printf_DebMax("Anzahl der Interfaces: %d\r\n", iNumIfc); 

	return TRUE;

	// common jump out if init fails
	getOut:
	if (list) List_Destroy((DWORD)list);

	Out_Debug_String("CVPCIC32D_Device_Init(), no PCIADA found.\r\n");

	return FALSE;
}

//-----------------------------------------------------------------
// called for each IOCTL
//
DWORD _stdcall CVPCIC32D_Device_IOctrl(DWORD  dwService, PDIOCPARAMETERS ptr) 
{
	DWORD error = 0;
 
    switch(dwService)
    {
		case(VPCIC32_GET_STATUS):
			error = vpcic32d_get_status(ptr);
			break;

		case (VPCIC32_CLEAR_STATUS):
			error = vpcic32d_clear_status(ptr);
			break;

		case(VPCIC32_ATTACH_CC32):
			error = vpcic32d_attach_cc32(ptr);
			break;

		case(VPCIC32_DETACH_CC32):
			error = vpcic32d_detach_cc32(ptr);
			break;

		case(DIOC_OPEN):
			error = vpcic32d_dioc_open(ptr);
			break;

		case(DIOC_CLOSEHANDLE):
			error = vpcic32d_dioc_closehandle(ptr);
			break;

		default:
			Debug_Printf_DebMax("dwService: 0x%08x\r\n", dwService);
			error = ERROR_INVALID_PARAMETER;
    } 
	
	return error;
}

//-----------------------------------------------------------------
// called when driver is removed from memory
//
BOOL _stdcall CVPCIC32D_Device_UNInit(void)
{
  VPCIC32_STATIC_STATUS *psVss;
  int i;
  
  Out_Debug_String("CVPCIC32D_Device_UNInit()\r\n\r\n");

  if (list)
  {
	  psVss = (VPCIC32_STATIC_STATUS *)List_Get_First(list);
  
	  while (psVss)
	  {
		  if (psVss->dwShared_base != (DWORD)BOGUSADDRESS)
		  {
			  if (PageDecommit(MAKE_PAGE(psVss->dwShared_base), VPCIC32_USR_PAGES, 0) == 0)
				  return ERROR_NO_SYSTEM_RESOURCES;
			  _PageFree((PVOID)psVss->dwShared_base, 0);   
				  
			  psVss->dwShared_base = (DWORD)BOGUSADDRESS;
		  }

		  psVss = (VPCIC32_STATIC_STATUS *)List_Get_Next(list, (VMMLISTNODE)psVss);
	  }

	  List_Destroy((DWORD)list);
  }

  return TRUE;
}


