
//-------------------------------------------------------------------------
// a simple test program to test the main functions of the 
// PCIVME drivers for WIN95 or WINNT from ARW Elektronik, Germany
//
// this example uses the 'pcivme_ni.dll' (link with ..\pcivme_ni\{debug|release}\pcivme_ni.lib)
//
// (c) 2000-2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           08.08.2000
// small cosmetic changes                          AR           08.06.2002
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <windows.h>
#include <stdlib.h>  // for atexit()
#include <conio.h>   // because of _getch()

#include "..\pcivme_ni\pcivme_ni.h"    // include the prototypes of the DLL calls
#include <vme.h>


//-------------------------------------------------------------------------
// DEFINES
//
#define _WINNT

#ifdef _WINNT
#define DEFDEVICENAME "\\\\.\\PCIVME:\\VMEMMxx"     // a device name 'template' for WINNT
#else  // no _WINNT
#define DEFDEVICENAME "\\\\.\\C:\\windows\\system\\VpcivmeD.vxd" // the same for WIN95/98
#endif 

#define DEFAULT_MODULE_NUMBER     1 // number of connected CC32 module
#define MY_ACCESS_ADDRESS 0x1000    // example address for access
#define BUFFER_SIZE       0x80000   // size of a buffer


#define MESSAGESTRING "Simple access to the PCIVME interface using the 'pcivme_ni.dll' - %s\n"


//-------------------------------------------------------------------------
// GLOBALS
//
unsigned short wbuffer[BUFFER_SIZE];// a big buffer for read or write blocks of data

void hlpMsg(void)
{
	printf(MESSAGESTRING, __DATE__);
    printf("usage: PCIVMETest [-?]\n");
	printf("       ? - get help\n");
	printf("\n");
	printf("This program addresses interface #%d\n", DEFAULT_MODULE_NUMBER);
	printf("\n");
}


void stopAtEnd(void)
{
	printf("Press any key to proceed.\n");
	_getch();
}

//-------------------------------------------------------------------------
// MAIN
//
int main(int argc, char *argv[])
{
	int result = 0;
	int  hHandle = 0;       // the handle to my device path
	unsigned char bbuf;     // holds the byte (8 bit) read or write arguments
	unsigned long lbuf;		// holds the long (32 bits) reads
	unsigned short wbuf;	// holds the short (16 bit) reads

	if (argc != 1)
	{
		hlpMsg();
		exit(0);
	}

	// to stop at each exit
	atexit(stopAtEnd);

	printf(MESSAGESTRING, __DATE__);

	/* open a path to a device. */
    result = VMEinit(DEFDEVICENAME, DEFAULT_MODULE_NUMBER, Std_NoPriv_Data, &hHandle);
	if (result)
	{
		printf("Can't open interface \"%s\" to VMEMM-module \"%d\"!\n", DEFDEVICENAME, DEFAULT_MODULE_NUMBER);
		return(result);
	}
	else
	{	
		// write a byte -------------------------------------
		bbuf = 0x55;
		VMEwrite(hHandle, MY_ACCESS_ADDRESS, 1 /* byte */, 1, &bbuf);

		// read a byte --------------------------------------
		VMEread(hHandle, MY_ACCESS_ADDRESS, 1 /* byte */, 1, &bbuf);

		// write a word -------------------------------------
		wbuf = 0xAAAA;
		VMEwrite(hHandle, MY_ACCESS_ADDRESS, 2 /* byte */, 1, &wbuf);

		// read a word --------------------------------------
		VMEread(hHandle, MY_ACCESS_ADDRESS, 2 /* byte */, 1, &wbuf);

		// write a long -------------------------------------
		lbuf = 0x66666666;
		VMEwrite(hHandle, MY_ACCESS_ADDRESS, 4 /* byte */, 1, &lbuf);

		// read a long --------------------------------------
		VMEread(hHandle, MY_ACCESS_ADDRESS, 4 /* byte */, 1, &lbuf);

		// get the current interrupt vector -----------------
		VMEinterrupt(hHandle, &bbuf);

		// read wordwise a buffer of data -------------------
		VMEread(hHandle, MY_ACCESS_ADDRESS, 2 /* byte */, BUFFER_SIZE, wbuffer);


		/* close the opened path */
		result = VMEclose(hHandle); 
		if (result)
		{
			printf("Can't close interface!\n");
			exit(result);
		}
	}

	return 0;
}