// a very simple test program to use with the WIN95 driver for the PCIVME interface
// of ARW Elektronik
//
// compiled with VC++ 5.0 as console application
//
// created                                               AR     21.07.2000


//----------- INCLUDES ----------------------------------------------------
#include <windows.h>
#include <stdio.h>
#include <..\vpcivmed\vpcivmed.h>    // header for win-95..98
#include <..\vpcivmed\vic.h>
#include <..\vpcivmed\vme.h>

//----------- DEFINES -----------------------------------------------------
#define VMEMM_MODULE_NUMBER 1
#define DRIVER_PATH         "\\\\.\\c:\\Windows\\system\\vpcivmed.vxd"
#define MAX_TIMEOUT_LOOPS   1000000  // maximum loops for waiting reset finished (poor)

//----------- CONSTANTS ---------------------------------------------------
// user initialisation table for pcivme
static VPCIVMED_INIT_COMMAND sUserInitStruct   = {VMEMM_MODULE_NUMBER, {{STOP, WORD_ACCESS, 0,    0}}};            
// user deinitialisation table
static VPCIVMED_INIT_COMMAND sUserDeInitStruct = {VMEMM_MODULE_NUMBER, {{STOP, WORD_ACCESS, 0,    0}}};

//----------- FUNCTIONS ---------------------------------------------------
HANDLE _open(char *driver_path)
{
	HANDLE  nHandle;
	DWORD	DIOC_count;			// count of returned bytes of DeviceIoControl
	DWORD	result;

    nHandle = CreateFile(driver_path, 
						0, 
						0, 
						NULL, 
						0, 
						FILE_FLAG_DELETE_ON_CLOSE, 
						NULL);
	
	if (nHandle != INVALID_HANDLE_VALUE) 
	{ 
		// init hardware (only one time after the first init it works OK)
		result = DeviceIoControl(nHandle, 
						VPCIVMED_INIT_HARDWARE,
						&sUserInitStruct, 
						(DWORD)sizeof(sUserInitStruct), 
						NULL,
						(DWORD)0, 
						&DIOC_count, 
						NULL);

		if (result) 
			return nHandle;

		CloseHandle(nHandle); 
	}

	return INVALID_HANDLE_VALUE;
}

void _close(HANDLE nHandle)
{
	if (nHandle != INVALID_HANDLE_VALUE)
		CloseHandle(nHandle); 
}

//-------------------------------------------------------------------------------------------
// generate a SYSRESET on the vmebus and re-init the interface
//
int  _reset(HANDLE nHandle)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_RESET_COMMAND sResetCmd;  // structure to make reset
  VPCIVMED_RESET_RESULT  sResetResult; // structure to get result 
  DWORD                  result;
  int                    timeout;    // control the waiting reset finished loop 

  sResetCmd.dwInterface    = VMEMM_MODULE_NUMBER;
  sResetCmd.wCommand       = GLOBAL_RESET_CMD;
 
  // make reset on VME-BUS
  result = DeviceIoControl(nHandle, VPCIVMED_RESET,
                     &sResetCmd, (DWORD)sizeof(sResetCmd), &sResetResult,
                            (DWORD)sizeof(sResetResult), &DIOC_count, NULL);	
  
  if (!result)
	return GetLastError();

  // poll if reset command finished
  timeout = MAX_TIMEOUT_LOOPS;
  do
  {
    sResetCmd.dwInterface    = VMEMM_MODULE_NUMBER;
    sResetCmd.wCommand       = POLL_RESET_CMD;
 
    result = DeviceIoControl(nHandle, VPCIVMED_RESET,
                     &sResetCmd, (DWORD)sizeof(sResetCmd), &sResetResult,
                            (DWORD)sizeof(sResetResult), &DIOC_count, NULL);
  } while ((result) && (sResetResult.wResult != 0) && (--timeout));

  if (!result)
	return GetLastError();

  if (!timeout)
	return -1;    // ?
  
  return 0;
}

void *_attachWindow(HANDLE nHandle, unsigned long offset, 
				              unsigned long size, unsigned short modifier)
{
	VPCIVMED_ADD_WINDOW sAWnd;    // attach-window structure
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	void  *WinBase;
	DWORD result;

	if (nHandle == INVALID_HANDLE_VALUE) return NULL;

	// attach a window into the VME space ----------------
	sAWnd.dwInterface = VMEMM_MODULE_NUMBER;
	sAWnd.size        = ((size + 0x0FFF) / 0x1000) * 0x1000; // truncate to chunks of 0x1000
	sAWnd.modifier    = modifier;
	sAWnd.base        = offset & HI_ADDRESS_MASK;
	result = DeviceIoControl(nHandle, VPCIVMED_ATTACH_WINDOW,
						&sAWnd, sizeof(sAWnd), &WinBase, sizeof(WinBase),
								 &DIOC_count, NULL);
	if (!result)
      return NULL;

	return WinBase;
}

// release a vme window
void *_detachWindow(HANDLE nHandle, void *WinBase)
{
	VPCIVMED_REMOVE_WINDOW sRWnd; // remove-window structure
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;

	if (nHandle == INVALID_HANDLE_VALUE) return WinBase;

	// detach the window into VME space
	sRWnd.dwInterface = VMEMM_MODULE_NUMBER;
	sRWnd.pdwLinAdr   = (DWORD *)WinBase;
	result = DeviceIoControl(nHandle, VPCIVMED_DETACH_WINDOW,
						&sRWnd, sizeof(sRWnd), &WinBase,
							sizeof(WinBase), &DIOC_count, NULL);

	if (!result)
		return WinBase;  // can't release
	else
		return NULL;     // released
}




//----------- MAIN --------------------------------------------------------
int main(int argc, char *argv[])
{
	HANDLE nHandle;
	int error;
	void *pvBase = NULL;
	unsigned short *wData;
	int i;

	// open path to interface
	nHandle = _open(DRIVER_PATH);
	if (nHandle == INVALID_HANDLE_VALUE)
	{
		error = GetLastError();
		printf("Can't open path to driver! (0x%08x)\n", error);
		return error;
	}

	// reset VME
	error = _reset(nHandle);
	if (error)
	{
		printf("Error @ VME Reset (0x%08x)\n", error);
		return error;
	}

	// close path (because of _reset)
	_close(nHandle);


	// open path again (because of _reset)
	nHandle = _open(DRIVER_PATH);
	if (nHandle == INVALID_HANDLE_VALUE)
	{
		error = GetLastError();
		printf("Can't open path to driver 2nd time! (0x%08x)\n", error);
		return error;
	}


	// attach a window into VME defined by offset, size and address modifier
	pvBase = _attachWindow(nHandle, 0, 0x1000000, Std_NoPriv_Data); 
	if (pvBase == NULL)
	{
		error = GetLastError();
		printf("Can't attach Window into VME at Address Modifier 0x%02x! (0x%08x)\n", Std_NoPriv_Data, error);
		return error;
	}

	// make writes to VME window (Address 0 to 0x1FFE as words in Std_NoPriv_Data)
	wData = (unsigned short *)pvBase;  // assign the virtual base to write pointer
	for (i = 0; i < 0xFFF; i++)         
		*wData++ = i;                  // write and increment word address

	// make reads from VME window and print result (Address 0 to 0x1FFE as words in Std_NoPriv_Data)
	wData = (unsigned short *)pvBase;  // assign the virtual base to write pointer
	for (i = 0; i < 0xFFF; i++)
		printf("Read back 0x%04x <-> 0x%04x.\n", i, *wData++); // read and increment word address

	// detach the window into VME
	_detachWindow(nHandle, pvBase);

	// close the path to the interface
	_close(nHandle);

	return 0;
}