//-------------------------------------------------------------------------
// a simple test program to test the main functions of the 
// WINNT-PCIVME driver from ARW Elektronik, Germany
//
// this example accesses dircetly the WINNT driver through 'ioctls' and 'read', 'write'
//
// (c) 2000-2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// $Log: pciVMETest.c,v $
// Revision 1.5  2002/10/27 21:32:24  klaus
// compatibility improved
//
// Revision 1.4  2002/10/27 19:23:11  klaus
// backward compatibilty problem for 2 Gbyte limit solved
//
// Revision 1.3  2002/10/27 18:20:35  klaus
// some includes point now to original locations
//
// Revision 1.2  2002/10/27 18:08:35  klaus
// CVS log added, WINNT file size 2 Gbyte circumvent
//
// what                                            who          when
// started                                         AR           16.04.2000
// waiting for user input to CatchInterrupts_VME.. AR           05.08.2001
//

//-------------------------------------------------------------------------
// please look into the pvmon sources for pcilibNt.c for information about
// advanced features of the programming with the PCIVME interface
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include <..\pcivme\pcivme.h>    // include the structure definitions and IOCTL codes
#include <..\pcivme\vme.h>       // some defines about VME address modifiers

//-------------------------------------------------------------------------
// DEFINES
//
#define DEFDEVICENAME "\\\\.\\PCIVME:\\VMEMM1"  // device name for the 1st VMEMM
#define MODULE_NUMBER 1                         // the same as for VMEMM<1>
#define MY_ACCESS_ADDRESS 0x1000                // example address for access
#define LIMIT_2GBYTE  0x80000000                // 2 GByte addressing limit of WINNT ...

//-------------------------------------------------------------------------
// GLOBALS
//
// user initialisation table for pcivme
static PCIVME_INIT_COMMAND sUserInitStruct   = {MODULE_NUMBER, {{STOP, WORD_ACCESS, 0,    0}}};
            
// user deinitialisation table
static PCIVME_INIT_COMMAND sUserDeInitStruct = {MODULE_NUMBER, {{STOP, WORD_ACCESS, 0,    0}}};
            
//-------------------------------------------------------------------------
// open my path to the interface
//
static HANDLE openPath(char *name)
{
    return (HANDLE)CreateFile( 
                     name, 
                     GENERIC_READ | GENERIC_WRITE, 
                     0, 
                     NULL, 
                     OPEN_EXISTING, 
                     FILE_ATTRIBUTE_NORMAL, 
                     NULL 
                     );
}

//-------------------------------------------------------------------------
// init the hardware with its default and user defined parameters
//
static DWORD initHardware(HANDLE hHandle, PCIVME_INIT_COMMAND *psUserInitStruct)
{
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl

	return DeviceIoControl(hHandle, PCIVME_INIT_HARDWARE,
                      psUserInitStruct, (DWORD)sizeof(*psUserInitStruct), NULL,
                          (DWORD)0, &DIOC_count, NULL);
}

//-------------------------------------------------------------------------
// the inverse to above
//
static DWORD deinitHardware(HANDLE hHandle, PCIVME_INIT_COMMAND *psUserInitStruct)
{
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl

	return DeviceIoControl(hHandle, PCIVME_DEINIT_HARDWARE,
                      psUserInitStruct, (DWORD)sizeof(*psUserInitStruct), NULL,
                          (DWORD)0, &DIOC_count, NULL);
}

//-------------------------------------------------------------------------
// init the interface
//
int Init_InterfaceNT(HANDLE *hHandle)
{
  int error = 0;
  int result;

  // start the driver ------------------
  if ((*hHandle = openPath(DEFDEVICENAME)) == (HANDLE)-1)
	  return GetLastError();

  result = initHardware(*hHandle, &sUserInitStruct);
  if (!result) 
	  return GetLastError();

  return 0;
}

//-------------------------------------------------------------------------------------------
// read bytes/words/longs
//
static void setAccessParameter(HANDLE hHandle, char modifier, char width, char increment, unsigned long *ulAddress)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
    PCIVME_ACCESS_COMMAND access_parameter;

	access_parameter.bAddressModifier	= modifier;
	access_parameter.bAccessType		= width;
	access_parameter.bIncrement			= increment;

	// take care of only 2 Gbyte addressing capabilities of WINNT ...
	if (*ulAddress >= LIMIT_2GBYTE)
	{
		*ulAddress -= LIMIT_2GBYTE;
		access_parameter.dwAccessBase = LIMIT_2GBYTE;
	}
	else
		access_parameter.dwAccessBase = 0;

	result = DeviceIoControl(hHandle, PCIVME_SET_ACCESS_PARA,
				&access_parameter, (DWORD)sizeof(access_parameter), NULL,
					   0, &DIOC_count, NULL);	 
}

char  ReadByteNT(HANDLE hHandle, unsigned long adr, unsigned short modifier, char *BusError)
{
  DWORD bytesRead;
  char  buffer;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(char), sizeof(char), &adr);
    
  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(char), &bytesRead, NULL))
	  *BusError = TRUE;

  return buffer;
}

short ReadWordNT(HANDLE hHandle, unsigned long adr, unsigned short modifier, char *BusError)
{
  DWORD bytesRead;
  short buffer;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(short), sizeof(short), &adr);

  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(short), &bytesRead, NULL))
	  *BusError = TRUE;

  return buffer;
}

long  ReadLongNT(HANDLE hHandle, unsigned long adr, unsigned short modifier, char *BusError)
{
  DWORD bytesRead;
  long  buffer;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(long), sizeof(long), &adr);
      
  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(long), &bytesRead, NULL))
	  *BusError = TRUE;

  return buffer;
}

//-------------------------------------------------------------------------------------------
// write a byte/word/long 
//
void  WriteByteNT(HANDLE hHandle, unsigned long adr, char value, unsigned short modifier, char *BusError)
{
  DWORD bytesWritten;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(char), sizeof(char), &adr);
	  
  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(char), &bytesWritten, NULL))
	  *BusError = TRUE;
}

void  WriteWordNT(HANDLE hHandle, unsigned long adr, short value, unsigned short modifier, char *BusError)
{
  DWORD bytesWritten;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(short), sizeof(short), &adr);
      
  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(short), &bytesWritten, NULL))
	  *BusError = TRUE;
}

void  WriteLongNT(HANDLE hHandle, unsigned long adr, long value, unsigned short modifier, char *BusError)
{
  DWORD bytesWritten;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(long), sizeof(long), &adr);
     
  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(long), &bytesWritten, NULL))
	  *BusError = TRUE;
}



//-------------------------------------------------------------------------
// the main program
//
int main(int argc, char *argv[])
{
	HANDLE hHandle = INVALID_HANDLE_VALUE;  // the handle to my device path
	int result;
	unsigned char byte;		// data to hold a read byte
	unsigned short word;	//				  read word
	unsigned long  dword;	//                read long
	char BusError;			// holds a pending BusError

	printf("A simple test program to test the PCIVME WINNT driver. (%s)\n", __DATE__);

	result = Init_InterfaceNT(&hHandle);  // open the device path to the 1st interface
	if (result)
	{
		printf("Can't Init_InterfaceNT() - returns error 0x%08x\n", result);
		return(result);
	}
	else
	{	
		printf("\nInitialisation is done. Please press <return> to continue.\n");
		getchar();

		// write a byte -------------------------------------
		WriteByteNT(hHandle, MY_ACCESS_ADDRESS, 0x55, Std_NoPriv_Data, &BusError);

		// read a byte --------------------------------------
		byte = ReadByteNT(hHandle, MY_ACCESS_ADDRESS, Std_NoPriv_Data, &BusError);

		// write a word -------------------------------------
		WriteWordNT(hHandle, MY_ACCESS_ADDRESS, (short)0xAAAA, Std_NoPriv_Data, &BusError);

		// read a word --------------------------------------
		word = ReadWordNT(hHandle, MY_ACCESS_ADDRESS, Std_NoPriv_Data, &BusError);

		// write a long -------------------------------------
		WriteLongNT(hHandle, MY_ACCESS_ADDRESS, 0x66666666, Std_NoPriv_Data, &BusError);

		// read a long --------------------------------------
		dword = ReadLongNT(hHandle, MY_ACCESS_ADDRESS, Std_NoPriv_Data, &BusError);


		// close device -------------------------------------
		result = deinitHardware(hHandle, &sUserDeInitStruct);

		if (hHandle != INVALID_HANDLE_VALUE)
			CloseHandle(hHandle);

		return 0;
	}
}