//-------------------------------------------------------------------------------------------
// os_wrapper.c - coordinates the driver for the windows versions
// (c) 1999-2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// $Log: os_wrapper.c,v $
// Revision 1.2  2002/10/27 17:32:17  klaus
// CVS log added, WINNT 2 Gbyte file size circumvent
//
// what                                                              who    when
// first steps                                                       AR     07.11.1999
//

//-------------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <pcilibNT.h>
#include <pcilib95.h>
#include <os_info.h>

//-------------------------------------------------------------------------------------------
// DEFINES
//

//-------------------------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct
{
	int   (*Init_Interface)(int);        
	void  (*DeInit_Interface)(int);         
	void  (*Reset_VME)(void);                               

	int   (*GetError)(void);                                
	void  (*ClearError)(void);                              

	char  (*ReadByte)(unsigned long,unsigned short);        
	void  (*WriteByte)(unsigned long,char,unsigned short);  
	short (*ReadWord)(unsigned long,unsigned short);        
	void  (*WriteWord)(unsigned long, short,unsigned short);
	long  (*ReadLong)(unsigned long,unsigned short);        
	void  (*WriteLong)(unsigned long,long,unsigned short);  

	void  (*InitInterrupt)(void);                       
	void  (*EnableInterrupt)(unsigned short);   
	void  (*DisableInterrupt)(unsigned short);  
	int   (*PollInterrupt)(unsigned short);     
	char  (*ReadVectorByte)(void);            
	short (*ReadVectorWord)(void);            
	long  (*ReadVectorLong)(void);            

	char  (*TAS)(unsigned long,unsigned short);      
	void  (*SetSfail)(void);                                       
	void  (*ClrSfail)(void);                                 
	unsigned short (*PollSfail)(void);                        

	void (*GetInterfaceInfo)(char);    
	void (*_SetModifier)(unsigned short);

	unsigned long (*_SetRegister)(unsigned long ,unsigned long);
	unsigned long (*_GetRegister)(unsigned long);

    char *cszDriverPath;                      
    int  nInterfaceNumber;                       
    int  WordMode;                           
} WRAPPER;

//-------------------------------------------------------------------------------------------
// LOCALS
//
static WRAPPER wpr;

//-------------------------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------------------------
// GLOBALS
//
char *cszDriverPath;                      
int  nInterfaceNumber = 0;                       
int  WordMode         = 0;                           

//-------------------------------------------------------------------------------------------
// FUNCTIONS
//
void PrepareForOS(void)
{
  if (IsWindowsNT()) 
  {
	printf("Windows-NT detected.\n");

	wpr.Init_Interface   = Init_InterfaceNT;
	wpr.Init_Interface   = Init_InterfaceNT;       
	wpr.DeInit_Interface = DeInit_InterfaceNT;         
	wpr.Reset_VME        = Reset_VMENT;                        

	wpr.GetError         = GetErrorNT;                        
	wpr.ClearError       = ClearErrorNT;                        

	wpr.ReadByte         = ReadByteNT;
	wpr.WriteByte	     = WriteByteNT;
	wpr.ReadWord         = ReadWordNT;
	wpr.WriteWord        = WriteWordNT;
	wpr.ReadLong         = ReadLongNT;
	wpr.WriteLong        = WriteLongNT;

	wpr.ReadVectorByte   = ReadVectorByteNT;       
	wpr.ReadVectorWord   = ReadVectorWordNT;         
	wpr.ReadVectorLong   = ReadVectorLongNT;          

	wpr.TAS              = TASNT;
	wpr.SetSfail         = SetSfailNT;                           
	wpr.ClrSfail         = ClrSfailNT;                      
	wpr.PollSfail        = PollSfailNT;               

	wpr.GetInterfaceInfo = GetInterfaceInfoNT;   
	wpr._SetModifier     = _SetModifierNT;

	wpr._SetRegister     = _SetRegisterNT;
	wpr._GetRegister     = _GetRegisterNT;

	cszDriverPath        = cszDriverPathNT;                      
  }
  else
  {
    if (IsWindows98()) printf("Windows98 detected.\n");
    if (IsWindows95()) printf("Windows95 detected.\n");

	wpr.Init_Interface   = Init_Interface95;
	wpr.Init_Interface   = Init_Interface95;       
	wpr.DeInit_Interface = DeInit_Interface95;         
	wpr.Reset_VME        = Reset_VME95;                        

	wpr.GetError         = GetError95;                        
	wpr.ClearError       = ClearError95;                        

	wpr.ReadByte         = ReadByte95;
	wpr.WriteByte	     = WriteByte95;
	wpr.ReadWord         = ReadWord95;
	wpr.WriteWord        = WriteWord95;
	wpr.ReadLong         = ReadLong95;
	wpr.WriteLong        = WriteLong95;

	wpr.ReadVectorByte   = ReadVectorByte95;       
	wpr.ReadVectorWord   = ReadVectorWord95;         
	wpr.ReadVectorLong   = ReadVectorLong95;          

	wpr.TAS              = TAS95;
	wpr.SetSfail         = SetSfail95;                           
	wpr.ClrSfail         = ClrSfail95;                      
	wpr.PollSfail        = PollSfail95;               

	wpr.GetInterfaceInfo = GetInterfaceInfo95;   
	wpr._SetModifier     = _SetModifier95;

	wpr._SetRegister     = _SetRegister95;
	wpr._GetRegister     = _GetRegister95;

	cszDriverPath        = cszDriverPath95;                      
  }
}

int   Init_Interface(int InterfaceNumber)
{
	return wpr.Init_Interface(InterfaceNumber);
}

void  DeInit_Interface(int InterfaceNumber)
{
	wpr.DeInit_Interface(InterfaceNumber);
}

void  Reset_VME(void)
{
	wpr.Reset_VME();
}

int   GetError(void)
{
	return wpr.GetError();
}

void  ClearError(void)
{
	wpr.ClearError();
}

char  ReadByte(unsigned long adr,unsigned short modf)
{
	return wpr.ReadByte(adr, modf);
}

void  WriteByte(unsigned long adr, char val, unsigned short modf)
{
	wpr.WriteByte(adr,val,modf);
}

short ReadWord(unsigned long adr,unsigned short modf)
{
	return wpr.ReadWord(adr, modf);
}

void  WriteWord(unsigned long adr, short val, unsigned short modf)
{
	wpr.WriteWord(adr, val, modf);
}

long  ReadLong(unsigned long adr, unsigned short modf)
{
	return wpr.ReadLong(adr, modf);
}

void  WriteLong(unsigned long adr, long val, unsigned short modf)
{
	wpr.WriteLong(adr, val, modf);
}

char  ReadVectorByte(void)
{
	return wpr.ReadVectorByte();
}

short ReadVectorWord(void)
{
	return wpr.ReadVectorWord();
}

long  ReadVectorLong(void)
{
	return wpr.ReadVectorLong();
}

char  TAS(unsigned long adr, unsigned short modf)
{
	return wpr.TAS(adr, modf);
}

void  SetSfail(void)
{
	wpr.SetSfail();
}

void  ClrSfail(void)
{
	wpr.ClrSfail();
}

unsigned short PollSfail(void)
{
	return wpr.PollSfail();
}

void GetInterfaceInfo(char type)
{
	wpr.GetInterfaceInfo(type);
}

void _SetModifier(unsigned short modifier)
{
	wpr._SetModifier(modifier);
}

unsigned long _SetRegister(unsigned long Address, unsigned long Value)
{
	return wpr._SetRegister(Address, Value);
}

unsigned long _GetRegister(unsigned long Address)
{
	return _GetRegister(Address);
}

