//-------------------------------------------------------------------------------------------
// pcilib.c - some functions to interface to the pvmon
// (c) 1999-2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// $Log: pcilib95.c,v $
// Revision 1.2  2002/10/27 17:32:17  klaus
// CVS log added, WINNT 2 Gbyte file size circumvent
//
// what                                                              who    when
// finished first release to use with the PCIVME interface of ARW    AR     22.03.1999
// polling mode for systems with too less free IRQs introduced		 AR     04.04.1999
// some changes in error handling and reset functions                AR     18.04.1999
//

//-------------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <vpcivmed.h>
#include <VIC.H>
#include <VME.H>
#include <pcilib95.h>
#include <mbuffer.h>

//-------------------------------------------------------------------------------------------
// DEFINES
//
#define DEVPATH "\\\\.\\c:\\Windows\\system\\vpcivmed.vxd"  // default path to vpcivmed.vxd driver
#define DEFAULT_SIZE      0x01000000                      // default VME window size
#define DEFAULT_ADDRESS   0x00000000                      // default VME window base address
#define DEFAULT_MODIFIER  Std_NoPriv_Data                 // default address modifier
#define DEFAULT_TYPE      sizeof(char);                   // default data BUS access width
#define MAX_TIMEOUT_LOOPS 100000                          // maximum loops for waiting reset finished

//-------------------------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct
{
  void           *base;   
  unsigned long  offset;
  unsigned long  size;
  unsigned short modifier;
  short          active;
} MYWINDOW;


//-------------------------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------------------------
// GLOBALS
//

// the path of the vpcivmed driver
char cszDriverPath95[100]; 

//--------------------------------------------------------------------------
// LOCALS
// the handle of the opend path
static HANDLE Handle = INVALID_HANDLE_VALUE;  

// user initialisation table for pcivme
static VPCIVMED_INIT_COMMAND sUserInitStruct   = {0, {{STOP, WORD_ACCESS, 0,    0}}};
            
// user deinitialisation table
static VPCIVMED_INIT_COMMAND sUserDeInitStruct = {0, {{STOP, WORD_ACCESS, 0,    0}}};
            
// window into vme
static MYWINDOW myWindow = {0,DEFAULT_ADDRESS,DEFAULT_SIZE,DEFAULT_MODIFIER,0};

static short BusError = 0;

static int NoIRQ = 0;      // IRQs are available and allowed

//-------------------------------------------------------------------------------------------
// FUNCTIONS
//

//-------------------------------------------------------------------------------------------
// callback function for all irqs
//
static void _stdcall IrqHandler(DWORD levelAndVector)
{
    VPCIVMED_IRQ_CONTROL sIrqControl;
	WORD level  = (WORD)(levelAndVector >> 16);
	WORD vector = (WORD)(levelAndVector & 0xffff);

	// different handling for bus errors -----------------------------------
	if ((level == 7) && (vector == 7))
      BusError = 1;
	else
      if (!((level ==6) && (vector == 6)))
 	    AddIRQtoBuffer(level, vector);

	// re-enable interrupts #1 ---------------------------------------------
	sIrqControl.dwInterface    = nInterfaceNumber;
	sIrqControl.wEnable        = 1;  // enable
    DeviceIoControl(Handle, VPCIVMED_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, NULL, NULL);

	// this is a fron panel push button rest
    if ((level == 6) && (vector == 6))
	{
      AddMsgAsStringtoBuffer("Push Button Reset pressed and handled!"); 
      Reset_VME95();
	}
}

//-------------------------------------------------------------------------------------------
// replacement function if no irqs are available
//
static void PollIrqHandler(void)
{
	VPCIVMED_VECTOR_COMMAND sVectorCommand;
	VPCIVMED_VECTOR_LEVEL   sVectorLevel;
	DWORD                   DIOC_count;
	int result;

	sVectorCommand.dwInterface = nInterfaceNumber;
	sVectorCommand.wAction     = READ_VECTOR;
	sVectorCommand.wType        = 1;

    result = DeviceIoControl(Handle, VPCIVMED_READ_VECTOR,
                        &sVectorCommand, sizeof(sVectorCommand), 
						&sVectorLevel, sizeof(sVectorLevel), &DIOC_count, NULL);
  

	if (!result)
    {
      AddMsgtoBuffer(0, GetLastError());
	  return;
    }

	// different handling for bus errors -----------------------------------
	if ((sVectorLevel.wLevel == 7) && ((sVectorLevel.dwStatusID & 0xFF) == 7))
      BusError = 1;
	else
      if (sVectorLevel.wLevel)
 	    AddIRQtoBuffer(sVectorLevel.wLevel, (short)(sVectorLevel.dwStatusID & 0xFF));
}

//-------------------------------------------------------------------------------------------
// function to get access to a vme window
//
static void *attachWindow(unsigned long size, unsigned long offset, unsigned short modifier)
{
  VPCIVMED_ADD_WINDOW sAWnd;    // attach-window structure
  DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
  void  *WinBase;
  DWORD result;

  // attach a window into the VME space ----------------
  sAWnd.dwInterface = nInterfaceNumber;
  sAWnd.size        = ((size + 0x0FFF) / 0x1000) * 0x1000; // truncate to chunks of 0x1000
  sAWnd.modifier    = modifier;
  sAWnd.base        = offset & 0xFFFFF000;
  result = DeviceIoControl(Handle, VPCIVMED_ATTACH_WINDOW,
                        &sAWnd, sizeof(sAWnd), &WinBase, sizeof(WinBase),
                                 &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(1,GetLastError());

  myWindow.base     = WinBase;
  myWindow.offset   = offset & 0xFFFFF000;
  myWindow.size     = size;
  myWindow.modifier = modifier;
  myWindow.active   = 1;

  return myWindow.base;
}


//-------------------------------------------------------------------------------------------
// release a vme window
//
static void detachWindow(void *base)
{
  VPCIVMED_REMOVE_WINDOW sRWnd; // remove-window structure
  DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
  DWORD result;
  void  *WinBase;

  // detach the window into VME space
  sRWnd.dwInterface = nInterfaceNumber;
  sRWnd.pdwLinAdr   = (DWORD *)base;
  result = DeviceIoControl(Handle, VPCIVMED_DETACH_WINDOW,
                        &sRWnd, sizeof(sRWnd), &WinBase,
                            sizeof(WinBase), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(2,GetLastError());

  myWindow.active = 0;
}

//-------------------------------------------------------------------------------------------
// checks if the access is outside the current window
//
static void checkWindow(unsigned long offset, short elementSize)
{
  if ((offset   >=  myWindow.offset ) && 
	  (offset   <= (myWindow.offset + myWindow.size - elementSize)) &&
	  (myWindow.active))
	return;

  // remove old window into vme ---------------------------------------
  if (myWindow.active)
	detachWindow(myWindow.base);

  myWindow.base     = attachWindow(myWindow.size, offset, myWindow.modifier);
}


//-------------------------------------------------------------------------------------------
// set the current modifier to a new value
//
void _SetModifier95(unsigned short modifier)
{
  DWORD size;
  unsigned long offset = 0;

  if (modifier == myWindow.modifier)
	return;

  // remove old window into vme --------------------------------------
  if (myWindow.active)
	detachWindow(myWindow.base);

  // special handling for some address modifiers ---------------------
  switch (modifier)
  {
    case Short_Sup:
	case Short_NoPriv:
		 offset = 0;
		 size   = 0x10000;
		 break;
	case Std_NoPriv_Data:
    case Std_Sup_Data:
	case Std_NoPriv_Prog:
	case Std_Sup_Prog:
		 offset = 0;
		 size   = 0x01000000;
		 break;

    default:
		 size = DEFAULT_SIZE;  
		 break;
  }

  myWindow.base     = attachWindow(size, offset, modifier);
}


//-------------------------------------------------------------------------------------------
// set interface special registers - VIC68A
//
unsigned long _SetRegister95(unsigned long Address, unsigned long Value)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_VIC68A_ACTION sAction;    // structure to access vic chip
  DWORD                  result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = (unsigned short)Address;
  sAction.wAccessMode      = VIC68A_WRITE;
  sAction.bContent         = (unsigned char)Value & 0xff;
 
  result = DeviceIoControl(Handle, VPCIVMED_ACCESS_VIC68A,
                     &sAction, sizeof(sAction), &sAction,
                            sizeof(sAction), &DIOC_count, NULL);	 

  if (!result)
    AddMsgtoBuffer(24,GetLastError());

  return sAction.bContent & 0xff;
}

//-------------------------------------------------------------------------------------------
// function to get special interface registers - VIC68A
//
unsigned long _GetRegister95(unsigned long Address)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_VIC68A_ACTION sAction;    // structure to access vic chip
  DWORD                  result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = (unsigned short)Address;
  sAction.wAccessMode      = VIC68A_READ;
  sAction.bContent         = 0;
 
  result = DeviceIoControl(Handle, VPCIVMED_ACCESS_VIC68A,
                     &sAction, sizeof(sAction), &sAction,
                            sizeof(sAction), &DIOC_count, NULL);	 

  if (!result)
    AddMsgtoBuffer(25,GetLastError());

  return sAction.bContent & 0xff;
}

//-------------------------------------------------------------------------------------------
// test if the driver is connected to a VMEMM
//
static int Connect(int nIfcNum) 
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_STATIC_STATUS sPCIss;     // structure to hold status information
  VPCIVMED_STANDARD_COMMAND sSc;
  DWORD result;

  sSc.dwInterface = nIfcNum;


  // try to get static status ----------------
  result = DeviceIoControl(Handle, VPCIVMED_GET_STATIC_STATUS,
                      &sSc, (DWORD)sizeof(sSc), &sPCIss,
                           (DWORD)sizeof(sPCIss), &DIOC_count, NULL);
  
  if (!result) return GetLastError();

  if (DIOC_count != sizeof(VPCIVMED_STATIC_STATUS)) return ERROR_INVALID_PARAMETER;

  if (!(sPCIss.wConnected)) return ERROR_BAD_UNIT;

  if (!(sPCIss.wNumIRQs)) NoIRQ = 1;

  return ERROR_SUCCESS;
}


//-------------------------------------------------------------------------------------------
// init the interface with the current properties
//
int Init_Interface95(int nIfcNum)
{
  int error = 0;
  DWORD DIOC_count;              // count of returned bytes of DeviceIoControl
  DWORD result;

  VPCIVMED_IRQ_HANDLER     sIrqHandler;  // structure for irq handler handling
  VPCIVMED_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs

  myWindow.active = 0;

  strcpy(cszDriverPath95, DEVPATH);

  // load the VXD dynamically ------------------
  if (Handle == INVALID_HANDLE_VALUE)
  {
    Handle = CreateFile(cszDriverPath95,0,0,NULL,0,FILE_FLAG_DELETE_ON_CLOSE,NULL);

    if (Handle == INVALID_HANDLE_VALUE) return GetLastError();
  }

  sUserInitStruct.dwInterface = nIfcNum;
  result = DeviceIoControl(Handle, VPCIVMED_INIT_HARDWARE,
                      &sUserInitStruct, (DWORD)sizeof(sUserInitStruct), NULL,
                          (DWORD)0, &DIOC_count, NULL);
  if (!result) return GetLastError();

  // install the irq-handler ----------------------------------------------
  sIrqHandler.dwInterface  = nIfcNum;
  sIrqHandler.dwIrqHandler = (DWORD)IrqHandler;
  result = DeviceIoControl(Handle, VPCIVMED_INSTALL_IRQ_HANDLER,
                        &sIrqHandler, (DWORD)sizeof(sIrqHandler), NULL,
                            0, NULL, NULL);
  if (!result) 
  {
	int error;

	error = GetLastError();
	if (error != ERROR_IRQ_BUSY) 
	  return error;
	else
	  AddMsgAsStringtoBuffer("No Interrupt available (Install IRQ Handler)!\n");
  }

  // enable interrupts ----------------------------------------------------
  sIrqControl.dwInterface    = nIfcNum;
  sIrqControl.wEnable        = 1;  // enable both
  result = DeviceIoControl(Handle, VPCIVMED_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, NULL, NULL);
  if (!result) 
  {
	int error;

	error = GetLastError();
	if (error != ERROR_IRQ_BUSY) 
	  return error;
	else
	  AddMsgAsStringtoBuffer("No Interrupt available (Enable Interrupts)!\n");
  }

  return Connect(nIfcNum);  // be careful - result sense inverted here
}


//-------------------------------------------------------------------------------------------
// deinit the current interface
//
void  DeInit_Interface95(int nIfcNum)
{
  DWORD result;
  DWORD DIOC_count;              // count of returned bytes of DeviceIoControl

  VPCIVMED_IRQ_HANDLER     sIrqHandler;  // structure for irq handler handling
  VPCIVMED_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs

  if (Handle == INVALID_HANDLE_VALUE)
	return;

  // disable interrupts    ----------------------------------------------
  sIrqControl.dwInterface    = nIfcNum;
  sIrqControl.wEnable        = 0;  // disable both
  result = DeviceIoControl(Handle, VPCIVMED_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, NULL, NULL);

  if (!result)
  {
    AddMsgtoBuffer(7,GetLastError());
    return;
  }


  // deinstall the irq-handler --------------------------------------------
  sIrqHandler.dwInterface  = nIfcNum;
  sIrqHandler.dwIrqHandler = (DWORD)BOGUSADDRESS;
  result = DeviceIoControl(Handle, VPCIVMED_INSTALL_IRQ_HANDLER,
                        &sIrqHandler, (DWORD)sizeof(sIrqHandler), NULL,
                            0, NULL, NULL);
  if (!result)
  {
    AddMsgtoBuffer(8,GetLastError());
    return;
  }

  // deinitialize the hardware --------------------------
  sUserDeInitStruct.dwInterface = nIfcNum;
  result = DeviceIoControl(Handle, VPCIVMED_DEINIT_HARDWARE,
                    &sUserDeInitStruct, (DWORD)sizeof(sUserDeInitStruct), NULL,
                           0, &DIOC_count, NULL);	 

  if (!result)
  {
    AddMsgtoBuffer(9,GetLastError());
    return;
  }

  // close device -------------------------------------
  if (Handle != INVALID_HANDLE_VALUE)
    CloseHandle(Handle);

  Handle = INVALID_HANDLE_VALUE;
  myWindow.active = 0;
}


//-------------------------------------------------------------------------------------------
// check if an error occured
//
int GetError95(void)
{
  if (NoIRQ)
	PollIrqHandler();  // poll it
  else
    SleepEx(0,TRUE);   // wait for a callback to arrive

  return BusError;
}


//-------------------------------------------------------------------------------------------
// clear a pending error
//
void ClearError95(void)
{
  BusError = 0;
}

//-------------------------------------------------------------------------------------------
// read bytes/words/longs longs are split into 2 word if a word interface is selected
//
char  ReadByte95(unsigned long adr, unsigned short modifier)
{
  checkWindow(adr, sizeof(char));

  return *(char *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset));
}

short ReadWord95(unsigned long adr, unsigned short modifier)
{
  checkWindow(adr, sizeof(short));

  return *(short *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset));
}

long  ReadLong95(unsigned long adr, unsigned short modifier)
{
  // special handling if WORD-MODE VME range ---------------------
  if (WordMode)
  {
    long  lRet;
    short sRet;

    sRet = ReadWord95(adr, modifier);
	lRet = sRet;
	lRet <<= 16;
	sRet = ReadWord95(adr + 2, modifier);
	lRet |= sRet;

	return lRet;
  }

  checkWindow(adr, sizeof(long));

  return *(long *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset));
}

//-------------------------------------------------------------------------------------------
// write a byte/word/long and split a long access into 2 word accesses if word data path
//
void  WriteByte95(unsigned long adr, char value, unsigned short modifier)
{
  checkWindow(adr, sizeof(char));

  *(char *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset)) = value;
}

void  WriteWord95(unsigned long adr, short value, unsigned short modifier)
{
  checkWindow(adr, sizeof(short));

  *(short *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset)) = value;
}

void  WriteLong95(unsigned long adr, long value, unsigned short modifier)
{
  // special handling if WORD-MODE VME range ---------------------
  if (WordMode)
  {
	  WriteWord95(adr    , (short)(value >> 16), modifier);
	  WriteWord95(adr + 2, (short)(value      ), modifier);
	  return;
  }

  checkWindow(adr, sizeof(long));

  *(long *)((DWORD)myWindow.base + (adr - (DWORD)myWindow.offset)) = value;
}

//-------------------------------------------------------------------------------------------
// check for a pending SYSFAIL 
//
unsigned short PollSfail95(void)
{
  unsigned char reg;

  reg = (unsigned char)_GetRegister95(EGICR);

  return (reg & 0x08) ? 0 : 1;
}

//-------------------------------------------------------------------------------------------
// clear a interface set SYSFAIL
//
void  ClrSfail95(void)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_VIC68A_ACTION sAction;    // structure to access vic chip
  DWORD result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = ICR7;
  sAction.wAccessMode      = VIC68A_OR;
  sAction.bContent         = 0x80;
 
  result = DeviceIoControl(Handle, VPCIVMED_ACCESS_VIC68A,
                     &sAction, (DWORD)sizeof(sAction), &sAction,
                            (DWORD)sizeof(sAction), &DIOC_count, NULL);	 
  if (!result)
	AddMsgtoBuffer(14,GetLastError());
}

//-------------------------------------------------------------------------------------------
// set a SYSFAIL
//
void  SetSfail95(void)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_VIC68A_ACTION sAction;    // structure to access vic chip
  DWORD result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = ICR7;
  sAction.wAccessMode      = VIC68A_AND;
  sAction.bContent         = 0x3F;
 
  result = DeviceIoControl(Handle, VPCIVMED_ACCESS_VIC68A,
                     &sAction, (DWORD)sizeof(sAction), &sAction,
                            (DWORD)sizeof(sAction), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(15,GetLastError());	 
}

//-------------------------------------------------------------------------------------------
// read a interrupt vector as byte/word/long (if supported)
//
char  ReadVectorByte95(void)
{
  VPCIVMED_VECTOR_COMMAND   sVectorCommand;
  VPCIVMED_VECTOR_LEVEL     sVectorLevel; // vector + interrupt-level structure
  DWORD                     DIOC_count;   // count of returned bytes of DeviceIoControl
  DWORD                     result;

  sVectorCommand.dwInterface = nInterfaceNumber;
  sVectorCommand.wAction     = READ_VECTOR;
  sVectorCommand.wType       = BYTE_ACCESS;

  // look if an interrupt is pending ------------------ 
  result = DeviceIoControl(Handle, VPCIVMED_READ_VECTOR,
                        &sVectorCommand, sizeof(sVectorCommand), &sVectorLevel,
                            (DWORD)sizeof(sVectorLevel), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(16,GetLastError());

  return (char)sVectorLevel.dwStatusID;
}

short ReadVectorWord95(void)
{
  VPCIVMED_VECTOR_COMMAND   sVectorCommand;
  VPCIVMED_VECTOR_LEVEL     sVectorLevel; // vector + interrupt-level structure
  DWORD                     DIOC_count;   // count of returned bytes of DeviceIoControl
  DWORD                     result;

  sVectorCommand.dwInterface = nInterfaceNumber;
  sVectorCommand.wAction     = READ_VECTOR;
  sVectorCommand.wType       = WORD_ACCESS;

  // look if an interrupt is pending ------------------ 
  result = DeviceIoControl(Handle, VPCIVMED_READ_VECTOR,
                        &sVectorCommand, sizeof(sVectorCommand), &sVectorLevel,
                            (DWORD)sizeof(sVectorLevel), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(17,GetLastError());

  return (short)sVectorLevel.dwStatusID;
}

long  ReadVectorLong95(void)
{
  VPCIVMED_VECTOR_COMMAND   sVectorCommand;
  VPCIVMED_VECTOR_LEVEL     sVectorLevel; // vector + interrupt-level structure
  DWORD                     DIOC_count;   // count of returned bytes of DeviceIoControl
  DWORD                     result;

  if (WordMode)
    return (long)ReadVectorWord95();
 
  sVectorCommand.dwInterface = nInterfaceNumber;
  sVectorCommand.wAction     = READ_VECTOR;
  sVectorCommand.wType       = LONG_ACCESS;

  // look if an interrupt is pending ------------------ 
  result = DeviceIoControl(Handle, VPCIVMED_READ_VECTOR,
                        &sVectorCommand, sizeof(sVectorCommand), &sVectorLevel,
                            (DWORD)sizeof(sVectorLevel), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(18,GetLastError());

  return sVectorLevel.dwStatusID;
}

//-------------------------------------------------------------------------------------------
// emulate a 68K TAS (read/modify/write) instruction
//
char  TAS95(unsigned long adr, unsigned short modifier)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_TAS_STRUCT    sTAS;       // structure to do a Test and Set
  DWORD                  result;

  sTAS.dwInterface      = nInterfaceNumber;
  sTAS.wModifier        = modifier;
  sTAS.dwAddress        = adr;
  sTAS.bContent         = 0x80;
 
  result = DeviceIoControl(Handle, VPCIVMED_TAS,
                     &sTAS, (DWORD)sizeof(sTAS), &sTAS,
                            (DWORD)sizeof(sTAS), &DIOC_count, NULL);	 
  if (!result)
	AddMsgtoBuffer(19,GetLastError());

  return (sTAS.bContent == 0x80) ? 1 : 0;  // check if a read reads a lock
}

//-------------------------------------------------------------------------------------------
// generate a SYSRESET on the vmebus and re-init the interface
//
void  Reset_VME95(void)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  VPCIVMED_RESET_COMMAND sResetCmd;  // structure to make reset
  VPCIVMED_RESET_RESULT  sResetResult; // structure to get result 
  DWORD                  result;
  int                    timeout;    // control the waiting reset finished loop 

  sResetCmd.dwInterface    = nInterfaceNumber;
  sResetCmd.wCommand       = GLOBAL_RESET_CMD;
 
  // make reset on VME-BUS
  result = DeviceIoControl(Handle, VPCIVMED_RESET,
                     &sResetCmd, (DWORD)sizeof(sResetCmd), &sResetResult,
                            (DWORD)sizeof(sResetResult), &DIOC_count, NULL);	
  
  if (!result)
	AddMsgtoBuffer(20,GetLastError());

  // poll if reset command finished
  timeout = MAX_TIMEOUT_LOOPS;
  do
  {
    sResetCmd.dwInterface    = nInterfaceNumber;
    sResetCmd.wCommand       = POLL_RESET_CMD;
 
    result = DeviceIoControl(Handle, VPCIVMED_RESET,
                     &sResetCmd, (DWORD)sizeof(sResetCmd), &sResetResult,
                            (DWORD)sizeof(sResetResult), &DIOC_count, NULL);
  } while ((result) && (sResetResult.wResult != 0) && (--timeout));

  if (!result)
	AddMsgtoBuffer(20,GetLastError());

  if (!timeout)
	AddMsgAsStringtoBuffer("Timeout while awiting for SYSRESET finished!");
  
  DeInit_Interface95(nInterfaceNumber);
  Init_Interface95(nInterfaceNumber);

  if (GetError95()) ClearError95();
}


//-------------------------------------------------------------------------------------------
// get - and print out - dynamic status information from the interface
//
static void GetDynamicStatus(void)
{
  int   i;
  unsigned char reg;

  // enabled interrupts
  for (i = 0; i < 7; i++)
  {
    reg = (unsigned char)_GetRegister95(VICR1 + (i * 4));

	printf("VME Interrupt %d is %s and maps to level %d\n", i+1, \
		       (reg & 0x80) ? "disabled" : "enabled ", reg & 0x07); 
  }

  printf("\n");
}

//-------------------------------------------------------------------------------------------
// get - and print out - dynamic status information from the interface
//
static void GetVIC68ARegs(void)
{
  int   i;

  // read and printout all VIC68A registers ----------------------
  printf("VIC68 register:value\n");
  for (i = 0; i < 14; i++)
  {
	printf("0x%02x:0x%02x    0x%02x:0x%02x    0x%02x:0x%02x    0x%02x:0x%02x\n", 
		i * 4 +   3, (unsigned char)_GetRegister95(i * 4 +    3), 
		i * 4 +  59, (unsigned char)_GetRegister95(i * 4 +   59), 
		i * 4 + 115, (unsigned char)_GetRegister95(i * 4 +  115), 
		i * 4 + 171, (unsigned char)_GetRegister95(i * 4 +  171)); 
  }

  printf("\n");
}

//-------------------------------------------------------------------------------------------
// get - and printout - static status information about the interface
//
static void GetStaticStatus(void)
{
    DWORD      DIOC_count;             // count of returned bytes of DeviceIoControl
    VPCIVMED_STATIC_STATUS sPCIss;     // structure to hold status information
    VPCIVMED_STANDARD_COMMAND sSc;
	DWORD                     result;

	sSc.dwInterface = nInterfaceNumber;

    // try to get static status ----------------
    result = DeviceIoControl(Handle, VPCIVMED_GET_STATIC_STATUS,
                        &sSc, sizeof(sSc), &sPCIss,
                            (DWORD)sizeof(sPCIss),
                                 &DIOC_count, NULL);
    if (!result)
	  AddMsgtoBuffer(22,GetLastError());

	if (!DIOC_count)
    {
      printf("No information for VMEMM No. %d available!\n\n", nInterfaceNumber);
	  return;
	}

  	printf("Number of VMEMM   : %d\n", nInterfaceNumber);
	printf("Memory Windows    : %d\n", sPCIss.wNumMemWindows);
	printf("IO Ports          : %d\n", sPCIss.wNumIOPorts);
	printf("IRQs              : %d\n", sPCIss.wNumIRQs);
	printf("DMAs              : %d\n", sPCIss.wNumDMAs);
    printf("Base of Memory    : 0x%08x\n", sPCIss.dUSR_MemBase);
    printf("Length of Memory  : %d\n", sPCIss.dUSR_MemLength);
	printf("Connected?        : %s\n", (sPCIss.wConnected) ? "yes" : "no");
	printf("Driver link count : %d\n", sPCIss.dwLinkCount);
	printf("\n");
}

//-------------------------------------------------------------------------------------------
// get - and print out - information about all found PCIADA
//
static void GetPCIADAStatus(void)
{
    DWORD      DIOC_count;             // count of returned bytes of DeviceIoControl
    VPCIVMED_PCIADA_STATUS    sVps;    // structure to hold status information
    VPCIVMED_STANDARD_COMMAND sSc;     // dummy only
	DWORD                     result;
    int i;

	sSc.dwInterface = 0;

    // try to get static status ----------------
    result = DeviceIoControl(Handle, VPCIVMED_GET_PCIADA_STATUS,
                        &sSc, sizeof(sSc), &sVps,
                            (DWORD)sizeof(sVps),
                                 &DIOC_count, NULL);
    if (!result)
	  AddMsgtoBuffer(23,GetLastError());

	if (DIOC_count != sizeof(sVps))
    {
      printf("Information not available!\n");
	  return;
	}

    printf("Driver Path/Name  : %s\n", cszDriverPath95);
    printf("\n");
	printf("Driver Version    : %d.%d\n", sVps.wVersion >> 8, sVps.wVersion & 0xff);
    printf("Number of PCIADA  : %d\n",    sVps.wNumberOfInterfaces);

	for (i = 0; i < sVps.wNumberOfInterfaces; i++)
	{
		printf("------------------%d-------\n", i);
		printf("Status of PCIADA  : %d\n", sVps.sPCIAda[i].dIfcStatus); 

		printf("Version of FPGA   : %d\n", sVps.sPCIAda[i].wFPGAVersion);
		printf("Number of VMEMM   : %d\n", sVps.sPCIAda[i].wModuleNumber);
		printf("WordMode/SysCtrl  : %s/%s\n", (sVps.sPCIAda[i].wWordMode)   ? "yes" : "no",
			                                  (sVps.sPCIAda[i].wSysControl) ? "yes" : "no");

		printf("Connected?        : %s\n", (sVps.sPCIAda[i].wConnected) ? "yes" : "no");
	}
	printf("\n");
}

//-------------------------------------------------------------------------------------------
// print out some interface special info
//
void GetInterfaceInfo95(char type)
{
  switch (type)
  {
    case '0': GetPCIADAStatus();
  	          break;
    case '1': GetStaticStatus();
	          break;
    case '2': GetDynamicStatus();
	          break;
    case '3': GetVIC68ARegs();
		      break;
  }
}

//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
