//-------------------------------------------------------------------------------------------
// pcilib.c - some functions to interface to the pvmon
// (c) 1999-2002 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// $Log: pcilibNT.c,v $
// Revision 1.4  2002/10/27 21:32:13  klaus
// compatibility improved
//
// Revision 1.3  2002/10/27 19:23:22  klaus
// backward compatibilty problem for 2 Gbyte limit solved
//
// Revision 1.2  2002/10/27 17:32:17  klaus
// CVS log added, WINNT 2 Gbyte file size circumvent
//
// what                                                              who    when
// finished first release to use with the PCIVME interface of ARW    AR     22.03.1999
// polling mode for systems with too less free IRQs introduced		 AR     04.04.1999
// some changes in error handling and reset functions                AR     18.04.1999
//

//-------------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <winioctl.h>
#include <pcivme.h>
#include "VIC.H"
#include "VME.H"
#include <pcilibNT.h>
#include "mbuffer.h"

//-------------------------------------------------------------------------------------------
// DEFINES
//
#define DEVPATH " - No input neccessary!"                    // not used
#define DEFDEVICENAME "\\\\.\\PCIVME:\\VMEMMxx"
#define DEFAULT_ADDRESS   0x00000000                      // default VME window base address
#define DEFAULT_MODIFIER  Std_NoPriv_Data                 // default address modifier
#define DEFAULT_TYPE      sizeof(char)                    // default data BUS access width
#define MAX_TIMEOUT_LOOPS 100000                          // maximum loops for waiting reset finished
#define LIMIT_2GBYTE  0x80000000                          // 2 GByte addressing limit of WINNT ...

//-------------------------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct
{
  ULONG		dwCurrentAddress;
  USHORT	wCurrentModifier;
  UCHAR		bCurrentAccessWidth;
} MY_ACCESS_PARAMETER;

//-------------------------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------------------------
// GLOBALS
//

// the path of the vpcivmed driver
char cszDriverPathNT[100]; 

//--------------------------------------------------------------------------
// LOCALS
// the handle of the opend path
static HANDLE hHandle    = INVALID_HANDLE_VALUE; 
static HANDLE hIrqHandle = INVALID_HANDLE_VALUE;  // for irq thread
static HANDLE hThread    = INVALID_HANDLE_VALUE; 

// user initialisation table for pcivme
static PCIVME_INIT_COMMAND sUserInitStruct   = {0, {{STOP, WORD_ACCESS, 0,    0}}};
            
// user deinitialisation table
static PCIVME_INIT_COMMAND sUserDeInitStruct = {0, {{STOP, WORD_ACCESS, 0,    0}}};
            
// window into vme
static short BusError = 0;

// if no IRQs are supported
static int NoIRQ = 0;      // IRQs are available and allowed

// a simple access cache
static MY_ACCESS_PARAMETER my_access_para = {DEFAULT_ADDRESS, DEFAULT_MODIFIER, DEFAULT_TYPE};

// storage for the number of PCIADA 
static int nIfcNumStat = 0;

//-------------------------------------------------------------------------------------------
// FUNCTIONS
//


//-------------------------------------------------------------------------------------------
// replacement function if no irqs are available
//
static int GetVector(HANDLE hHandle_)
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_VECTOR_REQUEST  request;
	PCIVME_VECTOR_RESPONSE response;
	OVERLAPPED sOverlapped;
	int		nError = ERROR_SUCCESS;

	sOverlapped.hEvent    = NULL;
	response.wCount       = 0;
	request.bPoll         = FALSE;
	request.wRequestCount = 1;
	result = DeviceIoControl(hHandle_, PCIVME_READ_VECTOR,
                &request, sizeof(request), &response, sizeof(response), 
				   &DIOC_count, &sOverlapped);	 

	if (!result) nError = GetLastError();

	// a real error
	if ((nError != ERROR_IO_PENDING) && (nError != ERROR_SUCCESS)) 
	{
		AddMsgtoBuffer(33,nError);
		return 0;
	}

	// my IO is pending - wait for it
	if (nError == ERROR_IO_PENDING)
	{
		result = GetOverlappedResult(hHandle_, &sOverlapped, &DIOC_count, TRUE);
		
		if (!result) 
		{
			nError = GetLastError();   // if it was cancelled
			if (nError != ERROR_OPERATION_ABORTED) AddMsgtoBuffer(34, nError);
			return 0;
		}
	}

	// have got immediate response
	if (response.wCount)
 		   AddIRQtoBuffer(response.bOverflow, (short)(response.bStatusID & 0xFF));

	return response.wCount;
}

static DWORD WINAPI myThread(LPVOID hIrqHandle)
{
	do
	{
		if (!GetVector((HANDLE)hIrqHandle)) ExitThread(0);
	} while (1);

	return 0;
}

//-------------------------------------------------------------------------------------------
// set the current modifier to a new value
//
void _SetModifierNT(unsigned short modifier)
{
	// do nothing
}


//-------------------------------------------------------------------------------------------
// set interface special registers - VIC68A
//
unsigned long _SetRegisterNT(unsigned long Address, unsigned long Value)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  PCIVME_VIC68A_ACTION   sAction;    // structure to access vic chip
  DWORD                  result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = (unsigned short)Address;
  sAction.wAccessMode      = VIC68A_WRITE;
  sAction.bContent         = (unsigned char)Value & 0xff;
 
  result = DeviceIoControl(hHandle, PCIVME_ACCESS_VIC68A,
                     &sAction, sizeof(sAction), &sAction,
                            sizeof(sAction), &DIOC_count, NULL);	 

  if (!result)
    AddMsgtoBuffer(24,GetLastError());

  return sAction.bContent & 0xff;
}

//-------------------------------------------------------------------------------------------
// function to get special interface registers - VIC68A
//
unsigned long _GetRegisterNT(unsigned long Address)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  PCIVME_VIC68A_ACTION   sAction;    // structure to access vic chip
  DWORD                  result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = (unsigned short)Address;
  sAction.wAccessMode      = VIC68A_READ;
  sAction.bContent         = 0;
 
  result = DeviceIoControl(hHandle, PCIVME_ACCESS_VIC68A,
                     &sAction, sizeof(sAction), &sAction,
                            sizeof(sAction), &DIOC_count, NULL);	 

  if (!result)
    AddMsgtoBuffer(25,GetLastError());

  return sAction.bContent & 0xff;
}

//-------------------------------------------------------------------------------------------
// test if the driver is connected to a VMEMM
//
static int Connect(int nIfcNum) 
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_STATIC_STATUS staticStatus;

	result = DeviceIoControl(hHandle, PCIVME_GET_STATIC_STATUS,
			NULL, 0, &staticStatus, (DWORD)sizeof(staticStatus), 
			   &DIOC_count, NULL);	 

	if (!result) return GetLastError();

	if (DIOC_count != sizeof(PCIVME_STATIC_STATUS)) return ERROR_INVALID_PARAMETER;

	if (!(staticStatus.wConnected)) return ERROR_BAD_UNIT;

	if (!(staticStatus.wNumIRQs)) NoIRQ = 1;

	return ERROR_SUCCESS;
}

//-------------------------------------------------------------------------
// create a DeviceName out of cszDeviceName and nIfcNum
static char *DeviceName(int nIfcNum)
{
	static char buffer[255];
	static char DevPath[255];
	char *ptr = buffer;

	strcpy(buffer, DEFDEVICENAME);

	while (*ptr) ptr++;
	
	do
	{
		ptr--;
	} while (*ptr != 'M');

	ptr++;
	
	if (nIfcNum >= 10)
	{
		*ptr++   = '1';
		nIfcNum -= 10;
	}

	*ptr++ = '0' + nIfcNum;
	*ptr = 0;

	// a little help for the users
	strcpy(DevPath, buffer);
	strcat(DevPath, DEVPATH);
	strncpy(cszDriverPathNT, DevPath, 99);

    return buffer;
}

//-------------------------------------------------------------------------
// open my path to the interface
static HANDLE openPath(char *name)
{
    return (HANDLE)CreateFile( 
                     name, 
                     GENERIC_READ | GENERIC_WRITE, 
                     0, 
                     NULL, 
                     OPEN_EXISTING, 
                     FILE_ATTRIBUTE_NORMAL, 
                     NULL 
                     );
}

//-------------------------------------------------------------------------
// open my path to the interface
static HANDLE openPathOverlapped(char *name)
{
    return (HANDLE)CreateFile( 
                     name, 
                     GENERIC_READ | GENERIC_WRITE, 
                     0, 
                     NULL, 
                     OPEN_EXISTING, 
                     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, 
                     NULL 
                     );
}

//-------------------------------------------------------------------------
// init the hardware with its default and user defined parameters
static DWORD initHardware(HANDLE handle, PCIVME_INIT_COMMAND *psUserInitStruct)
{
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl

	return DeviceIoControl(handle, PCIVME_INIT_HARDWARE,
                      psUserInitStruct, (DWORD)sizeof(*psUserInitStruct), NULL,
                          (DWORD)0, &DIOC_count, NULL);
}

//-------------------------------------------------------------------------
// the inverse to above
static DWORD deinitHardware(HANDLE handle, PCIVME_INIT_COMMAND *psUserInitStruct)
{
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl

	return DeviceIoControl(handle, PCIVME_DEINIT_HARDWARE,
                      psUserInitStruct, (DWORD)sizeof(*psUserInitStruct), NULL,
                          (DWORD)0, &DIOC_count, NULL);
}

//-------------------------------------------------------------------------------------------
// init the interface with the current properties
//
int Init_InterfaceNT(int nIfcNum)
{
  int error = 0;
  DWORD result;
  PCIVME_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs
  DWORD DIOC_count;
  DWORD	threaId;

  strcpy(cszDriverPathNT, DEVPATH);

  nIfcNumStat = nIfcNum;  // store for later use in ..info

  // start the driver ------------------
  if (hHandle == INVALID_HANDLE_VALUE)
  {
		if ((hHandle = openPath(DeviceName(nIfcNum))) == (HANDLE)-1)
			return GetLastError();
  }

  sUserInitStruct.dwInterface = nIfcNum;
  result = initHardware(hHandle, &sUserInitStruct);
  if (!result) return GetLastError();

  // enable interrupts ----------------------------------------------------
  sIrqControl.wEnable        = 1;  // enable the IRQ queue for hHandle
  result = DeviceIoControl(hHandle, PCIVME_CONTROL_INTERRUPTS,
						&sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
							0, &DIOC_count, NULL);
  if (!result) 
  {
	 int error;

	 error = GetLastError();
	 if (error != ERROR_IRQ_BUSY) 
		return error;
	 else
		AddMsgAsStringtoBuffer("No Interrupt available (Enable Interrupts)!\n");
  }

  if (hIrqHandle == INVALID_HANDLE_VALUE)
  {
	if ((hIrqHandle = openPathOverlapped(DeviceName(nIfcNum))) == (HANDLE)-1)
			return GetLastError();

    // enable interrupts ----------------------------------------------------
    sIrqControl.wEnable        = 1;  // enable the Irq queue for hIrqHandle
    result = DeviceIoControl(hIrqHandle, PCIVME_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, &DIOC_count, NULL);
    if (!result) 
	{
	  int error;

	  error = GetLastError();
	  if (error != ERROR_IRQ_BUSY) 
	    return error;
	  else
	    AddMsgAsStringtoBuffer("No Interrupt available (Enable Interrupts)!\n");
	}

    // create a thread which waits for a GetVector call to finish
    hThread = CreateThread(NULL, 0x1000, myThread, hIrqHandle, 0, &threaId);
    if (hThread == NULL) printf("No thread created!\n");
  }

  my_access_para.wCurrentModifier = 0xff;  // force new cache entry

  return Connect(nIfcNum);  // be careful - result sense inverted here
}


//-------------------------------------------------------------------------------------------
// deinit the current interface
//
void  DeInit_InterfaceNT(int nIfcNum)
{
  DWORD result;
  PCIVME_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs
  DWORD DIOC_count;

  if ((hHandle == INVALID_HANDLE_VALUE) || 
	      (hIrqHandle == INVALID_HANDLE_VALUE))
	return;

  if (hIrqHandle != INVALID_HANDLE_VALUE)
  {
    // disable interrupts    ----------------------------------------------
    sIrqControl.wEnable        = 0;  // disable hIrqHandle Irq queue
    result = DeviceIoControl(hIrqHandle, PCIVME_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, &DIOC_count, NULL);

    if (!result)
	{
      AddMsgtoBuffer(7,GetLastError());
      return;
	}

	// let the unblock go
	Sleep(100);

    // close Irq device -----------------------------------
    CloseHandle(hIrqHandle);
    hIrqHandle = INVALID_HANDLE_VALUE;
  }

  // disable interrupts    ----------------------------------------------
  sIrqControl.wEnable        = 0;  // disable hHandle Irq queue
  result = DeviceIoControl(hHandle, PCIVME_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, &DIOC_count, NULL);

  if (!result)
  {
    AddMsgtoBuffer(7,GetLastError());
    return;
  }

  // deinitialize the hardware --------------------------
  sUserDeInitStruct.dwInterface = nIfcNum;
  result = deinitHardware(hHandle, &sUserDeInitStruct);

  if (!result)
  {
    AddMsgtoBuffer(9,GetLastError());
    return;
  }

  // close device -------------------------------------
  if (hHandle != INVALID_HANDLE_VALUE)
    CloseHandle(hHandle);

  hHandle = INVALID_HANDLE_VALUE;
}


//-------------------------------------------------------------------------------------------
// check if an error occured
//
int GetErrorNT(void)
{
  return BusError;
}


//-------------------------------------------------------------------------------------------
// clear a pending error
//
void ClearErrorNT(void)
{
  BusError = 0;
}

//-------------------------------------------------------------------------------------------
// read bytes/words/longs longs are split into 2 word if a word interface is selected
//
static void setAccessParameter(HANDLE hHandle, UCHAR modifier, UCHAR width, UCHAR increment, ULONG *ulAddress)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
  PCIVME_ACCESS_COMMAND access_parameter;

	// take care of only 2 Gbyte addressing capabilities of WINNT ...
	if (*ulAddress >= LIMIT_2GBYTE)
	{
		*ulAddress -= LIMIT_2GBYTE;
		access_parameter.dwAccessBase = LIMIT_2GBYTE;
	}
	else
		access_parameter.dwAccessBase = 0;

	if ((my_access_para.wCurrentModifier != modifier) ||
		   (my_access_para.bCurrentAccessWidth != width))
	{
		access_parameter.bAddressModifier	= modifier;
		access_parameter.bAccessType		  = width;
		access_parameter.bIncrement			  = increment;

		result = DeviceIoControl(hHandle, PCIVME_SET_ACCESS_PARA,
					&access_parameter, (DWORD)sizeof(access_parameter), NULL,
						   0, &DIOC_count, NULL);	 

		if (!result)
		{
			AddMsgtoBuffer(10,GetLastError());
			return;
		}

		my_access_para.wCurrentModifier		= modifier;
		my_access_para.bCurrentAccessWidth  = width;
	}
}

char  ReadByteNT(unsigned long adr, unsigned short modifier)
{
  DWORD bytesRead;
  char  buffer;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(char), sizeof(char), &adr);
  if (my_access_para.dwCurrentAddress != adr)
    SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(char), &bytesRead, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(char);

  return buffer;
}

short ReadWordNT(unsigned long adr, unsigned short modifier)
{
  DWORD bytesRead;
  short buffer;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(short), sizeof(short), &adr);
  if (my_access_para.dwCurrentAddress != adr)
	  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(short), &bytesRead, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(short);

  return buffer;
}

long  ReadLongNT(unsigned long adr, unsigned short modifier)
{
  DWORD bytesRead;
  long  buffer;

  // special handling if WORD-MODE VME range ---------------------
  if (WordMode)
  {
    long  lRet;
    short sRet;

    sRet = ReadWordNT(adr, modifier);
	lRet = sRet;
	lRet <<= 16;
	sRet = ReadWordNT(adr + 2, modifier);
	lRet |= sRet;

	return lRet;
  }

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(long), sizeof(long), &adr);
  if (my_access_para.dwCurrentAddress != adr)
      SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!ReadFile(hHandle, &buffer, sizeof(long), &bytesRead, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(long);

  return buffer;
}

//-------------------------------------------------------------------------------------------
// write a byte/word/long and split a long access into 2 word accesses if word data path
//
void  WriteByteNT(unsigned long adr, char value, unsigned short modifier)
{
  DWORD bytesWritten;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(char), sizeof(char), &adr);
  if (my_access_para.dwCurrentAddress != adr)
	  SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(char), &bytesWritten, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(char);
}

void  WriteWordNT(unsigned long adr, short value, unsigned short modifier)
{
  DWORD bytesWritten;

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(short), sizeof(short), &adr);
  if (my_access_para.dwCurrentAddress != adr)
      SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(short), &bytesWritten, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(short);
}

void  WriteLongNT(unsigned long adr, long value, unsigned short modifier)
{
  DWORD bytesWritten;

  // special handling if WORD-MODE VME range ---------------------
  if (WordMode)
  {
	  WriteWordNT(adr    , (short)(value >> 16), modifier);
	  WriteWordNT(adr + 2, (short)(value      ), modifier);
	  return;
  }

  setAccessParameter(hHandle, (unsigned char)modifier, sizeof(long), sizeof(long), &adr);
  if (my_access_para.dwCurrentAddress != adr)
      SetFilePointer(hHandle, adr, NULL, FILE_BEGIN);
  if (!WriteFile(hHandle, &value, sizeof(long), &bytesWritten, NULL))
	  BusError = TRUE;
  else 
	  my_access_para.dwCurrentAddress += sizeof(long);
}

//-------------------------------------------------------------------------------------------
// check for a pending SYSFAIL 
//
unsigned short PollSfailNT(void)
{
  unsigned char reg;

  reg = (unsigned char)_GetRegisterNT(EGICR);

  return (reg & 0x08) ? 0 : 1;
}

//-------------------------------------------------------------------------------------------
// clear a interface set SYSFAIL
//
void  ClrSfailNT(void)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  PCIVME_VIC68A_ACTION   sAction;    // structure to access vic chip
  DWORD result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = ICR7;
  sAction.wAccessMode      = VIC68A_OR;
  sAction.bContent         = 0x80;
 
  result = DeviceIoControl(hHandle, PCIVME_ACCESS_VIC68A,
                     &sAction, (DWORD)sizeof(sAction), &sAction,
                            (DWORD)sizeof(sAction), &DIOC_count, NULL);	 
  if (!result)
	AddMsgtoBuffer(14,GetLastError());
}

//-------------------------------------------------------------------------------------------
// set a SYSFAIL
//
void  SetSfailNT(void)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  PCIVME_VIC68A_ACTION   sAction;    // structure to access vic chip
  DWORD result;

  sAction.dwInterface      = nInterfaceNumber;
  sAction.wRegisterAddress = ICR7;
  sAction.wAccessMode      = VIC68A_AND;
  sAction.bContent         = 0x3F;
 
  result = DeviceIoControl(hHandle, PCIVME_ACCESS_VIC68A,
                     &sAction, (DWORD)sizeof(sAction), &sAction,
                            (DWORD)sizeof(sAction), &DIOC_count, NULL);
  if (!result)
	AddMsgtoBuffer(15,GetLastError());	 
}

//-------------------------------------------------------------------------------------------
// read a interrupt vector as byte/word/long (if supported)
//
#define POLL_COUNT 64       // get vectors up to ... count
char  ReadVectorByteNT(void)
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_VECTOR_REQUEST  request;
	struct
	{
		PCIVME_VECTOR_RESPONSE response;
		char				   vectors[(POLL_COUNT - 1)];
	} sResponse;

	sResponse.response.wCount = 0;
	request.bPoll         = TRUE;
	request.wRequestCount = POLL_COUNT;
	result = DeviceIoControl(hHandle, PCIVME_READ_VECTOR,
                &request, sizeof(request), &sResponse, sizeof(sResponse), 
				   &DIOC_count, NULL);	 

	if (!result) 
		AddMsgtoBuffer(33, GetLastError());
	else
	{
		int i;
		char *ptr = &sResponse.response.bStatusID;

		printf("Got %d vectors:\n", sResponse.response.wCount);
		for (i = 0; i < sResponse.response.wCount; )
		{
			printf("0x%02x ", *ptr++);
			if (!(++i % 8)) printf("\n");  // wrap
		}
		printf("\n");
		printf("%s vectors still in queue.\n", sResponse.response.wPendingCount);
		printf("%s queue overflow was detected.\n", (sResponse.response.bOverflow) ? "A" : "No");
	}

	return 0;
}

short ReadVectorWordNT(void)
{
  printf("Word read of a vector not available!\n"); 
  return 0;
}

long  ReadVectorLongNT(void)
{
  printf("Longword read of a vector not available!\n");
  return 0;
}

//-------------------------------------------------------------------------------------------
// emulate a 68K TAS (read/modify/write) instruction
//
char  TASNT(unsigned long adr, unsigned short modifier)
{
  DWORD                  DIOC_count; // count of returned bytes of DeviceIoControl
  PCIVME_TAS_STRUCT      sTAS;       // structure to do a Test and Set
  DWORD                  result;

  sTAS.dwInterface      = nInterfaceNumber;
  sTAS.wModifier        = modifier;
  sTAS.dwAddress        = adr;
  sTAS.bContent         = 0x80;
 
  result = DeviceIoControl(hHandle, PCIVME_TAS,
                     &sTAS, (DWORD)sizeof(sTAS), &sTAS,
                            (DWORD)sizeof(sTAS), &DIOC_count, NULL); 
  if (!result)
  {
	  int Error = GetLastError();

	  if (Error == ERROR_NOACCESS)
		BusError = TRUE;
	  else
		AddMsgtoBuffer(19,Error);
  }

  return (sTAS.bContent == 0x80) ? 1 : 0;  // check if a read reads a lock
}

//-------------------------------------------------------------------------------------------
// generate a SYSRESET on the vmebus and re-init the interface
//
void  Reset_VMENT(void)
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_RESET_COMMAND reset_command;
	PCIVME_RESET_RESULT  reset_result;
    PCIVME_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs

	// deinit irq-path before --------------------------
    if (hIrqHandle != INVALID_HANDLE_VALUE)
	{
      // disable interrupts    ----------------------------------------------
      sIrqControl.wEnable        = 0;  // disable both
      result = DeviceIoControl(hIrqHandle, PCIVME_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, &DIOC_count, NULL);

      if (!result)
	  {
        AddMsgtoBuffer(36,GetLastError());
        return;
	  }

	  // let the unblock go
	  Sleep(100);

      // close Irq device -----------------------------------
      CloseHandle(hIrqHandle);
      hIrqHandle = INVALID_HANDLE_VALUE;
	}

	reset_command.wCommand = VME_RESET_CMD;
	reset_result.wResult   = 0;
	result = DeviceIoControl(hHandle, PCIVME_RESET,
                &reset_command, sizeof(reset_command), 
				&reset_result,  sizeof(reset_result), 
				   &DIOC_count, NULL);	 

	if (!result) 
		printf("PCIVME_RESET error: 0x%08x\n",GetLastError());
	else
	{
		while (reset_result.wResult)
		{
			reset_command.wCommand = POLL_RESET_CMD;
			result = DeviceIoControl(hHandle, PCIVME_RESET,
						&reset_command, sizeof(reset_command), 
						&reset_result,  sizeof(reset_result), 
							&DIOC_count, NULL);	 
			Sleep(10);
		}
	}

	my_access_para.wCurrentModifier = 0xff; // force modifier cache new

	// open IRQ path again ----------
	if (hIrqHandle == INVALID_HANDLE_VALUE)
	{
		DWORD	threaId;

		if ((hIrqHandle = openPathOverlapped(DeviceName(nInterfaceNumber))) == (HANDLE)-1)
		{
          AddMsgtoBuffer(37,GetLastError());
          return;
		}


    // enable interrupts ----------------------------------------------------
    sIrqControl.wEnable        = 1;  // enable both
    result = DeviceIoControl(hIrqHandle, PCIVME_CONTROL_INTERRUPTS,
                        &sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
                            0, &DIOC_count, NULL);
    if (!result) 
	{
	   AddMsgAsStringtoBuffer("No Interrupt available (Enable Interrupts)!\n");
	   return;
	}

    // create a thread which waits for a GetVector call to finish
    hThread = CreateThread(NULL, 0x1000, myThread, hIrqHandle, 0, &threaId);
    if (hThread == NULL) printf("No thread created!\n");
  }
}


//-------------------------------------------------------------------------------------------
// get - and print out - dynamic status information from the interface
//
static void GetDynamicStatus(void)
{
  int   i;
  unsigned char reg;

  // enabled interrupts
  for (i = 0; i < 7; i++)
  {
    reg = (unsigned char)_GetRegisterNT(VICR1 + (i * 4));

	printf("VME Interrupt %d is %s and maps to level %d\n", i+1, \
		       (reg & 0x80) ? "disabled" : "enabled ", reg & 0x07); 
  }

  printf("\n");
}

//-------------------------------------------------------------------------------------------
// get - and print out - dynamic status information from the interface
//
static void GetVIC68ARegs(void)
{
  int   i;

  // read and printout all VIC68A registers ----------------------
  printf("VIC68 register:value\n");
  for (i = 0; i < 14; i++)
  {
	printf("0x%02x:0x%02x    0x%02x:0x%02x    0x%02x:0x%02x    0x%02x:0x%02x\n", 
		i * 4 +   3, (unsigned char)_GetRegisterNT(i * 4 +    3), 
		i * 4 +  59, (unsigned char)_GetRegisterNT(i * 4 +   59), 
		i * 4 + 115, (unsigned char)_GetRegisterNT(i * 4 +  115), 
		i * 4 + 171, (unsigned char)_GetRegisterNT(i * 4 +  171)); 
  }

  printf("\n");
}

//-------------------------------------------------------------------------------------------
// get - and printout - static status information about the interface
//
static void GetStaticStatus(void)
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_STATIC_STATUS staticStatus;

	result = DeviceIoControl(hHandle, PCIVME_GET_STATIC_STATUS,
                NULL, 0, &staticStatus, (DWORD)sizeof(staticStatus), 
				   &DIOC_count, NULL);	 

	if (!result) 
		printf("PCIVME_GET_STATIC_STATUS error: 0x%08x\n",GetLastError());
	else
	{
		if (DeviceName(nIfcNumStat) != NULL)
		  printf("Driver-Path:\t%s\n", DeviceName(nIfcNumStat));
		else 
		  printf("Driver-Path:\t%s\n", "unknown");

		printf("Driver-Version:\t%d.%d\n", (staticStatus.dwDriverVersion >> 16) & 0xffff, 
			                                staticStatus.dwDriverVersion & 0xffff);
		printf("Driver-Variant:\t%d\n",	staticStatus.dwDriverVariant);

		printf("LinkCount:\t%d\n",     staticStatus.dwLinkCount); // how often this interface is requested  

		printf("PCI-Windows:\t%d\n",   staticStatus.wNumMemWindows);
		printf("PCI-IOPorts:\t%d\n",   staticStatus.wNumIOPorts);
		printf("PCI-Irqs:\t%d\n",      staticStatus.wNumIRQs);
		printf("PCI-DMAs:\t%d\n",      staticStatus.wNumDMAs);
		printf("Module-Type:\t%d\n",   staticStatus.wModuleType);
		printf("FPGA-Version:\t%d\n",  staticStatus.wFPGAVersion);
		printf("Module-Number:\t%d\n", staticStatus.wModuleNumber);
		printf("Word-Mode:\t%s\n",         (staticStatus.wWordMode) ? "yes" : "no");
        printf("System-Ctrl:\t%s\n", (staticStatus.wSysControl) ? "yes" : "no");
		printf("Connected:\t%s\n",         (staticStatus.wConnected)  ? "yes" : "no");
	}
}

//-------------------------------------------------------------------------------------------
// get - and print out - information about all found PCIADA
//
static void GetPCIADAStatus(void)
{
	DWORD	result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCIVME_DYNAMIC_STATUS dynamicStatus;

	result = DeviceIoControl(hHandle, PCIVME_GET_DYNAMIC_STATUS,
                NULL, 0, &dynamicStatus, (DWORD)sizeof(dynamicStatus), 
				   &DIOC_count, NULL);	 

	if (!result) 
		printf("PCIVME_GET_DYNAMIC_STATUS error: 0x%08x\n",GetLastError());
	else
	{
		printf("VMEMM connected:\t%s\n", (dynamicStatus.wVMEMM_connected) ? "yes" : "no");
		printf("VMEMM enabled:\t\t%s\n",   (dynamicStatus.wVMEMM_enable)    ? "yes" : "no");
		printf("PCIADA-Irq enabled:\t%s\n", (dynamicStatus.wPCIADAIrq) ? "yes" : "no");
		printf("VMEMM-Irq enabled:\t%s\n",  (dynamicStatus.wVMEMMIrq)  ? "yes" : "no");
	}
}

//-------------------------------------------------------------------------------------------
// print out some interface special info
//
void GetInterfaceInfoNT(char type)
{
  switch (type)
  {
    case '0': GetPCIADAStatus();
  	          break;
    case '1': GetStaticStatus();
	          break;
    case '2': GetDynamicStatus();
	          break;
    case '3': GetVIC68ARegs();
		      break;
  }
}

//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
