//-------------------------------------------------------------------------------------------
// pcivme_ni_95.c - a ni labview dll skeleton for the ARW pcivme interface, win95+98
//
// (c) 1999-2002 ARW Elektronik, Germany
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// what                                                              who    when
// first steps                                                       AR     07.11.1999
//

//-------------------------------------------------------------------------------------------
// INCLUDES
//
#include <windows.h>
#include <..\vpcivmed\vpcivmed.h>    // header for win-95..98
#include <..\vpcivmed\vic.h>
#include <..\vpcivmed\vme.h>
#include <klist.h>
#include <pcivme_ni_95.h>


//-------------------------------------------------------------------------------------------
// DEFINES
//
#define DEFDEVICENAME "\\\\.\\c:\\Windows\\system\\vpcivmed.vxd"

//-------------------------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct
{
  unsigned long  nInterfaceNumber;
  void           *base;   
  unsigned long  offset;
  unsigned long  size;
  unsigned short modifier;
  short          active;
} MYWINDOW;

typedef struct // a element associated to a open path (between VMEinit and VMEclose)
{
	HANDLE		nHandle;
	MYWINDOW	myWindow;
} OPEN_PATH_95;

#pragma pack(push, 1)
typedef struct                          // received from the callback
{
	unsigned char  vector;				// the vector coded as LSB
	unsigned char  reserved0;	
	unsigned char  level;				// next the interrupt level 
	unsigned char  reserved1;			
} IRQ_INFO;
#pragma pack(pop)

// this will change in the future
typedef struct
{
	int			  nHandle;
	void (_stdcall *IrqHandler)(IRQ_INFO irq_info);
} IRQ_DESC;

//-------------------------------------------------------------------------------------------
// PROTOTYPES
//
static void _stdcall IrqHandler0(IRQ_INFO irq_info);
static void _stdcall IrqHandler1(IRQ_INFO irq_info);
static void _stdcall IrqHandler2(IRQ_INFO irq_info);
static void _stdcall IrqHandler3(IRQ_INFO irq_info);
static void _stdcall IrqHandler4(IRQ_INFO irq_info);
static void _stdcall IrqHandler5(IRQ_INFO irq_info);
static void _stdcall IrqHandler6(IRQ_INFO irq_info);
static void _stdcall IrqHandler7(IRQ_INFO irq_info);
static void _stdcall IrqHandler8(IRQ_INFO irq_info);
static void _stdcall IrqHandler9(IRQ_INFO irq_info);
static void _stdcall IrqHandler10(IRQ_INFO irq_info);
static void _stdcall IrqHandler11(IRQ_INFO irq_info);
static void _stdcall IrqHandler12(IRQ_INFO irq_info);
static void _stdcall IrqHandler13(IRQ_INFO irq_info);
static void _stdcall IrqHandler14(IRQ_INFO irq_info);
static void _stdcall IrqHandler15(IRQ_INFO irq_info);

//-------------------------------------------------------------------------------------------
// LOCALS
//

// user initialisation table for pcivme
static VPCIVMED_INIT_COMMAND sUserInitStruct   = {2, {{STOP, WORD_ACCESS, 0,    0}}};            
// user deinitialisation table
static VPCIVMED_INIT_COMMAND sUserDeInitStruct = {2, {{STOP, WORD_ACCESS, 0,    0}}};

// the list of Path specific data (a element lives between VMEinit and VMEclose) 
static LIST liPathList		= (LIST)NULL;

// the indicators of a raised BusError
BOOLEAN BusError[VPCIVMED_MAX_VMEMM] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

// a fast lookup for irq-handlers (this will change in the future)
static IRQ_DESC irq_desc[VPCIVMED_MAX_VMEMM] = {	
	{-1, IrqHandler0},
	{-1, IrqHandler1},
	{-1, IrqHandler2},
	{-1, IrqHandler3},
	{-1, IrqHandler4},
	{-1, IrqHandler5},
	{-1, IrqHandler6},
	{-1, IrqHandler7},
	{-1, IrqHandler8},
	{-1, IrqHandler9},
	{-1, IrqHandler10},
	{-1, IrqHandler11},
	{-1, IrqHandler12},
	{-1, IrqHandler13},
	{-1, IrqHandler14},
	{-1, IrqHandler15}};

//-------------------------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------------------------
// GLOBALS
//

//-------------------------------------------------------------------------------------------
// FUNCTIONS
//
// not only delete a element - even remove the whole list if it is empty
static void removeListElement(OPEN_PATH_95 *open_path)
{
	List_Delete((LPVOID)open_path);

	// remove the list if the last item was deleted ----
	if ((liPathList != NULL) && (List_IsEmpty(liPathList) == TRUE)) 
	{
		List_Destroy(&liPathList);
		liPathList = (LIST)NULL;
	}
}

static void _stdcall IrqHandler(IRQ_INFO irq_info, int nHandle)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
    VPCIVMED_IRQ_CONTROL sIrqControl;

	if ((nHandle != 0) && (nHandle != -1))
	{
		// different handling for bus errors -----------------------------------
		if ((irq_info.vector == 7) || (irq_info.vector == 1))
		  BusError[open_path->myWindow.nInterfaceNumber] = 1;
		else
		{
			// at this time: don't knoe what to do with vectors
		}

		// re-enable interrupts #1 ---------------------------------------------
		sIrqControl.dwInterface    = open_path->myWindow.nInterfaceNumber;
		sIrqControl.wEnable        = 1;  // enable
		DeviceIoControl(open_path->nHandle, VPCIVMED_CONTROL_INTERRUPTS,
							&sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
								0, NULL, NULL);
	}
}

//--------------------------------------------------------------------------
// 16 different handlers (because of different irq sources)
// this will be changed in future versions
//
static void _stdcall IrqHandler0(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[0].nHandle);
}

static void _stdcall IrqHandler1(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[1].nHandle);
}

static void _stdcall IrqHandler2(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[2].nHandle);
}

static void _stdcall IrqHandler3(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[3].nHandle);
}

static void _stdcall IrqHandler4(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[4].nHandle);
}

static void _stdcall IrqHandler5(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[5].nHandle);
}

static void _stdcall IrqHandler6(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[6].nHandle);
}

static void _stdcall IrqHandler7(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[7].nHandle);
}

static void _stdcall IrqHandler8(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[8].nHandle);
}

static void _stdcall IrqHandler9(IRQ_INFO irq_info)
{
	IrqHandler(irq_info,  irq_desc[9].nHandle);
}

static void _stdcall IrqHandler10(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[10].nHandle);
}

static void _stdcall IrqHandler11(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[11].nHandle);
}

static void _stdcall IrqHandler12(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[12].nHandle);
}

static void _stdcall IrqHandler13(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[13].nHandle);
}

static void _stdcall IrqHandler14(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[14].nHandle);
}

static void _stdcall IrqHandler15(IRQ_INFO irq_info)
{
	IrqHandler(irq_info, irq_desc[15].nHandle);
}

// create a DeviceName out of cszDeviceName and nIfcNum
static char *DeviceName(const char *cszDeviceName, int nIfcNum)
{
	static char buffer[255];
	char *ptr = buffer;

	if (cszDeviceName == NULL)
		strcpy(buffer, DEFDEVICENAME);
	else
		strcpy(buffer, cszDeviceName);

    return buffer;
}

// attach a window to the opened path
static void *attachWindow(int nHandle, unsigned long offset, unsigned short modifier)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_ADD_WINDOW sAWnd;    // attach-window structure
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	void  *WinBase;
	DWORD result;
	DWORD size;

	if (nHandle == -1) return NULL;

	// determine a size to a given modifier
	switch (modifier & 0x30)
	{
		case 0x00: 
		case 0x10: size = 0x08000000; break;
		case 0x20: size = 0x00010000; break;
		case 0x30: size = 0x01000000; break;
	}

	// attach a window into the VME space ----------------
	sAWnd.dwInterface = open_path->myWindow.nInterfaceNumber;
	sAWnd.size        = ((size + 0x0FFF) / 0x1000) * 0x1000; // truncate to chunks of 0x1000
	sAWnd.modifier    = modifier;
	sAWnd.base        = offset & HI_ADDRESS_MASK;
	result = DeviceIoControl(open_path->nHandle, VPCIVMED_ATTACH_WINDOW,
						&sAWnd, sizeof(sAWnd), &WinBase, sizeof(WinBase),
								 &DIOC_count, NULL);
	if (!result)
      return NULL;

	open_path->myWindow.base     = WinBase;
	open_path->myWindow.offset   = offset & HI_ADDRESS_MASK;
	open_path->myWindow.size     = size;
	open_path->myWindow.modifier = modifier;
	open_path->myWindow.active   = 1;

	return open_path->myWindow.base;
}

// release a vme window
static int detachWindow(int nHandle)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_REMOVE_WINDOW sRWnd; // remove-window structure
	DWORD DIOC_count;             // count of returned bytes of DeviceIoControl
	DWORD result;
	void  *WinBase;

	if (nHandle == -1) return -1;

	// detach the window into VME space
	sRWnd.dwInterface = open_path->myWindow.nInterfaceNumber;
	sRWnd.pdwLinAdr   = (DWORD *)open_path->myWindow.base;
	result = DeviceIoControl(open_path->nHandle, VPCIVMED_DETACH_WINDOW,
						&sRWnd, sizeof(sRWnd), &WinBase,
							sizeof(WinBase), &DIOC_count, NULL);

	if (!result)
		return GetLastError();
	else
	{
		open_path->myWindow.active = 0;
		return 0;
	}
}

// checks if the access is outside the current window
static void *checkWindow(int nHandle, unsigned long offset, unsigned long ulCount, unsigned long elementSize)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	void *pvBase;

	if ((offset   >=  open_path->myWindow.offset ) && 
	    (offset   <= (open_path->myWindow.offset + open_path->myWindow.size - (elementSize * ulCount))) &&
	    (open_path->myWindow.active))
	return open_path->myWindow.base;

  // remove old window into vme ---------------------------------------
  if (open_path->myWindow.active) detachWindow(nHandle);

  pvBase = attachWindow(nHandle, offset, open_path->myWindow.modifier);

  if (pvBase == NULL)
	  return NULL;
  else
  {
	  open_path->myWindow.base = pvBase;
	  return pvBase;
  }	
}

// fast copy routines - faster with assembled functions ---
static void memcpyByte(void *dst, void *src, register unsigned long count)
{
	register char *d = (char *)dst;
	register char *s = (char *)src;

	while (count--) *d++ = *s++; 
}

static void memcpyShort(void *dst, void *src, register unsigned long count)
{
	register short *d = (short *)dst;
	register short *s = (short *)src;

	while (count--) *d++ = *s++; 
}

static void memcpyLong(void *dst, void *src, register unsigned long count)
{
	register long *d = (long *)dst;
	register long *s = (long *)src;

	while (count--) *d++ = *s++; 
}

//----------------------------------------------------------------------------------
// exportable functions
//
int VMEinit95(const char *cszDeviceName, unsigned short nVMEMM, unsigned char ubAddressModifier, int *pnHandle)
{
	OPEN_PATH_95 *open_path;
	DWORD	DIOC_count;			// count of returned bytes of DeviceIoControl
	DWORD	result;
	void    *baseAdr;

	VPCIVMED_IRQ_HANDLER     sIrqHandler;  // structure for irq handler handling
	VPCIVMED_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs

    if (liPathList == NULL)		// create a list to hold the paths and its variables
	{
		liPathList = List_Create();
		if (liPathList == (LIST)NULL)
			return GetLastError();
	}
	
	open_path = (OPEN_PATH_95 *)List_NewFirst(liPathList, sizeof(OPEN_PATH_95));

	*pnHandle = -1;

    open_path->nHandle = CreateFile( 
						DeviceName(cszDeviceName, nVMEMM), 
						0, 
						0, 
						NULL, 
						0, 
						FILE_FLAG_DELETE_ON_CLOSE, 
						NULL);
	
	if (open_path->nHandle != INVALID_HANDLE_VALUE) 
	{ 
		open_path->myWindow.nInterfaceNumber = nVMEMM;

		// init hardware (only one time after the first init it works OK)
		sUserInitStruct.dwInterface = open_path->myWindow.nInterfaceNumber;
		result = DeviceIoControl(open_path->nHandle, 
						VPCIVMED_INIT_HARDWARE,
						&sUserInitStruct, 
						(DWORD)sizeof(sUserInitStruct), 
						NULL,
						(DWORD)0, 
						&DIOC_count, 
						NULL);

		if (!result) 
		{
			result = GetLastError();
			CloseHandle(open_path->nHandle); 
			removeListElement(open_path);
			return result;
		}

		open_path->myWindow.active = 0;

		// set the current access parameters ------------------
		baseAdr = attachWindow((int)open_path, 0, ubAddressModifier);

		if (baseAdr == NULL) 
		{
			result = GetLastError();
			CloseHandle(open_path->nHandle); 
			removeListElement(open_path);
			return result;
		}

		sIrqHandler.dwInterface  = open_path->myWindow.nInterfaceNumber;
		sIrqHandler.dwIrqHandler = 
			(DWORD)irq_desc[open_path->myWindow.nInterfaceNumber].IrqHandler;
		result = DeviceIoControl(open_path->nHandle, VPCIVMED_INSTALL_IRQ_HANDLER,
							&sIrqHandler, (DWORD)sizeof(sIrqHandler), NULL,
								0, NULL, NULL);
		if (!result) 
		{
			result = GetLastError();
			CloseHandle(open_path->nHandle); 
			removeListElement(open_path);
			return result;
		}

		// enable interrupts ----------------------------------------------------
		sIrqControl.dwInterface    = open_path->myWindow.nInterfaceNumber;
		sIrqControl.wEnable        = 1;  // enable interrupt
		result = DeviceIoControl(open_path->nHandle, VPCIVMED_CONTROL_INTERRUPTS,
							&sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
								0, NULL, NULL);
		if (!result) 
		{
			result = GetLastError();
			CloseHandle(open_path->nHandle); 
			removeListElement(open_path);
			return result;
		}
 
		*pnHandle											    = (int)open_path;
		irq_desc[open_path->myWindow.nInterfaceNumber].nHandle	= (int)open_path;

		return 0;
	}
	else 
	{
		result = GetLastError();
		removeListElement(open_path);
		return result;
	}
}

int VMEread95(int nHandle, unsigned long ulAddress, unsigned char ubAccessWidth, unsigned long ulElementCount, void *pvBuffer)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;

	if (nHandle == -1) return -1;

    if (checkWindow(nHandle, ulAddress, ulElementCount, ubAccessWidth) != NULL)
	{
		void *srcAdr = (void *)((unsigned long)open_path->myWindow.base 
										+ (ulAddress  - open_path->myWindow.offset));
		
		switch (ubAccessWidth)
		{
			case sizeof(char):	memcpyByte(pvBuffer, srcAdr, ulElementCount);  break;
			case sizeof(short): memcpyShort(pvBuffer, srcAdr, ulElementCount); break;
			case sizeof(long):  memcpyLong(pvBuffer, srcAdr, ulElementCount);  break;
			default: return -1;
		}

		SleepEx(0, TRUE);  // give the interrupt-thread a chance
		if (BusError[open_path->myWindow.nInterfaceNumber])
		{
			BusError[open_path->myWindow.nInterfaceNumber] = 0;
			return STATUS_ACCESS_VIOLATION;
		}
		else
			return 0;
	}

	return -1;
}

int VMEwrite95(int nHandle, unsigned long ulAddress, unsigned char ubAccessWidth, unsigned long ulElementCount, void *pvBuffer)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;

	if (nHandle == -1) return -1;

    if (checkWindow(nHandle, ulAddress, ulElementCount, ubAccessWidth) != NULL)
	{
		void *dstAdr = (void *)((unsigned long)open_path->myWindow.base 
										+ (ulAddress  - open_path->myWindow.offset));
		
		switch (ubAccessWidth)
		{
			case sizeof(char):	memcpyByte(dstAdr, pvBuffer, ulElementCount);  break;
			case sizeof(short): memcpyShort(dstAdr, pvBuffer, ulElementCount); break;
			case sizeof(long):  memcpyLong(dstAdr, pvBuffer, ulElementCount);  break;
			default: return -1;
		}

		SleepEx(0, TRUE);  // give the interrupt-thread a chance
		if (BusError[open_path->myWindow.nInterfaceNumber])
		{
			BusError[open_path->myWindow.nInterfaceNumber] = 0;
			return STATUS_ACCESS_VIOLATION;
		}
		else
			return 0;
	}

	return -1;
}

int VMEaccessVIC95(int nHandle, unsigned char ubAccessMode, unsigned short uwAddress, unsigned char *ubContent)
{
	DWORD	 DIOC_count;     // count of returned bytes of DeviceIoControl
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_VIC68A_ACTION sAction;    // structure to access vic chip
	DWORD                  result;

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	sAction.dwInterface      = open_path->myWindow.nInterfaceNumber;
	sAction.wRegisterAddress = uwAddress;
	sAction.wAccessMode      = ubAccessMode;
	sAction.bContent         = *ubContent;

	result = DeviceIoControl(open_path->nHandle, VPCIVMED_ACCESS_VIC68A,
					 &sAction, sizeof(sAction), &sAction,
							sizeof(sAction), &DIOC_count, NULL);	 

	*ubContent = sAction.bContent;

	if (!result) 
		return GetLastError();
	else
		return 0;
}

int VMEreset95(int nHandle)
{
	DWORD	 DIOC_count;     // count of returned bytes of DeviceIoControl
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	DWORD	result;
	VPCIVMED_RESET_COMMAND reset_command;
	VPCIVMED_RESET_RESULT  reset_result;

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	reset_command.dwInterface      = open_path->myWindow.nInterfaceNumber;
	reset_command.wCommand = VME_RESET_CMD;
	reset_result.wResult   = 0;
	result = DeviceIoControl(open_path->nHandle, VPCIVMED_RESET,
                &reset_command, sizeof(reset_command), 
				&reset_result,  sizeof(reset_result), 
				   &DIOC_count, NULL);	 

	if (!result) 
		return GetLastError();
	else
	{
		result = 1;

		while (reset_result.wResult && result)
		{
			reset_command.dwInterface   = open_path->myWindow.nInterfaceNumber;
			reset_command.wCommand		= POLL_RESET_CMD;
			result = DeviceIoControl(open_path->nHandle, VPCIVMED_RESET,
						&reset_command, sizeof(reset_command), 
						&reset_result,  sizeof(reset_result), 
							&DIOC_count, NULL);	 
			Sleep(10);
		}
	}

	if (!result)
		return GetLastError();
	else
		return 0;
}

int VMETAS95(int nHandle, unsigned long ulAddress, unsigned char *ubResult)
{
	DWORD	 DIOC_count;     // count of returned bytes of DeviceIoControl
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_TAS_STRUCT    sTAS;       // structure to do a Test and Set
	DWORD                  result;

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	sTAS.dwInterface	  = open_path->myWindow.nInterfaceNumber;
	sTAS.wModifier        = open_path->myWindow.modifier;
	sTAS.dwAddress        = ulAddress;
	sTAS.bContent         = 0x80;

	result = DeviceIoControl(open_path->nHandle, VPCIVMED_TAS,
					 &sTAS, (DWORD)sizeof(sTAS), &sTAS,
							(DWORD)sizeof(sTAS), &DIOC_count, NULL); 

	*ubResult = sTAS.bContent; 

	if (!result)
		return GetLastError();
	else
	{
		SleepEx(0, TRUE);  // give the interrupt-thread a chance
		if (BusError[open_path->myWindow.nInterfaceNumber])
		{
			BusError[open_path->myWindow.nInterfaceNumber] = 0;
			return STATUS_ACCESS_VIOLATION;
		}
	}

	return 0;
}

int VMEinterrupt95(int nHandle, unsigned char *ubVector)
{
	return 0;
}

int VMEsysfailGet95(int nHandle, BOOLEAN *bResult)
{
	DWORD	 DIOC_count;     // count of returned bytes of DeviceIoControl
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_VIC68A_ACTION  sAction;    // structure to access vic chip
	DWORD result;

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	sAction.dwInterface		 = open_path->myWindow.nInterfaceNumber;
	sAction.wRegisterAddress = EGICR;
	sAction.wAccessMode      = VIC68A_READ;
	sAction.bContent         = 0;

	result = DeviceIoControl(open_path->nHandle, VPCIVMED_ACCESS_VIC68A,
					 &sAction, sizeof(sAction), &sAction,
							sizeof(sAction), &DIOC_count, NULL);	 


	*bResult = (sAction.bContent & 0x08) ? FALSE : TRUE;

	if (!result)
		return GetLastError();
	else
		return 0;
}

int VMEsysfailSet95(int nHandle, BOOLEAN bForce)
{
	DWORD	 DIOC_count;     // count of returned bytes of DeviceIoControl
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	VPCIVMED_VIC68A_ACTION  sAction;    // structure to access vic chip
	DWORD result;

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	sAction.dwInterface		 = open_path->myWindow.nInterfaceNumber;
	sAction.wRegisterAddress = ICR7;
	sAction.wAccessMode      = (bForce == TRUE) ? VIC68A_AND : VIC68A_OR;
	sAction.bContent         = (bForce == TRUE) ? 0x3F		 : 0x80;

	result = DeviceIoControl(open_path->nHandle, VPCIVMED_ACCESS_VIC68A,
					 &sAction, (DWORD)sizeof(sAction), &sAction,
							(DWORD)sizeof(sAction), &DIOC_count, NULL);
	if (!result)
		return GetLastError();
	else
		return 0;
}

int VMEclose95(int nHandle)
{
	OPEN_PATH_95 *open_path = (OPEN_PATH_95 *)nHandle;
	DWORD result;
	DWORD DIOC_count;              // count of returned bytes of DeviceIoControl

	VPCIVMED_IRQ_HANDLER     sIrqHandler;  // structure for irq handler handling
	VPCIVMED_IRQ_CONTROL     sIrqControl;  // enable and disable of irqs

	if (nHandle == -1) return ERROR_PATH_NOT_FOUND;

	// disable interrupts    ----------------------------------------------
	sIrqControl.dwInterface    = open_path->myWindow.nInterfaceNumber;
	sIrqControl.wEnable        = 0;  // disable both
	result = DeviceIoControl(open_path->nHandle, VPCIVMED_CONTROL_INTERRUPTS,
						&sIrqControl, (DWORD)sizeof(sIrqControl), NULL,
							0, NULL, NULL);
	if (!result)
	{
		// don't know what to do
	}
	else
	{
		// deinstall the irq-handler --------------------------------------------
		sIrqHandler.dwInterface  = open_path->myWindow.nInterfaceNumber;
		sIrqHandler.dwIrqHandler = (DWORD)BOGUSADDRESS;
		result = DeviceIoControl(open_path->nHandle, VPCIVMED_INSTALL_IRQ_HANDLER,
							&sIrqHandler, (DWORD)sizeof(sIrqHandler), NULL,
								0, NULL, NULL);
		if (!result)
		{
			// don't know what to do
		}
		else
			irq_desc[open_path->myWindow.nInterfaceNumber].nHandle = -1;
	}

	// detatch window
	detachWindow(nHandle);

	// deinitialize the hardware --------------------------
	sUserDeInitStruct.dwInterface = open_path->myWindow.nInterfaceNumber;
	result = DeviceIoControl(open_path->nHandle, VPCIVMED_DEINIT_HARDWARE,
					&sUserDeInitStruct, (DWORD)sizeof(sUserDeInitStruct), NULL,
						   0, &DIOC_count, NULL);	 

	// close device -------------------------------------
	CloseHandle(open_path->nHandle);
	removeListElement(open_path);

	return 0;
}

