//-------------------------------------------------------------------------
//
// a universal menuing functions - menu functions itself
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and
// modify it unless this header  is  not modified or deleted. No warranty is given that
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// first steps						AR     22.11.97
//
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <alloc.h>
#include <menu.h>
#include <screen.h>
#include <tastatur.h>

//---------------- einige feste lokationen auf dem bildschirm -----
#define X_POS_MENU   1
#define X_POS_PRINT  (sc_maxxpos() - 20)

#define Y_POS_HEADER 1
#define Y_POS_MENU   4
#define Y_POS_QUIT   (sc_maxypos() - 2)
#define Y_POS_STATUS (sc_maxypos())

//----------------- rund um die statuszeile ------------------------
#define STATUS_ZEILEN_LAENGE 40

static char status_zeile[STATUS_ZEILEN_LAENGE];

void set_status_zeile(char *zeile)
{
  int len = strlen(zeile);
  if (len >= STATUS_ZEILEN_LAENGE) len = STATUS_ZEILEN_LAENGE - 1;

  strncpy(status_zeile, zeile, STATUS_ZEILEN_LAENGE);
  status_zeile[len] = 0;
}

static void print_status_zeile(void)
{
  sc_goto(X_POS_MENU, Y_POS_STATUS);
  sc_clreol();
  sc_high();
  printf("%s", status_zeile);
  sc_low();
  set_status_zeile("");                          // leer neu besetzen
}

static void print_header_zeile(char *zeile)
{
  if (zeile != NULL)
  {
    sc_goto(X_POS_MENU, Y_POS_HEADER);
    sc_high();
    printf("%s", zeile);
    sc_low();
  }
}

//-------------------------------------------------------------------
// ein einfaches menu
//
void mn_menu(char *menu_header, int (*entry)(void), MN_MENU_ENTRY menues[])
{
  MN_MENU_ENTRY *mn;
  int  abbruch = 0;
  int  zeichen;
  char eingabe;
  int  found;
  int  i;

  // entry funktion ausfuehren ------------
  sc_clear();

  if (entry())
    return;

  set_status_zeile("");
  ts_leer_les();

  do
  {
    sc_clear();
    sc_home();

    print_header_zeile(menu_header);

    i = Y_POS_MENU;
    mn = menues;
    while (mn->wahl_zeichen)
    {
      if (tolower(mn->wahl_zeichen) != 'q')
      {
	sc_goto(X_POS_MENU,i);
	sc_clreol();

	if (mn->wahl_zeichen != -1)
	  printf("<%c> - %s", mn->wahl_zeichen, mn->wahl_string);
	else
	  printf("      %s",mn->wahl_string);   // keine wahl, nur status

	if (mn->print != NULL)
	{
	  sc_goto(X_POS_PRINT,i);
	  mn->print();
	}
	printf("\n");
	i++;
      }
      mn++;
    }

    sc_goto(X_POS_MENU,Y_POS_QUIT);
    sc_clreol();
    mn = menues;
    while (mn->wahl_zeichen)
    {
      if (tolower(mn->wahl_zeichen) == 'q')
	printf("<%c> - %s\n", mn->wahl_zeichen, mn->wahl_string);
      mn++;
    }

    //----------- statuszeile ausgeben -----------------
    print_status_zeile();   // ausgeben

    //--------- eingabe abfragen -----------------------
    sc_goto(X_POS_MENU,Y_POS_STATUS);
    eingabe = tolower(ts_getch());

    //--------- aktion ausfuehren ----------------------
    i     = Y_POS_MENU;
    mn    = menues;
    found = 0;
    while (mn->wahl_zeichen)
    {
      if (mn->wahl_zeichen != -1)  // reine anzeige hat keine mn->aktion()
      {
	if ((tolower(mn->wahl_zeichen) == eingabe) ||
	   ((tolower(mn->wahl_zeichen) == 'q') && (eingabe == ESCAPE)))
	{
	  sc_goto(X_POS_PRINT,i);
	  abbruch = mn->aktion(); // pos von mn->print()
	  found   = 1;
	  break;
	}
      }
      mn++;
      i++;
    }

    //------------ status_string erzeugen ------------------------------
    if ((found == 0) && (eingabe != BLANK) && (eingabe != RETURN))
    {
      char buffer[STATUS_ZEILEN_LAENGE];

      sprintf(buffer,"Illegal Input: '%c'!", eingabe);
      set_status_zeile(buffer);
    }

  } while (abbruch == 0);
}

//----------------------------------------------------------------------
// eine abfrage zwischen 2 alternativen
// abbruch liefert 0 zurueck
//
int mn_abfrage_bool(char *frage_header, char *frage,
				    char true_wert, char false_wert,
							int *ergebnis)
{
  char eingabe;

  while (1)
  {
    sc_clear();
    sc_home();

    print_header_zeile(frage_header);

    sc_goto(X_POS_MENU, Y_POS_MENU);
    printf("%s [%c/%c]\n",frage,true_wert,false_wert);
    sc_goto(X_POS_PRINT, Y_POS_MENU);

    print_status_zeile();

    //--------- eingabe abfragen -----------------------
    eingabe = tolower(ts_getche());

    if (eingabe == tolower(true_wert))
    {
      *ergebnis = 1;
      return 1;
    }

    if (eingabe == tolower(false_wert))
    {
      *ergebnis = 0;
      return 1;
    }

    if (eingabe == ESCAPE)
      return 0;
    else
    {
      char *buff;

      if ((buff = (char*)malloc(STATUS_ZEILEN_LAENGE)) == NULL)
      {
	set_status_zeile("Malloc failed!");
	return 0;
      }

      sprintf(buff,"Illegal Input: '%c'!", eingabe);
      set_status_zeile(buff);

      free(buff);
      return 0;
    }
  }
}

//--------------------------------------------------------------------
// eingabe vor ort in einem menu (an der print-position)
// abbruch liefert 0 zurueck
//
int mn_abfrage_input_inplace(char *format, void *buffer)
{
  char *buf;
  int  x_pos, y_pos;

  sc_getpos(&x_pos,&y_pos);

  if ((buf = (char*)malloc(MAXBUFSIZE)) == NULL)
  {
    set_status_zeile("Malloc failed!");
    return 0;
  }

  print_status_zeile();

  if(ts_edit_field(x_pos, y_pos, buf))
  {
    if (sscanf(buf,format,buffer) != 1)
    {
      free(buf);
      return 0;
    }
    else
    {
      free(buf);
      return 1;
    }
  }
  else
  {
    set_status_zeile("Input break.");
    free(buf);
    return 0;
  }
}

//----------------------------------------------------------------------
// fragt die eingabe ohne weitere interpretation ab, alle editiermoeglichk.
// vorhanden
// abbruch liefert 0 zurueck
//
int mn_abfrage_input_text_inplace(void *buffer)
{
  int x_pos, y_pos;

  sc_getpos(&x_pos, &y_pos);

  print_status_zeile();

  if(ts_edit_field(x_pos, y_pos, buffer))
    return 1;
  else
  {
    set_status_zeile("Input break.");
    return 0;
  }
}

//----------------------------------------------------------------------
// eine allgemeine abfrage nach beliebigem formatierten input
// es ist nur das lesen eines elementes mglich
// abbruch liefert 0 zurueck
//
int mn_abfrage_input_element(char *frage_header, char *frage,
				       char *format, void *buffer)
{
  char *buf;

  if ((buf = (char*)malloc(MAXBUFSIZE)) == NULL)
  {
    set_status_zeile("Malloc failed!");
    return 0;
  }

  sc_clear();
  sc_home();

  print_header_zeile(frage_header);

  sc_goto(X_POS_MENU, Y_POS_MENU);
  printf("%s ",frage);

  print_status_zeile();

  sc_goto(X_POS_PRINT, Y_POS_MENU);

  if (mn_abfrage_input_inplace(format, buffer))
  {
    free(buf);
    return 1;
  }
  else
  {
    char *buff;

    if ((buff = (char*)malloc(STATUS_ZEILEN_LAENGE)) == NULL)
    {
      set_status_zeile("Malloc failed!");
      free(buf);
      return 0;
    }

    if (frage_header != NULL)
    {
      sprintf(buff,"Fehler bei \"%s\"!", frage_header);
      set_status_zeile(buff);
    }
    else
      set_status_zeile("Fehler bei Abfrage!");

    free(buff);
    free(buf);
    return 0;
  }
}

//-------------------------------------------------------------------
// eine abfrage mit mehreren auswahlen
// das ergebnis liefert das eingabezeichen zurueck
// abbruch liefert 0 zurueck
//
int mn_abfrage_multiple(char *header_zeile, MN_WAHL_ENTRY altern[],
							   int *ergebnis)
{
  int i;
  MN_WAHL_ENTRY *al = altern;
  int found = 0;
  char eingabe;

  if ((al == NULL) || (al->wahl_zeichen == 0))         // keine wahl?
  {
    set_status_zeile("Keine Auswahl mglich!");
    return 0;
  }

  do
  {
    sc_clear();

    print_header_zeile(header_zeile);

    i = Y_POS_MENU;
    al = altern;
    while (al->wahl_zeichen != 0)
    {
      sc_goto(X_POS_MENU, i);
      printf("<%c> - %s", al->wahl_zeichen, al->wahl_string);
      i++;
      al++;
    }

    print_status_zeile();

    //--------- eingabe abfragen -----------------------
    sc_goto(X_POS_MENU,Y_POS_STATUS);
    eingabe = tolower(ts_getch());

    //--------- aktion ausfuehren ----------------------
    if (eingabe == ESCAPE)
      return 0;

    al = altern;
    while (al->wahl_zeichen != 0)
    {
      if (tolower(al->wahl_zeichen) == eingabe)
      {
	found = 1;
	break;
      }
      al++;
    }

    if ((eingabe != RETURN) && (eingabe != BLANK) && (found == 0))
    {
      char buff[STATUS_ZEILEN_LAENGE];

      sprintf(buff, "Illegal Input: '%c'!", eingabe);
      set_status_zeile(buff);
    }

  } while (found == 0);

  *ergebnis = eingabe;    // gebe das ergebnis zurueck

  return 1;
}

//---------------------------------------------------------------------
// zeigt repetierend einen status zeilenweise an
// pause = delay zwischen anzeigen in msec,
// 0 = anhalten nach jedem durchlauf
// die angabe leerer strings ("") erlauben die zyklische ausfuehrung
// beliebiger funktionen
//
void mn_show_status(char *header, int (*entry)(void),
				      MN_SHOW_ENTRY status[], int pause)
{
  MN_SHOW_ENTRY *stat;
  int i;
  char eingabe;

  sc_clear();

  print_header_zeile(header);

  i = Y_POS_MENU;
  stat = status;
  while (stat->show_text != NULL)
  {
    sc_goto(X_POS_MENU + 6,i);
    printf("%s\n" ,stat->show_text);   // keine wahl, nur status
    stat++;
    i++;
  }

  if (pause == 0)
  {
    sc_goto(X_POS_MENU, Y_POS_QUIT - 1);
    printf("<?> - Read again");
  }

  sc_goto(X_POS_MENU, Y_POS_QUIT);
  printf("<q> - Quit", header);

  print_status_zeile();

  do
  {
    if (entry())
      return;

    stat = status;
    i = Y_POS_MENU;
    while (stat->show_print != NULL)
    {
      sc_goto(X_POS_PRINT, i);
      sc_clreol();
      stat->show_print();
      stat++;
      i++;
    };


    if ((!pause) || ts_kbhit())
    {
      sc_goto(X_POS_MENU, Y_POS_STATUS);

      eingabe = tolower(ts_getch());

      if ((eingabe == ESCAPE) || (eingabe == 'q'))
	return;
    }

    if (pause)
      sc_delay(pause);

  } while (1);
}

//---------------------------------------------------------------------
// bringt eine seite mit kommandozeileneingabe zum benutzer
// q,quit,exit,bye und ESCAPE sind als kommando nicht erlaubt
// obige kommamdos brechen die eingabe ab
//
void mn_kommandozeile(char *header,
		       int (*entry)(void), IN_INTERPRET interpret[])
{
  IN_INTERPRET *in_in = interpret;
  char *buf;

  sc_clear();

  if (entry())
    return;

  if ((buf = (char*)malloc(MAXBUFSIZE)) == NULL)
  {
    set_status_zeile("Malloc failed!");
    return;
  }

  do
  {
    sc_clear();

    print_header_zeile(header);

    sc_goto(X_POS_MENU, Y_POS_MENU);

    if (mn_abfrage_input_text_inplace(buf))
    {
      if ((strcmp(buf, "q")    == 0) ||
	  (strcmp(buf, "quit") == 0) ||
	  (strcmp(buf, "exit") == 0) ||
	  (strcmp(buf, "bye")  == 0))
	break;

      if (in_interpret_input(buf, in_in) == 0)
	set_status_zeile("Unbekanntes Kommando (?)!");
    }
    else
      break;

  } while (1);

  free(buf);
}