//-------------------------------------------------------------------------
//
// a universal menuing functions - keyboard input functions
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and
// modify it unless this header  is  not modified or deleted. No warranty is given that
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// first steps						AR     2.11.97
//
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <malloc.h>
#include <screen.h>
#include <tastatur.h>

//------------------------------------------------------------------
#define HISTORYSIZE 10

static char *history[HISTORYSIZE];
static int with_history = 0;
static int hist_read;
static int hist_write;
static int hist_entries;

//-------------------------------------------------------------------------
// initialisiert die tastatur
void ts_init(void)
{
  char *buf;
  int i;

  if ((buf = (char *)malloc(MAXBUFSIZE * HISTORYSIZE)) == NULL)
    return;

  for (i = 0; i < HISTORYSIZE; i++)
  {
    history[i]  = buf + i * MAXBUFSIZE;
    *history[i] = 0;
  }

  with_history = 1;
  hist_write = hist_read = hist_entries = 0;
}

//--------------------------------------------------------------------------
// deinitialisiert die tastatur
void ts_deinit(void)
{
  if (with_history == 1)
    free(history[0]);

  with_history = 0;
  hist_write = hist_read = hist_entries = 0;
}

//--------------------------------------------------------------------------
// testet auf cursor-sequenzen, 1 wenn erkannt
//
int test_cursor_right(char *c)
{
  static char b = -1;

  if ((*c == 'M') && (b == 0))
    return 1;

  b = *c;
  return 0;
}

int test_cursor_left(char *c)
{
  static char b = -1;

  if ((*c == 'K') && (b == 0))
    return 1;

  b = *c;
  return 0;
}

int test_cursor_up(char *c)
{
  static char b = -1;

  if ((*c == 'H') && (b == 0))
    return 1;

  b = *c;
  return 0;
}

int test_cursor_down(char *c)
{
  static char b = -1;

  if ((*c == 'P') && (b == 0))
    return 1;

  b = *c;
  return 0;
}


//--------------------------------------------------------------------------
// schaut nach, ob eine taste gedrueckt ist
int ts_kbhit(void)
{
  return kbhit();
}

//--------------------------------------------------------------------------
// liest den tastatur-puffer leer
extern void ts_leer_les(void)
{
  while (ts_kbhit()) getch();
}

//--------------------------------------------------------------------------
// liest ohne CR eine Zeichen von stdin ohne echo
char ts_getch(void)
{
  int zeichen;

  zeichen = getch();
  if (zeichen == EOF)
    return 0;
  else
    return (char)zeichen;
}

//--------------------------------------------------------------------------
// liest ohne CR eine Zeichen von stdin mit echo
char ts_getche(void)
{
  int zeichen;

  zeichen = getche();
  if (zeichen == EOF)
    return 0;
  else
    return (char)zeichen;
}


//---------------------------------------------------------------------
// ein tastaturbuffer-editor
//
int ts_edit_field(int x_pos, int y_pos, void *buffer)
{
  int i;
  char buf[MAXBUFSIZE];
  char c;

  sc_goto(x_pos, y_pos);
  sc_clreol();

  i = -1;
  do
  {
    i++;
    sc_goto(x_pos + i, y_pos);
    c = buf[i] = ts_getche();

    if (c == BACKSPACE)
    {
      if (i > 0)
	i -= 2;
      else
	i--;
      printf(" \b");
    }

    if (with_history)
    {
      if (test_cursor_up(&c))
      {
	if (--hist_read < 0)
	  hist_read = hist_entries - 1;
	strcpy(buf, history[hist_read]);
	sc_goto(x_pos, y_pos);
	sc_clreol();
	printf(buf);
	i = strlen(buf) - 1;
      }

      if (test_cursor_down(&c))
      {
	if (++hist_read >= hist_entries)
	  hist_read = 0;
	strcpy(buf, history[hist_read]);
	sc_goto(x_pos, y_pos);
	sc_clreol();
	printf(buf);
	i = strlen(buf) - 1;
      }
    }
  } while ((c != ESCAPE) && (c != '\r') && (i < (MAXBUFSIZE - 2)));
  buf[i] = 0;

  if ((c == ESCAPE) || (i >= (MAXBUFSIZE - 1)))
    return 0;

  strcpy(buffer,buf);

  if (with_history)
  {
    if (strcmp(history[hist_read], buf))
    {
      hist_entries++;
      if (hist_entries >= HISTORYSIZE)
	hist_entries = HISTORYSIZE;
    }

    strcpy(history[hist_write], buf);
    hist_write++;
    if (hist_write >= HISTORYSIZE)
      hist_write = 0;
    hist_read = hist_write;
  }

  return 1;
}