//-------------------------------------------------------------------------
//
// pcibios support functions - main of a simple test program
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and
// modify it unless this header  is  not modified or deleted. No warranty is given that
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// Erste Version released                               AR     31.10.97
//

#define VERSION "1.0"

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <alloc.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <pcibios.h>
#include <pcicheck.h>
#include <calls.h>

#define DEVICE_ID 0x9050
#define VENDOR_ID 0x10B5

void helpmsg(void)
{
   printf("pci0: ein Programm zum Interpretieren der PCI-Information.\n");
   printf("Version: %s vom %s\n", VERSION, __DATE__);
   printf("\n");
   printf("          pci0 [-?] [-d=device] [-v=vendor]\n");
   printf("\n");
   printf("          -?         : Dies hier.\n");
   printf("          -d=device  : Eingabe der DEVICE_ID in hexadezimal.\n");
   printf("          -v=vendor  : Eingabe des VENDOR_ID in hexadezimal.\n");
   printf("\n");
}

int main(int argc, char *argv[])
{
  int error = 0;
  int i;
  unsigned short device_id = DEVICE_ID;
  unsigned short vendor_id = VENDOR_ID;
  char *cptr;

  // kommandozeile zerlegen
  for (i = 1; i < argc; i++)
  {
    cptr = argv[i];

    if ((*cptr == '-') | (*cptr == '/')) cptr++;

    switch (tolower(*cptr))
    {
      case '?': helpmsg();
		return 0;
      case 'd': cptr++;
		if (*cptr == '=') cptr ++;
		sscanf(cptr, "%x", &device_id);
		if (!device_id)
		{
		  printf("pci0: Falsches \"-d\"-Argument!\n");
		  return -7;
		}
		break;
      case 'v': cptr++;
		if (*cptr == '=') cptr ++;
		sscanf(cptr, "%x", &vendor_id);
		if (!vendor_id)
		{
		  printf("pci0: Falsches \"-v\"-Argument!\n");
		  return -8;
		}
		break;

      default:  printf("pci0: Unbekanntes Kommando \"%s\"!\n", cptr);
		return -6;
    }
  }

  // 1. existiert ein PCI-BIOS?
  {
    int main_version, sub_version, anzahl_pci_bus, mechanism;

    if (pcibios_present(&main_version, &sub_version, &anzahl_pci_bus, &mechanism))
    {
      printf("pci0: Kein PCI-BIOS vorhanden!\n");
      return -1;
    }
    else
    {
      printf("PCI-BIOS gefunden!\n");
      printf("Version %d.%d. BUS-Anzahl %d. Mechanismus %02x.\n",
	       main_version, sub_version, anzahl_pci_bus, mechanism);
    }
  }

  // 2. frage den config-space ab
  {
    int bus_number, device_number, function_number;
    PCI_CONFIG_SPACE config_space;
    unsigned long *lptr;

    error = find_pci_device(device_id, vendor_id, 0,
			  &bus_number, &device_number, &function_number);
    if (error)
    {
      printf("pci0: find_pci_device : \"%s\"!\n", interpret_error(error));
      return -2;
    }
    else
    {
      printf("Device 0x%04x/0x%04x/0 mit bus_number      %d\n", device_id, vendor_id, bus_number);
      printf("                           device_number   %d\n", device_number);
      printf("                           function_number %d gefunden.\n", function_number);
    }

    error = pci_get_configuration_space(bus_number, device_number,
					function_number, &config_space);
    if (error)
    {
      printf("pci0: pci_get_configuration : \"%s\"!\n", interpret_error(error));
      return -3;
    }

    {
      PCI_ADDRESS pci_address;
      char *cptr;
      int size = 2048;

      if ((cptr = (char *)malloc(size)) == NULL)
      {
	printf("pci0: Kein Speicher frei!\n");
	return -4;
      }

      type_config_space(&config_space, cptr, &size);
      printf("\n%s\n",cptr);   // auf den bildschirm bringen

      free(cptr);

      // 3. loese die adressen heraus
      error = resolve_address(bus_number, device_number, function_number, &pci_address);

      if (error)
      {
	printf("pci0: resolve_address : \"%s\"!\n", interpret_error(error));
	return -5;
      }

      for (i = 0; i < pci_address.spaces; i++)
      {
	printf("Adress-Eintrag Nr. %d:\n", i);
	printf("Dieser Bereich liegt im %s\n", (pci_address.loc[i].io_space) ? "IO-space" : "Memory-Space");
	printf("Die Basis-Adresse liegt bei 0x%08lx\n", pci_address.loc[i].physical_address);
	printf("Der Adressbereich ist 0x%08lx\n", pci_address.loc[i].physical_range);
	if (!pci_address.loc[i].io_space)
	  printf("Die Adressen sind %s\"prefetchable\"\n", (pci_address.loc[i].prefetchable) ? "" : "nicht ");
	printf("\n");
      }

      {
	 unsigned long i, j;
	 unsigned long far *adr;

	 adr = MK_FP((unsigned)(pci_address.loc[2].physical_address >> 4),
	             (unsigned)(pci_address.loc[2].physical_address & 0xf));
	 i = 1;

#ifdef TEST
	 while (i++)
	 {
	   writelong(adr, i);
	   if (readlong(adr) != i) printf("%ld->%ld\n",i,j);
	 }

	 {
	   unsigned long quelle[32];
	   long int      i = 32;
	   clock_t         old_time;

	   while (--i >= 0)
	     quelle[i] = 1L << (i % 32);

	   i = 0;
	   old_time = clock();
	   printf("Durchlaeufe: %ld, Zeit: %ld\n", i, (clock() - old_time) / CLK_TCK);
	   while (++i)
           {
	     movelong(adr,quelle,32);

	     if (! (i % 1000000))
	     {
	       printf("Durchlaeufe: %ld, Zeit: %g\n", i, (clock() - old_time) / CLK_TCK);
	       old_time = clock();
	     }
	   }
	 }
#endif
       }
    }
  }

  return 0;
}