//-------------------------------------------------------------------------
//
// pcibios support functions - functions to resolve PCI information
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and
// modify it unless this header  is  not modified or deleted. No warranty is given that
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// First parts born                                     AR     27.10.97
//
#include <stdio.h>
#include <string.h>
#include <pcicheck.h>
#include <pcibios.h>

//-----------------------------------------------------------------------
// laedt den configuration space von ... nach config_space
//
int pci_get_configuration_space(int bus_number, int device_number,
				  int function_number, PCI_CONFIG_SPACE *config_space)
{
  unsigned long *lptr;
  unsigned short i;
  int result;

  lptr = (unsigned long *)config_space;

  for (i = 0; i < sizeof(PCI_CONFIG_SPACE); i += sizeof(unsigned long))
  {
    if ((result = read_configuration_dword(bus_number,
					    device_number,
					     function_number,
					      i,
					       lptr++)) != PCI_SUCCESSFUL)
      break;
  }
  return result;
}

//---------------------------------------------------------------------
// legt die ASCII-Interpretation eines config_space nach buffer
//
void type_config_space(PCI_CONFIG_SPACE *c_s, char *buffer, int *size)
{
  char *ptr;
  int i;

  if (*size < 600)
  {
    *size = 0;
    return;
  }

  ptr = buffer;
  sprintf(ptr, "Device_ID / Vendor_ID  : 0x%04x / 0x%04x\n", c_s->device_id, c_s->vendor_id);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Status/Command         : 0x%04x / 0x%04x\n", c_s->status, c_s->command);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Class-Code/Revision-ID : 0x%06lx / 0x%02x\n", c_s->class_revision >> 8, c_s->class_revision & 0xff);
  ptr = strchr(buffer,0);
  sprintf(ptr, "BIST / Header-Type     : 0x%02x   /   0x%02x\n", c_s->bist, c_s->header_type);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Latency-Timer/Cache-L. : 0x%02x   /   0x%02x\n", c_s->latency_timer, c_s->cache_line_size);
  ptr = strchr(buffer,0);
  for (i = 0; i < 6; i++)
  {
    sprintf(ptr, "Base-Address-0         : 0x%08lx\n", c_s->base[i]);
    ptr = strchr(buffer,0);
  }
  sprintf(ptr, "Cardbus-CIS-Pointer    : 0x%08lx\n", c_s->cardbus_cis_pointer);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Subsystem-ID / -Vendor : 0x%04x / 0x%04x\n", c_s->subsystem_id, c_s->subsystem_vendor_id);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Expansion-ROM-base_adr.: 0x%08lx\n", c_s->expansion_rom_base_address);
  ptr = strchr(buffer,0);
  sprintf(ptr, "MaxLat / MinGnt        : 0x%02x   /   0x%02x\n", c_s->max_lat, c_s->min_gnt);
  ptr = strchr(buffer,0);
  sprintf(ptr, "Interrupt-Pin / -Line  : 0x%02x   /   0x%02x\n", c_s->interrupt_pin, c_s->interrupt_line);
  ptr = strchr(buffer,0);

  *size = (ptr + 1 - buffer);
}

//-----------------------------------------------------------------------
// ermittelt die adressen eines config_space
//
int resolve_address(int bus_number, int device_number,
		    int function_number, PCI_ADDRESS *pci_address)
{
  PCI_CONFIG_SPACE config_space;
  int i;
  int result;          // erfolg oder nicht?
  unsigned long range;         // adressbereich


  result = pci_get_configuration_space(bus_number, device_number,
					 function_number, &config_space);
  if (result) return result;

  pci_address->spaces = 0;

  for (i = 0; i <= 6; i++)  // alle base eintraege durchgehen
  {
    if (config_space.base[i])
    {
      pci_address->spaces++;

      if (config_space.base[i] & 1)  // io- oder memory-space?
      {
	pci_address->loc[i].io_space = 1;
	pci_address->loc[i].physical_address = config_space.base[i] & 0xfffffffc;
      }
      else
      {
	pci_address->loc[i].io_space = 0;
	pci_address->loc[i].prefetchable = (config_space.base[i] & 0x8) ? 1 : 0;
	pci_address->loc[i].physical_address = config_space.base[i] & 0xfffffff0;
      }

      result = write_configuration_dword(bus_number, device_number,
					 function_number,
					 0x10 + i * sizeof(unsigned long),
					 0xffffffff);
      if (result) return result;
      result = read_configuration_dword(bus_number, device_number,
					function_number,
					0x10 + i * sizeof(unsigned long),
					&range);
      if (result) return result;
      result = write_configuration_dword(bus_number, device_number,
					 function_number,
					 0x10 + i * sizeof(unsigned long),
					 config_space.base[i]);
      if (result) return result;

      pci_address->loc[i].physical_range = ~range | 0xf;
    }
  }

  return PCI_SUCCESSFUL;
}

//-----------------------------------------------------------------------
// ermittelt die interrupt line eines devices
//
int resolve_IRQ_line(int bus_number, int device_number,
			 int function_number, int *IRQ_line)
{
  PCI_CONFIG_SPACE config_space;
  int result;          // erfolg oder nicht?

  result = pci_get_configuration_space(bus_number, device_number,
					 function_number, &config_space);
  if (result) return result;

  *IRQ_line = config_space.interrupt_line;

  return PCI_SUCCESSFUL;
}


