//-------------------------------------------------------------------------
//
// main menu frame for test and demo program pcilook for the PCIVME
// interface of ARW Elektronik
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// first steps						AR     22.03.99
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <stdlib.h>
#include <menu.h>
#include <pcibios.h>
#include <pcicheck.h>
#include <pcilib.h>

//-------------------------------------------------------------------------
// DEFINES
//

//-------------------------------------------------------------------------
// TYPDEFS
//

//-------------------------------------------------------------------------
// EXTERNALS
//
extern char *progname;
extern int main_version, sub_version;  // von pci1
extern unsigned short device;
extern unsigned short vendor;

extern MN_MENU_ENTRY menu_1[];         // von menu1
extern int menu_1_entry(void);

extern MN_MENU_ENTRY menu_2[];         // von menu1
extern int menu_2_entry(void);

extern MN_MENU_ENTRY menu_3[];         // von menu3
extern int menu_3_entry(void);

extern MN_SHOW_ENTRY status_1[];       // von local configuration
extern int status_1_entry(void);

extern MN_SHOW_ENTRY status_2[];       // von status_2
extern int status_2_entry(void);

//-------------------------------------------------------------------------
// LOCALS
//
static PCIDEV *pcidev;
static int count;
static int select_index;

//-------------------------------------------------------------------------
// GLOBALS
//
PCIDEV *selected_device = NULL;     // zeigt auf ein device
PCI_SPACE *selected_space = NULL;   // zeigt auf den space im device

//-------------------------------------------------------------------------
// FUNCTIONS
//
static void print_selected_device(void)
{
  printf("%d of %d", select_index + 1, count);
}

static void print_pci_version(void)
{
  printf("%d.%d", main_version, sub_version);
}

static void print_bus_number(void)
{
  printf("%d", selected_device->bus_number);
}

static void print_device_number(void)
{
  printf("%d", selected_device->device_number);
}

static void print_function_number(void)
{
  printf("%d", selected_device->function_number);
}

static int action_next(void)
{
  if ((++select_index) >= count)
    select_index = 0;

  selected_device = &pcidev[select_index];
  selected_space  = &pcidev[select_index].pciadr.loc[2];

  return 0;
}

static int action_previous(void)
{
  if ((--select_index) < 0)
    select_index = count - 1;

  selected_device = &pcidev[select_index];
  selected_space  = &pcidev[select_index].pciadr.loc[2];

  return 0;
}

static int action_leave(void)
{
  int ergebnis = 0;

  if (mn_abfrage_bool("", "Really leave ?", 'y', 'n', &ergebnis))
    return ergebnis;
  else
    return 0;
}

static int action_address(void)
{
  mn_menu("Show addresses", menu_1_entry, menu_1);

  return 0;
}

static void print_selected_address(void)
{
  if (selected_space->io_space)
    printf("IO:  ");
  else
    printf("MEM: ");

  printf("0x%08lx..0x%08lx",
    selected_space->physical_address,
      selected_space->physical_address +
	selected_space->physical_range);
}

static int action_show_pci(void)
{
  mn_show_status("Show PCI registers", status_2_entry, status_2, 0);
  return 0;
}

static int action_local_config(void)
{
  mn_show_status("Show local configuration registers",
					status_1_entry, status_1, 0);
  return 0;
}

static int action_VMEMM(void)
{
  mn_menu("VMEMM features", menu_2_entry, menu_2);

  return 0;
}

static int action_VME(void)
{
  mn_menu("VME features", menu_3_entry, menu_3);

  return 0;
}

static void dummy(void)
{
}

// die menudeklaration -------------------------------------------------------
MN_MENU_ENTRY menu_0[] = {{-1 , "PCI-bios version number", print_pci_version, NULL},
			  {-1 , "Selected PCI-device", print_selected_device, NULL},
			  {-1 , "Bus-number of selected device", print_bus_number, NULL},
			  {-1 , "Device-number of selected device", print_device_number, NULL},
			  {-1 , "Function-number of selected device", print_function_number, NULL},
			  {-1 , "Selected VME interface window", print_selected_address, NULL},
			  {-1 , "", dummy, NULL},
			  {'n', "Select next device", NULL, action_next},
			  {'p', "Select previous device", NULL, action_previous},
			  {'c', "Show PCI-registers of selected device", NULL, action_show_pci},
			  {'a', "Show addresses of selected device", NULL, action_address},
			  {'l', "Show local configuration registers", NULL, action_local_config},
			  {'v', "VMEMM features", NULL, action_VMEMM},
			  {'x', "VME features", NULL, action_VME},
			  {'q', "Leave program", NULL, action_leave},
			  { 0 , NULL, NULL, NULL}};

int menu_0_entry(void)
{
  int error;

  error = SearchPCIADA(&pcidev, &count);

  if (error != PCI_SUCCESSFUL)
  {
    printf("%s: Error @ searching PCI-devices : \"%s\"!\n",
					progname, interpret_error(error));
    exit(-1);
  }

  if (!count)
  {
    printf("%s: No PCIADA device found!\n", progname);
    exit(-1);
  }

  select_index    = 0;
  selected_device = &pcidev[select_index];
  selected_space  = &pcidev[select_index].pciadr.loc[2];

  return 0;
}