//-------------------------------------------------------------------------
//
// 3rd submenu of a test program for the PCIVME interface of ARW Elektronik
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
// what							who    when
// first steps						AR     14.04.99
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <menu.h>
#include <pcibios.h>
#include <pcicheck.h>
#include <pcilib.h>

//-------------------------------------------------------------------------
// DEFINES
//

//-------------------------------------------------------------------------
// TYPDEFS
//

//-------------------------------------------------------------------------
// EXTERNALS
//
extern PCIDEV *selected_device;     // zeigt auf ein device

//-------------------------------------------------------------------------
// GLOBALS
//

//-------------------------------------------------------------------------
// LOCALS
//
static int path_size = 1;
static int address_modifier = 0x39;  // standard range
static unsigned long target_address = 0;
static unsigned long block_size = 1;
static unsigned long write_data = 0xAAAAAAAA;
static PCIVME_INTERFACE *Interface = NULL;

//-------------------------------------------------------------------------
// FUNCTIONS
//
static void dummy(void)
{
}

static int adjust_address(void)
{
  int adjust = 0;

  switch (path_size)
  {
    case 1: break;
    case 2: adjust = (target_address & 1) || (block_size & 1);
	    target_address = (target_address + 1) & 0xFFFFFFFE;
	    block_size     = block_size & 0xFFFFFFFE;
	    break;
    case 4: adjust = (target_address & 3) || (block_size & 3);
	    target_address = (target_address + 3) & 0xFFFFFFFC;
	    block_size     = block_size & 0xFFFFFFFC;
	    break;
  }
  return (adjust) ? 1 : 0;
}

static int action_address_modifier(void)
{
  char buffer[40];
  unsigned int temp;

  if (mn_abfrage_input_inplace("%s", buffer))
  {
    sscanf(buffer, "%x", &temp);

    if (temp <= 0x3F)
      address_modifier = temp;
    else
      set_status_zeile("\a0x00 <= address modifier <= 0x3F !");
  }

  return 0;
}

static void print_address_modifier(void)
{
  printf("0x%02x", address_modifier);
}

static int action_target_address(void)
{
  char buffer[40];

  if (mn_abfrage_input_inplace("%s", buffer))
    sscanf(buffer, "%lx", &target_address);

  if (adjust_address())
    set_status_zeile("\aAddress or block size adjust forced.");

  return 0;
}

static void print_target_address(void)
{
  printf("0x%08lx\n", target_address);
}

static int action_block_size(void)
{
  char buffer[40];

  if (mn_abfrage_input_inplace("%s", buffer))
    sscanf(buffer, "%ld", &block_size);

  if (adjust_address())
    set_status_zeile("\aAddress or block size adjust forced.");

  return 0;
}

static void print_block_size(void)
{
  printf("%lu", block_size);
}

static int action_path_size(void)
{
  char buffer[40];
  int ergebnis;

  MN_WAHL_ENTRY wahl[] = {{'b', "byte (8 bit)"},
			  {'w', "word (16 bit)"},
			  {'l', "long (32 bit)"},
			  {0, "end"}};

  if (mn_abfrage_multiple("path size", wahl, &ergebnis) == 0)
    return 0;

  switch (ergebnis)
  {
    case 'b': path_size = 1; break;
    case 'w': path_size = 2; break;
    case 'l': path_size = 4; break;
  }

  if (adjust_address())
    set_status_zeile("\aAddress or block size adjust forced.");

  return 0;
}

static void print_path_size(void)
{
  switch (path_size)
  {
    case 1: printf("byte (8 bit)"); break;
    case 2: printf("word (16 bit)"); break;
    case 4: printf("long (32 bit)"); break;
  }
}

static int action_init_interface(void)
{
  int dummy;
  int ModuleNo;

  // try to get module number of connected interface
  if (enable_access_to_VMEMM(selected_device) != PCI_SUCCESSFUL)
  {
    set_status_zeile("\aenable_access_to_VMEMM() failed!");
    return 1;
  }

  if (get_information_about_VMEMM(selected_device, &dummy,
	     &dummy, &ModuleNo, &dummy, &dummy) != PCI_SUCCESSFUL)
  {
    set_status_zeile("\aget_information_about_VMEMM() failed!");
    return 1;
  }

  if (disable_access_to_VMEMM(selected_device) != PCI_SUCCESSFUL)
  {
    set_status_zeile("\adisable_access_to_VMEMM() failed!");
    return 1;
  }

  Interface = Init_Interface(ModuleNo); // Inits the Interface

  if (Interface == (PCIVME_INTERFACE *)NULL)
  {
    set_status_zeile("\aInit_Interface() failed!");
    return 1;
  }

  return 0;
}

static int action_do_read(void)
{
  unsigned long i = block_size / path_size;

  switch (path_size)
  {
    case 1: while (i--) ReadByte(Interface, target_address + i, address_modifier);
	    break;
    case 2: while (i--) ReadWord(Interface, target_address + (i << 1), address_modifier);
	    break;
    case 4: while (i--) ReadLong(Interface, target_address + (i << 2), address_modifier);
	    break;
  }

  return 0;
}

static int action_do_write(void)
{
  unsigned long i = block_size / path_size;
  unsigned long erg;

  switch (path_size)
  {
    case 1: while (i--) WriteByte(Interface, target_address + i, write_data, address_modifier);
	    break;
    case 2: while (i--) WriteWord(Interface, target_address + (i << 1), write_data, address_modifier);
	    break;
    case 4: while (i--) WriteLong(Interface, target_address + (i << 2), write_data, address_modifier);
	    break;
  }

  return 0;
}

static int action_do_TAS(void)
{
  char ergebnis;

  ergebnis = TAS(Interface, target_address, address_modifier);

  if (ergebnis)
    set_status_zeile("Semafore was set.");
  else
    set_status_zeile("Semafore was not set.");

  return 0;
}

static int action_clear_TAS(void)
{
  WriteByte(Interface, target_address, 0, address_modifier);
  return 0;
}

static void print_semafore(void)
{
  printf("0x%02x", 0xff & ReadByte(Interface, target_address, address_modifier));
}

static int action_deinit_interface(void)
{
  DeInit_Interface(&Interface);
  return 0;
}

static int action_set_sysfail(void)
{
  SetSfail(Interface);
  return 0;
}

static int action_clear_sysfail(void)
{
  ClrSfail(Interface);
  return 0;
}

static void print_sysfail(void)
{
  printf("%s", (PollSfail(Interface)) ? "set" : "reset");
}

static int action_set_sysreset(void)
{
  Reset_VME(&Interface, VME_RESET_CMD);

  set_status_zeile("VME SYSRESET raised!");
  return 0;
}

void print_write_data(void)
{
  switch (path_size)
  {
    case 1: printf("0x%02x", write_data & 0xff); break;
    case 2: printf("0x%04x", write_data & 0xffff); break;
    case 4: printf("0x%08lx", write_data); break;
  }
}

static int action_write_data(void)
{
  char buffer[40];

  if (mn_abfrage_input_inplace("%s", buffer))
    sscanf(buffer, "%lx", &write_data);

  return 0;
}


static int action_leave(void)
{
  action_deinit_interface();
  return 1;
}

// die menudeklaration -------------------------------------------------------
MN_MENU_ENTRY menu_3[] = {{'a', "set address modifier", print_address_modifier, action_address_modifier},
			  {'b', "set target address", print_target_address, action_target_address},
			  {'s', "set block size", print_block_size, action_block_size},
			  {'p', "set path size", print_path_size, action_path_size},
			  {'d', "set write data", print_write_data, action_write_data},
			  {-1 , "", NULL, NULL},
			  {'r', "do read(s)", NULL, action_do_read},
			  {'w', "do write(s)", NULL, action_do_write},
			  {'t', "do a TAS @ target address", print_semafore, action_do_TAS},
			  {'c', "clear semafore @ target address", print_semafore, action_clear_TAS},
			  {'y', "set SYSFAIL", print_sysfail, action_set_sysfail},
			  {'z', "clear SYSFAIL", print_sysfail, action_clear_sysfail},
			  {'v', "generate SYSRESET", NULL, action_set_sysreset},
			  {'q', "Leave menu", NULL, action_leave},
			  { 0 , NULL, NULL, NULL}};

int menu_3_entry(void)
{
  return action_init_interface();
}
