//-------------------------------------------------------------------------------------------
// pcilib.c - interface library to the pcivme interface of ARW Elektronik
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header  is  not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// what                                                              who    when
// first steps (partial kompatibility to ATVME)                      AR     01.01.1999
// bug in bugfix fixed                                               AR     27.05.1999
// pollIrq/ReadVector.. modified to get a vector to a specific level AR     27.05.1999
//

//-------------------------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <malloc.h>
#include <vic.h>
#include <pciif.h>
#include <pcicheck.h>
#include <pcibios.h>
#include <pcilib.h>
#include <calls.h>

//-------------------------------------------------------------------------------------------
// DEFINES
//
#define DEVICE_VENDOR_ADR 0x00    // address of device vendor entry in PCR
#define SUBSYS_VENDOR_ADR 0x2c    // address of subsystem vendor entry in PCR
#define DEVICE_VENDOR     0x905010B5
#define SUBSYS_VENDOR     0x11679050  // entry of ...
#define DEVICE_ID         0x9050  // chip PLX9050
#define VENDOR_ID         0x10B5  // PLX

#define PCR_MEMORY_BUG    0x10    // 1st PCR address of potential bug
#define PCR_IO_BUG        0x14    // 2nd PCR address of potential bug
#define PCR_MEMORY_OK     0x20    // 1st PCR address of no bug
#define PCR_IO_OK         0x24    // 2nd PCR address of no bug

#define bugadr(x) pcidev->pciadr.loc[x].physical_address

#define BYTE  unsigned char
#define WORD  unsigned short
#define DWORD unsigned long

#define VMEMM_TEST_FAILED  100    // it's not a PCI fail

#define BYTE_ACCESS      (BYTE)1  // write byte wise
#define WORD_ACCESS      (BYTE)2  //       word
#define LONG_ACCESS      (BYTE)4  //       long

#define LCR              (BYTE)0    // destination is LCR register
#define IFR              (BYTE)1    // destination is VME-Interface register
#define VIC              (BYTE)2    // destination is VIC68A register
#define STOP             (BYTE)255  // this command stops the init machine

//-------------------------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct               // one command element to initialize interface or deinitialize
{
  unsigned char  range;      // 0 = lcr, 1 = vme-interface, -1 = stop, default = vme-if
  unsigned char  type;       // 1 = byte access, 2 = word access, 4 = dword access, default byte
  unsigned short offset;     // offset into interface address range for initialisation
  unsigned long  value;      // value to initialize
} INIT_ELEMENT;

//-------------------------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------------------------
// GLOBALS
//

//--------------------------------------------------------------------------
// LOCALS
//
static PCIDEV pciDev[PCIADAMAX];    // maximal soviele gleiche devices

static PCIDEV *PCIADA = NULL;
static int    nCountOfPCIADA = 0;

static const INIT_ELEMENT init_element[] =
	 {{LCR,  WORD_ACCESS, 0x4c, DISABLE_PCI_IRQS},// disable interrupts
          {LCR,  WORD_ACCESS, 0x50, RELEASE_VMEMM},   // enable interface

          {VIC,  BYTE_ACCESS, (WORD)0x03, 0xf8+1},    // VIICR

          {VIC,  BYTE_ACCESS, (WORD)0x07, 0x78+1},    // VICR1
          {VIC,  BYTE_ACCESS, (WORD)0x0b, 0x78+2},
          {VIC,  BYTE_ACCESS, (WORD)0x0f, 0x78+3},
          {VIC,  BYTE_ACCESS, (WORD)0x13, 0x78+4},
          {VIC,  BYTE_ACCESS, (WORD)0x17, 0x78+5},
          {VIC,  BYTE_ACCESS, (WORD)0x1b, 0x78+6},
          {VIC,  BYTE_ACCESS, (WORD)0x1f, 0x78+7},    // VICR7

          {VIC,  BYTE_ACCESS, (WORD)0x23, 0xf8+0},    // DSICR

          {VIC,  BYTE_ACCESS, (WORD)0x27, 0xf8+1},    // LICR1
          {VIC,  BYTE_ACCESS, (WORD)0x2b, 0xf8+2},
          {VIC,  BYTE_ACCESS, (WORD)0x2f, 0xf8+3},
          {VIC,  BYTE_ACCESS, (WORD)0x33, 0xf8+4},
          {VIC,  BYTE_ACCESS, (WORD)0x37, 0xf8+5},
          {VIC,  BYTE_ACCESS, (WORD)0x3b, 0x38+6},
          {VIC,  BYTE_ACCESS, (WORD)0x3f, 0x38+7},    // LICR7

          {VIC,  BYTE_ACCESS, (WORD)0x43, 0xf8+2},    // ICGS
          {VIC,  BYTE_ACCESS, (WORD)0x47, 0xf8+3},    // ICMS

          {VIC,  BYTE_ACCESS, (WORD)0x4b, 0xf8+6},    // EGICR

          {VIC,  BYTE_ACCESS, (WORD)0x4f, 0x08},      // ICGS-IVBR (!)
          {VIC,  BYTE_ACCESS, (WORD)0x53, 0x0c},      // ICMS-IVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x57, 0x00},      // LIVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x5b, 0x10},      // EGIVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x5f, 0x00},      // ICSR

          {VIC,  BYTE_ACCESS, (WORD)0x63, 0x00},      // ICR0
          {VIC,  BYTE_ACCESS, (WORD)0x67, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x6b, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x6f, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x73, 0x00},      // ICR4

          {VIC,  BYTE_ACCESS, (WORD)0x83, 0xfe},      // VIRSR

          {VIC,  BYTE_ACCESS, (WORD)0x87, 0x0f},      // VIVR1
          {VIC,  BYTE_ACCESS, (WORD)0x8b, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x8f, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x93, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x97, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x9b, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x9f, 0x0f},      // VIVR7

          {VIC,  BYTE_ACCESS, (WORD)0xa3, 0x3c},      // TTR

          {VIC,  BYTE_ACCESS, (WORD)0xb3, 0x40},      // ARCR
          {VIC,  BYTE_ACCESS, (WORD)0xb7, 0x29},      // AMSR
          {VIC,  BYTE_ACCESS, (WORD)0xd3, 0x00},      // RCR

          {IFR,  LONG_ACCESS, (WORD)ADRHL, 0xF0F0F0F0},  // ADR-H, ADR-L
          {IFR,  WORD_ACCESS, (WORD)CSR  , 0x0000},      // Contr-Reg

	  {VIC,  BYTE_ACCESS, (WORD)0x7f, 0x80},      // ICR7
	  {LCR,  WORD_ACCESS, 0x4c, DISABLE_PCI_IRQS},// disable interrupts
          {STOP, WORD_ACCESS, 0,     0}};

static const INIT_ELEMENT deinit_element[] =
         {{VIC,  BYTE_ACCESS, (WORD)0x7f, 0x00},      // ICR7 - sysfail
	  {LCR,  WORD_ACCESS, 0x4c, DISABLE_PCI_IRQS},// disable interrupts
	  {LCR,  WORD_ACCESS, 0x50, INHIBIT_VMEMM},   // disable interface
          {STOP, WORD_ACCESS, 0,    0}};


//-------------------------------------------------------------------------------------------
// FUNCTIONS
//

//-------------------------------------------------------------------------------------------
// set interface special registers - VIC68A
//
void SetRegister(PCIVME_INTERFACE *Interface,
			      unsigned short usRegisterOffset, unsigned char ucValue)
{
  *(char far *)((char far *)(Interface->pucVIC68A) + usRegisterOffset) = ucValue;
}

//-------------------------------------------------------------------------------------------
// function to get special interface registers - VIC68A
//
unsigned char GetRegister(PCIVME_INTERFACE *Interface, unsigned short usRegisterOffset)
{
  return *(char far *)((char far *)(Interface->pucVIC68A) + usRegisterOffset);
}

static int check_command(const INIT_ELEMENT *psInitElement)
{
  WORD range;
  WORD access_size;
  
  switch (psInitElement->range)
  {
    case LCR: range = 0x54;     break;
    case IFR: range = 0x0c;     break;
    case VIC: range = 0xe4;
              if ((psInitElement->offset & 3) != 3)
		return 0;
              break;
    default:  range = 0;        break; 
  }

  switch (psInitElement->type)
  {
    case LONG_ACCESS: if (psInitElement->offset & 3)
			return 0;
                      access_size = sizeof(DWORD);
                      break;
    case WORD_ACCESS: if (psInitElement->offset & 1)
			return 0;
                      access_size = sizeof( WORD);
                      break;
    case BYTE_ACCESS: access_size = sizeof( BYTE); break;
    default         : access_size = 0xFFFF;        break;
  }

  if ((psInitElement->offset + access_size) > range)
    return 0;       // ignore it
    
  return 1;
}

static void CmdMachine(PCIVME_INTERFACE *Interface, const INIT_ELEMENT *psInitElement)
{
  void far *adr;
  
  // standard initialisierungen  
  while (psInitElement->range != STOP)
  {
    if (check_command(psInitElement))
    {
      switch (psInitElement->range)
      {
	case LCR: adr =             Interface->pvPCIADA;  break;
	case VIC: adr = (void far *)Interface->pucVIC68A; break;
        case IFR:
	default:  adr = (void far *)Interface->pusStat;   break;
      }

      switch (psInitElement->type)
      {
	case LONG_ACCESS:
	  writelong((unsigned long far *)
	  ((BYTE far *)adr + psInitElement->offset), psInitElement->value);
          break;
        case WORD_ACCESS:
	  *(WORD far *) ((BYTE far *)adr + psInitElement->offset) =
						      psInitElement->value;
          break;
        case BYTE_ACCESS:
        default:
	  *(BYTE far *) ((BYTE far *)adr + psInitElement->offset) =
						      psInitElement->value;
          break;
      }
    }
    psInitElement++;
  }
}

static int ProgramInterface(PCIVME_INTERFACE *Interface, const INIT_ELEMENT *init_element)
{
  if (init_element == (INIT_ELEMENT *)NULL)
    return 0;

  CmdMachine(Interface, init_element);

  Interface->ucAdrMod = GetRegister(Interface, AMSR) & 0x3F;
  Interface->ulADRHL  = *(Interface->pulADRHL);
  return 1;
}

//-------------------------------------------------------------------------------------------
// check if this device is a PCIADA
//
static int check_for_PCIADA(PCIDEV *pcidev)
{
  int error;
  unsigned long result;

  error = read_configuration_dword(pcidev->bus_number, pcidev->device_number,
				     pcidev->function_number,
				       DEVICE_VENDOR_ADR, &result);
  if (error != PCI_SUCCESSFUL) return error;
  if (result != DEVICE_VENDOR) return PCI_DEVICE_NOT_FOUND;

  error = read_configuration_dword(pcidev->bus_number, pcidev->device_number,
				     pcidev->function_number,
				       SUBSYS_VENDOR_ADR, &result);
  if (error != PCI_SUCCESSFUL) return error;
  if (result != SUBSYS_VENDOR) return PCI_DEVICE_NOT_FOUND;

  return PCI_SUCCESSFUL;
}

//-------------------------------------------------------------------------------------------
// search all PCIVME devices on the PCI bus
//
int SearchPCIADA(PCIDEV *pcidev[], int *nInterfaceCount)
{
  int i;
  int error;

  *pcidev = pciDev;

  // pcidev strukturen initialisieren --------------
  for (i = 0; i < PCIADAMAX; i++)
  {
    pciDev[i].i = pciDev[i].bus_number =
       pciDev[i].device_number = pciDev[i].function_number = -1;
  }

  // alle pci-devices finden -----------------------
  i = 0;
  *nInterfaceCount = 0;
  do
  {
    error = find_pci_device(DEVICE_ID, VENDOR_ID, i,
	   &pciDev[i].bus_number, &pciDev[i].device_number, &pciDev[i].function_number);
    if (error == PCI_SUCCESSFUL)
    {
      if (check_for_PCIADA(pciDev) == PCI_SUCCESSFUL)
	pciDev[i].i = i;
      i++;
      (*nInterfaceCount)++;
    }
  } while ((error == PCI_SUCCESSFUL) && (i < PCIADAMAX));

  if ((error != PCI_DEVICE_NOT_FOUND) && (error != PCI_SUCCESSFUL))
    return error;

  // die adressen herauslesen --------------------------------------------
  i = 0;
  while (i < *nInterfaceCount)
  {
    error = resolve_address(pciDev[i].bus_number, pciDev[i].device_number,
			       pciDev[i].function_number, &pciDev[i].pciadr);
    if (error) return error;

    error = resolve_IRQ_line(pciDev[i].bus_number, pciDev[i].device_number,
		       pciDev[i].function_number, &pciDev[i].interrupt_line);
    if (error) return error;

    i++;
  };

  return PCI_SUCCESSFUL;
}

//-------------------------------------------------------------------------------------------
// fix the PLX 9050 LCR bit #7 BUG
//
void PLX9050BugFix(PCIDEV *pcidev)
{
  unsigned long   dwData;
  unsigned short  wData;

  if ((bugadr(0) & 0x80) || (bugadr(1) & 0x80))
  {
    read_configuration_dword(pcidev->bus_number,
			       pcidev->device_number,
				 pcidev->function_number,
				   PCR_MEMORY_BUG, &dwData);
    if (dwData & 0x80)
    {
      read_configuration_dword(pcidev->bus_number,
				 pcidev->device_number,
				   pcidev->function_number,
				     PCR_MEMORY_OK, &bugadr(0));

      write_configuration_dword(pcidev->bus_number,
				  pcidev->device_number,
				    pcidev->function_number,
				      PCR_MEMORY_BUG, bugadr(0));
    }
    else
      bugadr(0) = dwData;

    bugadr(0) &= 0xFFFFFFF0;  // mask away special PCI memory codings

    read_configuration_word(pcidev->bus_number,
			      pcidev->device_number,
				pcidev->function_number,
				  PCR_IO_BUG, &wData);
    if (wData & 0x80)
    {
      read_configuration_word(pcidev->bus_number,
				 pcidev->device_number,
				   pcidev->function_number,
				     PCR_IO_OK,(unsigned short *)&bugadr(1));

      write_configuration_word(pcidev->bus_number,
				  pcidev->device_number,
				    pcidev->function_number,
				      PCR_IO_BUG,(unsigned short)bugadr(1));
    }
    else
      bugadr(1) = wData;

    bugadr(1) &= 0xFFFFFFFC;  // mask away special PCI io codings
  }
}

//-------------------------------------------------------------------------------------------
// enable access to VMEMM
//
int enable_access_to_VMEMM(PCIDEV *pcidev)
{
  unsigned short io_base;

  if (pcidev->pciadr.loc[1].io_space)
  {
    io_base = (unsigned short)pcidev->pciadr.loc[1].physical_address;
    outpw(io_base + 0x50, RELEASE_VMEMM);
    return PCI_SUCCESSFUL;
  }
  else
    return PCI_DEVICE_NOT_FOUND;
}

//-------------------------------------------------------------------------------------------
// disable access to VMEMM
//
int disable_access_to_VMEMM(PCIDEV *pcidev)
{
  unsigned short io_base;

  if (pcidev->pciadr.loc[1].io_space)
  {
    io_base = (unsigned short)pcidev->pciadr.loc[1].physical_address;
    outpw(io_base + 0x50, INHIBIT_VMEMM);
    outpw(io_base + 0x4c, DISABLE_PCI_IRQS);
    return PCI_SUCCESSFUL;
  }
  else
    return PCI_DEVICE_NOT_FOUND;
}

//-------------------------------------------------------------------------------------------
// get information about the connected VMEMM
// the path to VMEMM must be opened ind advance
//
int get_information_about_VMEMM(PCIDEV *pcidev, int *WordMode,
             int *Sysctrl, int *ModuleNo, int *FPGAVer, int *ModuleTy)
{
  unsigned long memory_base;
  unsigned short far *pwVMEMMStat;
  unsigned short wVMEMMStat;

  memory_base = (unsigned long)pcidev->pciadr.loc[2].physical_address;
  if (memory_base > 0xfffff) return VMEMM_TEST_FAILED;

  pwVMEMMStat = (unsigned short far *)MK_FP(memory_base >> 4, CSR);
  wVMEMMStat  = *pwVMEMMStat;

  *WordMode = (wVMEMMStat & FLAG_WORD)    >> 2;
  *Sysctrl  = (wVMEMMStat & FLAG_SYSCTL)  >> 3;
  *ModuleNo = (wVMEMMStat & MASK_MODNR)   >> 4;
  *FPGAVer  = (wVMEMMStat & MASK_FPGA)    >> 8;
  *ModuleTy = (wVMEMMStat & MASK_MODTYPE) >> 12;

  return PCI_SUCCESSFUL;
}

//-------------------------------------------------------------------------------------------
// test if the driver is connected to a VMEMM
// starting from a PCIADA view
int test_connection_to_VMEMM(PCIDEV *pcidev)
{
  unsigned long memory_base;
  unsigned short far *pADRH;
  unsigned short far *pADRL;
  unsigned short i;
  unsigned short ADRHshadow;
  unsigned short ADRLshadow;

  if (pcidev->pciadr.loc[2].io_space)
    return PCI_DEVICE_NOT_FOUND;
  else
  {
    memory_base = (unsigned long)pcidev->pciadr.loc[2].physical_address;

    if (memory_base > 0xfffff) return VMEMM_TEST_FAILED;

    pADRH = MK_FP(memory_base >> 4, ADRH);
    pADRL = MK_FP(memory_base >> 4, ADRL);

    ADRHshadow = *pADRH;   // do not destroy previous contents
    ADRLshadow = *pADRL;

    for (i = 10000; i > 0; i--)
    {
      *pADRH = 0x5555;
      *pADRL = 0xAAAA;
      if (*pADRH != 0x5555) return VMEMM_TEST_FAILED;
      if (*pADRL != (0xAAAA & ~VME_ADR_MASK)) return VMEMM_TEST_FAILED;
      *pADRH = 0xAAAA;
      *pADRL = 0x5555;
      if (*pADRH != 0xAAAA) return VMEMM_TEST_FAILED;
      if (*pADRL != (0x5555 & ~VME_ADR_MASK)) return VMEMM_TEST_FAILED;
      *pADRH = 0xFFFF;
      *pADRL = 0x0000;
      if (*pADRH != 0xFFFF) return VMEMM_TEST_FAILED;
      if (*pADRL != (0x0000 & ~VME_ADR_MASK)) return VMEMM_TEST_FAILED;
      *pADRH = 0x0000;
      *pADRL = 0xFFFF;
      if (*pADRH != 0x0000) return VMEMM_TEST_FAILED;
      if (*pADRL != (0xFFFF & ~VME_ADR_MASK)) return VMEMM_TEST_FAILED;
    }

    *pADRH = ADRHshadow;  // restore previous contents
    *pADRL = ADRLshadow;

    return PCI_SUCCESSFUL;
  }
}

//-------------------------------------------------------------------------------------------
// init the interface with the current properties
//
PCIVME_INTERFACE *Init_Interface(int nInterfaceNumber)
{
  int error;
  int i;
  PCIDEV *workPCIADA;
  PCIVME_INTERFACE *workVMEMM = NULL;
  static int firstime = 1;
  unsigned short io_base;
  unsigned short CNTRLshadow;
  unsigned short far *pus;
  int ModuleNo;
  int dummy;
  void far *VMEMMbase;

  if (firstime)
  {
    // ask the bios which PCIDADA are available
    error = SearchPCIADA(&PCIADA, &nCountOfPCIADA);
    if ((error != PCI_SUCCESSFUL) || !nCountOfPCIADA)
      return NULL;

    // patch any PLX bug if neccessary
    workPCIADA = PCIADA;
    for (i = 0; i < nCountOfPCIADA; i++, workPCIADA++)
      PLX9050BugFix(workPCIADA);

    firstime = 0;
  }

  workPCIADA = PCIADA;
  for (i = 0; i < nCountOfPCIADA; i++, workPCIADA++)
  {
    io_base      = (unsigned short)workPCIADA->pciadr.loc[1].physical_address;
    
    // store previous contents
    CNTRLshadow  = inpw(io_base + 0x50);

    // open interface
    outpw(io_base + 0x50, RELEASE_VMEMM);

    // test connection to VMEMM
    error = test_connection_to_VMEMM(workPCIADA);
    if (error == PCI_SUCCESSFUL)
    {
      // search for nInterfaceNumber
      if (get_information_about_VMEMM(workPCIADA, &dummy,
		 &dummy, &ModuleNo, &dummy, &dummy) == PCI_SUCCESSFUL)
      {
	// if found: break
	if (ModuleNo == nInterfaceNumber)
	{
	  // allocate PCIVME_INTERFACE
	  workVMEMM = (PCIVME_INTERFACE *)malloc(sizeof(PCIVME_INTERFACE));
	  break;
	}
      }
    }
    // restore old interface state
    outpw(io_base + 0x50, CNTRLshadow);
  }

  // return if nothing found or allocate fail
  if (workVMEMM == (PCIVME_INTERFACE *)NULL)
    return workVMEMM;

  // calculate base address of VMEMM
  VMEMMbase = (void far *)MK_FP(workPCIADA->pciadr.loc[2].physical_address >> 4, 0);

  // fill PCIVME_INTERFACE structure with static elements
  workVMEMM->nInterface = nInterfaceNumber;    // number of VMEMM

  workVMEMM->nIntNo     = workPCIADA->interrupt_line;

  workVMEMM->pvPCIADA   = (void far *)MK_FP(workPCIADA->pciadr.loc[0].physical_address >> 4, 0);
  workVMEMM->pusCNTRL   = (unsigned short far *)((char far *)workVMEMM->pvPCIADA + 0x50);
  workVMEMM->pusINTCSR  = (unsigned short far *)((char far *)workVMEMM->pvPCIADA + 0x4c);

  workVMEMM->pusStat    = (unsigned short far *)((char far *)VMEMMbase + CSR);
  workVMEMM->pusIRQStat = (unsigned short far *)((char far *)VMEMMbase + VICRES);
  workVMEMM->pulADRHL   = (unsigned long  far *)((char far *)VMEMMbase + ADRHL);
  workVMEMM->pucVIC68A  = (unsigned char  far *)((char far *)VMEMMbase + VICBASE);
  workVMEMM->pucIRQVec  = (unsigned char  far *)((char far *)VMEMMbase + VECBASE);
  workVMEMM->pvVMEWnd   = (unsigned void  far *)((char far *)VMEMMbase + VMEBASE);

  workVMEMM->pucAdrMod  = workVMEMM->pucVIC68A + AMSR;

  workVMEMM->ucWordMode =
  workVMEMM->ucWordPath = (*(workVMEMM->pusStat) & FLAG_WORD) ? 1 : 0;
  
  workVMEMM->ulADRHL    = 0; // Interface isn't open
  workVMEMM->ucAdrMod   = 0;
  workVMEMM->nBusError  = 0; // no BusError is pending

  workVMEMM->nInterruptLine = workPCIADA->interrupt_line;

  // init the registers of PLX9050 and VIC68A
  if (!ProgramInterface(workVMEMM, init_element))
  {
    free(workVMEMM);
    workVMEMM = (PCIVME_INTERFACE *)NULL;
  }

  // return the pointer to the PCIVME_INTERFACE
  return workVMEMM;
}


//-------------------------------------------------------------------------------------------
// deinit the current interface
//
void DeInit_Interface(PCIVME_INTERFACE **Interface)
{
  // set pointer to PCIVME_INTERFACE to NULL
  if (*Interface != (PCIVME_INTERFACE *)NULL)
  {
    // deinit VIC68A and PLX9050
    ProgramInterface(*Interface, deinit_element);

    free(*Interface);
    *Interface = (PCIVME_INTERFACE *)NULL;
  }
}

//-------------------------------------------------------------------------------------------
// check if an bus error occured and clears it
//
int GetError(PCIVME_INTERFACE *Interface)
{
#ifdef INTERRUPT_SUPPORT
  int BusError = Interface->nBusError;

  Interface->nBusError = 0;
#else
  int BusError = 0;

  if (ReadVectorByte(Interface, 7) == 7) BusError = 1;
#endif
  return BusError;
}

//-------------------------------------------------------------------------------------------
// check the VME-Window and set base or address modifier if neccessary
//
static void checkWindow(PCIVME_INTERFACE *Interface, unsigned long ulAdr,
                                                        unsigned char ucModifier)
{
  // compare the shadows - its faster ---
  
  if ((ulAdr & ~VME_ADR_MASK) != Interface->ulADRHL)
  {
    Interface->ulADRHL = ulAdr & ~VME_ADR_MASK;
    writelong(Interface->pulADRHL, Interface->ulADRHL);
  }

  if (ucModifier != Interface->ucAdrMod)
  {
    Interface->ucAdrMod = *(Interface->pucAdrMod) = ucModifier & 0x3F;
  }  
}

//-------------------------------------------------------------------------------------------
// makes a segment:offset address into the VME Window
//
static void far *MK_VADR(PCIVME_INTERFACE *Interface, unsigned long ulAdr)
{
  return (void far *)((char *)(Interface->pvVMEWnd) + (ulAdr - Interface->ulADRHL));
}

//-------------------------------------------------------------------------------------------
// read bytes/words/longs longs are split into 2 word if a word interface is selected
//
char ReadByte(PCIVME_INTERFACE *Interface, unsigned long adr, unsigned char modifier)
{
  checkWindow(Interface, adr, modifier);
  return *(char far *)MK_VADR(Interface, adr);
}

short ReadWord(PCIVME_INTERFACE *Interface, unsigned long adr, unsigned char modifier)
{
  checkWindow(Interface, adr, modifier);
  return *(short far *)MK_VADR(Interface, adr);
}

long  ReadLong(PCIVME_INTERFACE *Interface, unsigned long adr, unsigned char modifier)
{
  // special handling if WORD-MODE VME range ---------------------
  if (Interface->ucWordPath)
  {
    long  lRet;
    short sRet;

    sRet = ReadWord(Interface, adr, modifier);
    lRet = sRet;
    lRet <<= 16;
    sRet = ReadWord(Interface, adr + 2, modifier);
    lRet |= sRet;

    return lRet;
  }

  checkWindow(Interface, adr, modifier);
  return readlong((unsigned long far *)MK_VADR(Interface, adr));
}

//-------------------------------------------------------------------------------------------
// write a byte/word/long and split a long access into 2 word accesses if word data path
//
void WriteByte(PCIVME_INTERFACE *Interface, unsigned long adr,
					 char value, unsigned char modifier)
{
  checkWindow(Interface, adr, modifier);
  *(char far *)MK_VADR(Interface, adr) = value;
}

void WriteWord(PCIVME_INTERFACE *Interface, unsigned long adr,
					 short value, unsigned char modifier)
{
  checkWindow(Interface, adr, modifier);
  *(short far *)MK_VADR(Interface, adr) = value;
}

void WriteLong(PCIVME_INTERFACE *Interface, unsigned long adr,
					 long value, unsigned char modifier)
{
  // special handling if WORD-MODE VME range ---------------------
  if (Interface->ucWordPath)
  {
    WriteWord(Interface, adr    , (short)(value >> 16), modifier);
    WriteWord(Interface, adr + 2, (short)(value      ), modifier);
    return;
  }

  checkWindow(Interface, adr, modifier);
  writelong((unsigned long far *)MK_VADR(Interface, adr), value);
}

//-------------------------------------------------------------------------------------------
// check for a pending SYSFAIL 
//
int PollSfail(PCIVME_INTERFACE *Interface)
{
  unsigned char reg;
  
  reg = (unsigned char)GetRegister(Interface, EGICR);

  return (reg & 0x08) ? 1 : 0;
}

//-------------------------------------------------------------------------------------------
// clear a interface set SYSFAIL
//
void ClrSfail(PCIVME_INTERFACE *Interface)
{
  unsigned char reg;

  reg = GetRegister(Interface, ICR7);
  reg |= 0x80;
  SetRegister(Interface, ICR7, reg);
}

//-------------------------------------------------------------------------------------------
// set a SYSFAIL
//
void SetSfail(PCIVME_INTERFACE *Interface)
{
  unsigned char reg;

  reg = GetRegister(Interface, ICR7);
  reg &= 0x3F;
  SetRegister(Interface, ICR7, reg);
}

//-------------------------------------------------------------------------------------------
// asks if a IRQ is pending
//
int PollIrq(PCIVME_INTERFACE *Interface)
{
  unsigned short IRQStat;

  if (*(Interface->pusINTCSR) & 0x0004)
  {  
    IRQStat = *(Interface->pusIRQStat);
    return IRQStat >> 1;  // return level
  }

  if (*(Interface->pusINTCSR) & 0x0020)
    return 8;  // virtual level 8

  return 0;
}

//-------------------------------------------------------------------------------------------
// read a interrupt vector as byte
//
char ReadVectorByte(PCIVME_INTERFACE *Interface, int level)
{
  unsigned short IRQStat;

  if (!level)  // get the highest level if level equal 0
  {
    if (*(Interface->pusINTCSR) & 0x0004)
    {  
      IRQStat = *(Interface->pusIRQStat);
      if (IRQStat)
        return *(char far*)(Interface->pucIRQVec + IRQStat);
    }

    if (*(Interface->pusINTCSR) & 0x0020)
    {
      if ((*(Interface->pusCNTRL) & 0x0180) == 0x0180) // interface initialized ?
      {
        *(Interface->pusCNTRL) &= ~0x0100;
        *(Interface->pusCNTRL) |=  0x0100;  // clear error
        return 1;
      }
    }
  }
  else
  {
    if (level <= 7)  // get a vector to a specific level
      return *(char far*)(Interface->pucIRQVec + (level << 1) + 1);

    if ((level == 8) && (*(Interface->pusINTCSR) & 0x0020))
    {
      if ((*(Interface->pusCNTRL) & 0x0180) == 0x0180) 
      {
        *(Interface->pusCNTRL) &= ~0x0100;
        *(Interface->pusCNTRL) |=  0x0100;  // clear error
        return 1;
      }
    }
  }

  return 0;
}

//-------------------------------------------------------------------------------------------
// emulate a 68K TAS (read/modify/write) instruction
//
char TAS(PCIVME_INTERFACE *Interface, unsigned long adr, unsigned short modifier)
{
  unsigned short status;
  unsigned char  orgByte;
  
  *(Interface->pusStat) |= 1;   // set RMC

  orgByte = ReadByte(Interface, adr, modifier);  // read locked
  WriteByte(Interface, adr, 0x80, modifier);     // write locked

  *(Interface->pusStat) &= ~1;  // clear RMC

  return (orgByte & 0x80) ? 1 : 0;
}

//-------------------------------------------------------------------------------------------
// global PCIADA interrupt disable
//
void global_IRQ_disable(PCIVME_INTERFACE *Interface)
{
  *(Interface->pusINTCSR) &= ~0x0040;
}

//-------------------------------------------------------------------------------------------
// global PCIADA interrupt disable
//
void global_IRQ_enable(PCIVME_INTERFACE *Interface)
{
  *(Interface->pusINTCSR) |= 0x0040;
}

//-------------------------------------------------------------------------------------------
// generate a SYSRESET on the vmebus and re-init the interface
//
void  Reset_VME(PCIVME_INTERFACE **Interface, short sResetType)
{
  unsigned char reg;
  int           InterfaceNumber;
  PCIVME_INTERFACE *Ifc = *Interface;
  unsigned long waitLoops = 1000000;
  unsigned short INTCSRshadow;

  InterfaceNumber = Ifc->nInterface;            // store interface number

  switch (sResetType)
  {
    case VME_RESET_CMD:    SetRegister(Ifc, SRR, 0xF0);
			   break; // reset VME
    case LOCAL_RESET_CMD:  *(Ifc->pusIRQStat) = 0x0005;
			   break; // local Reset

    default:
    case GLOBAL_RESET_CMD: *(Ifc->pusIRQStat) = 0x000a;
			   break; // global Reset
  }

  INTCSRshadow = *(Ifc->pusINTCSR);             // save INTCSR
  *(Ifc->pusINTCSR) &= ~0x0040;                 // disable global IRQ
  while (waitLoops--)
  {                                             // wait until reset finish
    SetRegister(Ifc, AMSR, 0x55);
    if (GetRegister(Ifc, AMSR) == 0x55) break;
  }

  *(Ifc->pusCNTRL) &= ~0x0100;                  // reset pending IRQ #2
  *(Ifc->pusCNTRL) |=  0x0100;

  *(Ifc->pusINTCSR) = INTCSRshadow;             // restore INTCSR

  DeInit_Interface(Interface);
  *Interface = Init_Interface(InterfaceNumber); // set new pointer to Interface
}

//-------------------------------------------------------------------------------------------
// get the number of the PCI associated interrupt on that machine
// take care: not the VME interrupt level!
//
short interrupt_number(PCIVME_INTERFACE *Interface)
{
  return Interface->nIntNo;
}

//-------------------------------------------------------------------------------------------
// set the VME path temporary to WORD
// (even if the interface isn't still forced to)
//
void SetVMEwordPath(PCIVME_INTERFACE *Interface)
{
  Interface->ucWordPath = 1;
}

//-------------------------------------------------------------------------------------------
// clear the VME path temporary to WORD (if it is possible)
//
void ClrVMEwordPath(PCIVME_INTERFACE *Interface)
{
  if (Interface->ucWordMode) return; // do nothing - not possible

  Interface->ucWordPath = 0;
}

//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
