#ifndef __PCILIB_H__
#define __PCILIB_H__

//-------------------------------------------------------------------------------------------
// pcilib.h - defaults and interface functions of pcilib.c
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
// 
//
// what                                                              who    when
// written for use with PCIVME interface                             AR     22.03.1999
// includes moved from pcilook.c                                     AR     27.05.1999
// interrupt_number included                                         AR     27.05.1999
// ReadVectorByte modified                                           AR     27.05.1999
//

//-----------------------------------------------------------------------------
// INCLUDES  
//
#include <pcibios.h>
#include <pcicheck.h>

//-----------------------------------------------------------------------------
// TYPEDEFS  
//
typedef struct
{
  int                 nInterface; // number of VMEMM
  unsigned char       ucWordMode; // it's a D16 interface only

  void           far *pvPCIADA;   // pointer to LCR base of PCIADA
  unsigned short far *pusCNTRL;   // pointer to PCIADA CNTRL register
  unsigned short far *pusINTCSR;  // pointer to PCIADA INTCSR register 
  
  unsigned short far *pusStat;    // pointer to Status of this VMEMM
  unsigned short far *pusIRQStat; // pointer to Status of this VMEMM
  unsigned long  far *pulADRHL;   // pointer to address register
  unsigned char  far *pucVIC68A;  // pointer to VIC68A base (first register @ 0x03)
  unsigned char  far *pucAdrMod;  // pointer to VIC68A address modifier register
  unsigned char  far *pucIRQVec;  // pointer to base of IRQ vector (first register @ 0x03)
  void           far *pvVMEWnd;   // Window in VME

  unsigned short     nIntNo;      // number of associated PC(i) interrupt

  unsigned long      ulADRHL;     // shadow of address register
  unsigned char      ucAdrMod;    // shadow of address modifier register

  int                nBusError;   // bus error notification
  int                nInterruptLine;  // which interrupt line is controlled

  unsigned char      ucWordPath;  // temporary switched to VME word path
} PCIVME_INTERFACE;

//-----------------------------------------------------------------------------
// DEFINES
//
#define PCIADAMAX  8        // don't change - max number of PCIADA

#define VME_RESET_CMD    0  // reset commands - VME only
#define LOCAL_RESET_CMD  1  // internal only
#define GLOBAL_RESET_CMD 2  // both

//-----------------------------------------------------------------------------
// PROTOTYPES
//

//-----------------------------------------------------------------------------
// First part of access functions - public for diagnostic use only
//
int  SearchPCIADA(PCIDEV *pcidev[], int *nInterfaceCount); // search all PCIADA
void PLX9050BugFix(PCIDEV *selected_device);   // PLX9050 Bug Fix
int  enable_access_to_VMEMM(PCIDEV *pcidev);   // open interface
int  disable_access_to_VMEMM(PCIDEV *pcidev);  // close it for security
int  test_connection_to_VMEMM(PCIDEV *pcidev); // pure DOS only - memory access
int  get_information_about_VMEMM(PCIDEV *pcidev, int *WordMode,
	     int *Sysctrl, int *ModuleNo, int *FPGAVer, int *ModuleTy);

//-----------------------------------------------------------------------------
// Second part of access funktions - for public use
//
PCIVME_INTERFACE *Init_Interface(int InterfaceNumber); // Inits the Interface
void  DeInit_Interface(PCIVME_INTERFACE **Interface);  // de-initialises Interface

char  ReadByte(PCIVME_INTERFACE *Interface, unsigned long Address,
			unsigned char Modifier);            // get byte from any address
void  WriteByte(PCIVME_INTERFACE *Interface, unsigned long Address,
			char Value,unsigned char Modifier); // write byte to any address
short ReadWord(PCIVME_INTERFACE *Interface, unsigned long Address,
			unsigned char Modifier);            // get word from any address
void  WriteWord(PCIVME_INTERFACE *Interface, unsigned long Address,
			short Value, unsigned char Modifier); // write word to any address
long  ReadLong(PCIVME_INTERFACE *Interface, unsigned long Address,
			unsigned char Modifier);            // read longword from any address
void  WriteLong(PCIVME_INTERFACE *Interface, unsigned long Address,
			long Value,unsigned char Modifier); // write longword to any address

int   PollIrq(PCIVME_INTERFACE *Interface);           // polls interrupt sources
char  ReadVectorByte(PCIVME_INTERFACE *Interface, int nLevel); // reads a vector byte from VME_LEVEL
int   GetError(PCIVME_INTERFACE *Interface);          // checks the ERROR flag

char  TAS(PCIVME_INTERFACE *Interface, unsigned long,unsigned short);  // 68K TAS (semfore) emulation
void  SetSfail(PCIVME_INTERFACE *Interface);          // set SYSFAIL
void  ClrSfail(PCIVME_INTERFACE *Interface);          // clear own SYSFAIL
int   PollSfail(PCIVME_INTERFACE *Interface);         // get SYSFAIL status

void  SetRegister(PCIVME_INTERFACE *Interface, unsigned short Address,
			unsigned char Value);         // write VIC68A
unsigned char GetRegister(PCIVME_INTERFACE *Interface, unsigned short Address);
						      // read VIC68A
						      // generate SYSRESET
void  Reset_VME(PCIVME_INTERFACE **Interface, short sResetType);

void  global_IRQ_disable(PCIVME_INTERFACE *Interface); // disable IRQs
void  global_IRQ_enable(PCIVME_INTERFACE *Interface);  // enable them

short interrupt_number(PCIVME_INTERFACE *Interface); // get PCI associated IRQ
void  SetVMEwordPath(PCIVME_INTERFACE *Interface);   // set VME word path
void  ClrVMEwordPath(PCIVME_INTERFACE *Interface);   // clear VME word path

#endif // __PCILIB_H__

//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------
