
//-------------------------------------------------------------------------
// WIN95 driver for PCIVME interface from ARW Elektronik, Germany ---------
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// without the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps                                     AR           12.01.98
// irq priority changed                            AR           20.02.99
// removed page_info                               AR           25.02.99
// WIN98 compatibility forced (fluch!)             AR           01.04.99
// init_irq_descriptor moved to pcivmei.c          AR           01.04.99
// version 2.04 released - WIN98 irq sharable      AR           11.04.99
// reset command added, release of version 2.5     AR           18.04.99
// release of version 2.06                         AR           27.05.99
// shared interrupts disallowed                    AR           02.02.2000
// release of version 3.00                         AR           02.02.2000
// fixed error in error return by interrupt alloc  AR           03.02.2000
// release of version 3.01                         AR           03.02.2000
// segmentation error fixed                        AR           07.02.2000
// release of version 3.03                         AR           07.02.2000
// removed error in CHECK_OUT_BUFFER               AR           11.03.2000
// release of version 3.05                         AR           11.02.2000
// release of version 3.08                         AR           20.03.2000
//
//

#define VERSION 0x0308      // version of this driver

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#include <vwin32.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <vmmreg.h>
#include <vxdwraps.h>
#include <debug.h>
#include <configmg.h>
#include <vpicd.h>
#include <winerror.h>

#include "vpcivmed.h"
#include "pciif.h"
#include "vic.h"
#include "vme.h"

#include "pcivmex.h"   // interface zu den hilfsfunktionen
#include "pcivmer.h"   // interface zu den registrierungsfunktionen
#include "pcivmep.h"   // interface zum pager
#include "pcivmei.h"   // interface to interrupt handling

//------------------------------------------------------------------------
// PRAGMAS
//
#undef  CURSEG
#define CURSEG() LCODE

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

//-------------------------------------------------------------------------
// DEFINES
//
#define PCINAME_95 "PCI\\VEN_10B5&DEV_9050"                        // start in devnode WIN95
#define PCINAME_98 "PCI\\VEN_10B5&DEV_9050&SUBSYS_11679050&REV_01" // start in devnode WIN98
#define PCINAME_99 "PCI\\VEN_10B5&DEV_9050&SUBSYS_11679050&REV_02" // start in devnode WIN98

//------ calculates the index into PCIADA lookup table --------------
#define GET_INDEX() {\
                      dwIndex = *(DWORD *)ptr->lpvInBuffer;\
                      wPCIADAIndex = checkInterfaces(wVMEMM, dwIndex);\
                      if (wPCIADAIndex >= VPCIVMED_MAX_PCIADA)\
                      {\
                        *(DWORD *)ptr->lpcbBytesReturned = 0;\
                        return ERROR_BAD_DEVICE;\
                      }\
                    }

//------ checks the available room in the output buffer -------------
#define CHECK_OUT_BUFFER(x) {\
                              if (ptr->cbOutBuffer < x)\
                              {\
                                *(DWORD *)ptr->lpcbBytesReturned = 0;\
                                return ERROR_INVALID_PARAMETER;\
                              }\
                            }

#define FIRST_IOCTL_CODE  VPCIVMED_INIT_HARDWARE
#define LAST_IOCTL_CODE   VPCIVMED_RESET                    
#define IOCTL_REVERSE(x) ((x >> 2) & 0xff)
                    
//------------------------------------------------------------------------
// TYPEDEFS
//

//-------------------------------------------------------------------------
// EXTERNALS
//
extern void  UnhookInvalidPageFault(void);
extern BOOL  HookInvalidPageFault(void);

//------------------------------------------------------------------------
// GLOBALS
//
WORD                    wDriverVersion                 = VERSION;
VPCIVMED_STATIC_STATUS  sVss[VPCIVMED_MAX_PCIADA]      = {0};
WORD                    wVMEMM[VPCIVMED_MAX_VMEMM]     = {0};
int                     iNumIfc                        = 0;
BOOL (*jump[LAST_IOCTL_CODE - FIRST_IOCTL_CODE + 1])(PDIOCPARAMETERS) = {0};

//------------------------------------------------------------------------
// FUNCTIONS
//

//-----------------------------------------------------------------
// jump entries
//
static DWORD vpcivmed_control_interrupts(PDIOCPARAMETERS ptr)
{
  VPCIVMED_IRQ_CONTROL   *psVIRQc;
  void                   *pvLcr;
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_control_interrupts()\r\n");

  GET_INDEX();       
      
  psVIRQc = (VPCIVMED_IRQ_CONTROL *)ptr->lpvInBuffer;
  pvLcr   = sVss[wPCIADAIndex].pvLcr;

  if (!(sVss[wPCIADAIndex].wNumIRQs))  // no interrupt line available
    return ERROR_IRQ_BUSY;
  
  if (psVIRQc->wEnable)
    EnableIRQ(pvLcr);
  else
    DisableIRQ(pvLcr);

  return ERROR_SUCCESS;
}        

static DWORD vpcivmed_read_vector(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  // Out_Debug_String("VPCIVMED_READ_VECTOR\r\n");

  GET_INDEX();
 
  *(DWORD *)ptr->lpcbBytesReturned = ptr->cbOutBuffer; 
 
  return get_vector_level(&sVss[wPCIADAIndex],
                (VPCIVMED_VECTOR_COMMAND *)ptr->lpvInBuffer,
                   (VPCIVMED_VECTOR_LEVEL *)ptr->lpvOutBuffer,
                                  (DWORD *)ptr->lpcbBytesReturned);
}        

static DWORD vpcivmed_access_VIC68A(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_access_VIC68A()\r\n");

  GET_INDEX();    
  
  *(DWORD *)ptr->lpcbBytesReturned = ptr->cbOutBuffer;

  return access_vic68a(&sVss[wPCIADAIndex],
       (VPCIVMED_VIC68A_ACTION *)ptr->lpvInBuffer,
              (VPCIVMED_VIC68A_ACTION *)ptr->lpvOutBuffer,
                               (DWORD *)ptr->lpcbBytesReturned);         
}

static DWORD vpcivmed_get_dynamic_status(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_get_dynamic_status()\r\n");

  GET_INDEX();
      
  *(DWORD *)ptr->lpcbBytesReturned = ptr->cbOutBuffer; 

  return get_dynamic_status(&sVss[wPCIADAIndex],
             (VPCIVMED_STANDARD_COMMAND *)ptr->lpvInBuffer,
                  (VPCIVMED_DYNAMIC_STATUS *)ptr->lpvOutBuffer,
                              (DWORD *)ptr->lpcbBytesReturned);
}
    
static DWORD vpcivmed_tas(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_tas()\r\n");

  GET_INDEX();

  *(DWORD *)ptr->lpcbBytesReturned = ptr->cbOutBuffer; 

  return make_tas(&sVss[wPCIADAIndex],
       (VPCIVMED_TAS_STRUCT *)ptr->lpvInBuffer,
                (VPCIVMED_TAS_STRUCT *)ptr->lpvOutBuffer,
                              (DWORD *)ptr->lpcbBytesReturned);
}

static DWORD vpcivmed_get_static_status(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_get_static_status()\r\n");

  GET_INDEX();

  *(DWORD *)ptr->lpcbBytesReturned = ptr->cbOutBuffer; 
 
  return get_static_status(&sVss[wPCIADAIndex],
           (VPCIVMED_STANDARD_COMMAND *)ptr->lpvInBuffer,
                 (VPCIVMED_STATIC_STATUS *)ptr->lpvOutBuffer,
                               (DWORD *)ptr->lpcbBytesReturned);
}

static DWORD vpcivmed_attach_window(PDIOCPARAMETERS ptr)
{
  VPCIVMED_STATIC_STATUS *psVss;
  VPCIVMED_ADD_WINDOW    *psVwnd;
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;
          
  Out_Debug_String("vpcivmed_attach_window(1)\r\n");

  GET_INDEX();
  CHECK_OUT_BUFFER(sizeof(DWORD *));

  Debug_Printf_DebMax("wPCIADAIndex: %d, ptr: 0x%08x\r\n", wPCIADAIndex, ptr); 

  psVss = &sVss[wPCIADAIndex];
      
  Debug_Printf_DebMax("psVss: 0x%08x\r\n", psVss); 

  psVwnd = (VPCIVMED_ADD_WINDOW *)ptr->lpvInBuffer;
      
  *(DWORD *)ptr->lpcbBytesReturned = sizeof(DWORD *);

  Debug_Printf_DebMax("returned: %d\r\n", *(DWORD *)ptr->lpcbBytesReturned); 

  if ((*(DWORD *)ptr->lpvOutBuffer = AddWindow(psVss, 
        psVwnd->modifier, psVwnd->base, psVwnd->size)) != BOGUSADDRESS)
    return ERROR_SUCCESS;
  else
    return ERROR_INVALID_PARAMETER;  
}

static DWORD vpcivmed_detach_window(PDIOCPARAMETERS ptr)
{
  VPCIVMED_STATIC_STATUS *psVss;
  VPCIVMED_REMOVE_WINDOW *pdRmv;
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;
        
  Out_Debug_String("vpcivmed_detach_window()\r\n");

  GET_INDEX();
  CHECK_OUT_BUFFER(sizeof(DWORD));

  psVss = &sVss[wPCIADAIndex];
  pdRmv = (VPCIVMED_REMOVE_WINDOW *)ptr->lpvInBuffer;

  *(DWORD *)ptr->lpcbBytesReturned = sizeof(DWORD *);
  if ((*(DWORD *)ptr->lpvOutBuffer = 
          RemoveWindow(psVss, pdRmv->pdwLinAdr)) != BOGUSADDRESS)
    return ERROR_SUCCESS;
  else
    return ERROR_INVALID_PARAMETER;
}

static DWORD vpcivmed_install_irq_handler(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_install_irq_handler()\r\n");

  GET_INDEX();      
  CHECK_OUT_BUFFER(0);

  if (!(sVss[wPCIADAIndex].wNumIRQs))  // no interrupt line available
    return ERROR_IRQ_BUSY;
    
  sIrqDesc[wPCIADAIndex].TCB_S        = Get_Cur_Thread_Handle();
  sIrqDesc[wPCIADAIndex].dwIrqHandler =
  (DWORD)((VPCIVMED_IRQ_HANDLER *)ptr->lpvInBuffer)->dwIrqHandler;   
  *(DWORD *)ptr->lpcbBytesReturned    = 0;

  return ERROR_SUCCESS;
}

static DWORD vpcivmed_init_hardware(PDIOCPARAMETERS ptr)
{
  void                   *pvLcr;
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;
  VPCIVMED_STATIC_STATUS *psVss;
        
  Out_Debug_String("vpcivmed_init_hardware()\r\n");

  *(DWORD *)ptr->lpcbBytesReturned = 0;

  VerifyHardware(wVMEMM, sVss, iNumIfc);   // get info from all VMEMMs

  GET_INDEX();

  psVss = &sVss[wPCIADAIndex];

  Debug_Printf_DebMax("VMEMM: %d PCIADA: %d\r\n", dwIndex, wPCIADAIndex); 
          
  if (!(psVss->dwLinkCount))
  {
    InitInterface((VPCIVMED_INIT_COMMAND *)ptr->lpvInBuffer, psVss->pvLcr, psVss->pvIfr);
    InitPageInfo(psVss);   

    if (psVss->wNumIRQs)       // only if a IRQ is available
      if (Enable_Phys_IRQ(psVss) == FALSE)
		  return ERROR_IRQ_BUSY;
  }
    
  psVss->dwLinkCount++;
  return ERROR_SUCCESS;
}

static DWORD vpcivmed_deinit_hardware(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;
  VPCIVMED_STATIC_STATUS *psVss;
        
  Out_Debug_String("vpcivmed_deinit_hardware()\r\n");

  *(DWORD *)ptr->lpcbBytesReturned = 0;

  GET_INDEX();
          
  psVss = &sVss[wPCIADAIndex];

  if (!(--(psVss->dwLinkCount)))
  {
    if (psVss->wNumIRQs)       // only if a IRQ is available
      Disable_Phys_IRQ(psVss);

    DeInitPageInfo(psVss);
    DeInitInterface((VPCIVMED_INIT_COMMAND *)ptr->lpvInBuffer, psVss->pvLcr, psVss->pvIfr);
  }

  return ERROR_SUCCESS;
}

static DWORD vpcivmed_get_pciada_status(PDIOCPARAMETERS ptr)
{
  return get_PCIADA_status(sVss, iNumIfc, wDriverVersion,
             (VPCIVMED_PCIADA_STATUS *)ptr->lpvOutBuffer,
                         (DWORD *)ptr->lpcbBytesReturned);
}

static DWORD vpcivmed_reset(PDIOCPARAMETERS ptr)
{
  DWORD                  dwIndex;
  WORD                   wPCIADAIndex;

  Out_Debug_String("vpcivmed_reset()\r\n");
  GET_INDEX();
          
  return make_reset(&sVss[wPCIADAIndex], (VPCIVMED_RESET_COMMAND *)ptr->lpvInBuffer,
                            (VPCIVMED_RESET_RESULT *)ptr->lpvOutBuffer, 
                                   (DWORD *)ptr->lpcbBytesReturned);
}

static DWORD vpcivmed_dioc_getversion(PDIOCPARAMETERS ptr)
{
  Out_Debug_String("vpcivmed_dioc_getversion()\r\n");

  if (ptr->cbOutBuffer < sizeof(WORD))
	  *(DWORD *)ptr->lpcbBytesReturned = 0;
  else
  {
	  *(DWORD *)ptr->lpcbBytesReturned = sizeof(WORD);
	  *(WORD  *)ptr->lpvOutBuffer      = VERSION;
  }

  return ERROR_SUCCESS;
}

static DWORD vpcivmed_dioc_closehandle(PDIOCPARAMETERS ptr)
{
  Out_Debug_String("vpcivmed_dioc_closehandle()\r\n");
  
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------
// initialize the ioctl jumptable
//
static void InitJumpTable(void)
{
  jump[IOCTL_REVERSE(VPCIVMED_INIT_HARDWARE)]       = vpcivmed_init_hardware;
  jump[IOCTL_REVERSE(VPCIVMED_DEINIT_HARDWARE)]     = vpcivmed_deinit_hardware;
  jump[IOCTL_REVERSE(VPCIVMED_ATTACH_WINDOW)]       = vpcivmed_attach_window;
  jump[IOCTL_REVERSE(VPCIVMED_DETACH_WINDOW)]       = vpcivmed_detach_window;
  jump[IOCTL_REVERSE(VPCIVMED_GET_STATIC_STATUS)]   = vpcivmed_get_static_status;
  jump[IOCTL_REVERSE(VPCIVMED_GET_DYNAMIC_STATUS)]  = vpcivmed_get_dynamic_status;
  jump[IOCTL_REVERSE(VPCIVMED_READ_VECTOR)]         = vpcivmed_read_vector;
  jump[IOCTL_REVERSE(VPCIVMED_ACCESS_VIC68A)]       = vpcivmed_access_VIC68A;
  jump[IOCTL_REVERSE(VPCIVMED_INSTALL_IRQ_HANDLER)] = vpcivmed_install_irq_handler;
  jump[IOCTL_REVERSE(VPCIVMED_CONTROL_INTERRUPTS)]  = vpcivmed_control_interrupts;
  jump[IOCTL_REVERSE(VPCIVMED_TAS)]                 = vpcivmed_tas;
  jump[IOCTL_REVERSE(VPCIVMED_GET_PCIADA_STATUS)]   = vpcivmed_get_pciada_status;
  jump[IOCTL_REVERSE(VPCIVMED_RESET)]               = vpcivmed_reset;
}

//-----------------------------------------------------------------
// called when driver is launched
//
BOOL _stdcall CVPCIVMED_Device_Init(void)
{
  DWORD retcode;
  
  Out_Debug_String("CVPCIVMED_Device_Init()\r\n");

  init_VMEMM_array(wVMEMM);
  init_static_status(sVss, VPCIVMED_MAX_PCIADA);
  init_irq_descriptor(sIrqDesc, VPCIVMED_MAX_PCIADA);
  InitPageMachine(sVss);

  iNumIfc = 0;  // the first both are alternate cause of operating system
  retcode = ScanRegistry(PCINAME_95, sVss, &iNumIfc);
  retcode = ScanRegistry(PCINAME_98, sVss, &iNumIfc);
  retcode = ScanRegistry(PCINAME_99, sVss, &iNumIfc);

  if (!iNumIfc) 
  {
	retcode = ERROR_CANTOPEN;    
	return FALSE;
  }
  
  Debug_Printf_DebMax("Anzahl der Interfaces: %d  (%d)\r\n", iNumIfc, sizeof(sVss[0])); 

  if (MakeLinearAddresses(sVss, iNumIfc) != ERROR_SUCCESS)
    return FALSE;
    
  if (HookInvalidPageFault())
    return FALSE;

  InitJumpTable();
  
  return TRUE;         
}

//-----------------------------------------------------------------
// called for each IOCTL
//
DWORD _stdcall CVPCIVMED_Device_IOctrl(DWORD  dwService, PDIOCPARAMETERS ptr) 
{
  register DWORD dwIoctlCode = dwService;
 
  if ((dwIoctlCode >= FIRST_IOCTL_CODE) && (dwIoctlCode <= LAST_IOCTL_CODE))
  {
    dwIoctlCode -= FIRST_IOCTL_CODE;
    dwIoctlCode >>= 2;

    return jump[dwIoctlCode](ptr);
  }
  else
  {
    switch(dwIoctlCode)
    {
     case(DIOC_GETVERSION):
        return vpcivmed_dioc_getversion(ptr);

     case(DIOC_CLOSEHANDLE):
        return vpcivmed_dioc_closehandle(ptr);

     default:
        Debug_Printf_DebMax("dwIOctlCode: 0x%08x\r\n", dwIoctlCode);
        return ERROR_INVALID_PARAMETER;
    }  
  }
}

//-----------------------------------------------------------------
// called when driver is removed from memory
//
BOOL _stdcall CVPCIVMED_Device_UNInit(void)
{
  int i;
  
  Out_Debug_String("CVPCIVMED_Device_UNInit()\r\n\r\n");

  for (i = 0; i < VPCIVMED_MAX_PCIADA; i++)
  {
    if (sVss[i].pvLcr != (void *)BOGUSADDRESS)
    {
      if (sVss[i].wNumIRQs)       // only if a IRQ is available
        Disable_Phys_IRQ(&sVss[i]);
    }
  }
  
  DeInitPageMachine();

  UnhookInvalidPageFault();  

  return TRUE;
}


