//-------------------------------------------------------------------------
// WIN95 driver for PCIVME interface from ARW Elektronik, Germany ---------
// all parts belonging to IRQs
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// resolved from pcivmec.c                         AR           01.03.99
// define BUG removed                              AR           01.04.99
// made interrupts sharable                        AR           11.04.99
// moved segmentation instructions                 AR           07.02.2000
// enableIrq changed bcause out of order exec      AR           16.04.2000
//

#define WITH_IRQ        // prelimenary removal of irqs - not for sale

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>
#include <vwin32.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <vmmreg.h>
#include <debug.h>
#include <vxdwraps.h>
#include <winerror.h>
#include <vpicd.h>

#include "vpcivmed.h"
#include "pciif.h"
#include "vic.h"
#include "vme.h"
#include "pcivmei.h"

//------------------------------------------------------------------------
// GLOBALS
//
IRQ_DESCRIPTOR   sIrqDesc[VPCIVMED_MAX_PCIADA]  = {0};

//------------------------------------------------------------------------
// FUNCTIONS
//

//-----------------------------------------------------------------
// init the irq descriptor
//
void init_irq_descriptor(IRQ_DESCRIPTOR asIrq[], int Anzahl)
{
  int i;
  IRQ_DESCRIPTOR *psId = (IRQ_DESCRIPTOR *)asIrq;
  
  for (i = 0; i < Anzahl; i++)
  {
    psId->dwIrqHandle  = 
    psId->dwIrqHandler = BOGUSADDRESS;

    psId++;
  }
}

//-------------------------------------------
// disable all irqs on PCIADA
//
void DisableIRQ(void *pvLCR)
{
  volatile WORD *wAdr = (WORD *)((DWORD)pvLCR + 0x4c);
  volatile dummy;

  Out_Debug_String("DisableIRQ()\r\n");

  *wAdr       &= ~0x40;  
  dummy        = *wAdr;
}

//-------------------------------------------
// enable all irqs on PCIADA
//
void EnableIRQ(void *pvLCR)
{
  volatile WORD *wAdr = (WORD *)((DWORD)pvLCR + 0x4c);
  volatile dummy;

  Out_Debug_String("EnableIRQ()\r\n");

#ifdef WITH_IRQ
  *wAdr       |= 0x40;
  dummy        = *wAdr;
#else
  DisableIRQ(pvLCR);
#endif
}

//-------------------------------------------
// clear the 2nd irq on PCIADA
//
void ClearIRQ2(void *pvLCR)
{
  volatile WORD *wAdr = (WORD *)((DWORD)pvLCR + 0x50);
  volatile dummy;

  *wAdr      &= ~0x100;
  *wAdr      |=  0x100;
  dummy       = *wAdr;
}

//-------------------------------------------
// read the vector and level of IRQ1
//
DWORD dwReadIrq1Src(VPCIVMED_STATIC_STATUS *psVss)
{
  BYTE  *bVector;
  WORD  *pwLevel = psVss->pwIRQStat;
  register DWORD dwResult;
  
  dwResult     = *pwLevel;
  bVector      = (BYTE *)((DWORD)psVss->pbVector + dwResult);
  dwResult    &= 0xe;
  dwResult   <<= 15;
  dwResult    |= *bVector;
  
  return dwResult;
}  

//-------------------------------------------
// interface to APC function ...
//
BOOL VXDINLINE __declspec(naked)
  _VWIN32_QueueUserApc(DWORD userproc, DWORD refdata, PTCB thread)
{
  VxDJmp(_VWIN32_QueueUserApc)
}

//-------------------------------------------
// the asynchrones procedure call
//
static void vAPC(DWORD dwIndex, DWORD dwMsg)
{
#ifdef WITH_IRQ
	IRQ_DESCRIPTOR *irqDesc = &sIrqDesc[dwIndex];

 if (irqDesc->dwIrqHandler != BOGUSADDRESS)
   _VWIN32_QueueUserApc(irqDesc->dwIrqHandler, dwMsg, irqDesc->TCB_S);
#endif
}

//-------------------------------------------
// process the isr
//
static BOOL Process_ISR(DWORD hirq, DWORD dwRefData)
{
  register VPCIVMED_STATIC_STATUS *psVss = 
                            (VPCIVMED_STATIC_STATUS *)dwRefData;
  register WORD test;

  void  *pvLcr				   = psVss->pvLcr;
  volatile WORD  *wAdr         = (WORD *)((DWORD)pvLcr + 0x4c);

  test = *wAdr;

  // Debug_Printf_DebMax("Process_ISR(0x%04x)\r\n", test); 

  if ((test & 0x28) == 0x28)      // PCI Irq is higher priority
  {
    ClearIRQ2(pvLcr);
#ifdef WITH_IRQ
    vAPC(psVss->dwInterface, 0x00080001);  // virtual level 8, vector 1
    VPICD_Phys_EOI(hirq);
#endif
    return TRUE;
  }

  if ((test & 0x05) == 0x05)       // VME Irq is lower priority
  {
    DisableIRQ(pvLcr);
#ifdef WITH_IRQ
    vAPC(psVss->dwInterface, dwReadIrq1Src(psVss));
    VPICD_Phys_EOI(hirq);
#endif
    return TRUE;
  }
  
  // VPICD_Phys_EOI(hirq);         // (don't) know what to do when no source ?? 
  return FALSE;
}


//-------------------------------------------
// interface to Process_isr
//
static void _declspec(naked)ISR_PCIVME(void)
{
  _asm
  {
    push  edx  ; referenzdaten
    push  eax  ; irq number
    call  Process_ISR
    add   esp,8
    cmp   eax,1
    ret
  }
}

//-------------------------------------------
// disable the interrupt in WIN95
//
BOOL Disable_Phys_IRQ(VPCIVMED_STATIC_STATUS *psVss)
{
  DWORD i = psVss->dwInterface;
  DWORD *pdwHandle = &sIrqDesc[i].dwIrqHandle;

  if (psVss->pvLcr != (void *)BOGUSADDRESS)
    DisableIRQ(psVss->pvLcr);

#ifdef WITH_IRQ
  if (*pdwHandle != BOGUSADDRESS)
  {
    Out_Debug_String("Physically_Mask()\r\n");
    VPICD_Physically_Mask(*pdwHandle);      
    VPICD_Force_Default_Behavior(*pdwHandle);
    *pdwHandle = BOGUSADDRESS;
  }
#endif
    
  sIrqDesc[i].dwIrqHandler = BOGUSADDRESS;
  
  return TRUE;
}

//-------------------------------------------
// enable the interrupts in WIN95
//
BOOL Enable_Phys_IRQ(VPCIVMED_STATIC_STATUS *psVss)
{
  DWORD i = psVss->dwInterface;
  IRQ_DESCRIPTOR *psIrqDesc = &sIrqDesc[i];

  if (psVss->pvLcr != (void *)BOGUSADDRESS)
  {
#ifdef WITH_IRQ
    psIrqDesc->vid.VID_IRQ_Number       = psVss->wLCR_IRQ;  
    psIrqDesc->vid.VID_Options          = VPICD_OPT_REF_DATA | VPICD_OPT_CAN_SHARE | VPICD_OPT_SHARE_PMODE_ONLY;	
    psIrqDesc->vid.VID_Hw_Int_Proc      = (DWORD)ISR_PCIVME;
    psIrqDesc->vid.VID_Virt_Int_Proc    = 0;	
    psIrqDesc->vid.VID_EOI_Proc         = 0;	
    psIrqDesc->vid.VID_Mask_Change_Proc = 0;	
    psIrqDesc->vid.VID_IRET_Proc        = 0;
    psIrqDesc->vid.VID_IRET_Time_Out    = 500;
    psIrqDesc->vid.VID_Hw_Int_Ref       = (DWORD)psVss;	

    psIrqDesc->dwIrqHandle  = VPICD_Virtualize_IRQ(&psIrqDesc->vid);
    psIrqDesc->dwIrqHandler = BOGUSADDRESS;

    if (psIrqDesc->dwIrqHandle != BOGUSADDRESS)
    {
       Out_Debug_String("Physically_Unmask()\r\n");
       VPICD_Physically_Unmask(psIrqDesc->dwIrqHandle);
       
       return TRUE;
    }
#else
	return TRUE;
#endif
    
    return FALSE;
  }
  else
    return FALSE;
}


DWORD get_vector_level(VPCIVMED_STATIC_STATUS *psPCIss,
                        VPCIVMED_VECTOR_COMMAND *psVc,
                          VPCIVMED_VECTOR_LEVEL *psPCIvl, DWORD *pdwLen)
{
  DWORD *pdwControl   = (DWORD *)((BYTE *)psPCIss->pvLcr + 0x50);
  DWORD *pdwIRQStatus = (DWORD *)((BYTE *)psPCIss->pvLcr + 0x4C);
  
  if (*pdwLen < sizeof(VPCIVMED_VECTOR_LEVEL))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }

  *pdwLen = sizeof(VPCIVMED_VECTOR_LEVEL);
  psPCIvl->dwInterface = psVc->dwInterface;
  psPCIvl->wPCIIrq     = (WORD)*pdwIRQStatus;

  if (*pdwControl & 0x100)   // interface enabled
  {
    switch(psVc->wAction)
    {
      case READ_CURRENT_LEVEL:
        if (*pdwIRQStatus & 0x0020)
          psPCIvl->wLevel = 8;
        else  
          psPCIvl->wLevel = (*(psPCIss->pwIRQStat) >> 1);
        return ERROR_SUCCESS;
        
      case READ_VECTOR:
      {
        BYTE *pbVector;
        WORD wLevel;

        if (*pdwIRQStatus & 0x0020)      // higher priority check
        {
          *pdwControl        &= ~0x100; // clear PCI interrupt, if set
          psPCIvl->wLevel     = 8;      // special level  8 assigned
          psPCIvl->dwStatusID = 1;      // special vector 1 assigned
          *pdwControl        |= 0x100;
          return ERROR_SUCCESS;
        }
          
        wLevel = *(psPCIss->pwIRQStat);
        if (!wLevel)
        {
          psPCIvl->wLevel      = 0;
          psPCIvl->dwStatusID  = 0;
          
          return ERROR_SUCCESS;
        }
        
        pbVector            = (BYTE *)((DWORD)(psPCIss->pbVector) + wLevel);
        psPCIvl->wLevel     = (wLevel >> 1);

        switch (psVc->wType)
        {
          case BYTE_ACCESS: psPCIvl->dwStatusID = (DWORD)*pbVector;
                            return ERROR_SUCCESS;
          case WORD_ACCESS: pbVector -= 1;
                            psPCIvl->dwStatusID = (DWORD)*(WORD *)pbVector;
                            return ERROR_SUCCESS;
          case LONG_ACCESS: pbVector -= 3;
                            psPCIvl->dwStatusID = *(DWORD *)pbVector;
                            return ERROR_SUCCESS;
        }
      }
    }
  }

  return ERROR_INVALID_PARAMETER;  
}                   

//----------------------------------------------------------------
