//-------------------------------------------------------------------------
// Driver for PCIVME Interface of ARW Elektronik --------------------------
// The page mechanism
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// What                                            Who          When
// first steps                                     AR           26.01.98
// reducing the amount of different structures     AR           25.01.99
// small changes at AddWindow()                    AR           17.02.2000
// debug info in addWindow removed                 AR           16.03.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

//-------------------------------------------------------------------------
// INCLUDES
//
#include <basedef.h>
#include <vmm.h>

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

#include <debug.h>
#include <vxdwraps.h>

//------------------------------------------------------------------------
// PRAGMAS
//
#undef  CURSEG
#define CURSEG() LCODE

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include "vpcivmed.h"
#include "pciif.h"
#include "vic.h"
#include "pcivmep.h"     

//-------------------------------------------------------------------------
// DEFINES
//
#define MAX_ADR_WINDOWS 8                       // die maximal zulaessige anzahl offener fenster

#define MAKE_PAGE(x) (((DWORD)x) >> 12)         // berechnet die page zu einer lin adresse
#define PAGE_MASK(x) (((DWORD)x) & 0xfffff000)  // maskiert die page_basis heraus

#define PAGE_BOGUSADDRESS (PAGE_MASK(BOGUSADDRESS)) // special for pages

//--------------------------------------------------------------------------
// TYPEDEFS
//
typedef struct adr_window
{
  DWORD lower_bound;                    // die lineare basis der reservierten pages
  DWORD upper_bound;                    // die letzte addresse innerhalb der res. pages
  DWORD vme_base;                       // die zugeh. basis auf vme-seite
  WORD  modifier;                       // die zugeh. address modifier
  WORD  link_count;                     // die anzahl der nutzer
  VPCIVMED_STATIC_STATUS *psVss;        // bezug zum interface
} ADR_WINDOW;

//-------------------------------------------------------------------------
// GLOBALS
//
ADR_WINDOW wnds[MAX_ADR_WINDOWS]          = {0};

//-------------------------------------------------------------------------
// functions which are not acailable in ddk
//
ULONG VXDINLINE PageReserve(ULONG page, ULONG npages, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push npages
  _asm push page
  VMMCall(_PageReserve);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

ULONG VXDINLINE PageCommitPhys(ULONG page, ULONG npages, ULONG physpg, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push physpg
  _asm push npages
  _asm push page
  VMMCall(_PageCommitPhys);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

ULONG VXDINLINE PageDecommit(ULONG page, ULONG npages, ULONG flags)
{
  ULONG result;
  
  Touch_Register(ECX);
  Touch_Register(EDX);
  
  _asm push flags
  _asm push npages
  _asm push page
  VMMCall(_PageDecommit);
  _asm add esp,12
  _asm mov result,eax

  return result;
}

//--------------------------------------------------------------------------
// FUNCTIONS
//
void InitPageMachine(VPCIVMED_STATIC_STATUS sVss[])
{
  int i;
  ADR_WINDOW *psAdrWindow;
  
  Out_Debug_String("InitPageMachine()\r\n");

  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    psAdrWindow->lower_bound = 
    psAdrWindow->upper_bound =
    psAdrWindow->vme_base    = BOGUSADDRESS;
    psAdrWindow->modifier    = 0xFF;
    psAdrWindow->link_count  = 0;
    psAdrWindow->psVss       = (VPCIVMED_STATIC_STATUS *)BOGUSADDRESS;
  }
}

//------------------------------------------------------------
// prueft bei einem page fault, ob ein page umschalten moeglich ist
//
BOOL CheckAndSetPage(struct IPF_Data *ipf)
{
  int i;
  DWORD vme_page;
  VPCIVMED_STATIC_STATUS *psVss;
  ADR_WINDOW *psAdrWindow = wnds;

  
  // Debug_Printf_DebMax("CheckAndSetPage() 0x%08x\r\n", ipf->IPF_LinAddr); 

  for (i = 0; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    if (psAdrWindow->lower_bound != BOGUSADDRESS)
    {
	  // Debug_Printf_DebMax("lower_bound: 0x%08x, upper_bound: 0x%08x\r\n", psAdrWindow->lower_bound, psAdrWindow->upper_bound); 
	  
      if ((ipf->IPF_LinAddr >= psAdrWindow->lower_bound) &&
                    (ipf->IPF_LinAddr < psAdrWindow->upper_bound))
      {
        psVss = psAdrWindow->psVss;
      
        // falls es eine dwActivePage gibt -> freigeben
        if (psVss->dwActivePage != PAGE_BOGUSADDRESS)
        {
          if (PageDecommit(MAKE_PAGE(psVss->dwActivePage), 1, 0) == 0)
            return FALSE;
          psVss->dwActivePage = PAGE_BOGUSADDRESS;  
        }

        // neue dwActivePage zuordnen
        if (PageCommitPhys(MAKE_PAGE(ipf->IPF_LinAddr), 1,
                  psVss->dwPagePhysVME, (PC_USER | PC_WRITEABLE)) == 0)
          return FALSE;

        psVss->dwActivePage = PAGE_MASK(ipf->IPF_LinAddr);

        // hier muss noch der modifier und die vme_basis gesetzt werden
        vme_page = ipf->IPF_LinAddr - psAdrWindow->lower_bound + psAdrWindow->vme_base;

        // hier kann durch caching noch optimiert werden!
        *(DWORD *)(psVss->pdwVMEAdr)  = vme_page;
        *(BYTE *) (psVss->pbModifier) = psAdrWindow->modifier;
      
        return TRUE;
      }
    }      
  }
  
  return FALSE;                            // page nicht gefunden
}

//--------------------------------------------------------------------------------------
// fuegt ein fenster zum vmebus hinzu (size + vme_base must be page aligned!)
//
DWORD AddWindow(VPCIVMED_STATIC_STATUS *psVss, WORD modifier, DWORD vme_base, DWORD size) 
{
  int i;
  ADR_WINDOW *psAdrWindow = wnds;
  
  Out_Debug_String("AddWindow()\r\n");

  // suche einen u.u. vorhandenen passenden eintrag
  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    if ((psAdrWindow->modifier == modifier) &&
         (psAdrWindow->vme_base == vme_base) &&
           ((psAdrWindow->upper_bound - psAdrWindow->lower_bound) >= size) &&  //AR01
             (psAdrWindow->psVss == psVss))
    {
      // einen passenden eintrag gefunden
      if (psAdrWindow->link_count == 0xFFFF)  // zu viele links?
        return BOGUSADDRESS;
	  else
	    psAdrWindow->link_count += 1;

      return psAdrWindow->lower_bound;
    }
  }

  // sonst suche einen freien eintrag
  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    if (psAdrWindow->lower_bound == BOGUSADDRESS)
      break;
  }  

  // keinen freien eintrag gefunden
  if (i >= MAX_ADR_WINDOWS)
    return BOGUSADDRESS;

  psAdrWindow->lower_bound = PageReserve(PR_SHARED, MAKE_PAGE(size), PR_FIXED | PR_4MEG);

  // hat wohl nicht geklappt?
  if (psAdrWindow->lower_bound == BOGUSADDRESS)
    return BOGUSADDRESS;

  if (psAdrWindow->link_count == 0xFFFF)  // zu viele links?
    return BOGUSADDRESS;
  else
  {
	  psAdrWindow->link_count += 1;
	  psAdrWindow->upper_bound = psAdrWindow->lower_bound + size;
	  psAdrWindow->vme_base    = vme_base;
	  psAdrWindow->modifier    = modifier;
	  psAdrWindow->psVss       = psVss;        // link window to interface

	  return psAdrWindow->lower_bound;
  }
}

//------------------------------------------------------------------------------
// entfernt ein fenster zum vmebus
//
DWORD RemoveWindow(VPCIVMED_STATIC_STATUS *psVss, DWORD *pdwLinAdr)
{
  int i;
  ADR_WINDOW *psAdrWindow;

  Out_Debug_String("RemoveWindow()\r\n");

  // mogeln gilt nicht
  if ((DWORD)pdwLinAdr == BOGUSADDRESS)
    return BOGUSADDRESS;
    
  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    if ((psAdrWindow->lower_bound == (DWORD)pdwLinAdr) && (psAdrWindow->psVss == psVss))
    {
      psAdrWindow->link_count--;

      // letzter nutzer, reservierte pages freigeben
      if (psAdrWindow->link_count == 0)
      {
        // sollte die dwActivePage innerhalb liegen -> entfernen
        if ((psVss->dwActivePage >= psAdrWindow->lower_bound) &&
                         (psVss->dwActivePage < psAdrWindow->upper_bound))
		{
		  // Debug_Printf_DebMax("PageDecommit(0x%08x)\r\n", MAKE_PAGE(psVss->dwActivePage)); 

          if (PageDecommit(MAKE_PAGE(psVss->dwActivePage), 1, 0) == 0)
            return BOGUSADDRESS;
          else
            psVss->dwActivePage = PAGE_BOGUSADDRESS;    
		}
        
		// Debug_Printf_DebMax("_PageFree(0x%08x)\r\n", psAdrWindow->lower_bound); 

        _PageFree((PVOID)psAdrWindow->lower_bound, 0);        
        psAdrWindow->psVss       = (VPCIVMED_STATIC_STATUS *)BOGUSADDRESS;  
        psAdrWindow->lower_bound = BOGUSADDRESS;
      }

	  Out_Debug_String("RemoveWindow().OK\r\n");

      return (DWORD)pdwLinAdr;
    }
  }
  
  return BOGUSADDRESS;
}

//------------------------------------------------------------------------------
// entfernt !alle! fenster zum vmebus
//
static DWORD RemoveAllWindows(void)
{
  int i;
  ADR_WINDOW *psAdrWindow;

  Out_Debug_String("RemoveAllWindows()\r\n");

  // alle reservierten pages zurueckgeben
  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    // active page entfernen --------------------------------
    if (psAdrWindow->psVss->dwActivePage != PAGE_BOGUSADDRESS)
    {
      if (PageDecommit(MAKE_PAGE(psAdrWindow->psVss->dwActivePage), 1, 0) == 0)
        return BOGUSADDRESS;
      else
        psAdrWindow->psVss->dwActivePage = PAGE_BOGUSADDRESS;
    }
  
    if (psAdrWindow->lower_bound != BOGUSADDRESS)
    {
      _PageFree((PVOID)psAdrWindow->lower_bound, 0);
      psAdrWindow->lower_bound = BOGUSADDRESS;
    }
  }  

  return 0;
}

//-------------------------------------------------------------
// initialisiert die fenster !eines! interfaces
//
void InitPageInfo(VPCIVMED_STATIC_STATUS *psVss)
{
  Out_Debug_String("InitPageInfo()\r\n");

  psVss->dwPagePhysVME = MAKE_PAGE((DWORD)psVss->dUSR_MemBase + VMEBASE);
  psVss->dwActivePage  = PAGE_BOGUSADDRESS;          
  *psVss->pdwVMEAdr    = PAGE_BOGUSADDRESS;
}

//-------------------------------------------------------------
// entfernt die fenster !eines! interfaces zum vmebus
//
DWORD DeInitPageInfo(VPCIVMED_STATIC_STATUS *psVss)
{
  int i;
  ADR_WINDOW *psAdrWindow;
  
  Out_Debug_String("InitPageInfo()\r\n");

  // active page entfernen --------------------------------
  if (psVss->dwActivePage != PAGE_BOGUSADDRESS)
  {
    if (PageDecommit(MAKE_PAGE(PAGE_MASK(psVss->dwActivePage)), 1, 0) == 0)
      return BOGUSADDRESS;
    else
      psVss->dwActivePage = PAGE_BOGUSADDRESS;
  }
  
  // alle reservierten pages zurueckgeben
  for (i = 0, psAdrWindow = wnds; i < MAX_ADR_WINDOWS; i++, psAdrWindow++)
  {
    if ((psAdrWindow->psVss == psVss) &&
               (psAdrWindow->lower_bound != BOGUSADDRESS))
    {
      _PageFree((PVOID)psAdrWindow->lower_bound, 0);
      psAdrWindow->lower_bound = BOGUSADDRESS;
    }
  }  

  return 0;
}

void DeInitPageMachine(void)
{
  Out_Debug_String("DeInitPageMachine()\r\n");

  RemoveAllWindows();  
}


