//-------------------------------------------------------------------------
// WIN95-Treiber for the  PCI-VME-Interface of ARW Elektronik -------------
// all about configuration and registry
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
//
// What                                            who          when
// taken out of pcivmec.c                          AR           27.06.98
// removed code for PLX9050BugFix                  AR           06.03.99
// PLX9052 inserted                                AR           12.03.99
// take out HARDWAREID2                            AR           12.03.99
// PLX9052 removed                                 AR           16.03.99
// GET_OUT_OF_REGISTRY for HW_REVISION + SUB_DEV   AR           22.03.99
// Corrections because of WIN98 ??                 AR           01.04.99
// second open in same pg. caused 0xffff0000 adrsd AR           07.01.2000
// moved segmentation instructions                 AR           07.02.2000
// check for moduleType added                      AR           11.03.2000
// change in interrupt line association            AR           20.03.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <vmmreg.h>
#include <debug.h>
#include <vxdwraps.h>
#include <configmg.h>

#include "pcivmer.h"    
#include "pciif.h"
#include "pcivmeb.h"

//------------------------------------------------------------------------
// PRAGMAS
//
#undef  CURSEG
#define CURSEG() LCODE

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

//-------------------------------------------------------------------------
// DEFINES
//
#define HW_REVISION "001"                    // muss >= hardware rev. sein

#ifdef GET_OUT_OF_REGISTRY
#define HARDWAREID  "PCI\\VEN_10B5&DEV_9050&SUBSYS_11679050"
#else
#define HARDWAREID (DWORD)0x11679050
#endif

#define REGSTR_MAX_LENGTH 255  // maximum expected length

#define VMEMM_WINDOW_SIZE 0x2000
#define VMEMM_MODULE_TYPE 0x1000

//-------------------------------------------------------------------------
// TYPEDEFS
//
typedef DWORD HKEY;

//-------------------------------------------------------------------------
// EXTERNALS
//

//-------------------------------------------------------------------------
// LOKALE FUNKTIONEN
//
static char *kstrcpy(char *dst, char *src)
{
  while (*src != 0)
    *dst++ = *src++;
  *dst = 0;

  return dst; 
}

static BOOL kstrcmp(char *st0, char *st1)
{
  while (*st0 != 0)
  {
    if (*st0++ != *st1++)
      return FALSE;
  }

  return TRUE; 
}

//-------------------------------------------------------------------------
// request the subsystem id
//
#ifdef GET_OUT_OF_REGISTRY
DWORD eval_SubSystem(DEVNODE devnode)
{
  DWORD cret;
  DWORD len = 255;
  char buff[255];

  // HardwareID holen ---------------------------------
  cret = CM_Read_Registry_Value(devnode, NULL, REGSTR_VAL_HARDWAREID,
                    REG_SZ, (void *)buff, &len, CM_REGISTRY_HARDWARE);

  if (cret != CR_SUCCESS) 
    return cret;

  if (kstrcmp(HARDWAREID, buff) == TRUE)
    return CR_SUCCESS;
  else  
    return CR_NO_SUCH_VALUE;
}
#endif

//-------------------------------------------------------------------------
// FUNKTIONEN
//
DWORD GetCMInfo(char *name, VPCIVMED_STATIC_STATUS *psVss)
{
  DEVNODE   dn0;
  CMCONFIG  ccBuffer;
  CONFIGRET cret;
  BYTE      buff[8];
  DWORD     len = 8;
  
  if((cret = CM_Locate_DevNode(&dn0, name, 0)) == CR_SUCCESS)
  {
    Out_Debug_String("CM_Locate_DevNode.\r\n");

#ifdef GET_OUT_OF_REGISTRY
    if ((cret = eval_SubSystem(dn0)) != CR_SUCCESS)
      return cret;
#endif

    if ((cret = CM_Get_Alloc_Log_Conf(&ccBuffer, dn0,
                                         CM_GET_ALLOC_LOG_CONF_ALLOC))
                                                           == CR_SUCCESS)
    {
      Out_Debug_String("CM_Get_Alloc_Log_Conf\r\n");

      psVss->wNumMemWindows = ccBuffer.wNumMemWindows;
      psVss->wNumIOPorts    = ccBuffer.wNumIOPorts;
      psVss->wNumIRQs       = ccBuffer.wNumIRQs;
      psVss->wNumDMAs       = ccBuffer.wNumDMAs;
      
#ifdef GET_OUT_OF_REGISTRY
      // revision lesen -----------------------------------
      cret = CM_Read_Registry_Value(dn0, NULL, REGSTR_VAL_HWREV,
                                          REG_SZ, (void *)buff, &len,
                                             CM_REGISTRY_HARDWARE);
#endif

      if (cret == CR_SUCCESS)                                     
      {
        DWORD dwValue;
        
        Out_Debug_String("CM_Read_Registry_Value.\r\n");

#ifdef GET_OUT_OF_REGISTRY
        buff[7] = 0;        
        kstrcpy(psVss->cszHWRevision, buff);
#else
        kstrcpy(psVss->cszHWRevision, HW_REVISION);
#endif

        // info extrahieren -------------------------------
        psVss->dLCR_MemBase   = ccBuffer.dMemBase[0];
        psVss->dLCR_MemLength = ccBuffer.dMemLength[0];
                                        
        psVss->wLCR_IOBase    = ccBuffer.wIOPortBase[0];
        psVss->wLCR_IOLength  = ccBuffer.wIOPortLength[0];     

        //psVss->wLCR_IRQ       = ccBuffer.bIRQRegisters[0];

        psVss->dUSR_MemBase   = ccBuffer.dMemBase[1];
        psVss->dUSR_MemLength = ccBuffer.dMemLength[1];

		if (psVss->dUSR_MemLength != VMEMM_WINDOW_SIZE) 
			return CR_NO_SUCH_VALUE;

		// interrupt ist special
		{
			DWORD dwPCRreg3c;

			dwReadPCR(psVss->dUSR_MemBase, 0x3C >> 2, &dwPCRreg3c);
			psVss->wLCR_IRQ = dwPCRreg3c & 0x000000ff;
		}

#ifndef GET_OUT_OF_REGISTRY
        dwReadPCR(psVss->dUSR_MemBase, 0x2C >> 2, &dwValue);
        if (dwValue != HARDWAREID) return CR_NO_SUCH_VALUE;
#endif
        // potential BUG fix ------------------------------
        PLX9050BugFix(psVss);

        Out_Debug_String("GetCMInfo().OK\r\n");

        return CR_SUCCESS;
      }
      else
        return cret;
    }
    else
      return cret;
  }
  else
    return cret;
}

//---------------------------------------------------------------------
// checks a connection with a small test pattern
//
static WORD TestConnection(VPCIVMED_STATIC_STATUS *psVss)
{  
  WORD wRet;
  WORD retCode;
  DWORD dwStore;
  WORD  *pwADRH   = (WORD *)((DWORD)(psVss->pvIfr) + (DWORD)ADRH);
  WORD  *pwADRL   = (WORD *)((DWORD)(psVss->pvIfr) + (DWORD)ADRL);
  DWORD *pdwADRHL = (DWORD *)((DWORD)(psVss->pvIfr) + (DWORD)ADRHL);
  int i;

  Out_Debug_String("TestConnection().Start\r\n");

  dwStore = *pdwADRHL;
  for (i = 0; i < 10000; i++)
  {
    *pwADRH = 0x5555;
    *pwADRL = 0xAAAA;
    wRet    = *pwADRH;
    if (wRet != 0x5555) return E_CON_ERROR;
    
    *pwADRH = 0xAAAA;
    *pwADRL = 0x5555;
    wRet    = *pwADRH;
    if (wRet != 0xAAAA) return E_CON_ERROR;

    *pwADRH = 0x0000;
    *pwADRL = 0xFFFF;
    wRet    = *pwADRH;
    if (wRet != 0x0000) return E_CON_ERROR;

    *pwADRH = 0xFFFF;
    *pwADRL = 0x0000;
    wRet    = *pwADRH;
    if (wRet != 0xFFFF) return E_CON_ERROR;
  }

  *pdwADRHL = dwStore;

  Out_Debug_String("TestConnection().OK\r\n");

  return E_NO_ERROR;
}


//------------------------------------------------------------------
// fill the VMEMM pointer array (new)
//
static void fill_VMEMM_array(WORD wVMEMM[], VPCIVMED_STATIC_STATUS cpsVss[])
{
  int i;
  VPCIVMED_STATIC_STATUS *psVss = (VPCIVMED_STATIC_STATUS *)cpsVss;

  Out_Debug_String("fill_VMEMM_array()\r\n");

  for (i = 0; i < VPCIVMED_MAX_VMEMM; i++)
    wVMEMM[i] = 0xffff;

  for (i = 0; i < VPCIVMED_MAX_PCIADA; i++)
  {
    if ((psVss->wConnected) &&
        (psVss->wModuleNumber < VPCIVMED_MAX_VMEMM))
    {
      Debug_Printf_DebMax("Found ModuleNumber %d\r\n",
                             psVss->wModuleNumber);
      wVMEMM[psVss->wModuleNumber] = i;
    }
    psVss++;    
  }
}

//------------------------------------------------------------------
// init VMEMM pointer array
//
void init_VMEMM_array(WORD wVMEMM[])
{
  int i;

  for (i = 0; i < VPCIVMED_MAX_VMEMM; i++) wVMEMM[i] = 0xffff;
}


//------------------------------------------------------------------
// init the structure contents
//
void init_static_status(VPCIVMED_STATIC_STATUS cpsVss[], int maxInterfaces)
{
  int i;
  VPCIVMED_STATIC_STATUS *psVss = (VPCIVMED_STATIC_STATUS *)cpsVss;
  
  for (i = 0; i < maxInterfaces; i++)     
  {
    psVss->dwInterface      = i; 
    psVss->dIfcStatus       = E_EMPTY;

    psVss->wNumMemWindows   = 0;  
    psVss->wNumIOPorts      = 0;
    psVss->wNumIRQs         = 0;
    psVss->wNumDMAs         = 0;
      
    psVss->dLCR_MemBase     = BOGUSADDRESS; 
    psVss->dLCR_MemLength   = 0;
    psVss->wLCR_IOBase      = BOGUSADDRESS;
    psVss->wLCR_IOLength    = 0;
  
    psVss->wLCR_IRQ         = 0xffff;

    psVss->dUSR_MemBase     = BOGUSADDRESS;
    psVss->dUSR_MemLength   = 0;

    psVss->wModuleType      = 0xffff;
    psVss->wFPGAVersion     = 0xffff;
    psVss->wModuleNumber    = 0xffff;

    psVss->wWordMode        = 0;

    psVss->wSysControl      = 0;
    psVss->wConnected       = 0;

    psVss->dwLinkCount      = 0;
  
    psVss->pvLcr            = (PVOID)BOGUSADDRESS; 
    psVss->pvIfr            = (PVOID)BOGUSADDRESS; 

    psVss->pwCSR            = (WORD *)BOGUSADDRESS;
    psVss->pwIRQStat        = (WORD *)BOGUSADDRESS;
    psVss->pbVector         = (BYTE *)BOGUSADDRESS;
    psVss->pdwVMEAdr        = (DWORD *)BOGUSADDRESS;
    psVss->pbModifier       = (BYTE *)BOGUSADDRESS;
    psVss->pvVME            = (void *)BOGUSADDRESS;

    psVss->dwPagePhysVME    = BOGUSADDRESS;

    psVss->psIrqDescriptor  = (void *)BOGUSADDRESS;

    psVss->dwActivePage     = BOGUSADDRESS & 0xfffff000;

    psVss->wReserve2        = 0;
    
    psVss->cszHWRevision[0] = 0;

    psVss++;
  }
}
 

//------------------------------------------------------------------
// looks the registry for interfaces
//
DWORD ScanRegistry(char *KeyName, VPCIVMED_STATIC_STATUS cpsVss[], int *piNumIfc)
{
  DWORD     ret = 0;
  HKEY      hKey;
  DWORD     i;
  char      buffer[REGSTR_MAX_LENGTH];
  CONFIGRET cret;
  VPCIVMED_STATIC_STATUS *psVss = cpsVss;
  int       maxIfc              = VPCIVMED_MAX_PCIADA;

  if (maxIfc == 0)
    return ERROR_SUCCESS;   // seltsam, aber korrekt

  buffer[0] = 0;
  ret = _GetRegistryKey(REGTYPE_ENUM, KeyName, REGKEY_OPEN, &hKey);

  Debug_Printf_DebMax("_GetRegistryKey(%s)\r\n", KeyName);

  if(ret == ERROR_SUCCESS)
  {
    i = 0;  
	psVss += *piNumIfc; // step forward (PLX8052)
    do
    {     
      ret = _RegEnumKey(hKey, i++,
              kstrcpy(kstrcpy(buffer, KeyName),"\\"),
                                  (DWORD)REGSTR_MAX_LENGTH);

      Debug_Printf_DebMax("_RegEnumKey(%s)\r\n",buffer);

      if (ret == ERROR_SUCCESS)
      {
        cret = GetCMInfo(buffer, psVss);

        if (cret == CR_SUCCESS)
        {
          psVss++;
          (*piNumIfc)++;
        }
      }
    } while ((ret == ERROR_SUCCESS) && (*piNumIfc < maxIfc));

    _RegCloseKey(hKey);

    return ERROR_SUCCESS;
  }

  return ret;
}

//--------------------------------------------------------------
// prueft die Verbindung zu VMEMM ------------------------------
//
WORD checkVMEMM(VPCIVMED_STATIC_STATUS *psVss)
{
  WORD wCntrl;
  WORD *pwLcrReg = (WORD *)((DWORD)(psVss->pvLcr) + 0x50);
  WORD error = 0;

  if (psVss->pwCSR == (WORD *)BOGUSADDRESS)
    return E_INCOMPATIBLE;

  wCntrl = *pwLcrReg; // store

  Debug_Printf_DebMax("CheckVMEMM %04x\r\n", wCntrl);

  *pwLcrReg  = RELEASE_VMEMM;

  if (wCntrl & 0x0800)
  {
	WORD  wVMEMMStatus;

	wVMEMMStatus  = *(psVss->pwCSR);

	if ((wVMEMMStatus & MASK_MODTYPE) == VMEMM_MODULE_TYPE)
	{
		if (TestConnection(psVss) == E_NO_ERROR)
		{      
		  psVss->wConnected    = 1;

		  psVss->wWordMode     = (wVMEMMStatus & FLAG_WORD) ? TRUE : FALSE;
		  psVss->wSysControl   = (wVMEMMStatus & FLAG_SYSCTL) ? TRUE : FALSE;
		  psVss->wModuleNumber = (wVMEMMStatus & MASK_MODNR) >> 4;
		  psVss->wFPGAVersion  = (wVMEMMStatus & MASK_FPGA)  >> 8;
		  psVss->wModuleType   = (wVMEMMStatus & MASK_MODTYPE) >> 12;
  		  psVss->dIfcStatus    = E_NO_ERROR;
		}
	}
  }

  *pwLcrReg = wCntrl;  // restore

  return error;
}  

//-----------------------------------------------------------------
// allocates linear addresses
//
DWORD MakeLinearAddresses(VPCIVMED_STATIC_STATUS cpsVss[], int iNumIfc)
{
  VPCIVMED_STATIC_STATUS *psVss = (VPCIVMED_STATIC_STATUS *)cpsVss;
  int i, j;
  char *t, *p;
  
  for (i = 0; i < iNumIfc; i++)
  {
    psVss->dIfcStatus = E_NO_ERROR;
    
    // 1. resource usage OK? ----------------     
    if (psVss->dIfcStatus == E_NO_ERROR)
    {
      if ((psVss->wNumIRQs       != 1) || /* ???? WIN98 ???? */
          (psVss->wNumDMAs       != 0) ||
          (psVss->wNumMemWindows != 3) ||
          (psVss->wNumIOPorts    != 2))
        psVss->dIfcStatus = E_INCOMPATIBLE;
    }

    // 2. map LCR -----------------------------
    if (psVss->dIfcStatus == E_NO_ERROR)
    {
      psVss->pvLcr =
        _MapPhysToLinear(psVss->dLCR_MemBase, psVss->dLCR_MemLength, 0);

      if (psVss->pvLcr == (PVOID)BOGUSADDRESS)
        psVss->dIfcStatus = E_NO_ADDRESS;      
    }

    // 3. map IFR -----------------------------  
    if (psVss->dIfcStatus == E_NO_ERROR)
    {
      psVss->pvIfr =
        _MapPhysToLinear(psVss->dUSR_MemBase, psVss->dUSR_MemLength, 0);
        
      if (psVss->pvIfr == (PVOID)BOGUSADDRESS)
        psVss->dIfcStatus = E_NO_ADDRESS;
    }

    // 4. initialize some fast access addresses --
    if (psVss->dIfcStatus == E_NO_ERROR)
    {
      psVss->pwCSR      = (WORD *) ((DWORD)(psVss->pvIfr) + CSR); 
      psVss->pbModifier = (BYTE *) ((DWORD)(psVss->pvIfr) + VICBASE + AMSR);
      psVss->pdwVMEAdr  = (DWORD *)((DWORD)(psVss->pvIfr) + ADRHL);
      psVss->pwIRQStat  = (WORD *) ((DWORD)(psVss->pvIfr) + VICRES);
      psVss->pbVector   = (BYTE *) ((DWORD)(psVss->pvIfr) + VECBASE);
      psVss->pvVME      = (void *) ((DWORD)(psVss->pvIfr) + VMEBASE);

      *(WORD *)((BYTE *)psVss->pvLcr + 0x4c) &= ~0x40;         // disable irq
      *(WORD *)((BYTE *)psVss->pvLcr + 0x50)  = INHIBIT_VMEMM; // inhibit
    }

    // 5. set status to not connected ------------
    if (psVss->dIfcStatus == E_NO_ERROR)
      psVss->dIfcStatus = E_NOT_CONNECTED;

    // the next one
    psVss++;
  }
  
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------
// checks the VMEMM and gets some information about
//
DWORD VerifyHardware(WORD wVMEMM[], VPCIVMED_STATIC_STATUS cpsVss[], int iNumIfc)
{
  VPCIVMED_STATIC_STATUS *psVss = cpsVss;
  int i;
  
  Out_Debug_String("VerifyHardware()\r\n");
  
  for (i = 0; i < iNumIfc; i++)
  {
    checkVMEMM(psVss);

    psVss++;
  }

  fill_VMEMM_array(wVMEMM, cpsVss);

  return E_NO_ERROR;
}


//-----------------------------------------------------------------
// returns the PCIADA index out of a VMEMM lookup  
//
WORD checkInterfaces(WORD wVMEMM[], DWORD dwIndex)
{
  if (dwIndex >= VPCIVMED_MAX_VMEMM)
    return VPCIVMED_MAX_PCIADA;
    
  return wVMEMM[dwIndex];
}  

//---------------------------- ENDE --------------------------------------
