//-------------------------------------------------------------------------
// helping functions for the PCIVME interface of ARW Elektronik -----------
// 
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// first steps                                     AR           26.01.98
// SYSFAIL interrupt mask changed                  AR           10.02.99
// VIC68A_WRITE_ONLY added                         AR           17.03.99
// VME reset functions added                       AR           18.04.99
// DISABLE_PCIADA_.. included                      AR           18.04.99
// segment ordering for const_seg changed          AR           07.01.2000
// const declarations removed (segmentation)       AR           07.02.2000
//

//-------------------------------------------------------------------------
// INCLUDES
//
#define WANTVXDWRAPS

#include <basedef.h>
#include <vmm.h>

#ifdef DEBLEVEL
#undef DEBLEVEL
#endif
#define DEBLEVEL DEBLEVELMAX

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG
#pragma check_stack(off)                   // No stack checking for routines in this module

#include <debug.h>
#include <vxdwraps.h>
#include <winerror.h>

#include "pcivmex.h"
#include "vpcivmed.h"
#include "pciif.h"

//-------------------------------------------------------------------------
// DEFINES
//

//-------------------------------------------------------------------------
// EXTERNALS
//
#ifdef BUG
extern DWORD Reg4CShadow;
extern DWORD Reg50Shadow;
#endif

//-------------------------------------------------------------------------
// GLOBALS
//

//------------------------------------------------------------------------
// PRAGMAS
//
#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

VPCIVMED_INIT_ELEMENT init_element[] =
         {{LCR,  WORD_ACCESS, 0x4c,       DISABLE_PCIADA_IRQS}, // disable interrupts
          {LCR,  WORD_ACCESS, 0x50,       RELEASE_VMEMM},       // enable interface

          {VIC,  BYTE_ACCESS, (WORD)0x03, 0xf8+1},    // VIICR

          {VIC,  BYTE_ACCESS, (WORD)0x07, 0x78+1},    // VICR1
          {VIC,  BYTE_ACCESS, (WORD)0x0b, 0x78+2},
          {VIC,  BYTE_ACCESS, (WORD)0x0f, 0x78+3},
          {VIC,  BYTE_ACCESS, (WORD)0x13, 0x78+4},
          {VIC,  BYTE_ACCESS, (WORD)0x17, 0x78+5},
          {VIC,  BYTE_ACCESS, (WORD)0x1b, 0x78+6},
          {VIC,  BYTE_ACCESS, (WORD)0x1f, 0x78+7},    // VICR7

          {VIC,  BYTE_ACCESS, (WORD)0x23, 0xf8+0},    // DSICR

          {VIC,  BYTE_ACCESS, (WORD)0x27, 0xf8+1},    // LICR1
          {VIC,  BYTE_ACCESS, (WORD)0x2b, 0xf8+2},
          {VIC,  BYTE_ACCESS, (WORD)0x2f, 0xf8+3},
          {VIC,  BYTE_ACCESS, (WORD)0x33, 0xf8+4},
          {VIC,  BYTE_ACCESS, (WORD)0x37, 0xf8+5},
          {VIC,  BYTE_ACCESS, (WORD)0x3b, 0x38+6},
          {VIC,  BYTE_ACCESS, (WORD)0x3f, 0x38+7},    // LICR7

          {VIC,  BYTE_ACCESS, (WORD)0x43, 0xf8+2},    // ICGS
          {VIC,  BYTE_ACCESS, (WORD)0x47, 0xf8+3},    // ICMS

          {VIC,  BYTE_ACCESS, (WORD)0x4b, 0xf8+6},    // EGICR

          {VIC,  BYTE_ACCESS, (WORD)0x4f, 0x08},      // ICGS-IVBR (!)
          {VIC,  BYTE_ACCESS, (WORD)0x53, 0x0c},      // ICMS-IVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x57, 0x00},      // LIVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x5b, 0x10},      // EGIVBR (!)

          {VIC,  BYTE_ACCESS, (WORD)0x5f, 0x00},      // ICSR

          {VIC,  BYTE_ACCESS, (WORD)0x63, 0x00},      // ICR0
          {VIC,  BYTE_ACCESS, (WORD)0x67, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x6b, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x6f, 0x00},
          {VIC,  BYTE_ACCESS, (WORD)0x73, 0x00},      // ICR4

          {VIC,  BYTE_ACCESS, (WORD)0x83, 0xfe},      // VIRSR

          {VIC,  BYTE_ACCESS, (WORD)0x87, 0x0f},      // VIVR1
          {VIC,  BYTE_ACCESS, (WORD)0x8b, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x8f, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x93, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x97, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x9b, 0x0f},
          {VIC,  BYTE_ACCESS, (WORD)0x9f, 0x0f},      // VIVR7

          {VIC,  BYTE_ACCESS, (WORD)0xa3, 0x3c},      // TTR

          {VIC,  BYTE_ACCESS, (WORD)0xb3, 0x40},      // ARCR
          {VIC,  BYTE_ACCESS, (WORD)0xb7, 0x29},      // AMSR
          {VIC,  BYTE_ACCESS, (WORD)0xd3, 0x00},      // RCR

          {IFR,  LONG_ACCESS, (WORD)ADRHL, 0xF0F0F0F0},  // ADR-H, ADR-L
          {IFR,  WORD_ACCESS, (WORD)CSR  , 0x0000},      // Contr-Reg

          {VIC,  BYTE_ACCESS, (WORD)0x7f, 0x80},      // ICR7

          {LCR,  WORD_ACCESS, 0x4c, DISABLE_PCIADA_IRQS},  // disable interrupts
          
          {STOP, WORD_ACCESS, 0,     0}};

VPCIVMED_INIT_ELEMENT deinit_element_pre[] =
         {{VIC,  BYTE_ACCESS, (WORD)0x7f, 0x00},      // ICR7 - sysfail
          {LCR,  WORD_ACCESS, 0x4c, DISABLE_PCIADA_IRQS},  // disable interrupts
          {STOP, WORD_ACCESS, 0,    0}};

VPCIVMED_INIT_ELEMENT deinit_element_post[] =
         {{LCR,  WORD_ACCESS, 0x50, INHIBIT_VMEMM},        // disable interface
          {STOP, WORD_ACCESS, 0,    0}};

//--------------------------------------------------------------------------
// LOCAL FUNCTIONS
//
static char *kstrcpy(char *dst, char *src)
{
  while (*src != 0)
    *dst++ = *src++;
  *dst = 0;

  return dst; 
}

//--------------------------------------------------------------------------
// FUNCTIONS
//
BOOL check_command(const VPCIVMED_INIT_ELEMENT *psInitElement)
{
  WORD range;
  WORD access_size;
  
  switch (psInitElement->range)
  {
    case LCR: range = 0x54;     break;
    case IFR: range = 0x0c;     break;
    case VIC: range = 0xe4;
              if ((psInitElement->offset & 3) != 3)
                return FALSE;
              break;
    default:  range = 0;        break; 
  }

  switch (psInitElement->type)
  {
    case LONG_ACCESS: if (psInitElement->offset & 3)
                        return FALSE;
                      access_size = sizeof(DWORD);
                      break;
    case WORD_ACCESS: if (psInitElement->offset & 1)
                        return FALSE;
                      access_size = sizeof( WORD);
                      break;
    case BYTE_ACCESS: access_size = sizeof( BYTE); break;
    default         : access_size = 0xFFFF;        break;
  }

  if ((psInitElement->offset + access_size) > range)
    return FALSE;       // ignore it
    
  return TRUE;
}

static void CmdMachine(const VPCIVMED_INIT_ELEMENT *psInitElement,
                                                 PVOID pvLcr, PVOID pvIfr)
{
  PVOID adr;
  
  // standard initialisierungen  
  while (psInitElement->range != STOP)
  {
    if (check_command(psInitElement))
    {
      switch (psInitElement->range)
      {
        case LCR: adr = pvLcr; break;
        case VIC: adr = (PVOID)((DWORD)pvIfr + (WORD)VICBASE); break;
        case IFR:
        default:  adr = pvIfr; break;  
      }

#ifdef BUG
          if ((adr == pvLcr) && (psInitElement->offset == 0x4C))
            Reg4CShadow = psInitElement->value;
          if ((adr == pvLcr) && (psInitElement->offset == 0x50))
            Reg50Shadow = psInitElement->value;
#endif 

      switch (psInitElement->type)
      {
        case LONG_ACCESS:
          *(DWORD *)((BYTE *)adr + psInitElement->offset) = psInitElement->value;
          break;
        case WORD_ACCESS:
          *(WORD *) ((BYTE *)adr + psInitElement->offset) = psInitElement->value;
          break;
        case BYTE_ACCESS:
        default:
          *(BYTE *) ((BYTE *)adr + psInitElement->offset) = psInitElement->value;
          break;
      }
    }
    psInitElement++;
  }
}

BOOL InitInterface(const VPCIVMED_INIT_COMMAND *psInitCmd,
                                                    PVOID pvLcr, PVOID pvIfr)
{
  VPCIVMED_INIT_ELEMENT *psVie;
  
  Debug_Printf_DebMax("InitInterface(0x%08x,0x%08x)\r\n", pvLcr, pvIfr);

  if ((psInitCmd == (VPCIVMED_INIT_COMMAND *)BOGUSADDRESS) ||
      (pvLcr == ((PVOID)BOGUSADDRESS)                      ||
      (pvIfr == (PVOID)BOGUSADDRESS)))
    return FALSE;

  psVie = (VPCIVMED_INIT_ELEMENT *)psInitCmd->sVie;
  CmdMachine(init_element , pvLcr, pvIfr);    // standard initialisierungen  
  CmdMachine(psVie,         pvLcr, pvIfr);    // benutzer initialisierungen
  
  return TRUE;
}

BOOL DeInitInterface(const VPCIVMED_INIT_COMMAND *psDeInitCmd,
                                                     PVOID pvLcr, PVOID pvIfr)
{
  VPCIVMED_INIT_ELEMENT *psVie;

  Debug_Printf_DebMax("DeInitInterface(0x%08x,0x%08x)\r\n", pvLcr, pvIfr);

  if (pvLcr == ((PVOID)BOGUSADDRESS) || (pvIfr == (PVOID)BOGUSADDRESS))
    return FALSE;

  psVie = (VPCIVMED_INIT_ELEMENT *)psDeInitCmd->sVie;
  CmdMachine(deinit_element_pre,   pvLcr, pvIfr); // standard de-initialisierungen  
  CmdMachine(psVie,                pvLcr, pvIfr); // benutzer de-initialisierungen
  CmdMachine(deinit_element_post , pvLcr, pvIfr); // disable interface
  
  return TRUE;
}

static void memcpy(void *dst, void *src, DWORD len)
{
  char *d = (char *)dst;
  char *s = (char *)src;

  while (len--) *d++ = *s++;
}

DWORD get_dynamic_status(VPCIVMED_STATIC_STATUS *psPCIss,
                          VPCIVMED_STANDARD_COMMAND *psSc, 
                             VPCIVMED_DYNAMIC_STATUS *pPCIds, DWORD *pdwLen)
{
  WORD *pwControl   = (WORD *)((char *)(psPCIss->pvLcr) + 0x50);
  WORD *pwIRQStatus = (WORD *)((char *)(psPCIss->pvLcr) + 0x4C);
  
  if (*pdwLen < sizeof(VPCIVMED_DYNAMIC_STATUS))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }

  pPCIds->dwInterface      = psSc->dwInterface;
  pPCIds->wVMEMM_connected = (*pwControl   & 0x800) ? 1 : 0;
  pPCIds->wVMEMM_enable    = (*pwControl   & 0x100) ? 1 : 0;
  pPCIds->wPCIADAIrq       = (*pwIRQStatus & 0x028) ? 1 : 0;
  pPCIds->wVMEMMIrq        = (*pwIRQStatus & 0x005) ? 1 : 0;
    
  *pdwLen = sizeof(VPCIVMED_DYNAMIC_STATUS);
  return ERROR_SUCCESS;
}

DWORD get_static_status(VPCIVMED_STATIC_STATUS *psPCIssSrc,
                         VPCIVMED_STANDARD_COMMAND *psSc, 
                            VPCIVMED_STATIC_STATUS *psPCIssDest, DWORD *pdwLen)
{ 
  if (*pdwLen < sizeof(VPCIVMED_STATIC_STATUS))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }
  
  memcpy(psPCIssDest, psPCIssSrc, sizeof(VPCIVMED_STATIC_STATUS));
  *pdwLen = sizeof(VPCIVMED_STATIC_STATUS);
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------
// fills the status of PCIADA
//
DWORD get_PCIADA_status(VPCIVMED_STATIC_STATUS sVss[],
                          int iNumIfc, WORD wDriverVersion,
                               VPCIVMED_PCIADA_STATUS *psVps, DWORD *pdwLen)
{
  int i;
  
  if (*pdwLen < sizeof(VPCIVMED_PCIADA_STATUS))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }
  
  *pdwLen = sizeof(VPCIVMED_PCIADA_STATUS);

  psVps->dwDummy             = 0x6b686974;
  psVps->wVersion            = wDriverVersion;
  psVps->wNumberOfInterfaces = iNumIfc;
  
  for (i = 0; i < VPCIVMED_MAX_PCIADA; i++)
  {
    psVps->sPCIAda[i].dIfcStatus        = sVss[i].dIfcStatus;
    psVps->sPCIAda[i].dwLinkCount       = sVss[i].dwLinkCount;

    psVps->sPCIAda[i].wModuleType       = sVss[i].wModuleType;
    psVps->sPCIAda[i].wFPGAVersion      = sVss[i].wFPGAVersion;
    psVps->sPCIAda[i].wModuleNumber     = sVss[i].wModuleNumber;
    psVps->sPCIAda[i].wWordMode         = sVss[i].wWordMode;  

    psVps->sPCIAda[i].wSysControl       = sVss[i].wSysControl;
    psVps->sPCIAda[i].wConnected        = sVss[i].wConnected; 

    kstrcpy(sVss[i].cszHWRevision, psVps->sPCIAda[i].cszHWRevision);
  }

  return ERROR_SUCCESS;
}



DWORD access_vic68a(VPCIVMED_STATIC_STATUS *psPCIss,
                          VPCIVMED_VIC68A_ACTION *psIn,
                               VPCIVMED_VIC68A_ACTION *psOut, DWORD *pdwLen)
{
  char bByte;
  char *pbAddress;
  
  if (*pdwLen < sizeof(VPCIVMED_VIC68A_ACTION))
  {
     *pdwLen = 0;
     return ERROR_INVALID_PARAMETER;
  }

  psOut->dwInterface = psPCIss->dwInterface;
  
  if ((psIn->wRegisterAddress > (WORD)SRR) ||
                    ((psIn->wRegisterAddress & 0x03) != 3))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }

  *pdwLen = sizeof(VPCIVMED_VIC68A_ACTION);
  pbAddress = (BYTE  *)((char *)psPCIss->pvIfr +
                           (DWORD)VICBASE + (DWORD)psIn->wRegisterAddress);
  
  switch(psIn->wAccessMode)
  {
    case VIC68A_READ:  bByte = *pbAddress;
                       break;
    case VIC68A_WRITE: *pbAddress = psIn->bContent;
                       bByte      = *pbAddress;
                       break;
    case VIC68A_WRITE_ONLY:
                       *pbAddress = psIn->bContent;
                       break;
    case VIC68A_OR:    bByte      = *pbAddress;
                       bByte     |= psIn->bContent;
                       *pbAddress = bByte;
                       bByte      = *pbAddress;
                       break;
    case VIC68A_AND:   bByte      = *pbAddress;
                       bByte     &= psIn->bContent;
                       *pbAddress = bByte;
                       bByte      = *pbAddress;
                       break;

    default:           *pdwLen = 0;
                       return ERROR_INVALID_PARAMETER;                   
  }  

  psOut = psIn;
  psOut->bContent = bByte;
  
  return ERROR_SUCCESS;
}                   


DWORD make_tas(VPCIVMED_STATIC_STATUS *psPCIss,
                          VPCIVMED_TAS_STRUCT *psIn,
                               VPCIVMED_TAS_STRUCT *psOut,
                                     DWORD *pdwLen)
{
  BYTE  *adr;
  DWORD vmeAdr;
  WORD  csr;
  BYTE  modifier;
  BYTE  bContent;

  Debug_Printf_DebMax("TAS Adr: 0x%08lx, Mod: 0x%02x , Cont: 0x%02x\r\n",
                            psIn->dwAddress, psIn->wModifier, psIn->bContent); 

  if (*pdwLen < sizeof(VPCIVMED_TAS_STRUCT))
  {
    *pdwLen = 0;
    return ERROR_INVALID_PARAMETER;
  }
  
  // calculate some addresses
  adr         = (BYTE *) ((DWORD)psPCIss->pvVME + (psIn->dwAddress & VME_ADR_MASK));

  // save old contents
  modifier               = *(psPCIss->pbModifier);
  vmeAdr                 = *(psPCIss->pdwVMEAdr);
  csr                    = *(psPCIss->pwCSR);
  bContent               = psIn->bContent;

  // set new contents
  *(psPCIss->pbModifier) = (BYTE)psIn->wModifier & 0x3f;
  *(psPCIss->pdwVMEAdr)  = psIn->dwAddress;
  *(psPCIss->pwCSR)      = csr | FLAG_RMC;

  // do the read - modify - write
  psIn->bContent         = *adr;                           
  *adr                   = bContent;  

  // restore old contents
  *(psPCIss->pwCSR)      = csr;  
  *(psPCIss->pdwVMEAdr)  = vmeAdr;
  *(psPCIss->pbModifier) = modifier;                      
  
  *pdwLen = sizeof(VPCIVMED_TAS_STRUCT);  
  return ERROR_SUCCESS;
}                   

DWORD make_reset(VPCIVMED_STATIC_STATUS *psPCIss,  VPCIVMED_RESET_COMMAND *psCommand,
                                         VPCIVMED_RESET_RESULT *psResult, DWORD *pdwLen)
{
  WORD *pwControl   = (WORD *)((char *)(psPCIss->pvLcr) + 0x50);
  WORD *pwIRQStatus = (WORD *)((char *)(psPCIss->pvLcr) + 0x4C);
  WORD wIRQStatus;
  WORD wControl;
  char *pbReset;
    
  // set default result return size and contents
  *pdwLen = sizeof(*psResult);
  psResult->wResult = 0xffff;
  psResult->dwInterface = psCommand->dwInterface;
  
  // am I connected and switched on??
  if ((*pwControl & 0x0980) != 0x0980) return ERROR_NOT_READY;

  // do command
  switch (psCommand->wCommand)
  {
    case VME_RESET_CMD:    *(psPCIss->pbModifier) = 0;
                           pbReset = (BYTE  *)((char *)psPCIss->pvIfr +
                                                   (DWORD)VICBASE + (DWORD)SRR);
                           *pbReset = 0xf0;  // make VME reset                         
                           break;
    case LOCAL_RESET_CMD:  *(psPCIss->pbModifier) = 0;
                           *(psPCIss->pwIRQStat)  = LOCAL_RESET;
                           break;
    case GLOBAL_RESET_CMD: *(psPCIss->pbModifier) = 0;
                           *(psPCIss->pwIRQStat)  = GLOBAL_RESET;
                           break;
    case POLL_RESET_CMD:   break;

    default:               return ERROR_INVALID_PARAMETER;
  }

  // save IRQ status of PCIADA and switch off PCIADA interrupts
  wIRQStatus = *pwIRQStatus;
  *pwIRQStatus &= ~0x0040;
  
  // always poll reset status - access will sometimes generate PCIADA #2 interrupt
  if (*(psPCIss->pbModifier) == 0x00)   
    psResult->wResult = 0;

  // reset any pending PCIADA interrupt #2  
  wControl    = *pwControl;
  *pwControl &= ~0x0100;
  *pwControl  = wControl;
     
  // restore IRQStatus
  *pwIRQStatus = wIRQStatus;
  
  return ERROR_SUCCESS;
}                                         

//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------

