#ifndef __PCIVMEH_H__

//----------------------------------------------------------------------------------------
// PCIVMEH.H, shared between applications and VPCIVMED driver
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// What                                                              Who  When
// first steps                                                       AR   24.01.98
// added direct read write access to vic68a chip registers           AR   12.07.98
// rename PCR_* into LCR_*                                           AR   19.07.98 
// TAS included                                                      AR   17.02.99
// Corrections about interrupt handling                              AR   20.02.99
// changes about PCIADA status                                       AR   25.02.99
// changes of IOCTL codes beacause of compatibility to WIN NT        AR   12.03.99
// PLX 9052 removed out of VPCIVMED_STATIC_STRUCT                    AR   16.03.99
// VIC68A_WRITE_ONLY added                                           AR   17.03.99
// extention for VME reset                                           AR   18.04.99
// release of version 2.5 for driver                                 AR   18.04.99
//

//----------------------------------------------------------------------------------------
// constants to be used to access certain features of the PCIVME interface
//
#define VPCIVMED_CTL_CODE(x)         (0x80002000 | (x << 2)) // compatibility to WIN-NT

#define VPCIVMED_INIT_HARDWARE       (VPCIVMED_CTL_CODE( 0)) // initializes the hardware with given parameters
#define VPCIVMED_DEINIT_HARDWARE     (VPCIVMED_CTL_CODE( 1)) // uninitializes the hardware
#define VPCIVMED_ATTACH_WINDOW       (VPCIVMED_CTL_CODE( 2)) // requests a base address to a vme window
#define VPCIVMED_DETACH_WINDOW       (VPCIVMED_CTL_CODE( 3)) // frees a vme window
#define VPCIVMED_GET_STATIC_STATUS   (VPCIVMED_CTL_CODE( 4)) // asks for INTERFACE structure
#define VPCIVMED_GET_DYNAMIC_STATUS  (VPCIVMED_CTL_CODE( 5)) // asks for dynamic status
#define VPCIVMED_READ_VECTOR         (VPCIVMED_CTL_CODE( 6)) // reads the level and vector of IRQ
#define VPCIVMED_ACCESS_VIC68A       (VPCIVMED_CTL_CODE( 7)) // access vic68a register
#define VPCIVMED_INSTALL_IRQ_HANDLER (VPCIVMED_CTL_CODE( 8)) // installs a handler funktion
#define VPCIVMED_CONTROL_INTERRUPTS  (VPCIVMED_CTL_CODE( 9)) // enable, disable of interrupts
#define VPCIVMED_TAS                 (VPCIVMED_CTL_CODE(10)) // make test and set
#define VPCIVMED_GET_PCIADA_STATUS   (VPCIVMED_CTL_CODE(11)) // get the status of PCIADA(s) only
#define VPCIVMED_RESET               (VPCIVMED_CTL_CODE(12)) // make a reset to VME or global

//----------------------------------------------------------------------------------------
// possible return codes
//
#define BOGUSADDRESS 0xffffffff      // Returned by MS routines

//----------------------------------------------------------------------------------------
// some built in limits
//
#define VPCIVMED_MAX_PCIADA     4 // maximam count of supported PCI interfaces
#define VPCIVMED_MAX_VMEMM     16 // maximum number of supported VMEMMs
#define VPCIVMED_MAX_WINDOWS    8 // maximum number of windows into VME 

//----------------------------------------------------------------------------------------
// switches and masks
//

// switches for VPCIVMED_INIT_COMMANDs --------------------------
#define LCR   (BYTE)0    // destination is LCR register
#define IFR   (BYTE)1    // destination is VME-Interface register
#define VIC   (BYTE)2    // destination is VIC68A register
#define STOP  (BYTE)255  // this command stops the init machine

#define BYTE_ACCESS (BYTE)1    // write byte wise
#define WORD_ACCESS (BYTE)2    //       word
#define LONG_ACCESS (BYTE)4    //       long

// switches for VPCIVMED_ACCESS_VIC68A --------------------------
#define VIC68A_READ       0    // read only access
#define VIC68A_WRITE      1    // write and read back access
#define VIC68A_OR         2    // read, bitwise 'or' content and read back access
#define VIC68A_AND        3    // read, bitwise 'and' content and read back access
#define VIC68A_WRITE_ONLY 4    // do not read back after write

// switches for VPCIVMED_VECTOR_CMD -----------------------------
#define READ_CURRENT_LEVEL 0   // try to get the current irq level
#define READ_VECTOR        1   // (if level == 0) read vector @ current LEVEL else @ level

// switches for the VPCIVMED_RESET ------------------------------
#define VME_RESET_CMD      0   // raise a VME reset only
#define LOCAL_RESET_CMD    1   // raise a local reset only
#define GLOBAL_RESET_CMD   2   // raise a global reset
#define POLL_RESET_CMD     3   // ask if reset is finished

// address masks for the pager - to use for offset and size @ window alignment -----------
#define HI_ADDRESS_MASK   (DWORD)0xFFFFF000    // masks the high part of a vme address
#define LO_ADDRESS_MASK   (~HI_ADDRESS_MASK)   // masks the low  part of a vme address
#define ONE_PAGE_SIZE     (LO_ADDRESS_MASK + 1)  // size of 1 page (hardware related)

// macros to calculate the real base and the real size of demand pages -------------------
#define PAGE_BASE(base)   (base & HI_ADDRESS_MASK)  // makes an aligned base for a page
#define PAGE_SIZE(base, size) (((base + size + LO_ADDRESS_MASK) / ONE_PAGE_SIZE) * ONE_PAGE_SIZE)

//----------------------------------------------------------------------------------------
// ERROR RETURNS in dIfcStatus
//
#define E_NO_ERROR         0   // all OK
#define E_INCOMPATIBLE     1   // incompatible hardware
#define E_NO_ADDRESS       2   // cant get lcr or ifr addresses
#define E_NOT_CONNECTED    3   // no VMEMM hardware connected
#define E_CON_ERROR        4   // data transfer failure
#define E_EMPTY            -1  // no PCI interface associated

//----------------------------------------------------------------------------------------
// shared structures between PCIVME-IF and Application - COMMANDS
//
typedef struct
{
  DWORD dwInterface;         // some command only need this input into requests
} VPCIVMED_STANDARD_COMMAND;

typedef struct               // one command element to initialize interface or deinitialize
{
  BYTE     range;            // 0 = lcr, 1 = vme-interface, -1 = stop, default = vme-if
  BYTE     type;             // 1 = byte access, 2 = word access, 4 = dword access, default byte
  WORD     offset;           // offset into interface address range for initialisation
  DWORD    value;            // value to initialize
} VPCIVMED_INIT_ELEMENT;

typedef struct
{
  DWORD    dwInterface;      // targets to interface number
  VPCIVMED_INIT_ELEMENT sVie[1];  // at least one zero element must be the last
} VPCIVMED_INIT_COMMAND;

typedef struct
{
  DWORD    dwInterface;       // targets to interface number ...
  DWORD    base;              // offset into VME address range. (base + size) must be less than 
  DWORD    size;              // 128 Mbytes for ext, 16 Mbytes for std, 64k for short
  WORD     modifier;          // VME address modifier for this window  
} VPCIVMED_ADD_WINDOW;

typedef struct
{
  DWORD dwInterface;          // targets to interface number ...
  DWORD *pdwLinAdr;           // linear address of window to remove
} VPCIVMED_REMOVE_WINDOW;

typedef struct
{
  DWORD    dwInterface;       // targets to interface number ...
  DWORD    dwAddress;         // tas to address
  WORD     wModifier;         // VME address modifier for this window  
  BYTE     bContent;          // byte content to store and get back
} VPCIVMED_TAS_STRUCT;

typedef struct
{
  DWORD dwInterface;          // targets to interface number ...
  WORD  wRegisterAddress;     // address offset of vic68a register
  WORD  wAccessMode;          // read, write, or, and
  BYTE  bContent;             // content to write, and, or
} VPCIVMED_VIC68A_ACTION;

typedef struct
{
  DWORD dwInterface;           // targets to the interface number ...
  DWORD dwIrqHandler;          // void (*IrqHandler)(DWORD) = User Handler
} VPCIVMED_IRQ_HANDLER;        // BOGUSADDRESS deinstalls

typedef struct
{
  DWORD dwInterface;           // targets to the interface number ...
  WORD  wEnable;               // a 1 enables, a 0 disables
} VPCIVMED_IRQ_CONTROL;

typedef struct
{
  DWORD dwInterface;           // targets to interface number ...
  WORD  wAction;               // read current irq level, read vector @ level
  WORD  wType;                 // must be set to 1
} VPCIVMED_VECTOR_COMMAND;

typedef struct
{
  DWORD dwInterface;           // targets to interface number ...
  WORD  wCommand;
} VPCIVMED_RESET_COMMAND; 

//----------------------------------------------------------------------------------------
// shared structures between PCIVME-IF and Application - RESPONSE
//

// includes static information about driver parameters ------
typedef struct                // achtung: sehr empfindlich auf alignmet
{
  DWORD dwInterface;          // comes from the interface No.  
  DWORD dIfcStatus;           // verwendbar? passt zum treiber? OK?
  DWORD dwLinkCount;          // how often this interface is requested  

  WORD  wNumMemWindows;       // aus der aktuellen konfiguration
  WORD  wNumIOPorts;
  WORD  wNumIRQs;
  WORD  wNumDMAs;
      
  DWORD dLCR_MemBase;         // aus der aktuellen konfiguration
  DWORD dLCR_MemLength;
  
  WORD  wLCR_IOBase;
  WORD  wLCR_IOLength;
  WORD  wLCR_IRQ;
  WORD  wReserve1;

  DWORD dUSR_MemBase;
  DWORD dUSR_MemLength;

  WORD  wModuleType;          // von der angeschlossenen hardware gelesen
  WORD  wFPGAVersion;
  WORD  wModuleNumber;
  WORD  wWordMode;

  WORD  wSysControl;
  WORD  wConnected;
  
  PVOID pvLcr;                // physikalische adresse des lcr
  PVOID pvIfr;                // physikalische adresse des interfaces ohne vme

                              // aus performance gruenden einige adressen
  WORD  *pwCSR;               // zeiger auf csr register
  WORD  *pwIRQStat;           // zeiger auf irq status
  BYTE  *pbVector;            // zeiger auf die vector lese register
  DWORD *pdwVMEAdr;           // zeiger zum VME adress register
  BYTE  *pbModifier;          // zeiger zum address modifier register
  void  *pvVME;               // pointer into VME window
  DWORD dwPagePhysVME;        // physical page number of the VME window
  
  void  *psIrqDescriptor;     // pointer to associated irq descriptor
  DWORD dwActivePage;         // the current active page of this interface

  WORD  wReserve2;            
  
  char  cszHWRevision[10];
} VPCIVMED_STATIC_STATUS;

typedef struct
{
  DWORD dwInterface;          // comes from the interface No.
  
  WORD  wVMEMM_connected;     // status: VMEMM is connected and powered
  WORD  wVMEMM_enable;        // status: VMEMM access is enabled
  WORD  wPCIADAIrq;           // status: PCIADA timeout IRQ pending
  WORD  wVMEMMIrq;            // status: VMEMM IRQ pending
} VPCIVMED_DYNAMIC_STATUS;

typedef struct
{
  DWORD dwInterface;          // comes from the interface No.
  
  DWORD dwStatusID;           // interrupt-vector (byte, word, long)
  WORD  wLevel;               // interrupt-level
  WORD  wPCIIrq;              // pending PCIADA Irq detected and cleared
} VPCIVMED_VECTOR_LEVEL;

typedef struct
{
  DWORD dwDummy;              // nothing usefull in here
  WORD  wVersion;             // Version of driver
  WORD  wNumberOfInterfaces;  // number of detected PCIADA
  struct
  {
    DWORD dIfcStatus;         // connection status of PCIADA-VMEMM
    DWORD dwLinkCount;        // how often this interface is requested  

    WORD  wModuleType;        // if connected: type of connected module
    WORD  wFPGAVersion;       // if connected: Version of VMEMM FPGA
    WORD  wModuleNumber;      // if connected: Number of Connected VMEMM
    WORD  wWordMode;          // if connected: Mode of operation

    WORD  wSysControl;        // if connected: VMEMM sysctl status
    WORD  wConnected;         // connected or not

    WORD  wDummy;
    char  cszHWRevision[10];  // revision of PCI interface
  } sPCIAda[VPCIVMED_MAX_PCIADA]; // status of each one  
} VPCIVMED_PCIADA_STATUS;

typedef struct
{
  DWORD dwInterface;           // targets to interface number ...
  WORD  wResult;
} VPCIVMED_RESET_RESULT;       // polling result: in progress if (wResult != 0) 

#define __PCIVMEH_H__
#endif

