
/*
 * Library for PCI-VMEbus Interface from
 * ARW Elektronik
 * 
 * Author: Dirk Muehlenberg
 * $Revision: 1.29 $
 * $Date: 1999/12/17 13:35:17 $
 * $Id: arwvme.c,v 1.29 1999/12/17 13:35:17 mathes Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <malloc.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <asm/page.h>

#include <pwd.h>

#include "arwvme.h"

/* ------------------------------------------------------------------------- */

static char Arwvme_version_string[] =
  "Arwvme Linux Version $Revision: 1.29 $ $Date: 1999/12/17 13:35:17 $";

char *ArwvmeGetVersionString ( void )
 {
  return Arwvme_version_string;
}

/* ------------------------------------------------------------------------- */

int ArwvmeOpen ( int module )
 {
  static char buffer[ARWVME_DRIVER_INFO_SIZE];
  unsigned int driver_version;
  int loop = 5, fd = -1;

  sprintf ( buffer, "/dev/arwvme%d", module );
  while ( fd < 0 && loop )
   {
    fd = open ( buffer, O_RDWR );
    if ( fd < 0 ) sleep(1);  /* wait for module loading ... */
    loop--;
  }
  if ( fd < 0 )
   {
    perror ( "open" );
    return -1;
  }

  if ( ioctl ( fd, ARWVME_GET_DRIVER_VERSION, &driver_version ) < 0 )
   {
    perror ( "ARWVME_GET_DRIVER_VERSION" );
    return -1;
  }
  
#ifndef ARWVME_DRIVER_VERSION    
# include "driver/arwvme_version.h"
#endif
   if ( driver_version != ARWVME_DRIVER_VERSION )
     {
      printf ( "%s%s%s0x%x%s0x%x",
               "Version mismatch between actually loaded driver",
               " and expected driver version: \n"
	       "Loaded= ", driver_version, 
	       "expected= \n", ARWVME_DRIVER_VERSION );

      close ( fd );
      exit(1);
    }

#if 1
  ioctl ( fd, ARWVME_GET_DRIVER_INFO, (unsigned long)buffer );
  puts ( buffer ); 
#endif

  return fd;
}

/* ------------------------------------------------------------------------- */

int ArwvmeCheckConnection ( int fd )
 {
  int ret;
  static char buf[256];
  if (ioctl(fd, ARWVME_CHECK_CONNECTION, &ret) < 0) {
    sprintf(buf,"ioctl(%d, CHECK_CONNECTION)",fd);
    perror(buf);
    ret = 0;
  }
  return ret;
}
 
/* ------------------------------------------------------------------------- */

int ArwvmeShutdown ( int fd )
 {
  return close(fd);
}

/* ------------------------------------------------------------------------- */

int ArwvmeInitialize ( int fd )
/* restricted to superuser and the group of the device */
 {
  return ioctl(fd, ARWVME_INITIALIZE, 0);
}

/* ------------------------------------------------------------------------- */

int ArwvmeResetVme ( int fd )
/* restricted to superuser and the group of the device */
 {
  return ioctl ( fd, ARWVME_SYSRESET, 0 );
}

/* ------------------------------------------------------------------------- */

int ArwvmeSetMode ( int fd, unsigned int mask )
 {
  ArwvmeADA ada;
  ada.addr_type = mask & ARWVME_A_MASK;
  ada.data_type = mask & ARWVME_D_MASK;
  ada.acc_type  = mask & ( ARWVME_ACC_MASK | ARWVME_PRIV_MASK );

  return ioctl(fd, ARWVME_SET_AM, (unsigned long)&ada);
}

/* ------------------------------------------------------------------------- */

ssize_t ArwvmeRead ( int fd, void *buf, size_t count )
 {
  int ret,old_cnt = count;
  while (count > 0) {
    ret = read(fd, buf, count);
    if (ret < 0) break;
    count -= ret;
    sched_yield();
  }
  return old_cnt-count;
}

/* ------------------------------------------------------------------------- */

ssize_t ArwvmeWrite ( int fd, void *buf, size_t count )
 {
  int ret,old_cnt = count;
  while (count > 0) {
    ret = write(fd, buf, count);
    if (ret < 0) break;
    count -= ret;
    sched_yield();
  }
  return old_cnt-count;
}

/* ------------------------------------------------------------------------- */

int ArwvmeCheckSysError ( int fd )
 {
  if (ArwvmeGetSysErrInfo(fd)->module_number > -1)
    return TRUE;
  return FALSE;
}

/* ------------------------------------------------------------------------- */

ArwvmeSysErrInfo *ArwvmeGetSysErrInfo ( int fd )
 {
  static ArwvmeSysErrInfo ret;
  ioctl(fd, ARWVME_ERROR_CHECK, &ret);
  return &ret;
}

/* ------------------------------------------------------------------------- */

ArwvmeIntAck *ArwvmeGetNextPendingInt ( int fd )
/* gets the next interrupt from the pendingInts queue */
/* the struct contains the int vector and a timelabel */
 {
  ArwvmeIntAck *ret = NEW(ArwvmeIntAck);

  if (ioctl(fd, ARWVME_GET_INTERRUPT, ret) < 0) {
    free ( ret );
    return NULL;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

int ArwvmeRegisterInterrupt ( int fd, ArwvmeAckType at, unsigned int rora_reg, 
                                      int vec , unsigned int acc_mask )
 {
  ArwvmeIntAck ack;

  ack.ack_type = at;
  if (at != ArwvmeRoak) {
    ack.rora_reg = rora_reg;
    ack.acc_mask = acc_mask;
  } else {    
    ack.rora_reg = 0x0;
    ack.acc_mask = 0x0;  
  }
  ack.int_vec = vec;
  ack.int_tl = 0;

  return ioctl ( fd, ARWVME_REGISTER_INTERRUPT, &ack );
}

/* ------------------------------------------------------------------------- */

int ArwvmeUnregisterInterrupt ( int fd, int vec )
 {
  return ioctl ( fd, ARWVME_UNREGISTER_INTERRUPT, &vec );
}

/* ------------------------------------------------------------------------- */

int ArwvmeSelect ( int fd, struct timeval *timeout )
 {
  fd_set in;

  FD_ZERO(&in);
  FD_SET(fd, &in);

  return select ( fd+1, &in, NULL, NULL, timeout );
}

/* ------------------------------------------------------------------------- */

int ArwvmeEnableInterrupt ( int fd )
/* PCIADA + VMEMM interrupts */
 {
  /* --- do nothing, interrupts are enabled by default --- */

  return ioctl ( fd, ARWVME_ENABLE_INTERRUPT , 0 );
}

/* ------------------------------------------------------------------------- */

int ArwvmeDisableInterrupt ( int fd )
 {
  /* --- do nothing, interrupts disabled at cleanup --- */ 

  return ioctl ( fd, ARWVME_DISABLE_INTERRUPT, 0 );
}

/* ------------------------------------------------------------------------- */

int ArwvmeSignal ( int fd, void (*sig_handler)(int) )
 {
  pid_t pid;
  struct sigaction sa;
  
  if ( !sig_handler ) return -1;
  if ( sig_handler == SIG_IGN || sig_handler == SIG_DFL )
   {
    sig_handler = SIG_IGN;
    pid = 0;
  }
  else
    pid = getpid();
    
  sa.sa_handler = sig_handler;
  sigemptyset ( &sa.sa_mask );
  sa.sa_flags = SA_RESTART;
  sa.sa_restorer = NULL;
  sigaction ( SIGUSR1, &sa, NULL );

  return ioctl ( fd, ARWVME_ATTACH_SIGNAL, &pid );
}

/* ------------------------------------------------------------------------- */

int ArwvmeCtrlArbiter ( int fd, ArwvmeArbModes arb_mode, int brl, int fair )
 {
  ArwvmeArbiter arb;
  arb.arb_mode = arb_mode;
  arb.br_lev = brl;
  arb.fairness = fair;

  return ioctl ( fd, ARWVME_CTRL_ARBITER, &arb );
}

/* ------------------------------------------------------------------------- */

int ArwvmeSetVMEbusTO ( int fd, int to )
 {
  if ( to <= 4 )
    to = 0;
  else if ( to <= 16 )
    to = 1;
  else if ( to <= 32 )
    to = 2;
  else if ( to <= 64 )
    to = 3;
  else if ( to <= 128 )
    to = 4;
  else if ( to <= 256 )
    to = 5;
  else if ( to <= 512 )
    to = 6;
  else
    to = 7;
  
  return ioctl ( fd, ARWVME_SET_VMEBUS_TO, &to );
}

/* ------------------------------------------------------------------------- */

int ArwvmeUntieFpos ( int fd )
 {
  return ioctl ( fd, ARWVME_UNTIE_FPOS, 0 );
}

/* ------------------------------------------------------------------------- */

int ArwvmeTieFpos ( int fd )
 {
  return ioctl ( fd, ARWVME_TIE_FPOS, 0 );
}

/* ------------------------------------------------------------------------- */

int ArwvmeSetVmeWindow ( int fd, unsigned int addr )
/*
 * this function fixes the start in the VME space,
 * this offset will be added in the nopage handler
 * s. ArwvmeMMap()
 */
 {
  return ioctl ( fd, ARWVME_SET_VME_WINDOW, &addr ); 
}

/* ------------------------------------------------------------------------- */

void *ArwvmeMMap ( int fd, int size )
/*
 * mmaps win_base of fd with size win_size to a buffer (ret). 
 * The driver actually does NO mapping, but will install a
 * page fault handler, performing the VMEADDR updates thus
 * returning the correct (physical) page. 
 */
 {
  void *ret;
  off_t win_base;

  if ( ioctl(fd, ARWVME_GET_VME_BASE, &win_base) ) return NULL;

  ret = (unsigned char *)mmap( 0, size,
                               PROT_READ | PROT_WRITE,
                               MAP_SHARED,
                               fd, win_base );

  if ( (long)ret < 0L ) return NULL;

  return ret;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
