
/*
 * File   : atvme_mm.c
 *
 * Purpose: Library functions for VME bus data accesses
 *          using the mmap() system call !!!
 *
 * $Revision: 1.4 $
 * $Date: 1999/12/17 13:35:17 $
 * $Id: arwvme_mm.c,v 1.4 1999/12/17 13:35:17 mathes Exp $
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <linux/arwvme.h>

#include "arwvme.h"
#include "vme.h"


int ArwvmeDefaultDevice = -1;

static unsigned int Arwvme_vme_base = 0xffffffff;

/* ------------------------------------------------------------------------- */

int ArwvmeSimpleOpen ( void )
 {                                                          /* thjm 13.12.99 */
  if ( ArwvmeDefaultDevice < 0 )
    ArwvmeDefaultDevice = ArwvmeOpen ( ARWVME_DEFAULT_MODULE );
  
  if ( ArwvmeDefaultDevice > 0 )
    return 0;
  else
    return ArwvmeDefaultDevice;
}

/* ------------------------------------------------------------------------- */

static void *MapNewRange ( VmeDesc *vme )
 {                                                          /* thjm 08.11.99 */
  unsigned int new_page = (unsigned int)vme->addr & 0xfffff000;
  static void *vme_page_ptr = NULL;

  if ( new_page != Arwvme_vme_base )
   {
    ArwvmeSetVmeWindow ( vme->device_fd, new_page );
    if ( vme_page_ptr )
     {
      munmap ( vme_page_ptr, 0x1000 );
      vme_page_ptr = NULL;
    }
    if ( !vme_page_ptr )
     {
      vme_page_ptr = ArwvmeMMap ( vme->device_fd, 0x1000 );
      if ( vme_page_ptr == NULL )
        perror ( "arwvme_mm.c: could not mmap()" );
    }
    
    Arwvme_vme_base = new_page;
  }
  
  return vme_page_ptr;
}

/* ------------------------------------------------------------------------- */

unsigned char ArwvmeReadChar ( VmeDesc *vme )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned char *data_ptr;
  unsigned char *data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D8 );
  data_ptr = (unsigned char *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned char *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    return *data_ptr;
  }
  else
    return 0;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteChar ( VmeDesc *vme, unsigned char data )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned char *data_ptr;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D8 );
  data_ptr = (unsigned char *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned char *)
                 ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    *data_ptr = data;
  }
}  

/* ------------------------------------------------------------------------- */

unsigned short ArwvmeReadShort ( VmeDesc *vme )
 {                                                          /* thjm 08.11.99 */
  volatile unsigned short *data_ptr;
  unsigned short data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  data_ptr = (unsigned short *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned short *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0x0fff));
    return *data_ptr;
  }
  else
    return 0;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteShort ( VmeDesc *vme, unsigned short data )
 {                                                          /* thjm 08.11.99 */
  volatile unsigned short *data_ptr;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  data_ptr = (unsigned short *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned short *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0x0fff));
    *data_ptr = data;
  }
}

/* ------------------------------------------------------------------------- */

void ArwvmeFifoReadShort ( VmeDesc *vme, 
                           unsigned short *data, unsigned int count )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned short *data_ptr;
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  data_ptr = (unsigned short *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned short *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    for ( i=0; i<count; i++)
      *data++ = *data_ptr;
  }
}

/* ------------------------------------------------------------------------- */

void ArwvmeBlockReadShort ( VmeDesc *vme, 
                           unsigned short *data, unsigned int count )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned short *data_ptr;
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  data_ptr = (unsigned short *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned short *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    for ( i=0; i<count; i++)
      *data++ = *data_ptr++;
  }
}

/* ------------------------------------------------------------------------- */

unsigned int ArwvmeReadInt ( VmeDesc *vme )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned int *data_ptr;
  unsigned int data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  data_ptr = (unsigned int *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned int *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0x0fff));
    return *data_ptr;
  }
  else
    return 0;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteInt ( VmeDesc *vme, unsigned int data )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned int *data_ptr;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  data_ptr = (unsigned int *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned int *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0x0fff));
    *data_ptr = data;
  }
}

/* ------------------------------------------------------------------------- */

void ArwvmeFifoReadInt ( VmeDesc *vme, 
                         unsigned int *data, unsigned int count )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned int *data_ptr;
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  data_ptr = (unsigned int *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned int *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    for ( i=0; i<count; i++)
      *data++ = *data_ptr;
  }
}

/* ------------------------------------------------------------------------- */

void ArwvmeBlockReadInt ( VmeDesc *vme, 
                         unsigned int *data, unsigned int count )
 {                                                          /* thjm 11.11.99 */
  volatile unsigned int *data_ptr;
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  data_ptr = (unsigned int *)MapNewRange ( vme );
  if ( data_ptr != NULL )
   {
    data_ptr = (unsigned int *)
               ((unsigned long)data_ptr + ((unsigned long)vme->addr & 0xffff));
    for ( i=0; i<count; i++)
      *data++ = *data_ptr++;
  }
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
