
/*
 * File   : atvme_simple.c
 *
 * Purpose: Library functions for VME bus data accesses
 *          without using the mmap() system call !!!
 *
 * $Revision: 1.4 $
 * $Date: 1999/12/17 13:35:17 $
 * $Id: arwvme_simple.c,v 1.4 1999/12/17 13:35:17 mathes Exp $
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <linux/arwvme.h>

#include "arwvme.h"
#include "vme.h"

int ArwvmeDefaultDevice = -1;

/* ------------------------------------------------------------------------- */

int ArwvmeSimpleOpen ( void )
 {                                                          /* thjm 13.12.99 */
  if ( ArwvmeDefaultDevice < 0 )
    ArwvmeDefaultDevice = ArwvmeOpen ( ARWVME_DEFAULT_MODULE );
  
  if ( ArwvmeDefaultDevice > 0 )
    return 0;
  else
    return ArwvmeDefaultDevice;
}

/* ------------------------------------------------------------------------- */

unsigned char ArwvmeReadChar ( ArwvmeDesc *vme )
 {                                                          /* thjm 08.10.99 */
  unsigned char data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd, 
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D8 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeRead ( vme->device_fd, &data, sizeof ( unsigned char ) );
  
  return data;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteChar ( ArwvmeDesc *vme, unsigned char data )
 {                                                          /* thjm 08.10.99 */
  int acc = vme->acc & ARWVME_A_MASK;

  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D8 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeWrite ( vme->device_fd, &data, sizeof ( unsigned char ) );
}

/* ------------------------------------------------------------------------- */

void ArwvmeBlockWriteChar ( ArwvmeDesc *vme, 
                            unsigned char *data, unsigned int count )
 {                                                          /* thjm 13.12.99 */
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D8 );
  
  ArwvmeTieFpos ( vme->device_fd );
  
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );
  ArwvmeRead ( vme->device_fd, (char *)data, count * sizeof ( unsigned char ) );
}

/* ------------------------------------------------------------------------- */

unsigned short ArwvmeReadShort ( ArwvmeDesc *vme )
 {                                                          /* thjm 08.10.99 */
  unsigned short data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd, 
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeRead ( vme->device_fd, (char *)&data, sizeof ( unsigned short ) );
  
  return data;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteShort ( ArwvmeDesc *vme, unsigned short data )
 {                                                          /* thjm 08.10.99 */
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd, 
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeWrite ( vme->device_fd, (char *)&data, sizeof ( unsigned short ) );
}

/* ------------------------------------------------------------------------- */

void ArwvmeBlockReadShort ( ArwvmeDesc *vme, 
                            unsigned short *data, unsigned int count )
 {                                                          /* thjm 08.10.99 */
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  
  ArwvmeTieFpos ( vme->device_fd );
  
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );
  ArwvmeRead ( vme->device_fd, (char *)data, count * sizeof ( unsigned short ) );
}

/* ------------------------------------------------------------------------- */

void ArwvmeFifoReadShort ( ArwvmeDesc *vme, 
                           unsigned short *data, unsigned int count )
 {                                                          /* thjm 14.12.99 */
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D16 );
  
  ArwvmeUntieFpos ( vme->device_fd );
  
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );
  for ( i=0; i<count; i++ )
    ArwvmeRead ( vme->device_fd, (char *)data, sizeof ( unsigned short ) );
}

/* ------------------------------------------------------------------------- */

unsigned int ArwvmeReadInt ( ArwvmeDesc *vme )
 {                                                          /* thjm 08.10.99 */
  unsigned int data;
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeRead ( vme->device_fd, (char *)&data, sizeof ( unsigned int ) );
  
  return data;
}

/* ------------------------------------------------------------------------- */

void ArwvmeWriteInt ( ArwvmeDesc *vme, unsigned int data )
 {                                                          /* thjm 08.10.99 */
  int acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );

  ArwvmeWrite ( vme->device_fd, (char *)&data, sizeof ( unsigned int ) );
}

/* ------------------------------------------------------------------------- */

void ArwvmeBlockReadInt ( ArwvmeDesc *vme, 
                          unsigned int *data, unsigned int count )
 {                                                          /* thjm 08.10.99 */
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  
  ArwvmeTieFpos ( vme->device_fd );
  
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );
  ArwvmeRead ( vme->device_fd, (char *)data, count * sizeof ( unsigned int ) );
}

/* ------------------------------------------------------------------------- */

void ArwvmeFifoReadInt ( ArwvmeDesc *vme, 
                         unsigned int *data, unsigned int count )
 {                                                          /* thjm 14.12.99 */
  int i, acc = vme->acc & ARWVME_A_MASK;
  
  ArwvmeSetMode ( vme->device_fd,
                  ARWVME_NONPRIV | ARWVME_DATA | acc | ARWVME_D32 );
  
  ArwvmeUntieFpos ( vme->device_fd );
  
  lseek ( vme->device_fd, (unsigned int)vme->addr, SEEK_SET );
  for ( i=0; i<count; i++ )
    ArwvmeRead ( vme->device_fd, (char *)data, sizeof ( unsigned int ) );
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
