
/*
 * Readout of PCI config data via PCI bios
 * the routine GetPCIConfigHeader() has no
 * hardware specific code, so you can use
 * it for any PCI hardware on a computer
 * running Linux.
 *
 * Author: Dirk Muehlenberg
 *
 * $Revision: 1.7 $ $Date: 1999/05/21 12:32:52 $
*/

#include "arwpci.h"

PCIConfigHeader *GetPCIConfigHeader(short vendor_id, short device_id, unsigned short index)
/* fills the configration header */
{
  PCIConfigHeader *pci_ch;
  int ret,i;
/* PCI Bios present ? */
  if (!pcibios_present()) {
    printk("GetPCIConfigHeader(): no pcibios present!\n");
    return (PCIConfigHeader *)NULL;
  }
  pci_ch = ANEW(PCIConfigHeader);
  if (!pci_ch) return NULL;

  pci_ch->vendor_id = vendor_id;
  pci_ch->device_id = device_id;
  pci_ch->index = index;
/* hardware in any pci slot ? */
  ret = pcibios_find_device(pci_ch->vendor_id, pci_ch->device_id, 
                           index, 
                           &pci_ch->busnr, &pci_ch->devnr);
  switch (ret) {
    case PCIBIOS_DEVICE_NOT_FOUND:
      FREE(pci_ch);
      printk("GetPCIConfigHeader(): no device with index %d found!\n", index);
      return (PCIConfigHeader *)NULL;
    case PCIBIOS_BAD_VENDOR_ID:
      FREE(pci_ch);
      printk("GetPCIConfigHeader(): invalid vendor id!\n");
      return (PCIConfigHeader *)NULL;
    case PCIBIOS_FUNC_NOT_SUPPORTED:
      FREE(pci_ch);
      printk("GetPCIConfigHeader(): pcibios func not available!\n");
      return (PCIConfigHeader *)NULL;
    case PCIBIOS_SUCCESSFUL:
    break;  
    default:
      FREE(pci_ch);
      printk("GetPCIConfigHeader(): unknown error!\n");
      return (PCIConfigHeader *)NULL;
  }
  /* now read class code */
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr, 
                                 PCI_CLASS_DEVICE, &pci_ch->class_code);
  if (ret == PCIBIOS_FUNC_NOT_SUPPORTED) {
    FREE(pci_ch);
    printk("GetPCIConfigHeader(): pcibios func not available!\n");
    return (PCIConfigHeader *)NULL;
  }
  ret = pcibios_read_config_byte(pci_ch->busnr, pci_ch->devnr, 
                                 PCI_REVISION_ID, &pci_ch->rev_id);

/* now get the addresses from config space */
  for (i=0;i<6;i++) {
    ret = pcibios_read_config_dword(pci_ch->busnr, pci_ch->devnr,
                                  PCI_BASE_ADDRESS_0+i*4, 
                                  pci_ch->base_address+i);
    printk("GetPCIConfigHeader(): found address%d: 0x%x\n",i,pci_ch->base_address[i]); 
  }

/* now read irq line */
  ret = pcibios_read_config_byte(pci_ch->busnr, pci_ch->devnr,
                                  PCI_INTERRUPT_LINE, &pci_ch->irq);
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr,
                                 PCI_SUBSYSTEM_ID, &pci_ch->subsystem_id);
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr,
                                 PCI_SUBSYSTEM_VENDOR_ID, &pci_ch->subsystem_vendor_id);

/* add more here if you like */
  return pci_ch;
}
