
/*
 * File:   arwvme.h
 *
 * Purpose: Header file for arwvme.c
 *
 * $Id: arwvme.h,v 1.21 1999/11/11 13:59:16 mathes Exp $
 * $Revision: 1.21 $
 *
 */
 
#ifndef LINUX_ARWVME_H
#define LINUX_ARWVME_H

/*
 * these are the declarations
 * exported from the driver
 */

#define ARWVME_DRIVER_INFO_SIZE 1024
/*
 * The next value gives the maximum
 * amount of transferrable bytes per
 * rw cycle. This prevents from hogging
 * the VMEbus. The value of PAGE_SIZE << 2
 * means, that one cycle takes around 2 jiffies
 * which is the heart beat of the kernel. 
 */
#define ARWVME_RW_MAX_COUNT     (PAGE_SIZE << 2)

/*
 * user supplies address-, data- and access-type
 */
typedef struct {
  unsigned short addr_type,
                 data_type,
                 acc_type;
} ArwvmeADA;

/*
 * Interrupt acknowledge types known by the irq handler
 */ 
typedef enum {ArwvmeRoak, ArwvmeRoraByUser, ArwvmeRoraByHandler} ArwvmeAckType;

/*
 * interrupt info exported to the user via ioctl()
 */
typedef struct {
  int int_vec;
  unsigned long int_tl;
  ArwvmeAckType ack_type;
  unsigned int rora_reg;
  unsigned int acc_mask;
} ArwvmeIntAck;

/*
 * Bus error (timeout) info exported to the user via ioctl()
 */
typedef enum {
  ARWVME_PCIADA_TO = 1, 
  ARWVME_RESET_BUTTON = 6, 
  ARWVME_VMEBUS_TO = 7, 
  ARWVME_ACFAIL = 16, 
  ARWVME_SYSFAIL = 19
} ArwvmeSysVectors;

typedef struct {
  unsigned long jiffies;
  int           module_number;
  unsigned int  vme_addr;
  ArwvmeSysVectors reason;
} ArwvmeSysErrInfo;

typedef enum { ArwvmeArbPri, ArwvmeArbRR } ArwvmeArbModes;

typedef struct {
  ArwvmeArbModes arb_mode;
  int br_lev;
  int fairness;
} ArwvmeArbiter;

#define ARWVME_PRIV_MASK          0x0300
#define ARWVME_SUP                0x0100
#define ARWVME_NONPRIV            0x0200

#define ARWVME_ACC_MASK           0x3000
#define ARWVME_BLOCK              0x1000
#define ARWVME_PROG               0x2000
#define ARWVME_DATA               0x3000

#define ARWVME_A_MASK             0x07
#define ARWVME_A16                0x01
#define ARWVME_A24                0x02
#define ARWVME_A32                0x03
#define ARWVME_A64                0x04

#define ARWVME_D_MASK             0x30
#define ARWVME_D8                 0x10
#define ARWVME_D16                0x20
#define ARWVME_D32                0x30

/*
 * ioctl() command codes ...
 */

#ifndef ARWVME_IOCTL_BASE
# define ARWVME_IOCTL_BASE         'd' 
#endif

#define ARWVME_INITIALIZE          _IO(ARWVME_IOCTL_BASE, 0x00)
#define ARWVME_SYSRESET            _IO(ARWVME_IOCTL_BASE, 0x01)
#define ARWVME_CHECK_CONNECTION    _IOR(ARWVME_IOCTL_BASE, 0x02, int)

#define ARWVME_SET_AM              _IOW(ARWVME_IOCTL_BASE, 0x03, ArwvmeADA)
#define ARWVME_ENABLE_INTERRUPT    _IO(ARWVME_IOCTL_BASE, 0x04)
#define ARWVME_DISABLE_INTERRUPT   _IO(ARWVME_IOCTL_BASE, 0x05)
#define ARWVME_REGISTER_INTERRUPT  _IOW(ARWVME_IOCTL_BASE, 0x06, ArwvmeIntAck)
#define ARWVME_UNREGISTER_INTERRUPT _IOW(ARWVME_IOCTL_BASE, 0x07, int)
#define ARWVME_GET_INTERRUPT       _IOR(ARWVME_IOCTL_BASE, 0x08, ArwvmeIntAck)
#define ARWVME_SET_VME_WINDOW      _IOW(ARWVME_IOCTL_BASE, 0x09, unsigned int)
#define ARWVME_ERROR_CHECK         _IOR(ARWVME_IOCTL_BASE, 0x0A, ArwvmeSysErrInfo)
#define ARWVME_TIE_FPOS            _IO(ARWVME_IOCTL_BASE, 0x0B)
#define ARWVME_UNTIE_FPOS          _IO(ARWVME_IOCTL_BASE, 0x0C)
#define ARWVME_ATTACH_SIGNAL       _IOW(ARWVME_IOCTL_BASE, 0x0E, pid_t)
#define ARWVME_GET_VME_SIZE        _IOR(ARWVME_IOCTL_BASE, 0x0F, unsigned int)
#define ARWVME_GET_VME_BASE        _IOR(ARWVME_IOCTL_BASE, 0x10, off_t)
#define ARWVME_CTRL_ARBITER        _IOW(ARWVME_IOCTL_BASE, 0x11, ArwvmeArbiter)
#define ARWVME_SET_VMEBUS_TO       _IOW(ARWVME_IOCTL_BASE, 0x12, int) 

#define ARWVME_GET_DRIVER_INFO     _IOR(ARWVME_IOCTL_BASE, 0x13, char[ARWVME_DRIVER_INFO_SIZE])

#define ARWVME_GET_DRIVER_VERSION  _IOR(ARWVME_IOCTL_BASE, 0x14, unsigned int)

#define ARWVME_MAX_INTS_QUEUED    50

/* --- obsolete ioctl()s --- */

#define ARWVME_POST_INTERRUPT	   ARWVME_REGISTER_INTERRUPT
#define ARWVME_REMOVE_INTERRUPT    ARWVME_UNREGISTER_INTERRUPT

#endif
