
/*
 * File   : list.c
 *
 * Purpose: Provide list management functions to the driver,
 *          like C++ templates.
 *
 * $Id: list.c,v 1.6 1999/10/07 09:17:26 mathes Exp $
 */
 
#include "list.h"

Node *newNode(void)
{
  Node *mynode;
      
  mynode = ANEW(Node);
  if (!mynode) return NULL;
  mynode->pred = NULL;
  mynode->succ = NULL;
  mynode->data = NULL;
   
  return mynode;
}

Node *addNode(Node *pos, void *insert, enum NodePosition p)
{
  Node *mynode;
  mynode = ANEW(Node);
  if (!mynode) return NULL;
  if (p == CurrentLeft) {
    mynode->pred = pos;
    mynode->succ = pos->succ;
    pos->succ = mynode;
  } else {
    mynode->pred = pos->pred;
    pos->pred = mynode;
    mynode->succ = pos;
  }
  mynode->data = insert;
  return mynode;
}

Node *delNodeAndData(Node *pos)
{
  Node *tail = NULL;
  if (!pos->pred) return NULL;
  /* can't delete head */
  if (!pos->succ) {
    tail = pos->pred;
    tail->succ = NULL;
  } else {
    pos->pred->succ = pos->succ;
    pos->succ->pred = pos->pred;
  }
  FREE(pos->data);
  FREE(pos);
  return tail;
}

Node *delNode(Node *pos)
{
  Node *tail = NULL;
  if (!pos) return NULL;
  if (!pos->pred) return NULL;
  /* can't delete head */
  if (!pos->succ) {
    tail = pos->pred;
    tail->succ = NULL;
  } else {
    pos->pred->succ = pos->succ;
    pos->succ->pred = pos->pred;
  }
  FREE(pos);
  return tail;
}
/* ------------------------------ */

List *newList(void)
{
  List *ret = ANEW(List);
  if (!ret) return NULL;
  ret->head = ret->tail = newNode();
  if (!ret->head) {
    FREE(ret);
    return NULL;
  }
  ret->nodes = 0;
  return ret;
}

void deleteList(List *l, void (*delete)(void *))
{
  if (l) {
    resetList(l,delete);
    FREE(l->head);
    FREE(l);
  }
}

Node *addTail(List *list, void *insert)
{
  Node *tail;
  if (!list) return NULL;
  tail = addNode(list->tail,insert,CurrentLeft);
  if (!tail) {
    printk("addTail(): addNode failed!\n");
    return NULL;
  }
  list->tail = tail;
  list->nodes++;
  return tail;
}

Node *addHead(List *list, void *insert)
{
  Node *node;
  if (!list) return NULL;
  if (!list->head->succ) {
    return addTail(list,insert);
  } else {
    node = addNode(list->head,insert,CurrentRight);
  }
  if (node) list->nodes++;
  return node;
}

Node *searchList(List *list, void *c, int (*comp)(void *, void *))
{
  register Node *n; 

  if (!list) return NULL;
  n = list->head->succ;
  while (n) {
    if (!comp(c,n->data)) {
      return n;
    } 
    n = n->succ;
  }  
  return NULL;
}

void delNodeAndDataInList(List *l, Node *pos)
{
  Node *n;
  if (!l) return;
  if ((n = delNodeAndData(pos))) l->tail = n;
  l->nodes--;
}

void delNodeInList(List *l, Node *pos)
{
  Node *n;
  if (!l) return;
  if ((n = delNode(pos))) l->tail = n;
  l->nodes--;
}

int emptyList(List *l)
{
  if (!l) return 1;
  return l->head->succ ? 0 : 1;
}

void delTailAndData(List *l)
{
  Node *t;
  if (!l) return;
  t = l->tail;
  if (emptyList(l)) return;
  l->tail = l->tail->pred;
  l->tail->succ = NULL;
  FREE(t->data);
  FREE(t);
  l->nodes--;
}

void delTail(List *l)
{
  Node *t;
  if (!l) return;
  t = l->tail;
  if (emptyList(l)) return;
  l->tail = l->tail->pred;
  l->tail->succ = NULL;
  FREE(t);
  l->nodes--;
}

void resetList(List *l, void (*delete)(void *))
{
  register Node *n;
  register void *data;

  if (!l) return;
  n = l->tail;
  while (n->pred) {
    data = n->data;
    if (delete) {
      delete(data);
    } else {
      FREE(data);
    }
#if 0
    n = delNode(n);
#endif
    delNode(n);
    n = n->pred;
  }
  l->tail = l->head;
  l->nodes = 0;
}

Node *getFirstNode(List *l)
{
  if (!l) return NULL;
  return l->head->succ;
}

Node *getLastNode(List *l)
{
  if (!l) return NULL;
  return l->tail;
}

void delFirstNode(List *l)
{
  if (!l) return;
  delNodeInList(l,getFirstNode(l));
}

void delFirstNodeAndData(List *l)
{
  if (!l) return;
  delNodeAndDataInList(l,getFirstNode(l));
}

void delLastNode(List *l)
{
  if (!l) return;
  delNodeInList(l,getLastNode(l));
}

void delLastNodeAndData(List *l)
{
  if (!l) return;
  delNodeAndDataInList(l,getLastNode(l));
}

#ifndef USE_MACROS
int getNumOfNodesInList(List *l)
{
  if (!l) return 0;
  return l->nodes;
}
#endif

#ifdef MODULE
#define printf printk
#endif

void printList(List *l, char *(*StructToString)(void *))
{
  register Node *n = l->head;
  int i=1;
  if (!l) return;
  if (!n->succ) {
    printf("List empty!\n");
    return;
  }
  while (n->succ) {
    printf("Node %d: %s",i++,StructToString(n->succ->data));
    n = n->succ;
  }
  printf("\n");
}

#undef printf
