#ifndef _LIST_H
#define _LIST_H

typedef struct _node {
  struct _node *pred, *succ;
  void *data;
} Node;

enum NodePosition {CurrentLeft, CurrentRight};

Node *initNode(void);
Node *addNode(Node *, void *, enum NodePosition);
Node *delNode(Node *);

typedef struct  _list {
  Node *head, *tail;
  int nodes;
} List;

List *newList(void);
void deleteList(List *, void (*)(void *));
Node *addTail(List *, void *);
Node *addHead(List *, void *);
Node *searchList(List *, void *, int (*)(void *, void *));
void delNodeInList(List *, Node *);
void delNodeAndDataInList(List *, Node *);
void delTail(List *);
void delTailAndData(List *);
void resetList(List *, void (*)(void *));
void printList(List *, char *(*)(void *));
#ifndef USE_MACROS
int getNumOfNodesInList(List *);
#else
#define getNumOfNodesInList(l) l->nodes
#endif
int emptyList(List *);
/* for lifo-, fifo operations */
Node *getFirstNode(List *);
Node *getLastNode(List *);
void delFirstNode(List *);
void delFirstNodeAndData(List *);
void delLastNode(List *);
void delLastNodeAndData(List *);

#ifdef MODULE

#ifndef ANEW
#define ANEW(type) (type *)kmalloc(sizeof(type), GFP_ATOMIC)
/* GFP_ATOMIC = not waiting for a free page */
#define FREE(ptr) kfree_s(ptr, sizeof(*ptr))
#endif

#else

#ifndef ANEW
#define ANEW(type) (type *)malloc(sizeof(type))
#define FREE free
#endif

#endif

#endif
