
/*
 * File   : test_v462.c
 *
 * Purpose: Test program for ARWVME interface kit
 *          testing the V462 GATE 
 *
 *
 * $Revision: 1.20 $
 * $Date: 1999/11/11 13:59:05 $
 * $Id: test_v462.c,v 1.20 1999/11/11 13:59:05 mathes Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sched.h>
#include <arwvme.h>
#include <time.h>

#include <sys/times.h>
#include <sched.h>

#define TTT 10
#define GATE_BASE 0x00FA0000
#define GATE_RANGE 0xFF

#define SWAP_WORD(a) (((a >> 16)) | ((a & 0xFFFF) << 16))

#define TEST_INTERRUPT		0

static int arwvme_fd = -1;
static int sig_cnt;
static int sig_sem;

unsigned int getBCD(int n)
{
  int i;
  unsigned int bcd = 0;
  for (i=0;i<2*sizeof(int);i++)
  {
    bcd += (n%10) << (i<<2);
    n /= 10;
    if (!n) break;
  }
  return bcd;
}

void sig_handler(int sig)
{
  ArwvmeIntAck *int_ack;
  ArwvmeSysErrInfo *sei;
  switch (sig) {
    case SIGUSR1:
      puts("sig_handler::SIGUSR1");
      int_ack = ArwvmeGetNextPendingInt(arwvme_fd);
      if (int_ack) {
        printf("Int 0x%x occured at %Ld\n",int_ack->int_vec,int_ack->int_tl);
        free(int_ack);
      }
    break;
    case SIGBUS:
      sei = ArwvmeGetSysErrInfo(arwvme_fd);
sig_cnt++;
sig_sem = 0;
      if (sei->module_number > 0) {
#if 0
       printf("VMEBus Sys error: reason = %d, tl = %ld, addr = 0x%lx\n", sei->reason, sei->jiffies, sei->vme_addr);
#endif
      } else {
        printf("Buserror sent from kernel!\n");
        exit(1);
      }
    break;
    default:
      puts("sig_handler::default");
  }
}

/* CSR access via mmap */
#define derefCSR(base)     *(unsigned short *)base
#define GetError           (derefCSR(base) & 0x8000)
#define TestChanInVmeMode0 (derefCSR(base) & 0x2000)
#define TestChanInVmeMode1 (derefCSR(base) & 0x4000)
#define TestGatePresent0   (derefCSR(base) & 0x0800)
#define TestGatePresent1   (derefCSR(base) & 0x1000)
#define StartGate0         (derefCSR(base) |= 0x0200)
#define StartGate1         (derefCSR(base) |= 0x0400)

#define getVersion(base)   ((*(unsigned short *)(base+0xFE) & 0xF000) >> 12)
#define getSeries(base)    (*(unsigned short *)(base+0xFE) & 0x0FFF)
#define getManufact(base)  ((*(unsigned short *)(base+0xFC) & 0xFE00) >> 10)
#define getModType(base)   (*(unsigned short *)(base+0xFC) & 0x003F)       
#define identifyV462(base) (*(unsigned short *)(base+0xFC) == 0x80A)

int readV462InfosViaMMap(unsigned char *base, char *buf)
{
  int p = 0;
  p = sprintf(buf, "V462 Module Identifier Words\n");
  p += sprintf(buf+p, "Version: 0x%x, Serial Number: 0x%x\n",
                    getVersion(base),getSeries(base));
  p += sprintf(buf+p, "Manufacturer Nr.: 0x%x, Type: 0x%x\n",
                    getManufact(base), getModType(base));
  p += sprintf(buf+p, identifyV462(base) ? "Module is V462\n" : "Module unknown\n");
  return p;
}

void startGate0(unsigned int offs)
{
  unsigned short csr;
  lseek(arwvme_fd, offs+2, 0);
  read(arwvme_fd, &csr, 2);
  csr |= 0x200;
  lseek(arwvme_fd, offs+2, 0);
  write(arwvme_fd, &csr, 2);
}

void stopGate0(unsigned int offs)
{
  unsigned short csr;
  lseek(arwvme_fd, offs+2, 0);
  read(arwvme_fd, &csr, 2);
  csr &= ~0x200;
  lseek(arwvme_fd, offs+2, 0);
  write(arwvme_fd, &csr, 2);
}

void writeGateViaMMap(unsigned short *base, unsigned int usec, int chan)
{
  if (chan < 0 || chan > 1) return;
  *(base+2+2*chan) = usec & 0xFFFF;
  *(base+1+2*chan) = (usec & 0xFFFF0000) >> 16;
}

void writeGate(unsigned int offs, unsigned int usec, int chan)
{
  if (chan < 0 || chan > 1) return;
  lseek(arwvme_fd, offs+2+chan * 4, 0);
  write(arwvme_fd, &usec, 4);
}

int readGateViaMMap(unsigned short *base, int chan)
{
  unsigned int bcd;
  bcd = *(base + 2+ chan*2);
  bcd |= *(base + 1 + chan*2) << 16;
  return bcd;
}

int readGate(unsigned int offs, int chan)
{
  unsigned int gate;
  if (chan < 0 || chan > 1) return 0xFFFFFFFF;
  lseek(arwvme_fd, offs+2+chan*4, 0);
  read(arwvme_fd, &gate, 4);
  return gate; 
}

int readV462Infos(unsigned int offs)
{
  unsigned short at_0xfc = 0;
  unsigned short at_0xfe = 0;
  lseek(arwvme_fd, offs+0xFC, 0);
  read(arwvme_fd, &at_0xfc, 2);
  lseek(arwvme_fd, offs+0xFE, 0);
  read(arwvme_fd, &at_0xfe, 2);
  printf("at_0xfc: 0x%x\n",at_0xfc);
  printf("at_0xfe: 0x%x\n",at_0xfe);
}

void main ( int argc, char *argv[] )
 {
  const int ROUNDS = (1<<20);
  clock_t start,help;
  unsigned char *ret,buf[ROUNDS];
  int i;
  unsigned short a;
  int offset = 0;
  int channel = 0;
  ArwvmeSysErrInfo *bei;
  long cls[ROUNDS];
  struct sched_param p; 
  static struct timespec ts = {1,0};
  struct timespec rem;
  
  if (argc < 2) channel = 0;
  else channel = atoi(argv[1]);
  if (channel < 0 || channel > 1) channel = 0;
  printf ( "ArwvmeOpen: %d %d\n", arwvme_fd = ArwvmeOpen(2), errno );
  if (arwvme_fd < 0) exit(1);

  printf ( "ArwvmeCheckConnection: %d\n", ArwvmeCheckConnection(arwvme_fd) );
  
  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A24 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);
/* this is the only possible mode of the v462 */

#if ( TEST_INTERRUPT == 1 )
  ArwvmeEnableInterrupt(arwvme_fd);
  /* register interrupt handler and enable int */

    printf ( "Testing interrupts ...\n" );

    ArwvmeEnableInterrupt(arwvme_fd);
    /* very important: choose the correct AckType for your board! */ 
    ArwvmePostInterrupt(arwvme_fd, ArwvmeRoak, 0x0, 0x40, 0x0);
    /* AckType, Rora Register, Vector */

    ArwvmeSignal(arwvme_fd, sig_handler);
    /* catch the VME system errors */
    signal(SIGBUS, sig_handler); 
#endif

#ifdef TEST_MMAP
  /* --- test for mmap() call --- */
  ret = ArwvmeMMap(arwvme_fd, 0x1000);
  printf("ret: %p\n",ret);
  /* choose appropriate page on VME bus */
  if (ret) {
    ArwvmeSetVmeWindow(arwvme_fd, GATE_BASE);
    /* accessing V462 */
    readV462InfosViaMMap(ret,buf);
    printf("V462 Identifier Strings:\n%s\n",buf);
    ArwvmeSetVmeWindow(arwvme_fd, GATE_BASE);

    p.sched_priority = sched_get_priority_max(SCHED_RR);
    sched_setscheduler(0,SCHED_RR,&p);

    start = times(NULL);
    for (i=0;i<ROUNDS;i++) {
      *(ret + 0x2) = 1; /* hopefully generating a SIGBUS */
      nanosleep(&ts,&rem);
      cls[i] = (ts.tv_sec-rem.tv_sec)*1000000000L+(ts.tv_nsec-rem.tv_nsec);
    }

    p.sched_priority = 0;
    sched_setscheduler(0,SCHED_OTHER,&p);

    for (i=0;i<ROUNDS;i++) {
      if (!(i%60)) printf("\n");
      printf("%d ",cls[i]);
    }
  } else {
    printf("Sorry, no remap!\n");
  }
#endif
  readV462Infos(GATE_BASE);
  memset(buf, (channel) ? 0x77 : 0x11, (ROUNDS));
  ArwvmeUntieFpos(arwvme_fd);
  if (lseek(arwvme_fd, GATE_BASE+2, 0) < 0) {
     printf("lseek: %s\n",sys_errlist[errno]);
     exit(1);
  }
    p.sched_priority = sched_get_priority_max(SCHED_RR);
    sched_setscheduler(0,SCHED_RR,&p);

  start = times(NULL);
  i = ArwvmeWrite(arwvme_fd, buf, ROUNDS);
  printf("written %d bytes\n",i);
  printf("time: %d\n", times(NULL)-start);  
    p.sched_priority = 0;
    sched_setscheduler(0,SCHED_OTHER,&p);

  ArwvmeShutdown(arwvme_fd);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
