
/*
 * File   : test_v560.c
 *
 * Purpose: Test program for ARWVME interface kit
 *          testing the V560 scaler 
 *
 *
 * $Revision: 1.4 $
 * $Date: 1999/03/02 10:48:04 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <arwvme.h>
#include <time.h>

#include <sys/times.h>

#define MOD1	1
#define SCALER_BASE 0x00FF0000
#define SCALER_RANGE 0xFF

#define SWAP_WORD(a) (((a >> 16)) | ((a & 0xFFFF) << 16))

#define TEST_INTERRUPT

#define TEST_MMAP

static int arwvme_fd = -1;

void sig_handler(int sig)
{
  ArwvmeIntAck *int_ack;
  switch (sig) {
    case SIGUSR1:
      puts("sig_handler::SIGUSR1");
      int_ack = ArwvmeGetNextPendingInt(arwvme_fd);
      if (int_ack) {
        printf("Int 0x%X occured at %Ld\n",int_ack->int_vec,int_ack->int_tl);
        free(int_ack);
      }
    break;
    default:
      puts("sig_handler::default");
  }
}

int readV560Infos(unsigned char *map, char *buf)
{
  int p = 0;
  int version, serialno, manufactno, moduletype;
  unsigned short fe;
  unsigned short fc;
  unsigned short fa;
  if (!map) return 0;
  fe = *(unsigned short *)(map+0xFE);
  fc = *(unsigned short *)(map+0xFC);
  fa = *(unsigned short *)(map+0xFA);
  version = (fe & 0xF000) >> 12;
  serialno = fe & 0x0FFF;
  manufactno = (fc & 0xFC00) >> 10;
  moduletype = fc & 0x03FF;
  p += sprintf(buf+p,"Version: %x\n",version);
  p += sprintf(buf+p,"Serial #: %x\n",serialno);
  p += sprintf(buf+p,"Manuf. #: %x\n", manufactno);
  p += sprintf(buf+p,"Mod. Typ: %x\n", moduletype);
  return p;
}

int readScalerD16(unsigned char *map, int which)
{
  int s;
  if (which < 0 || which > 15) return -1;
  s = *(unsigned short *)(map+0x10+4*which) << 16;
  s += *(unsigned short *)(map+0x12+4*which);
  return s;
}

int readScalerD32(unsigned char *map, int which)
{
  int s;
  if (which < 0 || which > 15) return -1;
  lseek(arwvme_fd, SCALER_BASE+0x10+4*which,SEEK_SET);
  read(arwvme_fd, (char *)&s,4);	
  return s;
}

void clearVeto(unsigned char *map)
{
  *(unsigned short *)(map+0x54) = 1;
}

void clearScalers(unsigned char *map)
{
  *(unsigned short *)(map+0x50) = 1;
}

void enableInt(unsigned char *map)
{
  *(unsigned short *)(map+0x08) = 1;
}

void disableInt(unsigned char *map)
{
  *(unsigned short *)(map+0x0A) = 1;
}

main ( void )
 {
  struct tms ti;
  clock_t cl;
  unsigned char *ret,buf[1024];
  struct timeval tv = {0,100};
  ArwvmeIntAck *int_ack;
  int i,w,p;
  unsigned int lumpf;
  
  printf ( "ArwvmeOpen: %d %d\n", arwvme_fd = ArwvmeOpen(MOD1), errno );
  if (arwvme_fd < 0) exit(1);

  printf ( "ArwvmeCheckConnection: %d\n", ArwvmeCheckConnection(arwvme_fd) );
  
  printf ( "ArwvmeResetVme: %d %d\n", ArwvmeResetVme(arwvme_fd), errno );

  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A24 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);

  /* --- test for mmap() call --- */
  
  /* assign mem range [0xD0000,0xDFFFF] of ATADA VME-Window to user buffer ret */
  /* do the mapping only once! */ 

  ret = ArwvmeMMap(arwvme_fd, 0x1000);
  /* choose appropriate 64K page on VME bus */
  ArwvmeSetVmeWindow(arwvme_fd, SCALER_BASE);
  /* accessing V560 scaler */
  if (ret) {
    p = readV560Infos(ret,buf);
    printf("V560 Identifier Strings:\n%s\n",buf);
    clearVeto(ret);
    clearScalers(ret);
  /* --- test for select() call --- */
  
  /* register interrupt handler and enable int */

    printf ( "Testing interrupts ...\n" );

    printf("Int Lev Reg: %x\n",*(unsigned short *)(ret + 0x06) & 0x7);
    *(unsigned short *)(ret + 0x04) = 0xAA;
    printf("Int Vec Reg: %x\n",*(unsigned short *)(ret + 0x04));
    *(unsigned short *)(ret + 0x0E) = 0xFF;
    printf("Int En Reg: %x\n",*(unsigned short *)(ret + 0x0E));
    printf("Statusreg 0x58: %x\n",*(unsigned short *)(ret + 0x58));

    ArwvmeEnableInterrupt(arwvme_fd);
    /* very important: choose the correct AckType for your board! */ 
    ArwvmePostInterrupt(arwvme_fd, ArwvmeRoraByHandler, SCALER_BASE+0x0C, 0xAA, 
                       ARWVME_A24 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);
    /* AckType, Rora Register, Vector */

    enableInt(ret);

    i = 0;
    ArwvmeSignal(arwvme_fd, sig_handler);
    while (i<0x80000000) {
      sleep(1);
      printf("read: %x\n", i = readScalerD32(ret, 15));
    }
#if 0
    while (ArwvmeSelect(arwvme_fd, &tv)) {
      int_ack = ArwvmeGetNextPendingInt(arwvme_fd);
      if (int_ack) {
        printf("Int 0x%X occured at %Ld\n",int_ack->int_vec,int_ack->int_tl);
        free(int_ack);
      } else {
        printf("Sorry, no int available!\n");
      }
    }
#endif
  } else {
    printf("Sorry, no remap!\n");
  }

  ArwvmeShutdown(arwvme_fd);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
