
/*
 * Project: Vme interface(s)
 *
 * File   : vme.h
 *
 * Purpose: General header file for vme functions
 *          So when it is intended to use another Vme-Controller:
 *          ==> just change a #define !
 *
 * $Date: 1999/12/14 16:32:08 $
 * $Id: vme.h,v 0.16 1999/12/14 16:32:08 mathes Exp $
 * $Revision: 0.16 $
 */
 
#ifndef _vme_h_
#define _vme_h_

#undef _VME_

#ifdef ATVME                       /* --- ??? --- */
# include <linux/vme.h>
#endif

#ifdef __WATCOMC__
# undef TVC
# define ATVME
#endif


                /* ----------------------------------------- */
                /*                                           */
                /*   Transputer based Vme Controller (TVC)   */
                /*                                           */
                /* ----------------------------------------- */

#ifdef TVC
# define _VME_

#include <vmetvc.h>

#define VME_A16      TVC_A16
#define VME_A24      TVC_A24
#define VME_A32      TVC_A32
#define VME_D8       TVC_D8
#define VME_D16      TVC_D16
#define VME_D32      TVC_D32

#define VME_BLT      0


#define VmeDesc VMELOC

#define VmeInit()                       VmeInit()

#define VmeGetVersionString()           "VmeTvc Version"

#define VmeWriteChar(_vme,_data)        VmeOutChar(_vme,_data)
#define VmeReadChar(_vme)               VmeInChar(_vme)
#define VmeWriteShort(_vme,_data)       VmeOutShort(_vme,_data)
#define VmeReadShort(_vme)              VmeInShort(_vme)
#define VmeWriteInt(_vme,_data)         VmeOutInt(_vme,_data)
#define VmeReadInt(_vme)                VmeInInt(_vme)

#define InitVmeDesc(_x,_addr,_acc)      { _x.addr = (void *)(_addr); \
                                          _x.acc = (_acc); }
#define SetVmeAddr(_x,_addr)            { _x.addr = (void *)(_addr); }
#define SetVmeAccessType(_x,_acc)        { _x.acc = (_acc); }

#endif /* ------ TVC ------ */

                /* ----------------------------------------- */
                /*                                           */
                /*    WIENER PC/AT-to-Vme interface ATVME    */
                /*                                           */
                /* ----------------------------------------- */
#ifdef ATVME
# define _VME_

#include <atvme.h>

#define VME_A16      ATVME_A16
#define VME_A24      ATVME_A24
#define VME_A32      ATVME_A32
#define VME_D8       ATVME_D8
#define VME_D16      ATVME_D16
#define VME_D32      ATVME_D32

#define VME_BLT      0

#define InitVmeDesc(_x,_addr,_acc)      { _x.addr = (void *)(_addr); \
                                          _x.acc = (_acc); }
#define SetVmeAddr(_x,_addr)            { _x.addr = (void *)(_addr); }
#define SetVmeAccessType(_x,_acc)       { _x.acc = (_acc); }

#define VmeInit()                       AtvmeOpen()
#define VmeReset()			AtvmeResetVme()

#define VmeGetVersionString()           AtvmeGetVersionString()

#define VmeGetFd()                      AtvmeGetFd()

#define VmeWriteChar(_vme,_data)        AtvmeWriteChar(_vme,_data)
#define VmeReadChar(_vme)               AtvmeReadChar(_vme)
#define VmeWriteShort(_vme,_data)       AtvmeWriteShort(_vme,_data)
#define VmeReadShort(_vme)              AtvmeReadShort(_vme)
#define VmeWriteInt(_vme,_data)         AtvmeWriteInt(_vme,_data)
#define VmeReadInt(_vme)                AtvmeReadInt(_vme)

#define VmeBlockReadShort(_vme,_data,_count) \
					AtvmeBlockReadShort(_vme,_data,_count)
#define VmeFifoReadShort(_vme,_data,_count) \
					AtvmeFifoReadShort(_vme,_data,_count)

#define VmeSetSysFail()		        AtvmeSetSysFail()
#define VmeClearSysFail()	        AtvmeClearSysFail()
#define VmeCheckSysFail()	        AtvmeCheckSysFail()

/* --- for interrupts --- */

#define VmeGetInterruptFd()             AtvmeGetInterruptFd()

typedef IntAckStruct VmeIntAckStruct;

#define VmeEnableInterrupt()            AtvmeEnableInterrupt()
#define VmeDisableInterrupt()           AtvmeDisableInterrupt()
#define VmeRegisterInterrupt(_ack_type,_ack_adr,_int_vec,_am) \
	AtvmeRegisterInterrupt(_ack_type,_ack_adr,_int_vec,_am)
#define VmeUnregisterInterrupt(_int_vec) \
        AtvmeUnregisterInterrupt(_int_vec)
/* --- two obsolete calls --- */
# define VmeAttachInterrupt             AtvmeRegisterInterrupt
# define VmeRemoveInterrupt             AtvmeUnregisterInterrupt
/* --- --- */
#define VmeSelect(_x)			AtvmeSelect(_x)
#define VmeGetNextPendingInt()		AtvmeGetNextPendingInt()

typedef enum
 { 
  VME_ROAK = AtvmeRoak, 
  VME_RORA_BY_USER = AtvmeRoraByUser, 
  VME_RORA_BY_HANDLER = AtvmeRoraByHandler

} VmeAckType;

#endif  /* ------ ATVME ------ */

                /* ----------------------------------------- */
                /*                                           */
                /*    SMS Bit3 PCI/Vme interface Model 617   */
                /*                                           */
                /* ----------------------------------------- */

#ifdef BIT3_617
# define _VME_

#include <bit3_vme.h>

#define VME_A16         BIT3_VME_A16
#define VME_A24         BIT3_VME_A24
#define VME_A32      	BIT3_VME_A32
#define VME_D8          BIT3_VME_D8
#define VME_D16         BIT3_VME_D16
#define VME_D32         BIT3_VME_D32

#define VME_BLT         BIT3_VME_BLT

#define VmeDesc				Bit3VmeDesc
#define InitVmeDesc(_x,_addr,_acc)	Bit3InitVmeDesc(_x,_addr,_acc)
#define SetVmeAddr(_x,_addr)		Bit3SetVmeAddr(_x,_addr)

#define VmeInit()			Bit3VmeOpen()

#define VmeGetVersionString()		Bit3VmeGetVersionString()

#define VmeReadChar(_vme)		Bit3VmeReadChar(_vme)
#define VmeWriteChar(_vme,_data)	Bit3VmeWriteChar(_vme,_data)

#define VmeReadShort(_vme)		Bit3VmeReadShort(_vme)
#define VmeWriteShort(_vme,_data)	Bit3VmeWriteShort(_vme,_data)
#define VmeBlockReadShort(_vme,_data,_count)	\
					Bit3VmeBlockReadShort(_vme,_data,_count)
#define VmeFifoReadShort(_vme,_data,_count)	\
					Bit3VmeFifoReadShort(_vme,_data,_count)

#define VmeBlockReadInt(_vme,_data,_count)	\
					Bit3VmeBlockReadInt(_vme,_data,_count)
#define VmeFifoReadInt(_vme,_data,_count)	\
					Bit3VmeFifoReadInt(_vme,_data,_count)

#endif  /* ----- BIT3_617 ----- */

                /* ----------------------------------------- */
                /*                                           */
                /*      Wiener PCI/Vme interface PCIVME      */
                /*                                           */
                /* ----------------------------------------- */

#ifdef ARWVME
# define _VME_

#include <arwvme.h>

#define VME_A16         ARWVME_A16
#define VME_A24         ARWVME_A24
#define VME_A32      	ARWVME_A32
#define VME_D8          ARWVME_D8
#define VME_D16         ARWVME_D16
#define VME_D32         ARWVME_D32

#define VmeGetFd()			ArwvmeDefaultDevice;

#define VmeDesc				ArwvmeDesc
#define InitVmeDesc(_x,_addr,_acc)	{ _x.addr = (void *)(_addr); \
                                          _x.acc = (_acc); \
					  _x.device_fd = ArwvmeDefaultDevice; }
#define SetVmeAddr(_x,_addr)		{ _x.addr = (void *)(_addr); }
#define GetVmeAddr(_x)			(char *)( _x.addr )

#define VmeInit()                       ArwvmeSimpleOpen()
#define VmeReset()			ArwvmeResetVme()

#define VmeGetVersionString()           ArwvmeGetVersionString()

#define VmeWriteChar(_vme,_data)        ArwvmeWriteChar(_vme,_data)
#define VmeReadChar(_vme)               ArwvmeReadChar(_vme)
#define VmeWriteShort(_vme,_data)       ArwvmeWriteShort(_vme,_data)
#define VmeReadShort(_vme)              ArwvmeReadShort(_vme)
#define VmeWriteInt(_vme,_data)         ArwvmeWriteInt(_vme,_data)
#define VmeReadInt(_vme)                ArwvmeReadInt(_vme)

#define VmeBlockWriteChar(_vme,_data,_count) \
					ArwvmeBlockWriteChar(_vme,_data,_count)
#define VmeBlockReadShort(_vme,_data,_count) \
					ArwvmeBlockReadShort(_vme,_data,_count)
#define VmeBlockReadInt(_vme,_data,_count) \
					ArwvmeBlockReadInt(_vme,_data,_count)

#define VmeFifoReadShort(_vme,_data,_count) \
					ArwvmeFifoReadShort(_vme,_data,_count)
#define VmeFifoReadInt(_vme,_data,_count) \
					ArwvmeFifoReadInt(_vme,_data,_count)

/* --- for interrupt handling --- */

#define VmeGetInterruptFd()		ArwvmeDefaultDevice;

typedef enum
 { 
  VME_ROAK = ArwvmeRoak, 
  VME_RORA_BY_USER = ArwvmeRoraByUser, 
  VME_RORA_BY_HANDLER = ArwvmeRoraByHandler

} VmeAckType;


#define VmeIntAckStruct			ArwvmeIntAck
#define VmeEnableInterrupt()            \
	ArwvmeEnableInterrupt(ArwvmeDefaultDevice)
#define VmeDisableInterrupt()           \
	ArwvmeDisableInterrupt(ArwvmeDefaultDevice)

#define VmeRegisterInterrupt(_ack_type,_ack_adr,_int_vec,_am) \
	ArwvmeRegisterInterrupt(ArwvmeDefaultDevice,\
	                        _ack_type,_ack_adr,_int_vec,_am)
#define VmeUnregisterInterrupt(_int_vec) \
        ArwvmeUnregisterInterrupt(ArwvmeDefaultDevice,_int_vec)

#define VmeSelect(_x)			ArwvmeSelect(ArwvmeDefaultDevice,_x)
#define VmeGetNextPendingInt()		\
	ArwvmeGetNextPendingInt(ArwvmeDefaultDevice)

#endif  /* ----- ARWVME ----- */

#endif  /* _vme_h_ */
