//****************************************************************************
// Copyright (C) 2000,2001  ARW Elektronik Germany
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// This product is not authorized for use as critical component in 
// life support systems without the express written approval of 
// ARW Elektronik Germany.
//  
// Please announce changes and hints to ARW Elektronik
//
// Maintainer(s): Klaus Hitschler (klaus.hitschler@gmx.de)
//
//****************************************************************************

//****************************************************************************
//
// askpci.c - a hardware independent tool to get 
//           information about searched pci-hardware
//
// $Log: askpci.c,v $
// Revision 1.4  2001/11/20 20:12:50  klaus
// included new header and CVS log
//
//
// derived from code originated from Dirk Muehlenberg               AR   18.02.2000
// MODVERSIONS included                                             AR   24.04.2000
//
//****************************************************************************


/*
      
   (c) 2000 ARW Elektronik

   this source code is published under GPL (Open Source). You can use, redistrubute and 
   modify it unless this header   is not modified or deleted. No warranty is given that 
   this software will work like expected.

*/

#include "common.h"  /* must be the first include */

#include <linux/pci.h>
#include <asm/types.h>
#include "list.h"     
#include "askpci.h"
 
#ifndef NULL
#define NULL 0
#endif

/* fills the configration header */
PCIConfigHeader *GetPCIConfigHeader ( __u16 vendor_id, __u16 device_id, short index)
{
  PCIConfigHeader *pci_ch = (PCIConfigHeader *)NULL;
#ifdef CONFIG_PCI  

  int ret, i;

  if (!pcibios_present())
  {
    printk("GetPCIConfigHeader(): no pcibios present!\n");
    return (PCIConfigHeader *)NULL;
  }

  pci_ch = (PCIConfigHeader *)kmalloc(sizeof(PCIConfigHeader), GFP_ATOMIC);
  if (!pci_ch) return (PCIConfigHeader *)NULL;

  pci_ch->vendor_id = vendor_id;
  pci_ch->device_id = device_id;
  pci_ch->index = index;
  
  ret = pcibios_find_device(pci_ch->vendor_id, pci_ch->device_id, index, &pci_ch->busnr, &pci_ch->devnr);

  switch (ret)
  {
    case PCIBIOS_DEVICE_NOT_FOUND:
      kfree_s(pci_ch, sizeof(*pci_ch)); // FREE(pci_ch);
      return (PCIConfigHeader *)NULL;
  
    case PCIBIOS_SUCCESSFUL: break;  

    default:
      kfree_s(pci_ch, sizeof(*pci_ch)); // FREE(pci_ch);
      printk("GetPCIConfigHeader(): PCI-error %d!\n", ret);
      return (PCIConfigHeader *)NULL;
  }
  
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr, 
                                 PCI_CLASS_DEVICE, &pci_ch->class_code);

  if (ret == PCIBIOS_FUNC_NOT_SUPPORTED)
  {
    kfree_s(pci_ch, sizeof(*pci_ch)); // FREE(pci_ch);
    printk("GetPCIConfigHeader(): pcibios func not available!\n");
    return (PCIConfigHeader *)NULL;
  }
  
  ret = pcibios_read_config_byte(pci_ch->busnr, pci_ch->devnr, 
                                 PCI_REVISION_ID, &pci_ch->rev_id);

  for (i=0;i<6;i++)
  {
    unsigned int val;
    unsigned int location = PCI_BASE_ADDRESS_0+i*4;
    
    ret = pcibios_read_config_dword(pci_ch->busnr, pci_ch->devnr,   /* get base */
                                  location, 
                                  &pci_ch->desc[i].base_address);

    ret = pcibios_write_config_dword(pci_ch->busnr, pci_ch->devnr,  /* get size */
                                  location, 
                                  0xFFFFFFFF);

    ret = pcibios_read_config_dword(pci_ch->busnr, pci_ch->devnr,
                                  location, 
                                  &val);

    ret = pcibios_write_config_dword(pci_ch->busnr, pci_ch->devnr,  /* restore */
                                  location, 
                                  pci_ch->desc[i].base_address);

    if (pci_ch->desc[i].base_address & PCI_BASE_ADDRESS_SPACE_IO) /* io space */
    {
    	pci_ch->desc[i].type = PCI_BASE_ADDRESS_SPACE_IO; 
	pci_ch->desc[i].prefetchable = 0;
	pci_ch->desc[i].base_address &= PCI_BASE_ADDRESS_IO_MASK;
	val &= PCI_BASE_ADDRESS_IO_MASK;
	pci_ch->desc[i].size = ~val + 1; 
    }
    else
    {
	pci_ch->desc[i].type = pci_ch->desc[i].base_address & PCI_BASE_ADDRESS_MEM_TYPE_MASK;
	pci_ch->desc[i].prefetchable = (pci_ch->desc[i].base_address & PCI_BASE_ADDRESS_MEM_PREFETCH) ? 1 : 0;
	pci_ch->desc[i].base_address &= PCI_BASE_ADDRESS_MEM_MASK; 
	val &= PCI_BASE_ADDRESS_MEM_MASK;
	pci_ch->desc[i].size = ~val + 1; 
    }				  				  
 }

  /* --- now read irq line --- */
  
  ret = pcibios_read_config_byte(pci_ch->busnr, pci_ch->devnr,
                                 PCI_INTERRUPT_LINE, &pci_ch->irq);
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr,
                                 PCI_SUBSYSTEM_ID, &pci_ch->subsystem_id);
  ret = pcibios_read_config_word(pci_ch->busnr, pci_ch->devnr,
                                 PCI_SUBSYSTEM_VENDOR_ID, &pci_ch->subsystem_vendor_id);
#endif  

  return pci_ch;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

